{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            copyright  2014 - 2015                                 }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}

unit FMX.TMSChart;

interface

{$I TMSDEFS.INC}

uses
  Classes, SysUtils, FMX.Types, Types, DateUtils, FMX.StdCtrls,
  Generics.Collections, Math, FMX.Controls, TypInfo, UITypes, UIConsts
  ,FMX.Graphics, System.Math.Vectors, FMX.TMSChartPersistence
  ;

const
  MAJ_VER = 2; // Major version nr.
  MIN_VER = 0; // Minor version nr.
  REL_VER = 3; // Release nr.
  BLD_VER = 3; // Build nr.
  {$HINTS OFF}
  TMSPlatforms = pidWin32 or pidWin64 or pidOSX32 or {$IF COMPILERVERSION > 32}pidOSX64 or pidiOSSimulator32 or pidiOSSimulator64{$ELSE}pidiOSSimulator{$IFEND} or {$IFDEF DELPHIXE8_LVL}pidiOSDevice32 or pidiOSDevice64{$ELSE}pidiOSDevice{$ENDIF} or {$IF COMPILERVERSION > 32}{$IF COMPILERVERSION > 34}pidAndroidArm32 or pidAndroidArm64{$ELSE}pidAndroid32Arm or pidAndroid64Arm{$IFEND}{$ELSE}pidAndroid{$IFEND};
  {$HINTS ON}
  INTERACTIONMARGIN = 5;

  //v1.0.0.0 : First release
  //v1.1.0.0 : New : XE6 Support
  //v1.1.0.1 : Improved : Trial version compiled in release mode
  //v1.2.0.0 : New: support for Pie, Sized Pie and Variable Radius Pie types
  //         : New: separate legend for each series
  //         : New: zooming and panning
  //v1.2.0.1 : Fixed: Issue with designtime initialization
  //v1.2.0.2 : Improved: Optimalization for large number of points
  //         : Fixed: Issue with designtime initialization in XE7
  //v1.2.1.0 : New : XE8 Support
  //v1.2.1.1 : Fixed : Issue with EndUpdate
  //v1.2.1.2 : Fixed : Issue with TList insert in XE8
  //v1.2.1.3 : Fixed : Issue with drawing digital line chart
  //v1.2.2.0 : New : Support for Delphi 10 Seattle Prof, Ent. Architect added
  //v1.3.0.0 : New : ctSpider chart type added
  //v1.4.0.0 : New : virtual mode
  //v1.4.0.1 : Fixed : Issue with retrieving correct number of annotations
  //v1.5.0.0 : New : ctBand support
  //         : New : AddParetoLine, AddLineSerie, AddTrendLine, AddTrendChannel, AddBandSerie, AddTrendChannelBand and AddMovingAverage support
  //         : Improved : Drawing of points when scrolling
  //v1.6.0.0 : New : RAD Studio 10.1 Berlin Support
  //v2.0.0.0 : New : OHLC, CandleStick and Boxplot support
  //         : New : Display centered Y-Axis and X-Axis at custom reference value
  //v2.0.1.0 : New : SaveToFile & LoadFromFile to persist chart settings
  //         : Fixed : Issue with displaying annotations in line types
  //v2.0.2.0 : New : OnSerieLegendItemClick and OnLegendItemClick events
  //         : Improved : Index property at TTMSFMXChartDrawXYValue and TTMSFMXChartDrawXYGridLine
  //v2.0.3.0 : New : DrawOrder property to change the order in which bars are drawn
  //v2.0.3.1 : Fixed : Issue with destroying series without clearing points in Android
  //v2.0.3.2 : Fixed : Issue with retrieving value with only one point
  //v2.0.3.3 : Fixed : Issue with persistence of settings

type
  TTMSFMXChart = class;

  TTMSFMXChartSerieType = (ctLine, ctDigitalLine, ctArea, ctStackedArea, ctStackedPercentageArea, ctBar, ctStackedBar, ctStackedPercentageBar,
    ctMarker, ctXYLine, ctXYMarker, ctPie, ctVariableRadiusPie, ctSizedPie, ctSpider, ctBand, ctOHLC, ctCandleStick, ctBoxPlot);

  TTMSFMXChartSelector = class(TControl)
  private
    FChartType: TTMSFMXChartSerieType;
    FSelected: Boolean;
    procedure SetChartType(const Value: TTMSFMXChartSerieType);
    procedure SetSelected(const Value: Boolean);
  protected
    procedure Paint; override;
    function GetText: String;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property ChartType: TTMSFMXChartSerieType read FChartType write SetChartType default ctLine;
    property Selected: Boolean read FSelected write SetSelected default False;
    property Cursor;
    property Position;
    property Width;
    property Height;
    property OnClick;
    property ClipChildren;
    {$IFDEF DELPHIXE7_LVL}
    property Size;
    {$ENDIF}
  end;

  TTMSFMXChartSerie = class;

  TTMSFMXChartPoint = class;

  TTMSFMXChartAnnotation = class;

  TTMSFMXChartAnnotationShape = (asRectangle, asCircle, asBalloon);

  TTMSFMXChartAnnotationArrow = (arLine, arStartArrow, arEndArrow, arDoubleArrow);

  TTMSFMXChartAnnotationBalloonDirection = (bdUpRight, bdUpLeft, bdDownRight, bdDownLeft);

  TTMSFMXChartAnnotationVirtual = record
    Index: Integer;
    Visible: Boolean;
    Text: String;
    AutoSize: Boolean;
    Annotation: TTMSFMXChartAnnotation;
    Width: Double;
    Height: Double;
    WordWrap: Boolean;
    Shape: TTMSFMXChartAnnotationShape;
    FontColor: TAlphaColor;
    LineColor: TAlphaColor;
    LineThickness: Double;
    ArrowOpacity: Double;
    LineOpacity: Double;
    CornerRadius: Double;
    Arrow: TTMSFMXChartAnnotationArrow;
    TextVerticalAlignment: TTextAlign;
    TextHorizontalAlignment: TTextAlign;
    OffsetX: Double;
    OffsetY: Double;
    BalloonArrowSize: Double;
    BalloonDirection: TTMSFMXChartAnnotationBalloonDirection;
    ArrowSize: Double;
    ArrowColor: TAlphaColor;
    StrokeOpacity: Double;
    FillOpacity: Double;
  end;

  TTMSFMXChartPointVirtual = record
    Index: Integer;
    Color: TAlphaColor;
    Tag: Integer;
    XValue: Double;
    XValueText: String;
    YValue, YValueSecond, YValueHigh, YValueClose, YValueLow, YValueOpen, YValueMedian: Double;
    YValueVariable: Double;
    LegendText: String;
    Explode: Double;
    Point: TTMSFMXChartPoint;
    SecondValue: Boolean;
  end;

  TTMSFMXChartDrawLine = record
    StackedValueStart: Double;
    StackedValueEnd: Double;
    StartPoint: TPointF;
    EndPoint: TPointF;
    StartReference: TTMSFMXChartPoint;
    EndReference: TTMSFMXChartPoint;
    StartVirtualReference: TTMSFMXChartPointVirtual;
    EndVirtualReference: TTMSFMXChartPointVirtual;
    Bottom: Boolean;
    DoAnimation: Boolean;
  end;

  TTMSFMXChartDrawMultiPoint = record
    PointHigh: TPointF;
    PointLow: TPointF;
    PointOpen: TPointF;
    PointMedian: TPointF;
    PointClose: TPointF;
    Rect: TRectF;
    Reference: TTMSFMXChartPoint;
    VirtualReference: TTMSFMXChartPointVirtual;
    DoAnimation: Boolean;    
  end;

  TTMSFMXChartDrawPoint = record
    Point: TPointF;
    Reference: TTMSFMXChartPoint;
    VirtualReference: TTMSFMXChartPointVirtual;
    DoAnimation: Boolean;
  end;

  TTMSFMXChartDrawRect = record
    Rect: TRectF;
    StackedValue: Double;
    Reference: TTMSFMXChartPoint;
    VirtualReference: TTMSFMXChartPointVirtual;
    DoAnimation: Boolean;
  end;

  TTMSFMXChartDrawSlice = record
    CenterPoint: TPointF;
    StartPoint: TPointF;
    EndPoint: TPointF;
    Radius: TPointF;
    InnerRadius: TPointF;
    StartAngle, SweepAngle: Double;
    Reference: TTMSFMXChartPoint;
    VirtualReference: TTMSFMXChartPointVirtual;
    DoAnimation: Boolean;
    AnimateEndPoint: Boolean;
  end;

  TTMSFMXChartAnnotation = class(TCollectionItem)
  private
    FTag: Integer;
    FDataString: String;
    FDataObject: TObject;
    FDataInteger: Integer;
    FPoint: TTMSFMXChartPoint;
    FAutoSize: Boolean;
    FLineThickness: Double;
    FTextVerticalAlignment: TTextAlign;
    FLineColor: TAlphaColor;
    FArrowColor: TAlphaColor;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FBalloonDirection: TTMSFMXChartAnnotationBalloonDirection;
    FWidth: Double;
    FStrokeCap: TStrokeCap;
    FShape: TTMSFMXChartAnnotationShape;
    FTextHorizontalAlignment: TTextAlign;
    FOffsetX: Double;
    FStrokeDash: TStrokeDash;
    FOffsetY: Double;
    FFont: TFont;
    FArrowSize: Double;
    FVisible: Boolean;
    FBalloonArrowSize: Double;
    FFillOpacity: Double;
    FFill: TBrush;
    FArrow: TTMSFMXChartAnnotationArrow;
    FText: string;
    FWordWrap: Boolean;
    FFontColor: TAlphaColor;
    FHeight: Double;
    FStrokeOpacity: Double;
    FStroke: TBrush;
    FCornerRadius: Double;
    FLineOpacity: Double;
    FArrowOpacity: Double;
    procedure SetArrow(const Value: TTMSFMXChartAnnotationArrow);
    procedure SetArrowColor(const Value: TAlphaColor);
    procedure SetArrowSize(const Value: Double);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetBalloonArrowSize(const Value: Double);
    procedure SetBalloonDirection(
      const Value: TTMSFMXChartAnnotationBalloonDirection);
    procedure SetFill(const Value: TBrush);
    procedure SetFillOpacity(const Value: Double);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
    procedure SetHeight(const Value: Double);
    procedure SetLineColor(const Value: TAlphaColor);
    procedure SetLineThickness(const Value: Double);
    procedure SetOffSetX(const Value: Double);
    procedure SetOffSetY(const Value: Double);
    procedure SetShape(const Value: TTMSFMXChartAnnotationShape);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetText(const Value: string);
    procedure SetTextHorizontalAlignment(const Value: TTextAlign);
    procedure SetTextVerticalAlignment(const Value: TTextAlign);
    procedure SetVisible(const Value: Boolean);
    procedure SetWidth(const Value: Double);
    procedure SetWordWrap(const Value: Boolean);
    procedure SetCornerRadius(const Value: Double);
    procedure SetLineOpacity(const Value: Double);
    procedure SetArrowOpacity(const Value: Double);
  protected
    procedure FontChanged(Sender: TObject);
    procedure Draw(ACanvas: TCanvas; ADrawPoint: TTMSFMXChartDrawPoint);
    procedure RepaintChart;
    procedure UpdateChart;
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    function Point: TTMSFMXChartPoint;
    function GetValue: TTMSFMXChartAnnotationVirtual;
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property DataObject: TObject read FDataObject write FDataObject;
    property DataString: String read FDataString write FDataString;
    property DataInteger: Integer read FDataInteger write FDataInteger;
  published
    property Tag: Integer read FTag write FTag default 0;
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property CornerRadius: Double read FCornerRadius write SetCornerRadius;
    property Visible: Boolean read FVisible write SetVisible;
    property Width: Double read FWidth write SetWidth;
    property Height: Double read FHeight write SetHeight;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property TextHorizontalAlignment: TTextAlign read FTextHorizontalAlignment write SetTextHorizontalAlignment default TTextAlign.Center;
    property TextVerticalAlignment: TTextAlign read FTextVerticalAlignment write SetTextVerticalAlignment default TTextAlign.Center;
    property Arrow: TTMSFMXChartAnnotationArrow read FArrow write SetArrow default arLine;
    property ArrowSize: Double read FArrowSize write SetArrowSize;
    property ArrowColor: TAlphaColor read FArrowColor write SetArrowColor default claBlack;
    property ArrowOpacity: Double read FArrowOpacity write SetArrowOpacity;
    property BalloonDirection: TTMSFMXChartAnnotationBalloonDirection read FBalloonDirection write SetBalloonDirection default bdDownLeft;
    property BalloonArrowSize: Double read FBalloonArrowSize write SetBalloonArrowSize;
    property LineThickness: Double read FLineThickness write SetLineThickness;
    property LineColor: TAlphaColor read FLineColor write SetLineColor default claBlack;
    property LineOpacity: Double read FLineOpacity write SetLineOpacity;
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property OffsetX: Double read FOffsetX write SetOffSetX;
    property OffsetY: Double read FOffsetY write SetOffSetY;
    property Shape: TTMSFMXChartAnnotationShape read FShape write SetShape default asRectangle;
    property Text: string read FText write SetText;
    property WordWrap: Boolean read FWordWrap write SetWordWrap default True;
  end;

  TTMSFMXChartAnnotations = class(TOwnedCollection)
  private
    FPoint: TTMSFMXChartPoint;
    function GetItem(Index: Integer): TTMSFMXChartAnnotation;
    procedure SetItem(Index: Integer; const Value: TTMSFMXChartAnnotation);
  public
    function Point: TTMSFMXChartPoint;
    constructor Create(APoint: TTMSFMXChartPoint);
    function Add: TTMSFMXChartAnnotation;
    function Insert(Index: Integer): TTMSFMXChartAnnotation;
    property Items[Index: Integer]: TTMSFMXChartAnnotation read GetItem write SetItem; default;
  end;

  TTMSFMXChartPoint = class(TCollectionItem)
  private
    FTag: Integer;
    FDataString: String;
    FDataObject: TObject;
    FDataInteger: Integer;
    FSerie: TTMSFMXChartSerie;
    FXValue: Double;
    FYValue: Double;
    FXValueText: String;
    FAnnotations: TTMSFMXChartAnnotations;
    FColor: TAlphaColor;
    FLegendText: String;
    FExplode: Double;
    FYValueVariable: Double;
    FYValueSecond: Double;
    FYValueLow: Double;
    FYValueOpen: Double;
    FYValueClose: Double;
    FYValueMedian: Double;
    procedure SetXValue(const Value: Double);
    procedure SetYValue(const Value: Double);
    procedure SetXValueText(const Value: String);
    procedure SetAnnotations(const Value: TTMSFMXChartAnnotations);
    procedure SetColor(const Value: TAlphaColor);
    procedure SetLegendText(const Value: String);
    procedure SetExplode(const Value: Double);
    procedure SetYValueVariable(const Value: Double);
    procedure SetYValueSecond(const Value: Double);
    procedure SetYValueClose(const Value: Double);
    procedure SetYValueLow(const Value: Double);
    procedure SetYValueOpen(const Value: Double);
    procedure SetYValueMedian(const Value: Double);
  protected
    procedure BitmapChanged(Sender: TObject);
    procedure UpdateChart;
    procedure UpdateSequential;
    procedure RepaintChart;
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    function GetValue: TTMSFMXChartPointVirtual;
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property DataObject: TObject read FDataObject write FDataObject;
    property DataString: String read FDataString write FDataString;
    property DataInteger: Integer read FDataInteger write FDataInteger;
  published
    property Annotations: TTMSFMXChartAnnotations read FAnnotations write SetAnnotations;
    property Color: TAlphaColor read FColor write SetColor default claNull;
    property Tag: Integer read FTag write FTag default 0;
    property XValue: Double read FXValue write SetXValue;
    property XValueText: String read FXValueText write SetXValueText;
    property YValue: Double read FYValue write SetYValue;
    property YValueSecond: Double read FYValueSecond write SetYValueSecond;
    property YValueVariable: Double read FYValueVariable write SetYValueVariable;
    property YValueLow: Double read FYValueLow write SetYValueLow;
    property YValueClose: Double read FYValueClose write SetYValueClose;
    property YValueHigh: Double read FYValue write SetYValue;
    property YValueOpen: Double read FYValueOpen write SetYValueOpen;
    property YValueMedian: Double read FYValueMedian write SetYValueMedian;
    property LegendText: String read FLegendText write SetLegendText;
    property Explode: Double read FExplode write SetExplode;
  end;

  TTMSFMXChartPoints = class(TOwnedCollection)
  private
    FSerie: TTMSFMXChartSerie;
    function GetItem(Index: Integer): TTMSFMXChartPoint;
    procedure SetItem(Index: Integer; const Value: TTMSFMXChartPoint);
  public
    function Serie: TTMSFMXChartSerie;
    constructor Create(ASerie: TTMSFMXChartSerie);
    function Add: TTMSFMXChartPoint;
    function Insert(Index: Integer): TTMSFMXChartPoint;
    property Items[Index: Integer]: TTMSFMXChartPoint read GetItem write SetItem; default;
  end;

  TTMSFMXChartDrawXYValueKind = (vkMajor, vkMinor);

  TTMSFMXChartDrawXYValue = record
    Index: Integer;
    ValueString: String;
    Value: Double;
    Kind: TTMSFMXChartDrawXYValueKind;
    TextWidth: Double;
    TextHeight: Double;
    ValuePositionYLeft, ValuePositionYCenter, ValuePositionYRight: TPointF;
    ValuePositionXTop, ValuePositionXCenter, ValuePositionXBottom: TPointF;
    TextPositionYLeft, TextPositionYCenter, TextPositionYRight: TPointF;
    TextPositionXTop, TextPositionXCenter, TextPositionXBottom: TPointF;
    ValueRadius: TPointF;
  end;

  TTMSFMXChartDrawXYGridLine = record
    Index: Integer;
    Value: Double;
    Kind: TTMSFMXChartDrawXYValueKind;
    XStartPoint, XEndPoint: TPointF;
    YStartPoint, YEndPoint: TPointF;
    ValueRadius: TPointF;
    ValueGridLine: Boolean;
  end;

  TTMSFMXChartDrawLabel = record
    ValueString: String;
    Value: Double;
    Reference: TTMSFMXChartPoint;
    VirtualReference: TTMSFMXChartPointVirtual;
    Point: TPointF;
    Rect: TRectF;
    TextWidth, TextHeight: Double;
  end;

  TTMSFMXChartDrawAnnotation = record
    AnnotationReference: TTMSFMXChartAnnotation;
    VirtualAnnotationReference: TTMSFMXChartAnnotationVirtual;
    PointReference: TTMSFMXChartPoint;
    VirtualPointReference: TTMSFMXChartPointVirtual;
    StartPoint, AnnotationPoint: TPointF;
    Rect: TRectF;
    TextWidth, TextHeight: Double;
  end;

  TTMSFMXChartSerieAutoRange = (arDisabled, arEnabled, arEnabledZeroBased, arCommon, arCommonZeroBased);

  TTMSFMXChartSerieMode = (smMathematical, smStatistical);

  TTMSFMXChartSerieBarWidthType = (bwtPercentage, bwtPixels);

  TTMSFMXChartSerieBar = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FWidth: Double;
    FSpacing: Double;
    FWidthType: TTMSFMXChartSerieBarWidthType;
    procedure SetSpacing(const Value: Double);
    procedure SetWidth(const Value: Double);
    procedure SetWidthType(const Value: TTMSFMXChartSerieBarWidthType);
  protected
    procedure UpdateChart;
    procedure RepaintChart;
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    constructor Create(ASerie: TTMSFMXChartSerie);
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property WidthType: TTMSFMXChartSerieBarWidthType read FWidthType write SetWidthType default bwtPercentage;
    property Width: Double read FWidth write SetWidth;
    property Spacing: Double read FSpacing write SetSpacing;
  end;

  TTMSFMXChartSerieMultiPoint = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FDecreaseStrokeColor: TAlphaColor;
    FIncreaseStrokeColor: TAlphaColor;
    FWidth: Double;
    FWidthType: TTMSFMXChartSerieBarWidthType;
    FIncreaseFillColor: TAlphaColor;
    FDecreaseFillColor: TAlphaColor;
    procedure SetDecreaseStrokeColor(const Value: TAlphaColor);
    procedure SetIncreaseStrokeColor(const Value: TAlphaColor);
    procedure SetWidth(const Value: Double);
    procedure SetWidthType(const Value: TTMSFMXChartSerieBarWidthType);
    procedure SetDecreaseFillColor(const Value: TAlphaColor);
    procedure SetIncreaseFillColor(const Value: TAlphaColor);
  protected
    procedure UpdateChart;
    procedure RepaintChart;
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    constructor Create(ASerie: TTMSFMXChartSerie);
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property WidthType: TTMSFMXChartSerieBarWidthType read FWidthType write SetWidthType default bwtPercentage;
    property Width: Double read FWidth write SetWidth;
    property IncreaseStrokeColor: TAlphaColor read FIncreaseStrokeColor write SetIncreaseStrokeColor default claDarkgreen;
    property DecreaseStrokeColor: TAlphaColor read FDecreaseStrokeColor write SetDecreaseStrokeColor default claDarkRed;
    property IncreaseFillColor: TAlphaColor read FIncreaseFillColor write SetIncreaseFillColor default claLightSeaGreen;
    property DecreaseFillColor: TAlphaColor read FDecreaseFillColor write SetDecreaseFillColor default claOrange;
  end;

  TTMSFMXChartSeriePiePosition = (ppTopLeft, ppTopCenter, ppTopRight, ppCenterLeft,
    ppCenterCenter, ppCenterRight, ppBottomLeft, ppBottomCenter, ppBottomRight);

  TTMSFMXChartSeriePie = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FPosition: TTMSFMXChartSeriePiePosition;
    FAutoSize: Boolean;
    FMargins: TBounds;
    FSize: Double;
    FStartAngle: Double;
    FSweepAngle: Double;
    FInnerSize: Double;
    FStacked: Boolean;
    procedure SetPosition(const Value: TTMSFMXChartSeriePiePosition);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetMargins(const Value: TBounds);
    procedure SetSize(const Value: Double);
    procedure SetStartAngle(const Value: Double);
    procedure SetSweepAngle(const Value: Double);
    procedure SetInnerSize(const Value: Double);
    procedure SetStacked(const Value: Boolean);
  protected
    procedure UpdateChart;
    procedure RepaintChart;
    procedure MarginsChanged(Sender: TObject);
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    constructor Create(ASerie: TTMSFMXChartSerie);
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property Position: TTMSFMXChartSeriePiePosition read FPosition write SetPosition default ppCenterCenter;
    property Size: Double read FSize write SetSize;
    property InnerSize: Double read FInnerSize write SetInnerSize;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property Margins: TBounds read FMargins write SetMargins;
    property StartAngle: Double read FStartAngle write SetStartAngle;
    property SweepAngle: Double read FSweepAngle write SetSweepAngle;
    property Stacked: Boolean read FStacked write SetStacked default False;
  end;

  TTMSFMXChartSerieMarkerShape = (msEllipse, msSquare, msDiamond, msTriangle, msBitmap, msCustom);

  TTMSFMXChartSerieMarkers = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FWidth: Double;
    FShape: TTMSFMXChartSerieMarkerShape;
    FHeight: Double;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFillOpacity: Double;
    FFill: TBrush;
    FStrokeOpacity: Double;
    FStroke: TBrush;
    FBitmap: TBitmap;
    FVisible: Boolean;
    procedure SetHeight(const Value: Double);
    procedure SetShape(const Value: TTMSFMXChartSerieMarkerShape);
    procedure SetWidth(const Value: Double);
    procedure SetFill(const Value: TBrush);
    procedure SetFillOpacity(const Value: Double);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetBitmap(const Value: TBitmap);
    procedure SetVisible(const Value: Boolean);
  protected
    procedure StrokeChanged(Sender: TObject);
    procedure FillChanged(Sender: TObject);
    procedure UpdateChart;
    procedure RepaintChart;
  public
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
    constructor Create(ASerie: TTMSFMXChartSerie);
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property Bitmap: TBitmap read FBitmap write SetBitmap;
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Height: Double read FHeight write SetHeight;
    property Shape: TTMSFMXChartSerieMarkerShape read FShape write SetShape default msEllipse;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property Visible: Boolean read FVisible write SetVisible default False;
    property Width: Double read FWidth write SetWidth;
  end;

  TTMSFMXChartXAxisPosition = (xpTop, xpCenter, xpBottom);
  TTMSFMXChartXAxisPositions = set of TTMSFMXChartXAxisPosition;

  TTMSFMXChartYAxisPosition = (ypLeft, ypCenter, ypRight);
  TTMSFMXChartYAxisPositions = set of TTMSFMXChartYAxisPosition;

  TTMSFMXChartDrawSlices = TList<TTMSFMXChartDrawSlice>;
  TTMSFMXChartDrawLines = TList<TTMSFMXChartDrawLine>;
  TTMSFMXChartDrawMultiPoints = TList<TTMSFMXChartDrawMultiPoint>;
  TTMSFMXChartDrawRects = TList<TTMSFMXChartDrawRect>;
  TTMSFMXChartDrawPoints = TList<TTMSFMXChartDrawPoint>;
  TTMSFMXChartDrawXYValues = TList<TTMSFMXChartDrawXYValue>;
  TTMSFMXChartDrawXYGridLines = TList<TTMSFMXChartDrawXYGridLine>;
  TTMSFMXChartDrawLabels = TList<TTMSFMXChartDrawLabel>;
  TTMSFMXChartDrawAnnotations = TList<TTMSFMXChartDrawAnnotation>;

  TTMSFMXChartElement = class(TPersistent)
  private
    FChart: TTMSFMXChart;
    FVisible: Boolean;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFill: TBrush;
    FStroke: TBrush;
    FFillOpacity: Double;
    FStrokeOpacity: Double;
    procedure SetVisible(const Value: Boolean);
    procedure SetFill(const Value: TBrush);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetFillOpacity(const Value: Double);
    procedure SetStrokeOpacity(const Value: Double);
  protected
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
    procedure Draw(ACanvas: TCanvas); virtual;
    procedure UpdateChart;
    procedure RepaintChart;
  public
    function Chart: TTMSFMXChart;
    procedure Assign(Source: TPersistent); override;
    constructor Create(AChart: TTMSFMXChart); virtual;
    destructor Destroy; override;
  published
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property Visible: Boolean read FVisible write SetVisible default True;
  end;

  TTMSFMXChartTitlePosition = (tipTop, tipBottom);
  TTMSFMXChartTitlePositions = set of TTMSFMXChartTitlePosition;

  TTMSFMXChartTitle = class(TTMSFMXChartElement)
  private
    FTextMargins: TBounds;
    FText: String;
    FPositions: TTMSFMXChartTitlePositions;
    FHeight: Double;
    FTextVerticalAlignment: TTextAlign;
    FTextHorizontalAlignment: TTextAlign;
    FFont: TFont;
    FFontColor: TAlphaColor;
    FBorder: Boolean;
    FLine: Boolean;
    procedure SetPositions(const Value: TTMSFMXChartTitlePositions);
    procedure SetText(const Value: String);
    procedure SetTextMargins(const Value: TBounds);
    procedure SetHeight(const Value: Double);
    procedure SetTextHorizontalAlignment(const Value: TTextAlign);
    procedure SetTextVerticalAlignment(const Value: TTextAlign);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
    procedure SetBorder(const Value: Boolean);
    procedure SetLine(const Value: Boolean);
  protected
    procedure Draw(ACanvas: TCanvas); override;
    procedure DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartTitlePosition);
    procedure TextMarginsChanged(Sender: TObject);
    procedure FontChanged(Sender: TObject);
  public
    function GetRectangle(APosition: TTMSFMXChartTitlePosition): TRectF;
    procedure Assign(Source: TPersistent); override;
    constructor Create(AChart: TTMSFMXChart); override;
    destructor Destroy; override;
  published
    property Border: Boolean read FBorder write SetBorder default False;
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property Height: Double read FHeight write SetHeight;
    property Line: Boolean read FLine write SetLine default True;
    property Positions: TTMSFMXChartTitlePositions read FPositions write SetPositions default [tipTop];
    property TextHorizontalAlignment: TTextAlign read FTextHorizontalAlignment write SetTextHorizontalAlignment default TTextAlign.Center;
    property TextVerticalAlignment: TTextAlign read FTextVerticalAlignment write SetTextVerticalAlignment default TTextAlign.Center;
    property TextMargins: TBounds read FTextMargins write SetTextMargins;
    property Text: String read FText write SetText;
  end;

  TTMSFMXChartXAxis = class(TTMSFMXChartElement)
  private
    FHeight: Double;
    FTopHeight, FCenterHeight, FBottomHeight: Double;
    FPositions: TTMSFMXChartXAxisPositions;
    FLine: Boolean;
    FBorder: Boolean;
    FAutoSize: Boolean;
    FDisplayAtReferenceValue: Boolean;
    FReferenceValueSeriesIndex: Integer;
    FReferenceValue: Double;
    procedure SetHeight(const Value: Double);
    procedure SetPositions(const Value: TTMSFMXChartXAxisPositions);
    procedure SetLine(const Value: Boolean);
    procedure SetBorder(const Value: Boolean);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetDisplayAtReferenceValue(const Value: Boolean);
    procedure SetReferenceValueSeriesIndex(const Value: Integer);
    procedure SetReferenceValue(const Value: Double);
    function IsReferenceValueStored: Boolean;
  protected
    procedure Draw(ACanvas: TCanvas); override;
    procedure DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition);
    procedure DrawValues(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie);
    procedure DrawKind(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition; AKind: TTMSFMXChartDrawXYValueKind; ASerie: TTMSFMXChartSerie);
    procedure DrawTitle(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie);
  public
    function GetHeight(APosition: TTMSFMXChartXAxisPosition): Double;
    function GetRectangle(APosition: TTMSFMXChartXAxisPosition): TRectF;
    function Chart: TTMSFMXChart;
    constructor Create(AChart: TTMSFMXChart); override;
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property DisplayAtReferenceValue: Boolean read FDisplayAtReferenceValue write SetDisplayAtReferenceValue default False;
    property ReferenceValueSeriesIndex: Integer read FReferenceValueSeriesIndex write SetReferenceValueSeriesIndex default 0;
    property ReferenceValue: Double read FReferenceValue write SetReferenceValue stored IsReferenceValueStored nodefault;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property Border: Boolean read FBorder write SetBorder default False;
    property Height: Double read FHeight write SetHeight;
    property Line: Boolean read FLine write SetLine default True;
    property Positions: TTMSFMXChartXAxisPositions read FPositions write SetPositions default [xpBottom];
  end;

  TTMSFMXChartYAxis = class(TTMSFMXChartElement)
  private
    FWidth, FLeftWidth, FCenterWidth, FRightWidth: Double;
    FPositions: TTMSFMXChartYAxisPositions;
    FLine: Boolean;
    FBorder: Boolean;
    FAutoSize: Boolean;
    FDisplayAtReferenceValue: Boolean;
    FReferenceValueSeriesIndex: Integer;
    FReferenceValue: Double;
    procedure SetWidth(const Value: Double);
    procedure SetPositions(const Value: TTMSFMXChartYAxisPositions);
    procedure SetLine(const Value: Boolean);
    procedure SetBorder(const Value: Boolean);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetDisplayAtReferenceValue(const Value: Boolean);
    procedure SetReferenceValueSeriesIndex(const Value: Integer);
    procedure SetReferenceValue(const Value: Double);
    function IsReferenceValueStored: Boolean;
  protected
    procedure Draw(ACanvas: TCanvas); override;
    procedure DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartYAxisPosition);
    procedure DrawValues(ACanvas: TCanvas; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie);
    procedure DrawKind(ACanvas: TCanvas; APosition: TTMSFMXChartYAxisPosition; AKind: TTMSFMXChartDrawXYValueKind; ASerie: TTMSFMXChartSerie);
    procedure DrawTitle(ACanvas: TCanvas; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie);
  public
    function GetWidth(APosition: TTMSFMXChartYAxisPosition): Double;
    function GetRectangle(APosition: TTMSFMXChartYAxisPosition): TRectF;
    constructor Create(AChart: TTMSFMXChart); override;
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property DisplayAtReferenceValue: Boolean read FDisplayAtReferenceValue write SetDisplayAtReferenceValue default False;
    property ReferenceValueSeriesIndex: Integer read FReferenceValueSeriesIndex write SetReferenceValueSeriesIndex default 0;
    property ReferenceValue: Double read FReferenceValue write SetReferenceValue stored IsReferenceValueStored nodefault;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property Border: Boolean read FBorder write SetBorder default False;
    property Line: Boolean read FLine write SetLine default True;
    property Positions: TTMSFMXChartYAxisPositions read FPositions write SetPositions default [ypLeft, ypRight];
    property Width: Double read FWidth write SetWidth;
  end;

  TTMSFMXChartLegendPosition = (lpTopLeft, lpTopCenter, lpTopRight, lpCenterLeft, lpCenterCenter,
    lpCenterRight, lpBottomLeft, lpBottomCenter, lpBottomRight);

  TTMSFMXChartLegend = class(TTMSFMXChartElement)
  private
    FTop: Double;
    FLeft: Double;
    FPosition: TTMSFMXChartLegendPosition;
    FFont: TFont;
    FFontColor: TAlphaColor;
    procedure SetLeft(const Value: Double);
    procedure SetPosition(const Value: TTMSFMXChartLegendPosition);
    procedure SetTop(const Value: Double);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
  protected
    procedure Draw(ACanvas: TCanvas); override;
    procedure FontChanged(Sender: TObject);
  public
    function GetRectangle: TRectF; virtual;
    constructor Create(AChart: TTMSFMXChart); override;
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
    function XYToItem(X, Y: Double): Integer; virtual;
  published
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property Left: Double read FLeft write SetLeft;
    property Position: TTMSFMXChartLegendPosition read FPosition write SetPosition default lpTopLeft;
    property Top: Double read FTop write SetTop;
  end;

  TTMSFMXChartSerieLegend = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FVisible: Boolean;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFillOpacity: Double;
    FFill: TBrush;
    FStrokeOpacity: Double;
    FStroke: TBrush;
    FFont: TFont;
    FFontColor: TAlphaColor;
    FTop: Double;
    FLeft: Double;
    FPosition: TTMSFMXChartLegendPosition;
    procedure SetVisible(const Value: Boolean);
    procedure SetFill(const Value: TBrush);
    procedure SetFillOpacity(const Value: Double);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
    procedure SetLeft(const Value: Double);
    procedure SetPosition(const Value: TTMSFMXChartLegendPosition);
    procedure SetTop(const Value: Double);
  protected
    function GetRectangle: TRectF; virtual;
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
    procedure RepaintChart;
    procedure UpdateChart;
    procedure FontChanged(Sender: TObject);
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(ASerie: TTMSFMXChartSerie);
    destructor Destroy; override;
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
  published
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property Left: Double read FLeft write SetLeft;
    property Position: TTMSFMXChartLegendPosition read FPosition write SetPosition default lpCenterRight;
    property Top: Double read FTop write SetTop;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property Visible: Boolean read FVisible write SetVisible default False;
  end;

  TTMSFMXChartSerieValueFormatType = (vftNormal, vftFloat, vftDateTime);

  TTMSFMXChartSerieXYValues = class;

  TTMSFMXChartSerieXYTitle = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FTextMargins: TBounds;
    FTextHorizontalAlignment: TTextAlign;
    FTextVerticalAlignment: TTextAlign;
    FText: String;
    FFont: TFont;
    FFontColor: TAlphaColor;
    procedure SetText(const Value: String);
    procedure SetTextMargins(const Value: TBounds);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
    procedure SetTextHorizontalAlignment(const Value: TTextAlign);
    procedure SetTextVerticalAlignment(const Value: TTextAlign);
  protected
    procedure UpdateChart;
    procedure RepaintChart;
    procedure TextMarginsChanged(Sender: TObject);
    procedure FontChanged(Sender: TObject);
  public
    function Serie: TTMSFMXChartSerie;
    function Chart: TTMSFMXChart;
    procedure Assign(Source: TPersistent); override;
    constructor Create(ASerie: TTMSFMXChartSerie); virtual;
    destructor Destroy; override;
  published
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property TextHorizontalAlignment: TTextAlign read FTextHorizontalAlignment write SetTextHorizontalAlignment default TTextAlign.Center;
    property TextVerticalAlignment: TTextAlign read FTextVerticalAlignment write SetTextVerticalAlignment default TTextAlign.Center;
    property TextMargins: TBounds read FTextMargins write SetTextMargins;
    property Text: String read FText write SetText;
  end;

  TTMSFMXChartSerieXYValues = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FMinorUnitFontColor: TAlphaColor;
    FMajorUnitFontColor: TAlphaColor;
    FMinorUnit: Double;
    FMajorUnit: Double;
    FMajorUnitTickMarkSize: Double;
    FMinorUnitFormat: String;
    FMinorUnitSpacing: Double;
    FMajorUnitFormat: String;
    FMajorUnitSpacing: Double;
    FMinorUnitFont: TFont;
    FMajorUnitFont: TFont;
    FMinorUnitTickMarkSize: Double;
    FMinorUnitFormatType: TTMSFMXChartSerieValueFormatType;
    FMajorUnitFormatType: TTMSFMXChartSerieValueFormatType;
    FMinorUnitTickMarkColor: TAlphaColor;
    FMajorUnitTickMarkColor: TAlphaColor;
    FAutoUnits: Boolean;
    FTitle: TTMSFMXChartSerieXYTitle;
    FSpiderAngle: Single;
    FSpiderValues: Boolean;
    procedure SetMajorUnitFontColor(const Value: TAlphaColor);
    procedure SetMajorUnit(const Value: Double);
    procedure SetMajorUnitFormat(const Value: String);
    procedure SetMajorUnitSpacing(const Value: Double);
    procedure SetMajorUnitTickMarkSize(const Value: Double);
    procedure SetMinorUnitFontColor(const Value: TAlphaColor);
    procedure SetMinorUnit(const Value: Double);
    procedure SetMinorUnitFormat(const Value: String);
    procedure SetMinorUnitSpacing(const Value: Double);
    procedure SetMajorUnitFont(const Value: TFont);
    procedure SetMinorUnitFont(const Value: TFont);
    procedure SetMinorUnitTickMarkSize(const Value: Double);
    procedure SetMajorUnitFormatType(
      const Value: TTMSFMXChartSerieValueFormatType);
    procedure SetMinorUnitFormatType(
      const Value: TTMSFMXChartSerieValueFormatType);
    procedure SetMajorUnitTickMarkColor(const Value: TAlphaColor);
    procedure SetMinorUnitTickMarkColor(const Value: TAlphaColor);
    procedure SetAutoUnits(const Value: Boolean);
    procedure SetTitle(const Value: TTMSFMXChartSerieXYTitle);
    procedure SetSpiderAngle(const Value: Single);
    procedure SetSpiderValues(const Value: Boolean);
  protected
    procedure UpdateChart;
    procedure RepaintChart;
    procedure FontChanged(Sender: TObject);
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(ASerie: TTMSFMXChartSerie); virtual;
    destructor Destroy; override;
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
  published
    property AutoUnits: Boolean read FAutoUnits write SetAutoUnits default True;
    property MajorUnit: Double read FMajorUnit write SetMajorUnit;
    property MajorUnitFont: TFont read FMajorUnitFont write SetMajorUnitFont;
    property MajorUnitFontColor: TAlphaColor read FMajorUnitFontColor write SetMajorUnitFontColor default claBlack;
    property MajorUnitFormat: String read FMajorUnitFormat write SetMajorUnitFormat;
    property MajorUnitFormatType: TTMSFMXChartSerieValueFormatType read FMajorUnitFormatType write SetMajorUnitFormatType default vftNormal;
    property MajorUnitSpacing: Double read FMajorUnitSpacing write SetMajorUnitSpacing;
    property MajorUnitTickMarkSize: Double read FMajorUnitTickMarkSize write SetMajorUnitTickMarkSize;
    property MajorUnitTickMarkColor: TAlphaColor read FMajorUnitTickMarkColor write SetMajorUnitTickMarkColor default claGray;
    property MinorUnit: Double read FMinorUnit write SetMinorUnit;
    property MinorUnitFont: TFont read FMinorUnitFont write SetMinorUnitFont;
    property MinorUnitFontColor: TAlphaColor read FMinorUnitFontColor write SetMinorUnitFontColor default claBlack;
    property MinorUnitFormat: String read FMinorUnitFormat write SetMinorUnitFormat;
    property MinorUnitFormatType: TTMSFMXChartSerieValueFormatType read FMinorUnitFormatType write SetMinorUnitFormatType default vftNormal;    
    property MinorUnitSpacing: Double read FMinorUnitSpacing write SetMinorUnitSpacing;
    property MinorUnitTickMarkSize: Double read FMinorUnitTickMarkSize write SetMinorUnitTickMarkSize;
    property MinorUnitTickMarkColor: TAlphaColor read FMinorUnitTickMarkColor write SetMinorUnitTickMarkColor default claGray;
    property Title: TTMSFMXChartSerieXYTitle read FTitle write SetTitle;
    property SpiderAngle: Single read FSpiderAngle write SetSpiderAngle;
    property SpiderValues: Boolean read FSpiderValues write SetSpiderValues default True;
  end;

  TTMSFMXChartSerieXValues = class(TTMSFMXChartSerieXYValues)
  private
    FPositions: TTMSFMXChartXAxisPositions;
    FAngle: Double;
    procedure SetPositions(const Value: TTMSFMXChartXAxisPositions);
    procedure SetAngle(const Value: Double);
  public
    constructor Create(ASerie: TTMSFMXChartSerie); override;
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property Angle: Double read FAngle write SetAngle;
    property AutoUnits default False;
    property Positions: TTMSFMXChartXAxisPositions read FPositions write SetPositions default [xpBottom];
  end;

  TTMSFMXChartSerieYValues = class(TTMSFMXChartSerieXYValues)
  private
    FPositions: TTMSFMXChartYAxisPositions;
    procedure SetPositions(const Value: TTMSFMXChartYAxisPositions);
  public
    constructor Create(ASerie: TTMSFMXChartSerie); override;
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
  published
    property Positions: TTMSFMXChartYAxisPositions read FPositions write SetPositions default [ypLeft];
  end;

  TTMSFMXChartSerieSpiderGridKind = (sgkCircles, sgkPolygon);

  TTMSFMXChartSerieXYGrid = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FMajorUnitStrokeOpacity: Double;
    FMajorUnitStroke: TBrush;
    FMajorUnitStrokeJoin: TStrokeJoin;
    FMajorUnitStrokeThickness: Double;
    FMajorUnitStrokeCap: TStrokeCap;
    FMajorUnitStrokeDash: TStrokeDash;
    FMinorUnitStrokeOpacity: Double;
    FMinorUnitStroke: TBrush;
    FMinorUnitStrokeJoin: TStrokeJoin;
    FMinorUnitStrokeThickness: Double;
    FMinorUnitStrokeCap: TStrokeCap;
    FMinorUnitStrokeDash: TStrokeDash;
    FVisible: Boolean;
    FExtended: Boolean;
    FSpiderKind: TTMSFMXChartSerieSpiderGridKind;
    FSpiderVisible: Boolean;
    FSpiderLegend: Boolean;
    procedure SetMajorUnitStroke(const Value: TBrush);
    procedure SetMajorUnitStrokeCap(const Value: TStrokeCap);
    procedure SetMajorUnitStrokeDash(const Value: TStrokeDash);
    procedure SetMajorUnitStrokeJoin(const Value: TStrokeJoin);
    procedure SetMajorUnitStrokeOpacity(const Value: Double);
    procedure SetMajorUnitStrokeThickness(const Value: Double);
    procedure SetMinorUnitStroke(const Value: TBrush);
    procedure SetMinorUnitStrokeCap(const Value: TStrokeCap);
    procedure SetMinorUnitStrokeDash(const Value: TStrokeDash);
    procedure SetMinorUnitStrokeJoin(const Value: TStrokeJoin);
    procedure SetMinorUnitStrokeOpacity(const Value: Double);
    procedure SetMinorUnitStrokeThickness(const Value: Double);
    procedure SetVisible(const Value: Boolean);
    procedure SetExtended(const Value: Boolean);
    procedure SetSpiderKind(const Value: TTMSFMXChartSerieSpiderGridKind);
    procedure SetSpiderVisible(const Value: Boolean);
    procedure SetSpiderLegend(const Value: Boolean);
  protected
    procedure StrokeChanged(Sender: TObject);
    procedure UpdateChart;
    procedure RepaintChart;
    procedure Draw(ACanvas: TCanvas; AXGrid: Boolean = True);
    procedure DrawKind(ACanvas: TCanvas; AKind: TTMSFMXChartDrawXYValueKind; AXGrid: Boolean = True);
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(ASerie: TTMSFMXChartSerie); virtual;
    destructor Destroy; override;
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
  published
    property Extended: Boolean read FExtended write SetExtended default True;
    property MajorUnitStroke: TBrush read FMajorUnitStroke write SetMajorUnitStroke;
    property MajorUnitStrokeOpacity: Double read FMajorUnitStrokeOpacity write SetMajorUnitStrokeOpacity;
    property MajorUnitStrokeThickness: Double read FMajorUnitStrokeThickness write SetMajorUnitStrokeThickness;
    property MajorUnitStrokeCap: TStrokeCap read FMajorUnitStrokeCap write SetMajorUnitStrokeCap default TStrokeCap.Flat;
    property MajorUnitStrokeDash: TStrokeDash read FMajorUnitStrokeDash write SetMajorUnitStrokeDash default TStrokeDash.Solid;
    property MajorUnitStrokeJoin: TStrokeJoin read FMajorUnitStrokeJoin write SetMajorUnitStrokeJoin default TStrokeJoin.Round;
    property MinorUnitStroke: TBrush read FMinorUnitStroke write SetMinorUnitStroke;
    property MinorUnitStrokeOpacity: Double read FMinorUnitStrokeOpacity write SetMinorUnitStrokeOpacity;
    property MinorUnitStrokeThickness: Double read FMinorUnitStrokeThickness write SetMinorUnitStrokeThickness;
    property MinorUnitStrokeCap: TStrokeCap read FMinorUnitStrokeCap write SetMinorUnitStrokeCap default TStrokeCap.Flat;
    property MinorUnitStrokeDash: TStrokeDash read FMinorUnitStrokeDash write SetMinorUnitStrokeDash default TStrokeDash.Solid;
    property MinorUnitStrokeJoin: TStrokeJoin read FMinorUnitStrokeJoin write SetMinorUnitStrokeJoin default TStrokeJoin.Round;
    property Visible: Boolean read FVisible write SetVisible default False;
    property SpiderVisible: Boolean read FSpiderVisible write SetSpiderVisible default True;
    property SpiderKind: TTMSFMXChartSerieSpiderGridKind read FSpiderKind write SetSpiderKind default sgkCircles;
    property SpiderLegend: Boolean read FSpiderLegend write SetSpiderLegend default False;
  end;

  TTMSFMXChartSerieLabelsPosition = (lpInside, lpOutside);

  TTMSFMXChartSerieLabelsMode = (lmNormal, lmStacked);

  TTMSFMXChartSerieLabels = class(TPersistent)
  private
    FSerie: TTMSFMXChartSerie;
    FVisible: Boolean;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFillOpacity: Double;
    FFill: TBrush;
    FStrokeOpacity: Double;
    FStroke: TBrush;
    FFormatType: TTMSFMXChartSerieValueFormatType;
    FFont: TFont;
    FFormat: String;
    FFontColor: TAlphaColor;
    FOffsetX: Double;
    FOffsetY: Double;
    FMode: TTMSFMXChartSerieLabelsMode;
    procedure SetVisible(const Value: Boolean);
    procedure SetFill(const Value: TBrush);
    procedure SetFillOpacity(const Value: Double);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetFont(const Value: TFont);
    procedure SetFontColor(const Value: TAlphaColor);
    procedure SetFormat(const Value: String);
    procedure SetFormatType(const Value: TTMSFMXChartSerieValueFormatType);
    procedure SetOffsetX(const Value: Double);
    procedure SetOffsetY(const Value: Double);
    procedure SetMode(const Value: TTMSFMXChartSerieLabelsMode);
  protected
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
    procedure RepaintChart;
    procedure UpdateChart;
    procedure FontChanged(Sender: TObject);
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(ASerie: TTMSFMXChartSerie);
    destructor Destroy; override;
    function Chart: TTMSFMXChart;
    function Serie: TTMSFMXChartSerie;
  published
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Font: TFont read FFont write SetFont;
    property FontColor: TAlphaColor read FFontColor write SetFontColor default claBlack;
    property Format: String read FFormat write SetFormat;
    property FormatType: TTMSFMXChartSerieValueFormatType read FFormatType write SetFormatType default vftNormal;
    property OffsetX: Double read FOffsetX write SetOffsetX;
    property OffsetY: Double read FOffsetY write SetOffsetY;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property Visible: Boolean read FVisible write SetVisible default False;
    property Mode: TTMSFMXChartSerieLabelsMode read FMode write SetMode default lmNormal;
  end;

  TTMSFMXChartSerie = class(TCollectionItem)
  private
    FSequential: Boolean;
    FPrevXValue: Double;
    FXAxisTopHeight, FXAxisCenterHeight, FXAxisBottomHeight: Double;
    FYAxisLeftWidth, FYAxisCenterWidth, FYAxisRightWidth: Double;
    FAnimateTimer: TTimer;
    FTag: Integer;
    FDataString: String;
    FDataObject: TObject;
    FDataInteger: Integer;
    FPoints: TTMSFMXChartPoints;
    FDrawSlices, FAnimateDrawSlices, FAnimateDrawSlicesReverse: TTMSFMXChartDrawSlices;
    FDrawLines, FAnimateDrawLines, FAnimateDrawLinesReverse: TTMSFMXChartDrawLines;
    FDrawRects, FAnimateDrawRects, FAnimateDrawRectsReverse: TTMSFMXChartDrawRects;
    FDrawPoints, FAnimateDrawPoints, FAnimateDrawPointsReverse: TTMSFMXChartDrawPoints;
    FDrawMultiPoints, FAnimateDrawMultiPoints, FAnimateDrawMultiPointsReverse: TTMSFMXChartDrawMultiPoints;
    FDrawXGridLines: TTMSFMXChartDrawXYGridLines;
    FDrawYGridLines: TTMSFMXChartDrawXYGridLines;
    FDrawXValues: TTMSFMXChartDrawXYValues;
    FDrawYValues: TTMSFMXChartDrawXYValues;
    FDrawLabels: TTMSFMXChartDrawLabels;
    FDrawAnnotations: TTMSFMXChartDrawAnnotations;
    FDrawPath: TPathData;
    FVisible: Boolean;
    FChart: TTMSFMXChart;
    FStartX, FEndX: Integer;
    FYMax: Double;
    FYMin: Double;
    FXMax: Double;
    FXMin: Double;
    FYScale: Double;
    FXScale: Double;
    FAutoXRange: TTMSFMXChartSerieAutoRange;
    FAutoYRange: TTMSFMXChartSerieAutoRange;
    FMinX: Double;
    FMinY: Double;
    FMaxX: Double;
    FMaxY: Double;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFill: TBrush;
    FStroke: TBrush;
    FChartType: TTMSFMXChartSerieType;
    FZeroReferenceValue: Double;
    FStrokeOpacity: Double;
    FFillOpacity: Double;
    FMode: TTMSFMXChartSerieMode;
    FBar: TTMSFMXChartSerieBar;
    FMultiPoint: TTMSFMXChartSerieMultiPoint;
    FAnimationFactor: Double;
    FAnimationFlow: Boolean;
    FLegendText: String;
    FShowInLegend: Boolean;
    FRevert: Boolean;
    FMarkers: TTMSFMXChartSerieMarkers;
    FYValues: TTMSFMXChartSerieYValues;
    FXGrid: TTMSFMXChartSerieXYGrid;
    FYGrid: TTMSFMXChartSerieXYGrid;
    FXValues: TTMSFMXChartSerieXValues;
    FMinXOffsetPercentage: Double;
    FMinYOffsetPercentage: Double;
    FMaxXOffsetPercentage: Double;
    FMaxYOffsetPercentage: Double;
    FLabels: TTMSFMXChartSerieLabels;
    FOffset3DX, FOffset3DY: Double;
    F3DXOffset, F3DYOffset: Double;
    FEnable3D: Boolean;
    FGroupIndex: Integer;
    FPie: TTMSFMXChartSeriePie;
    FLegend: TTMSFMXChartSerieLegend;
    procedure SetPoints(const Value: TTMSFMXChartPoints);
    procedure SetVisible(const Value: Boolean);
    procedure SetAutoXRange(const Value: TTMSFMXChartSerieAutoRange);
    procedure SetAutoYRange(const Value: TTMSFMXChartSerieAutoRange);
    procedure SetMaxX(const Value: Double);
    procedure SetMaxY(const Value: Double);
    procedure SetMinX(const Value: Double);
    procedure SetMinY(const Value: Double);
    procedure SetFill(const Value: TBrush);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetChartType(const Value: TTMSFMXChartSerieType);
    procedure SetZeroReferenceValue(const Value: Double);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetFillOpacity(const Value: Double);
    procedure SetMode(const Value: TTMSFMXChartSerieMode);
    procedure SetBar(const Value: TTMSFMXChartSerieBar);
    procedure SetLegendText(const Value: String);
    procedure SetShowInLegend(const Value: Boolean);
    procedure SetMarkers(const Value: TTMSFMXChartSerieMarkers);
    procedure SetXGrid(const Value: TTMSFMXChartSerieXYGrid);
    procedure SetXValues(const Value: TTMSFMXChartSerieXValues);
    procedure SetYGrid(const Value: TTMSFMXChartSerieXYGrid);
    procedure SetYValues(const Value: TTMSFMXChartSerieYValues);
    procedure SetMaxXOffsetPercentage(const Value: Double);
    procedure SetMaxYOffsetPercentage(const Value: Double);
    procedure SetMinXOffsetPercentage(const Value: Double);
    procedure SetMinYOffsetPercentage(const Value: Double);
    procedure SetLabels(const Value: TTMSFMXChartSerieLabels);
    procedure SetOffset3DX(const Value: Double);
    procedure SetOffset3DY(const Value: Double);
    procedure SetEnable3D(const Value: Boolean);
    procedure SetGroupIndex(const Value: Integer);
    procedure SetPie(const Value: TTMSFMXChartSeriePie);
    procedure SetLegend(const Value: TTMSFMXChartSerieLegend);
    procedure SetMultiPoint(const Value: TTMSFMXChartSerieMultiPoint);
  protected
    function IsStacked: Boolean;
    function IsBar: Boolean;
    function IsArea: Boolean;
    function IsPie: Boolean;
    function IsSpider: Boolean;
    function Is3DEnabled: Boolean;
    function Get3DOffsetX: Double;
    function Get3DOffsetY: Double;
    function FindXAxisTextForValue(AValue: Double): String;
    function GetStartCountPoint: Integer;
    function GetEndCountPoint: Integer;
    function GetStartPoint: Integer;
    function GetEndPoint: Integer;
    function GetBarCount: Integer;
    function GetBarOffset: Double;
    function GetBarWidth: Double;
    function GetMultiPointWidth: Double;
    function GetTotalBarWidth: Double;
    function GetPieRect(ASerie: Boolean = False): TRectF;
    procedure CalculateMinMaxX;
    procedure CalculateMinMaxY;
    procedure CalculateDrawPoints;
    procedure CalculateAnimation;
    procedure CalculateArea;
    procedure CalculateLabels;
    procedure CalculateAnnotations;
    procedure UpdateChart;
    procedure RepaintChart;
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
    procedure Draw(ACanvas: TCanvas; ADrawPoint: Integer = -1);
    procedure DrawMarkers(ACanvas: TCanvas);
    procedure DrawLabels(ACanvas: TCanvas);
    procedure DrawLegend(ACanvas: TCanvas);
    function GetLegendText(APoint: TTMSFMXChartPointVirtual): String;
    function GetSpiderLegendText(APoint: TTMSFMXChartPointVirtual): String;
    procedure DrawSpiderGrid(ACanvas: TCanvas);
    procedure DrawSpiderValues(ACanvas: TCanvas);
    procedure DrawAnnotations(ACanvas: TCanvas);
    procedure DrawArrow(ACanvas: TCanvas; ArrowColor: TAlphaColor; ArrowSize: Double; ArrowOpacity: Double; origin, target : TPointF);
    procedure AnimateProcess(Sender: TObject);
    function CalculateOffset: Boolean;
    function Offset: Boolean;
    function GetPointsTotal: Double;
    function GetPointsMax: Double;
    function GetStackedDrawRect(ADrawIndex: Integer; var ADrawRect: TTMSFMXChartDrawRect): Boolean;
    function GetStackedDrawLine(ADrawIndex: Integer; var ADrawLine: TTMSFMXChartDrawLine): Boolean;
    function GetStackedMax(APointIndex: Integer): Double;
  public
    function XYToLegendItem(X, Y: Double): Integer;
    function XYToPointVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
    function XYToBarVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
    function XYToSliceVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
    function XYToPoint(X, Y: Double): TTMSFMXChartPoint;
    function XYToBar(X, Y: Double): TTMSFMXChartPoint;
    function XYToSlice(X, Y: Double): TTMSFMXChartPoint;
    property Revert: Boolean read FRevert write FRevert;
    function DrawPoints: TTMSFMXChartDrawPoints;
    function DrawRects: TTMSFMXChartDrawRects;
    function DrawLines: TTMSFMXChartDrawLines;
    function DrawSlices: TTMSFMXChartDrawSlices;
    function DrawMultiPoints: TTMSFMXChartDrawMultiPoints;
    procedure Animate;
    function YToValue(AY: Double; APosition: Boolean = True): Double;
    function ValueToY(AValue: Double; APosition: Boolean = True): Double;
    function XToValue(AX: Double; APosition: Boolean = True): Double;
    function ValueToX(AValue: Double; APosition: Boolean = True): Double;
    function GetXOffset: Double;
    function Chart: TTMSFMXChart;
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property DataObject: TObject read FDataObject write FDataObject;
    property DataString: String read FDataString write FDataString;
    property DataInteger: Integer read FDataInteger write FDataInteger;
    function GetPointsCount: Integer;
    function GetPoint(AIndex: Integer): TTMSFMXChartPointVirtual;
    function XScale: Double;
    function YScale: Double;
    function YMax: Double;
    function YMin: Double;
    function XMax: Double;
    function XMin: Double;
    function AddPoint(AYValue: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddPoint(AYValue: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddSecondPoint(AYValue, AYValueSecond: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddSecondPoint(AYValue, AYValueSecond: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddVariablePoint(AYValue: Double; AYValueVariable: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddVariablePoint(AYValue: Double; AYValueVariable: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddXYPoint(AXValue, AYValue: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddXYPoint(AXValue, AYValue: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddMultiPoint(AYValueOpen, AYValueHigh, AYValueLow, AYValueClose: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddMultiPoint(AYValueOpen, AYValueHigh, AYValueLow, AYValueClose: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddMultiPoint(AYValueMinimum, AYValueQ1, AYValueMedian, AYValueQ3, AYValueMaximum: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
    function AddMultiPoint(AYValueMinimum, AYValueQ1, AYValueMedian, AYValueQ3, AYValueMaximum: Double; AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint; overload;
  published
    property AnimationFactor: Double read FAnimationFactor write FAnimationFactor;
    property AnimationFlow: Boolean read FAnimationFlow write FAnimationFlow default True;
    property AutoXRange: TTMSFMXChartSerieAutoRange read FAutoXRange write SetAutoXRange default arDisabled;
    property AutoYRange: TTMSFMXChartSerieAutoRange read FAutoYRange write SetAutoYRange default arEnabled;
    property Bar: TTMSFMXChartSerieBar read FBar write SetBar;
    property MultiPoint: TTMSFMXChartSerieMultiPoint read FMultiPoint write SetMultiPoint;
    property Pie: TTMSFMXChartSeriePie read FPie write SetPie;
    property ChartType: TTMSFMXChartSerieType read FChartType write SetChartType default ctLine;
    property Enable3D: Boolean read FEnable3D write SetEnable3D default False;
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property GroupIndex: Integer read FGroupIndex write SetGroupIndex default 0;
    property LegendText: String read FLegendText write SetLegendText;
    property Labels: TTMSFMXChartSerieLabels read FLabels write SetLabels;
    property Markers: TTMSFMXChartSerieMarkers read FMarkers write SetMarkers;
    property MaxX: Double read FMaxX write SetMaxX;
    property MaxY: Double read FMaxY write SetMaxY;
    property MaxXOffsetPercentage: Double read FMaxXOffsetPercentage write SetMaxXOffsetPercentage;
    property MaxYOffsetPercentage: Double read FMaxYOffsetPercentage write SetMaxYOffsetPercentage;
    property MinX: Double read FMinX write SetMinX;
    property MinY: Double read FMinY write SetMinY;
    property MinXOffsetPercentage: Double read FMinXOffsetPercentage write SetMinXOffsetPercentage;
    property MinYOffsetPercentage: Double read FMinYOffsetPercentage write SetMinYOffsetPercentage;
    property Mode: TTMSFMXChartSerieMode read FMode write SetMode default smMathematical;
    property Offset3DX: Double read FOffset3DX write SetOffset3DX;
    property Offset3DY: Double read FOffset3DY write SetOffset3DY;
    property Points: TTMSFMXChartPoints read FPoints write SetPoints;
    property Legend: TTMSFMXChartSerieLegend read FLegend write SetLegend;
    property ShowInLegend: Boolean read FShowInLegend write SetShowInLegend default True;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property Tag: Integer read FTag write FTag default 0;
    property Visible: Boolean read FVisible write SetVisible default True;
    property XGrid: TTMSFMXChartSerieXYGrid read FXGrid write SetXGrid;
    property XValues: TTMSFMXChartSerieXValues read FXValues write SetXValues;
    property YGrid: TTMSFMXChartSerieXYGrid read FYGrid write SetYGrid;
    property YValues: TTMSFMXChartSerieYValues read FYValues write SetYValues;
    property ZeroReferenceValue: Double read FZeroReferenceValue write SetZeroReferenceValue;
  end;

  TTMSFMXChartSeries = class(TOwnedCollection)
  private
    FChart: TTMSFMXChart;
    function GetItem(Index: Integer): TTMSFMXChartSerie;
    procedure SetItem(Index: Integer; const Value: TTMSFMXChartSerie);
  protected
    function GetBarStop: Integer;
    procedure Draw(ACanvas: TCanvas);
  public
    function Chart: TTMSFMXChart;
    constructor Create(AChart: TTMSFMXChart);
    function Add: TTMSFMXChartSerie;
    function Insert(Index: Integer): TTMSFMXChartSerie;
    property Items[Index: Integer]: TTMSFMXChartSerie read GetItem write SetItem; default;
  end;

  TTMSFMXChartBeforeDrawSerieLabel = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawLabel: TTMSFMXChartDrawLabel; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieLabel = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawLabel: TTMSFMXChartDrawLabel) of object;
  TTMSFMXChartBeforeDrawSerieGridLegendText = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLegendText: TTMSFMXChartDrawLabel; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieGridLegendText = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLegendText: TTMSFMXChartDrawLabel) of object;
  TTMSFMXChartBeforeDrawSerieAnnotation = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawAnnotation: TTMSFMXChartDrawAnnotation; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieAnnotation = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawAnnotation: TTMSFMXChartDrawAnnotation) of object;
  TTMSFMXChartGetSerieLabelVirtual = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var AValueString: String) of object;
  TTMSFMXChartGetSerieLegendTextVirtual = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var ALegendText: String) of object;
  TTMSFMXChartGetSerieLabel = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var AValueString: String) of object;
  TTMSFMXChartGetSerieLegendText = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var ALegendText: String) of object;
  TTMSFMXChartBeforeDrawChart = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawChart = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawBackground = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawBackground = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartBeforeDrawXValuesTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie; ATextRect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawXValuesTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie; ATextRect: TRectF) of object;
  TTMSFMXChartBeforeDrawYValuesTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie; ATextRect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawYValuesTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie; ATextRect: TRectF) of object;
  TTMSFMXChartDrawTitleText = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition; var AText: String; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartGetSerie3DOffset = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; var A3DOffset: Double) of object;
  TTMSFMXChartAfterDrawTitle = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition) of object;
  TTMSFMXChartBeforeDrawLegend = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawLegend = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawLegendIcon = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawLegendIcon = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawSerieLegend = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieLegend = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawSerieLegendIconVirtual = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieLegendIconVirtual = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawSerieLegendIcon = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieLegendIcon = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF) of object;
  TTMSFMXChartBeforeDrawXAxis = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawXAxis = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition) of object;
  TTMSFMXChartBeforeDrawSeries = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSeries = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF) of object;
  TTMSFMXChartAnimateSerieStarted = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie) of object;
  TTMSFMXChartAnimateSerieFinished = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie) of object;
  TTMSFMXChartBeforeDrawSerieXValue = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieXValue = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue) of object;
  TTMSFMXChartBeforeDrawSerieYValue = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieYValue = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue) of object;
  TTMSFMXChartBeforeDrawSerieXGridLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieXGridLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine) of object;
  TTMSFMXChartBeforeDrawSerieYGridLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieYGridLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine) of object;
  TTMSFMXChartGetSerieValue = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; AIndex: Integer; AKind: TTMSFMXChartDrawXYValueKind; AValue: Double; var AValueString: String) of object;
  TTMSFMXChartBeforeDrawYAxis = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawYAxis = procedure(Sender: TObject; ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition) of object;
  TTMSFMXChartBeforeDrawSerieMarker = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawPoint: TTMSFMXChartDrawPoint; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieMarker = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawPoint: TTMSFMXChartDrawPoint) of object;
  TTMSFMXChartBeforeDrawSerieBar = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawRect: TTMSFMXChartDrawRect; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieBar = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawRect: TTMSFMXChartDrawRect) of object;
  TTMSFMXChartBeforeDrawSerieLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawLine: TTMSFMXChartDrawLine; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartBeforeDrawSerieMultiPoint = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawMultiPoint: TTMSFMXChartDrawMultiPoint; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartBeforeDrawSerieSlice = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawSlice: TTMSFMXChartDrawSlice; var ADefaultDraw: Boolean) of object;
  TTMSFMXChartAfterDrawSerieLine = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawLine: TTMSFMXChartDrawLine) of object;
  TTMSFMXChartAfterDrawSerieMultiPoint = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawMultiPoint: TTMSFMXChartDrawMultiPoint) of object;
  TTMSFMXChartAfterDrawSerieSlice = procedure(Sender: TObject; ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawSlice: TTMSFMXChartDrawSlice) of object;
  TTMSFMXChartSerieClick = procedure(Sender: TObject; APoint: TTMSFMXChartPoint) of object;
  TTMSFMXChartSerieClickVirtual = procedure(Sender: TObject; APoint: TTMSFMXChartPointVirtual) of object;
  TTMSFMXChartGetNumberOfPoints = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; var ANumberOfPoints: Integer) of object;
  TTMSFMXChartGetPoint = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; AIndex: Integer; var APoint: TTMSFMXChartPointVirtual) of object;
  TTMSFMXChartCanSavePropertyEvent = procedure(Sender: TObject; AObject: TObject; APropertyName: string; APropertyType: TTypeKind; var ACanSave: Boolean) of object;
  TTMSFMXChartCanLoadPropertyEvent = procedure(Sender: TObject; AObject: TObject; APropertyName: string; APropertyType: TTypeKind; var ACanLoad: Boolean) of object;
  TTMSFMXChartSerieLegendItemClick = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; AIndex: Integer) of object;
  TTMSFMXChartLegendItemClick = procedure(Sender: TObject; AIndex: Integer) of object;

  TTMSFMXChartCustomizeAnnotationFont = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AFont: TFont) of object;
  TTMSFMXChartCustomizeAnnotationStroke = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AStroke: TStrokeBrush) of object;
  TTMSFMXChartCustomizeAnnotationFill = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AFill: TBrush) of object;
  TTMSFMXChartGetNumberOfAnnotations = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var ANumberOfAnnotations: Integer) of object;
  TTMSFMXChartGetAnnotation = procedure(Sender: TObject; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AIndex: Integer; var AAnnotation: TTMSFMXChartAnnotationVirtual) of object;

  TTMSFMXChartInteractionScaleMode = (smNone, smHorizontal, smVertical);

  TTMSFMXChartInteractionOptions = class(TPersistent)
  private
    FChart: TTMSFMXChart;
    FScaleMode: TTMSFMXChartInteractionScaleMode;
    FPanning: Boolean;
  public
    procedure Assign(Source: TPersistent); override;
    constructor Create(AChart: TTMSFMXChart);
    destructor Destroy; override;
    function Chart: TTMSFMXChart;
  published
    property ScaleMode: TTMSFMXChartInteractionScaleMode read FScaleMode write FScaleMode default smHorizontal;
    property Panning: Boolean read FPanning write FPanning default True;
  end;

  TTMSFMXChartPieQuadrant = (cpq1, cpq2, cpq3, cpq4);

  TTMSFMXChartSeriesDrawOrder = (csdoNormal, csdoBarsFirst, csdoBarsLast);

  [ComponentPlatformsAttribute(TMSPlatforms)]
  TTMSFMXChart = class(TControl, ITMSChartPersistence)
  private
    FVirtualMode: Boolean;
    FMouseDown, FMouseMoved: Boolean;
    FDownPt, FMovePt: TPointF;
    FTouchDistance: Integer;
    {$IFDEF MSWINDOWS}
    FTouchPoint: TPointF;
    {$ENDIF}
    FCache: Boolean;
    FCachedBitmap: TBitmap;
    FTotalOffset3DX: Double;
    FTotalOffset3DY: Double;
    FUpdateCount: Integer;
    FSeries: TTMSFMXChartSeries;
    FTitle: TTMSFMXChartTitle;
    FOnBeforeDrawTitle: TTMSFMXChartBeforeDrawTitle;
    FOnAfterDrawTitle: TTMSFMXChartAfterDrawTitle;
    FOnDrawTitleText: TTMSFMXChartDrawTitleText;
    FXAxis: TTMSFMXChartXAxis;
    FOnBeforeDrawXAxis: TTMSFMXChartBeforeDrawXAxis;
    FOnAfterDrawXAxis: TTMSFMXChartAfterDrawXAxis;
    FOnBeforeDrawYAxis: TTMSFMXChartBeforeDrawYAxis;
    FOnAfterDrawYAxis: TTMSFMXChartAfterDrawYAxis;
    FYAxis: TTMSFMXChartYAxis;
    FSeriesMargins: TBounds;
    FLegend: TTMSFMXChartLegend;
    FStrokeJoin: TStrokeJoin;
    FStrokeThickness: Double;
    FStrokeCap: TStrokeCap;
    FStrokeDash: TStrokeDash;
    FFill: TBrush;
    FStrokeOpacity: Double;
    FStroke: TBrush;
    FOnBeforeDrawSeries: TTMSFMXChartBeforeDrawSeries;
    FOnAfterDrawSeries: TTMSFMXChartAfterDrawSeries;
    FOnAnimateSerieFinished: TTMSFMXChartAnimateSerieFinished;
    FOnAnimateSerieStarted: TTMSFMXChartAnimateSerieStarted;
    FOnBeforeDrawSerieXValue: TTMSFMXChartBeforeDrawSerieXValue;
    FOnBeforeDrawSerieYValue: TTMSFMXChartBeforeDrawSerieYValue;
    FOnAfterDrawSerieXValue: TTMSFMXChartAfterDrawSerieXValue;
    FOnAfterDrawSerieYValue: TTMSFMXChartAfterDrawSerieYValue;
    FOnGetSerieXValue: TTMSFMXChartGetSerieValue;
    FOnGetSerieYValue: TTMSFMXChartGetSerieValue;
    FOnBeforeDrawSerieMarker: TTMSFMXChartBeforeDrawSerieMarker;
    FOnAfterDrawSerieMarker: TTMSFMXChartAfterDrawSerieMarker;
    FOnBeforeDrawSerieXGridLine: TTMSFMXChartBeforeDrawSerieXGridLine;
    FOnBeforeDrawSerieYGridLine: TTMSFMXChartBeforeDrawSerieYGridLine;
    FOnAfterDrawSerieXGridLine: TTMSFMXChartAfterDrawSerieXGridLine;
    FOnAfterDrawSerieYGridLine: TTMSFMXChartAfterDrawSerieYGridLine;
    FOnAfterDrawLegend: TTMSFMXChartAfterDrawLegend;
    FOnBeforeDrawLegend: TTMSFMXChartBeforeDrawLegend;
    FOnBeforeDrawBackground: TTMSFMXChartBeforeDrawBackground;
    FOnAfterDrawBackground: TTMSFMXChartAfterDrawBackground;
    FOnBeforeDrawChart: TTMSFMXChartBeforeDrawChart;
    FOnAfterDrawChart: TTMSFMXChartAfterDrawChart;
    FOnBeforeDrawXValuesTitle: TTMSFMXChartBeforeDrawXValuesTitle;
    FOnBeforeDrawYValuesTitle: TTMSFMXChartBeforeDrawYValuesTitle;
    FOnAfterDrawXValuesTitle: TTMSFMXChartAfterDrawXValuesTitle;
    FOnAfterDrawYValuesTitle: TTMSFMXChartAfterDrawYValuesTitle;
    FOnAfterDrawSerieLabel: TTMSFMXChartAfterDrawSerieLabel;
    FOnGetSerieLabel: TTMSFMXChartGetSerieLabel;
    FOnBeforeDrawSerieLabel: TTMSFMXChartBeforeDrawSerieLabel;
    FOnAfterDrawSerieLine: TTMSFMXChartAfterDrawSerieLine;
    FOnBeforeDrawSerieBar: TTMSFMXChartBeforeDrawSerieBar;
    FOnAfterDrawSerieBar: TTMSFMXChartAfterDrawSerieBar;
    FOnBeforeDrawSerieLine: TTMSFMXChartBeforeDrawSerieLine;
    FInteraction: Boolean;
    FOnSeriePointClick: TTMSFMXChartSerieClick;
    FOnSerieBarClick: TTMSFMXChartSerieClick;
    FClickMargin: Double;
    FOnBeforeDrawSerieAnnotation: TTMSFMXChartBeforeDrawSerieAnnotation;
    FOnAfterDrawSerieAnnotation: TTMSFMXChartAfterDrawSerieAnnotation;
    FFillOpacity: Double;
    FCachingQuality: TCanvasQuality;
    FCaching: Boolean;
    FOnAfterDrawSerieSlice: TTMSFMXChartAfterDrawSerieSlice;
    FOnBeforeDrawSerieSlice: TTMSFMXChartBeforeDrawSerieSlice;
    FOnSerieSliceClick: TTMSFMXChartSerieClick;
    FOnBeforeDrawSerieLegend: TTMSFMXChartBeforeDrawSerieLegend;
    FOnAfterDrawSerieLegend: TTMSFMXChartAfterDrawSerieLegend;
    FOnGetSerieLegendText: TTMSFMXChartGetSerieLegendText;
    FOnAfterDrawSerieLegendIcon: TTMSFMXChartAfterDrawSerieLegendIcon;
    FOnBeforeDrawSerieLegendIcon: TTMSFMXChartBeforeDrawSerieLegendIcon;
    FOnBeforeDrawLegendIcon: TTMSFMXChartBeforeDrawLegendIcon;
    FOnAfterDrawLegendIcon: TTMSFMXChartAfterDrawLegendIcon;
    FInteractionOptions: TTMSFMXChartInteractionOptions;
    FOnBeforeDrawSerieGridLegendText: TTMSFMXChartBeforeDrawSerieGridLegendText;
    FOnAfterDrawSerieGridLegendText: TTMSFMXChartAfterDrawSerieGridLegendText;
    FOnGetSerieSpiderLegendText: TTMSFMXChartGetSerieLegendText;
    FOnGetNumberOfPoints: TTMSFMXChartGetNumberOfPoints;
    FOnGetPoint: TTMSFMXChartGetPoint;
    FOnGetNumberOfAnnotations: TTMSFMXChartGetNumberOfAnnotations;
    FOnGetAnnotation: TTMSFMXChartGetAnnotation;
    FOnCustomizeAnnotationFill: TTMSFMXChartCustomizeAnnotationFill;
    FOnCustomizeAnnotationStroke: TTMSFMXChartCustomizeAnnotationStroke;
    FOnCustomizeAnnotationFont: TTMSFMXChartCustomizeAnnotationFont;
    FOnSeriePointClickVirtual: TTMSFMXChartSerieClickVirtual;
    FOnSerieBarClickVirtual: TTMSFMXChartSerieClickVirtual;
    FOnSerieSliceClickVirtual: TTMSFMXChartSerieClickVirtual;
    FOnGetSerieSpiderLegendTextVirtual: TTMSFMXChartGetSerieLegendTextVirtual;
    FOnGetSerieLabelVirtual: TTMSFMXChartGetSerieLabelVirtual;
    FOnGetSerieLegendTextVirtual: TTMSFMXChartGetSerieLegendTextVirtual;
    FOnAfterDrawSerieLegendIconVirtual: TTMSFMXChartAfterDrawSerieLegendIconVirtual;
    FOnBeforeDrawSerieLegendIconVirtual: TTMSFMXChartBeforeDrawSerieLegendIconVirtual;
    FOnBeforeDrawSerieMultiPoint: TTMSFMXChartBeforeDrawSerieMultiPoint;
    FOnAfterDrawSerieMultiPoint: TTMSFMXChartAfterDrawSerieMultiPoint;
    FOnCanSaveProperty: TTMSFMXChartCanSavePropertyEvent;
    FOnCanLoadProperty: TTMSFMXChartCanLoadPropertyEvent;
    FOnLegendItemClick: TTMSFMXChartLegendItemClick;
    FOnSerieLegendItemClick: TTMSFMXChartSerieLegendItemClick;
    FDrawOrder: TTMSFMXChartSeriesDrawOrder;
    procedure SetSeries(const Value: TTMSFMXChartSeries);
    function GetVersion: String;
    function GetVersionNr: Integer;
    procedure SetVersion(const Value: String);
    procedure SetTitle(const Value: TTMSFMXChartTitle);
    procedure SetXAxis(const Value: TTMSFMXChartXAxis);
    procedure SetYAxis(const Value: TTMSFMXChartYAxis);
    procedure SetSeriesMargins(const Value: TBounds);
    procedure SetLegend(const Value: TTMSFMXChartLegend);
    procedure SetFill(const Value: TBrush);
    procedure SetStroke(const Value: TBrush);
    procedure SetStrokeCap(const Value: TStrokeCap);
    procedure SetStrokeDash(const Value: TStrokeDash);
    procedure SetStrokeJoin(const Value: TStrokeJoin);
    procedure SetStrokeOpacity(const Value: Double);
    procedure SetStrokeThickness(const Value: Double);
    procedure SetFillOpacity(const Value: Double);
    function GetSerieByName(AName: string): TTMSFMXChartSerie;
    procedure SetCaching(const Value: Boolean);
    procedure SetCachingQuality(const Value: TCanvasQuality);
    procedure SetInteractionOptions(const Value: TTMSFMXChartInteractionOptions);
    procedure DoCanLoadProperty(AObject: TObject; APropertyName: string;
      APropertyType: TTypeKind; var ACanLoad: Boolean);
    procedure DoCanSaveProperty(AObject: TObject; APropertyName: string;
      APropertyType: TTypeKind; var ACanSave: Boolean);
    procedure SetDrawOrder(const Value: TTMSFMXChartSeriesDrawOrder);
  protected
    procedure Loaded; override;
    procedure RegisterRuntimeClasses; virtual;
    procedure UpdateRange(AXRange, AYRange: Double; AScale: Boolean);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Single); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Single); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Single); override;
    function CanSaveProperty(AObject: TObject; APropertyName: string; APropertyType: TTypeKind): Boolean;
    function CanLoadProperty(AObject: TObject; APropertyName: string; APropertyType: TTypeKind): Boolean;
    function GetTotal3DOffsetX: Double;
    function GetTotal3DOffsetY: Double;
    function HasPieChart: Boolean;
    function HasSpiderChart: Boolean;
    function HasStackedPieChart: Boolean;
    procedure FillChanged(Sender: TObject);
    procedure StrokeChanged(Sender: TObject);
    procedure CalculateXAxis;
    procedure CalculateXAxisHeight;
    procedure CalculateYAxis;
    procedure CalculateXScale;
    procedure CalculateYScale;
    procedure CalculateSeries;
    procedure CalculateMinMaxX;
    procedure CalculateMinMaxY;
    procedure Calculate3DOffset;
    procedure CalculateCommonRangeX;
    procedure CalculateCommonRangeY;
    procedure SeriesMarginsChanged(Sender: TObject);
    procedure UpdateChart;
    procedure RepaintChart;
    procedure DrawBackGround(ACanvas: TCanvas); virtual;
    procedure DrawXYGrid(ACanvas: TCanvas); virtual;
    procedure DoBeforeDrawYAxis(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawYAxis(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartYAxisPosition); virtual;
    procedure DoBeforeDrawXAxis(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawXAxis(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartXAxisPosition); virtual;
    procedure DoBeforeDrawTitle(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawTitle(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition); virtual;
    procedure DoBeforeDrawXValuesTitle(ACanvas: TCanvas; ARect: TRectF; ATextRect: TRectF; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawXValuesTitle(ACanvas: TCanvas; ARect: TRectF; ATextRect: TRectF; APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie); virtual;
    procedure DoBeforeDrawYValuesTitle(ACanvas: TCanvas; ARect: TRectF; ATextRect: TRectF; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawYValuesTitle(ACanvas: TCanvas; ARect: TRectF; ATextRect: TRectF; APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie); virtual;
    procedure DoBeforeDrawLegend(ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawLegend(ACanvas: TCanvas; ARect: TRectF); virtual;
    procedure DoBeforeDrawLegendIcon(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawLegendIcon(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF); virtual;
    procedure DoBeforeDrawSerieLegend(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieLegend(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF); virtual;
    procedure DoBeforeDrawSerieLegendIcon(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieLegendIcon(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF); virtual;
    procedure DoGetSerieLegendText(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var ALegendText: String); virtual;
    procedure DoGetSerieSpiderLegendText(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var ALegendText: String); virtual;
    procedure DoBeforeDrawSerieLegendIconVirtual(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieLegendIconVirtual(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF); virtual;
    procedure DoGetSerieLegendTextVirtual(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var ALegendText: String); virtual;
    procedure DoGetSerieSpiderLegendTextVirtual(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var ALegendText: String); virtual;
    procedure DoBeforeDrawBackground(ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawBackground(ACanvas: TCanvas; ARect: TRectF); virtual;
    procedure DoBeforeDrawChart(ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawChart(ACanvas: TCanvas; ARect: TRectF); virtual;
    procedure DoBeforeDrawSeries(ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSeries(ACanvas: TCanvas; ARect: TRectF); virtual;
    procedure DoBeforeDrawSerieMarker(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawPoint: TTMSFMXChartDrawPoint; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieMarker(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawPoint: TTMSFMXChartDrawPoint); virtual;
    procedure DoBeforeDrawSerieSlice(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawSlice: TTMSFMXChartDrawSlice; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieSlice(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawSlice: TTMSFMXChartDrawSlice); virtual;
    procedure DoBeforeDrawSerieLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawLine: TTMSFMXChartDrawLine; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawLine: TTMSFMXChartDrawLine); virtual;
    procedure DoBeforeDrawSerieMultiPoint(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawMultiPoint: TTMSFMXChartDrawMultiPoint; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieMultiPoint(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawMultiPoint: TTMSFMXChartDrawMultiPoint); virtual;
    procedure DoBeforeDrawSerieGridLegendText(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLegendText: TTMSFMXChartDrawLabel; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieGridLegendText(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLegendText: TTMSFMXChartDrawLabel); virtual;
    procedure DoBeforeDrawSerieLabel(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawLabel: TTMSFMXChartDrawLabel; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieLabel(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawLabel: TTMSFMXChartDrawLabel); virtual;
    procedure DoBeforeDrawSerieAnnotation(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawAnnotation: TTMSFMXChartDrawAnnotation; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieAnnotation(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawAnnotation: TTMSFMXChartDrawAnnotation); virtual;
    procedure DoBeforeDrawSerieBar(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawRect: TTMSFMXChartDrawRect; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieBar(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawRect: TTMSFMXChartDrawRect); virtual;
    procedure DoDrawTitleText(ACanvas: TCanvas; ARect: TRectF; APosition: TTMSFMXChartTitlePosition; var AText: String; var ADefaultDraw: Boolean); virtual;
    procedure DoAnimateSerieFinished(ASerie: TTMSFMXChartSerie); virtual;
    procedure DoAnimateSerieStarted(ASerie: TTMSFMXChartSerie); virtual;
    procedure DoGetSerieYValue(ASerie: TTMSFMXChartSerie; AIndex: Integer; AKind: TTMSFMXChartDrawXYValueKind; AValue: Double; var AValueString: String); virtual;
    procedure DoGetSerieLabel(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var AValueString: String); virtual;
    procedure DoGetSerieLabelVirtual(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var AValueString: String); virtual;
    procedure DoGetSerieXValue(ASerie: TTMSFMXChartSerie; AIndex: Integer; AKind: TTMSFMXChartDrawXYValueKind; AValue: Double; var AValueString: String); virtual;
    procedure DoBeforeDrawSerieYValue(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieYValue(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue); virtual;
    procedure DoBeforeDrawSerieXValue(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieXValue(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue); virtual;
    procedure DoBeforeDrawSerieYGridLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieYGridLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine); virtual;
    procedure DoBeforeDrawSerieXGridLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean); virtual;
    procedure DoAfterDrawSerieXGridLine(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine); virtual;
    procedure DoSeriePointClick(APoint: TTMSFMXChartPoint); virtual;
    procedure DoSerieBarClick(APoint: TTMSFMXChartPoint); virtual;
    procedure DoSerieSliceClick(APoint: TTMSFMXChartPoint); virtual;
    procedure DoSerieLegendItemClick(ASerie: TTMSFMXChartSerie; AIndex: Integer); virtual;
    procedure DoLegendItemClick(AIndex: Integer); virtual;
    procedure DoSeriePointClickVirtual(APoint: TTMSFMXChartPointVirtual); virtual;
    procedure DoSerieBarClickVirtual(APoint: TTMSFMXChartPointVirtual); virtual;
    procedure DoSerieSliceClickVirtual(APoint: TTMSFMXChartPointVirtual); virtual;
    procedure DoCustomizeAnnotationFont(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AFont: TFont); virtual;
    procedure DoCustomizeAnnotationStroke(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AStroke: TStrokeBrush); virtual;
    procedure DoCustomizeAnnotationFill(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual; AFill: TBrush); virtual;

    procedure DoGetNumberOfPoints(ASerie: TTMSFMXChartSerie; var ANumberOfPoints: Integer); virtual;
    procedure DoGetPoint(ASerie: TTMSFMXChartSerie; AIndex: Integer; var APoint: TTMSFMXChartPointVirtual); virtual;
    procedure DoGetNumberOfAnnotations(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; var ANumberOfAnnotations: Integer); virtual;
    procedure DoGetAnnotation(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; AIndex: Integer; var AAnnotation: TTMSFMXChartAnnotationVirtual); virtual;
    procedure MouseWheel(Shift: TShiftState; WheelDelta: Integer; var Handled: Boolean); override;
  public
    function XYToSerieLegendItem(X, Y: Double; var ASerie: TTMSFMXChartSerie): Integer;
    function XYToLegendItem(X, Y: Double): Integer;
    function XYToPoint(X, Y: Double): TTMSFMXChartPoint;
    function XYToBar(X, Y: Double): TTMSFMXChartPoint;
    function XYToSlice(X, Y: Double): TTMSFMXChartPoint;
    function XYToPointVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
    function XYToBarVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
    function XYToSliceVirtual(X, Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;

    function AddParetoLine(ASerieIndex, AStartx, AEndx: Integer; AUseCommonYRange: Boolean = True; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: Integer = 1): TTMSFMXChartSerie;
    function AddLineSerie(AStartX, AEndX: Integer; AStartValue, AEndValue: Double; AUseCommonYRange: Boolean = True; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: Integer = 1): TTMSFMXChartSerie;
    function AddTrendLine(ASerieIndex, AStartX, AEndX: Integer; AUseCommonYRange: Boolean = True; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: Integer = 1): TTMSFMXChartSerie;
    function AddBandSerie(AStartX, AEndX: integer; AStartValue1, AStartValue2, AEndValue1, AEndValue2: Double; AUseCommonYRange: Boolean = true; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: integer = 1; ASerieBandColor: TAlphaColor = claGray): TTMSFMXChartSerie;
    function AddTrendChannelBand(ASerieIndex, AStartX, AEndX: integer; AUseCommonYRange: Boolean = True; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: integer = 1; ASerieBandColor: TAlphaColor = claGray): TTMSFMXChartSerie;
    procedure AddTrendChannel(ASerieIndex, AStartX, AEndX: Integer; AUseCommonYRange: Boolean = True; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: Integer = 1);
    function AddMovingAverage(ASerieIndex, AStartX, AEndX, AWindow: Integer; AUseCommonYRange: Boolean = true; ASerieLineColor: TAlphaColor = claBlack; ASerieLineWidth: Integer = 1): TTMSFMXChartSerie;

    function GetSeriesRectangle: TRectF;
    function GetChartRectangle: TRectF;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    {$IF DEFINED(DELPHIXE7_LVL) AND NOT DEFINED(DELPHIXE8_LVL)}
    procedure BeginUpdate;
    procedure EndUpdate;
    {$ELSE}
    procedure BeginUpdate; override;
    procedure EndUpdate; override;
    {$ENDIF}
    procedure Paint; override;
    procedure DrawChart(ACanvas: TCanvas); virtual;
    procedure Resize; override;
    procedure InitSample;
    procedure Clear;
    property SerieByName[AName: string]: TTMSFMXChartSerie read GetSerieByName;
    procedure SaveToImage(AFileName: String);
    procedure SaveSettingsToFile(AFileName: string);
    procedure LoadSettingsFromFile(AFileName: string);
    procedure SaveSettingsToStream(AStream: TStream);
    procedure LoadSettingsFromStream(AStream: TStream);
    procedure CMGesture(var EventInfo: TGestureEventInfo); override;
  published
    property ClickMargin: Double read FClickMargin write FClickMargin;
    property InteractionOptions: TTMSFMXChartInteractionOptions read FInteractionOptions write SetInteractionOptions;
    property Fill: TBrush read FFill write SetFill;
    property FillOpacity: Double read FFillOpacity write SetFillOpacity;
    property Legend: TTMSFMXChartLegend read FLegend write SetLegend;
    property Interaction: Boolean read FInteraction write FInteraction default True;
    property Stroke: TBrush read FStroke write SetStroke;
    property StrokeOpacity: Double read FStrokeOpacity write SetStrokeOpacity;
    property StrokeThickness: Double read FStrokeThickness write SetStrokeThickness;
    property StrokeCap: TStrokeCap read FStrokeCap write SetStrokeCap default TStrokeCap.Flat;
    property StrokeDash: TStrokeDash read FStrokeDash write SetStrokeDash default TStrokeDash.Solid;
    property StrokeJoin: TStrokeJoin read FStrokeJoin write SetStrokeJoin default TStrokeJoin.Round;
    property SeriesMargins: TBounds read FSeriesMargins write SetSeriesMargins;
    property DrawOrder: TTMSFMXChartSeriesDrawOrder read FDrawOrder write SetDrawOrder default csdoBarsFirst;
    property Series: TTMSFMXChartSeries read FSeries write SetSeries;
    property Title: TTMSFMXChartTitle read FTitle write SetTitle;
    property Version: String read GetVersion write SetVersion;
    property XAxis: TTMSFMXChartXAxis read FXAxis write SetXAxis;
    property YAxis: TTMSFMXChartYAxis read FYAxis write SetYAxis;
    property Caching: Boolean read FCaching write SetCaching default False;
    property CachingQuality: TCanvasQuality read FCachingQuality write SetCachingQuality default TCanvasQuality.HighQuality;
    property OnAfterDrawSerieBar: TTMSFMXChartAfterDrawSerieBar read FOnAfterDrawSerieBar write FOnAfterDrawSerieBar;
    property OnAfterDrawSerieLine: TTMSFMXChartAfterDrawSerieLine read FOnAfterDrawSerieLine write FOnAfterDrawSerieLine;
    property OnAfterDrawSerieMultiPoint: TTMSFMXChartAfterDrawSerieMultiPoint read FOnAfterDrawSerieMultiPoint write FOnAfterDrawSerieMultiPoint;
    property OnAfterDrawSerieSlice: TTMSFMXChartAfterDrawSerieSlice read FOnAfterDrawSerieSlice write FOnAfterDrawSerieSlice;
    property OnAfterDrawSerieLabel: TTMSFMXChartAfterDrawSerieLabel read FOnAfterDrawSerieLabel write FOnAfterDrawSerieLabel;
    property OnAfterDrawSerieAnnotation: TTMSFMXChartAfterDrawSerieAnnotation read FOnAfterDrawSerieAnnotation write FOnAfterDrawSerieAnnotation;
    property OnBeforeDrawSerieBar: TTMSFMXChartBeforeDrawSerieBar read FOnBeforeDrawSerieBar write FOnBeforeDrawSerieBar;
    property OnBeforeDrawSerieLine: TTMSFMXChartBeforeDrawSerieLine read FOnBeforeDrawSerieLine write FOnBeforeDrawSerieLine;
    property OnBeforeDrawSerieMultiPoint: TTMSFMXChartBeforeDrawSerieMultiPoint read FOnBeforeDrawSerieMultiPoint write FOnBeforeDrawSerieMultiPoint;
    property OnBeforeDrawSerieSlice: TTMSFMXChartBeforeDrawSerieSlice read FOnBeforeDrawSerieSlice write FOnBeforeDrawSerieSlice;
    property OnBeforeDrawSerieLabel: TTMSFMXChartBeforeDrawSerieLabel read FOnBeforeDrawSerieLabel write FOnBeforeDrawSerieLabel;
    property OnBeforeDrawSerieGridLegendText: TTMSFMXChartBeforeDrawSerieGridLegendText read FOnBeforeDrawSerieGridLegendText write FOnBeforeDrawSerieGridLegendText;
    property OnAfterDrawSerieGridLegendText: TTMSFMXChartAfterDrawSerieGridLegendText read FOnAfterDrawSerieGridLegendText write FOnAfterDrawSerieGridLegendText;
    property OnBeforeDrawSerieAnnotation: TTMSFMXChartBeforeDrawSerieAnnotation read FOnBeforeDrawSerieAnnotation write FOnBeforeDrawSerieAnnotation;
    property OnBeforeDrawXValuesTitle: TTMSFMXChartBeforeDrawXValuesTitle read FOnBeforeDrawXValuesTitle write FOnBeforeDrawXValuesTitle;
    property OnBeforeDrawYValuesTitle: TTMSFMXChartBeforeDrawYValuesTitle read FOnBeforeDrawYValuesTitle write FOnBeforeDrawYValuesTitle;
    property OnAfterDrawXValuesTitle: TTMSFMXChartAfterDrawXValuesTitle read FOnAfterDrawXValuesTitle write FOnAfterDrawXValuesTitle;
    property OnAfterDrawYValuesTitle: TTMSFMXChartAfterDrawYValuesTitle read FOnAfterDrawYValuesTitle write FOnAfterDrawYValuesTitle;
    property OnBeforeDrawXAxis: TTMSFMXChartBeforeDrawXAxis read FOnBeforeDrawXAxis write FOnBeforeDrawXAxis;
    property OnAfterDrawXAxis: TTMSFMXChartAfterDrawXAxis read FOnAfterDrawXAxis write FOnAfterDrawXAxis;
    property OnBeforeDrawYAxis: TTMSFMXChartBeforeDrawYAxis read FOnBeforeDrawYAxis write FOnBeforeDrawYAxis;
    property OnAfterDrawYAxis: TTMSFMXChartAfterDrawYAxis read FOnAfterDrawYAxis write FOnAfterDrawYAxis;
    property OnBeforeDrawTitle: TTMSFMXChartBeforeDrawTitle read FOnBeforeDrawTitle write FOnBeforeDrawTitle;
    property OnAfterDrawTitle: TTMSFMXChartAfterDrawTitle read FOnAfterDrawTitle write FOnAfterDrawTitle;
    property OnBeforeDrawBackground: TTMSFMXChartBeforeDrawBackground read FOnBeforeDrawBackground write FOnBeforeDrawBackground;
    property OnAfterDrawBackground: TTMSFMXChartAfterDrawBackground read FOnAfterDrawBackground write FOnAfterDrawBackground;
    property OnBeforeDrawChart: TTMSFMXChartBeforeDrawChart read FOnBeforeDrawChart write FOnBeforeDrawChart;
    property OnAfterDrawChart: TTMSFMXChartAfterDrawChart read FOnAfterDrawChart write FOnAfterDrawChart;
    property OnBeforeDrawLegend: TTMSFMXChartBeforeDrawLegend read FOnBeforeDrawLegend write FOnBeforeDrawLegend;
    property OnAfterDrawLegend: TTMSFMXChartAfterDrawLegend read FOnAfterDrawLegend write FOnAfterDrawLegend;
    property OnBeforeDrawLegendIcon: TTMSFMXChartBeforeDrawLegendIcon read FOnBeforeDrawLegendIcon write FOnBeforeDrawLegendIcon;
    property OnAfterDrawLegendIcon: TTMSFMXChartAfterDrawLegendIcon read FOnAfterDrawLegendIcon write FOnAfterDrawLegendIcon;
    property OnBeforeDrawSerieLegend: TTMSFMXChartBeforeDrawSerieLegend read FOnBeforeDrawSerieLegend write FOnBeforeDrawSerieLegend;
    property OnAfterDrawSerieLegend: TTMSFMXChartAfterDrawSerieLegend read FOnAfterDrawSerieLegend write FOnAfterDrawSerieLegend;
    property OnBeforeDrawSerieLegendIcon: TTMSFMXChartBeforeDrawSerieLegendIcon read FOnBeforeDrawSerieLegendIcon write FOnBeforeDrawSerieLegendIcon;
    property OnAfterDrawSerieLegendIcon: TTMSFMXChartAfterDrawSerieLegendIcon read FOnAfterDrawSerieLegendIcon write FOnAfterDrawSerieLegendIcon;
    property OnBeforeDrawSerieLegendIconVirtual: TTMSFMXChartBeforeDrawSerieLegendIconVirtual read FOnBeforeDrawSerieLegendIconVirtual write FOnBeforeDrawSerieLegendIconVirtual;
    property OnAfterDrawSerieLegendIconVirtual: TTMSFMXChartAfterDrawSerieLegendIconVirtual read FOnAfterDrawSerieLegendIconVirtual write FOnAfterDrawSerieLegendIconVirtual;
    property OnCustomizeAnnotationFont: TTMSFMXChartCustomizeAnnotationFont read FOnCustomizeAnnotationFont write FOnCustomizeAnnotationFont;
    property OnCustomizeAnnotationFill: TTMSFMXChartCustomizeAnnotationFill read FOnCustomizeAnnotationFill write FOnCustomizeAnnotationFill;
    property OnCustomizeAnnotationStroke: TTMSFMXChartCustomizeAnnotationStroke read FOnCustomizeAnnotationStroke write FOnCustomizeAnnotationStroke;
    property OnGetNumberOfPoints: TTMSFMXChartGetNumberOfPoints read FOnGetNumberOfPoints write FOnGetNumberOfPoints;
    property OnGetPoint: TTMSFMXChartGetPoint read FOnGetPoint write FOnGetPoint;
    property OnGetNumberOfAnnotations: TTMSFMXChartGetNumberOfAnnotations read FOnGetNumberOfAnnotations write FOnGetNumberOfAnnotations;
    property OnGetAnnotation: TTMSFMXChartGetAnnotation read FOnGetAnnotation write FOnGetAnnotation;
    property OnDrawTitleText: TTMSFMXChartDrawTitleText read FOnDrawTitleText write FOnDrawTitleText;
    property OnGetSerieLabel: TTMSFMXChartGetSerieLabel read FOnGetSerieLabel write FOnGetSerieLabel;
    property OnGetSerieLegendText: TTMSFMXChartGetSerieLegendText read FOnGetSerieLegendText write FOnGetSerieLegendText;
    property OnGetSerieSpiderLegendText: TTMSFMXChartGetSerieLegendText read FOnGetSerieSpiderLegendText write FOnGetSerieSpiderLegendText;
    property OnGetSerieLabelVirtual: TTMSFMXChartGetSerieLabelVirtual read FOnGetSerieLabelVirtual write FOnGetSerieLabelVirtual;
    property OnGetSerieLegendTextVirtual: TTMSFMXChartGetSerieLegendTextVirtual read FOnGetSerieLegendTextVirtual write FOnGetSerieLegendTextVirtual;
    property OnGetSerieSpiderLegendTextVirtual: TTMSFMXChartGetSerieLegendTextVirtual read FOnGetSerieSpiderLegendTextVirtual write FOnGetSerieSpiderLegendTextVirtual;
    property OnBeforeDrawSerieMarker: TTMSFMXChartBeforeDrawSerieMarker read FOnBeforeDrawSerieMarker write FOnBeforeDrawSerieMarker;
    property OnAfterDrawSerieMarker: TTMSFMXChartAfterDrawSerieMarker read FOnAfterDrawSerieMarker write FOnAfterDrawSerieMarker;
    property OnBeforeDrawSeries: TTMSFMXChartBeforeDrawSeries read FOnBeforeDrawSeries write FOnBeforeDrawSeries;
    property OnAfterDrawSeries: TTMSFMXChartAfterDrawSeries read FOnAfterDrawSeries write FOnAfterDrawSeries;
    property OnAnimateSerieFinished: TTMSFMXChartAnimateSerieFinished read FOnAnimateSerieFinished write FOnAnimateSerieFinished;
    property OnAnimateSerieStarted: TTMSFMXChartAnimateSerieStarted read FOnAnimateSerieStarted write FOnAnimateSerieStarted;
    property OnBeforeDrawSerieXValue: TTMSFMXChartBeforeDrawSerieXValue read FOnBeforeDrawSerieXValue write FOnBeforeDrawSerieXValue;
    property OnBeforeDrawSerieYValue: TTMSFMXChartBeforeDrawSerieYValue read FOnBeforeDrawSerieYValue write FOnBeforeDrawSerieYValue;
    property OnAfterDrawSerieXValue: TTMSFMXChartAfterDrawSerieXValue read FOnAfterDrawSerieXValue write FOnAfterDrawSerieXValue;
    property OnAfterDrawSerieYValue: TTMSFMXChartAfterDrawSerieYValue read FOnAfterDrawSerieYValue write FOnAfterDrawSerieYValue;
    property OnBeforeDrawSerieXGridLine: TTMSFMXChartBeforeDrawSerieXGridLine read FOnBeforeDrawSerieXGridLine write FOnBeforeDrawSerieXGridLine;
    property OnBeforeDrawSerieYGridLine: TTMSFMXChartBeforeDrawSerieYGridLine read FOnBeforeDrawSerieYGridLine write FOnBeforeDrawSerieYGridLine;
    property OnAfterDrawSerieXGridLine: TTMSFMXChartAfterDrawSerieXGridLine read FOnAfterDrawSerieXGridLine write FOnAfterDrawSerieXGridLine;
    property OnAfterDrawSerieYGridLine: TTMSFMXChartAfterDrawSerieYGridLine read FOnAfterDrawSerieYGridLine write FOnAfterDrawSerieYGridLine;
    property OnGetSerieXValue: TTMSFMXChartGetSerieValue read FOnGetSerieXValue write FOnGetSerieXValue;
    property OnGetSerieYValue: TTMSFMXChartGetSerieValue read FOnGetSerieYValue write FOnGetSerieYValue;
    property OnSerieBarClick: TTMSFMXChartSerieClick read FOnSerieBarClick write FOnSerieBarClick;
    property OnSeriePointClick: TTMSFMXChartSerieClick read FOnSeriePointClick write FOnSeriePointClick;
    property OnSerieSliceClick: TTMSFMXChartSerieClick read FOnSerieSliceClick write FOnSerieSliceClick;
    property OnSerieBarClickVirtual: TTMSFMXChartSerieClickVirtual read FOnSerieBarClickVirtual write FOnSerieBarClickVirtual;
    property OnSerieLegendItemClick: TTMSFMXChartSerieLegendItemClick read FOnSerieLegendItemClick write FOnSerieLegendItemClick;
    property OnLegendItemClick: TTMSFMXChartLegendItemClick read FOnLegendItemClick write FOnLegendItemClick;
    property OnSeriePointClickVirtual: TTMSFMXChartSerieClickVirtual read FOnSeriePointClickVirtual write FOnSeriePointClickVirtual;
    property OnSerieSliceClickVirtual: TTMSFMXChartSerieClickVirtual read FOnSerieSliceClickVirtual write FOnSerieSliceClickVirtual;
    property OnCanSaveProperty: TTMSFMXChartCanSavePropertyEvent read FOnCanSaveProperty write FOnCanSaveProperty;
    property OnCanLoadProperty: TTMSFMXChartCanLoadPropertyEvent read FOnCanLoadProperty write FOnCanLoadProperty;
    property Align;
    property Anchors;
    property CanParentFocus;
    property ClipChildren default False;
    property ClipParent default False;
    property Cursor default crDefault;
    {$IFNDEF DELPHIXE7_LVL}
    property DesignVisible default True;
    {$ENDIF}
    property DisableFocusEffect default True;
    property DragMode default TDragMode.dmManual;
    property EnableDragHighlight default True;
    property Enabled default True;
    property Locked default False;
    property Height;
    property HitTest default True;
    property Margins;
    property Opacity;
    property Padding;
    property PopupMenu;
    property Position;
    property RotationAngle;
    property RotationCenter;
    property Scale;
    property TabOrder;
    property Visible default True;
    property Width;
    property OnApplyStyleLookup;
    property OnDragEnter;
    property OnDragLeave;
    property OnDragOver;
    property OnDragDrop;
    property OnDragEnd;
    property OnKeyDown;
    property OnKeyUp;
    property OnCanFocus;
    property OnClick;
    property OnDblClick;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnPainting;
    property OnPaint;
    property OnResize;
    {$IFDEF DELPHIXE7_LVL}
    property Size;
    {$ENDIF}
  end;


{$EXTERNALSYM Hiword}
function Hiword(L: DWORD): integer;
{$EXTERNALSYM LoWord}
function LoWord(L: DWORD): Integer;
{$EXTERNALSYM MakeWord}
function MakeWord(b1,b2: integer): integer;
{$EXTERNALSYM MakeLong}
function MakeLong(i1,i2: integer): integer;

procedure DrawChartIcon(ACanvas: TCanvas; AChartType: TTMSFMXChartSerieType; r: TRectF);
function ModifyRect(ARect: TRectF): TRectF;
function GetDefaultVirtualPoint: TTMSFMXChartPointVirtual;
function GetDefaultVirtualAnnotation: TTMSFMXChartAnnotationVirtual;

implementation

uses
  StrUtils;

function GetDefaultDrawRect: TTMSFMXChartDrawRect;
begin
  Result.Rect := RectF(0, 0, 0, 0);
  Result.StackedValue := 0;
  Result.Reference := nil;
  Result.VirtualReference := GetDefaultVirtualPoint;
  Result.DoAnimation := False;
end;

function GetDefaultDrawLine: TTMSFMXChartDrawLine;
begin
  Result.StackedValueStart := 0;
  Result.StackedValueEnd := 0;
  Result.StartPoint := PointF(0, 0);
  Result.EndPoint := PointF(0, 0);
  Result.StartReference := nil;
  Result.EndReference := nil;
  Result.StartVirtualReference := GetDefaultVirtualPoint;
  Result.EndVirtualReference := GetDefaultVirtualPoint;
  Result.Bottom := False;
  Result.DoAnimation := False;
end;

function GetDefaultVirtualAnnotation: TTMSFMXChartAnnotationVirtual;
begin
  Result.Index := -1;
  Result.AutoSize := True;
  Result.CornerRadius := 5;
  Result.LineOpacity := 1;
  Result.ArrowOpacity := 1;
  Result.Width := 75;
  Result.Height := 35;
  Result.LineThickness := 1;
  Result.LineColor := claBlack;
  Result.StrokeOpacity := 1;
  Result.FillOpacity := 1;
  Result.ArrowColor := claBlack;
  Result.ArrowSize := 10;
  Result.BalloonDirection := bdDownLeft;
  Result.BalloonArrowSize := 10;
  Result.Visible := True;
  Result.OffsetX := 0;
  Result.OffsetY := -10;
  Result.TextVerticalAlignment := TTextAlign.Center;
  Result.TextHorizontalAlignment := TTextAlign.Center;
  Result.Arrow := arLine;
  Result.FontColor := TAlphaColorRec.Black;
  Result.Shape := asRectangle;
  Result.WordWrap := True;
end;

function GetDefaultVirtualPoint: TTMSFMXChartPointVirtual;
begin
  Result.Index := -1;
  Result.Color := claNull;
  Result.Tag := 0;
  Result.XValue := 0;
  Result.YValue := 0;
  Result.YValueHigh := 0;
  Result.YValueClose := 0;
  Result.YValueOpen := 0;
  Result.YValueMedian := 0;
  Result.YValueLow := 0;
  Result.YValueSecond := 0;
  Result.XValueText := '';
  Result.YValueVariable := 0;
  Result.LegendText := '';
  Result.Explode := 0;
  Result.Point := nil;
end;

function GetPieQuadrant(ACenter: TPointF; APoint: TPointF): TTMSFMXChartPieQuadrant;
begin
  if (ACenter.X <= APoint.X) and (ACenter.Y >= APoint.Y) then
    Result := cpq1
  else if (ACenter.X < APoint.X) and (ACenter.Y < APoint.Y) then
    Result := cpq4
  else if (ACenter.X >= APoint.X) and (ACenter.Y >= APoint.Y) then
    Result := cpq2
  else
    Result := cpq3;
end;

function ModifyRect(ARect: TRectF): TRectF;
begin
  Result := RectF(int(ARect.Left) - 0.5, int(ARect.Top) - 0.5, int(ARect.Right) + 0.5, int(ARect.Bottom) + 0.5);
end;

procedure DrawChartIcon(ACanvas: TCanvas; AChartType: TTMSFMXChartSerieType; r: TRectF);
var
  rm, rb: TRectF;
  mw, mh: Double;
  pth: TPathData;
  c: TAlphaColor;
begin
  rm := r;
  case AChartType of
    ctXYLine:
    begin
      ACanvas.DrawLine(PointF(rm.Left, rm.Top + rm.Height / 4*2), PointF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 3), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 3), PointF(rm.Left + rm.Width / 2, rm.Top), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2, rm.Top), PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 5), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 5), rm.BottomRight, 1);
    end;
    ctLine:
    begin
      ACanvas.DrawLine(PointF(rm.Left, rm.Top + rm.Height / 3*2), PointF(rm.Left + rm.Width / 5, rm.Top + rm.Height / 3), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 5, rm.Top + rm.Height / 3), PointF(rm.Left + rm.Width / 5 * 2, rm.Bottom), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 5 * 2, rm.Bottom), PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 5), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 5), rm.BottomRight, 1);
    end;
    ctDigitalLine:
    begin
      ACanvas.DrawLine(PointF(rm.Left, rm.Top + rm.Height), PointF(rm.Left, rm.Top + rm.Height / 4), 1);
      ACanvas.DrawLine(PointF(rm.Left, rm.Top + rm.Height / 4), PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 4), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 4), PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 3 * 2), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 3 * 2), PointF(rm.Left + rm.Width, rm.Top + rm.Height / 3 * 2), 1);
      ACanvas.DrawLine(PointF(rm.Left + rm.Width, rm.Top + rm.Height / 3 * 2), rm.BottomRight, 1);
    end;
    ctXYMarker:
    begin
      mw := rm.Width / 4;
      mh := rm.Height / 4;
      ACanvas.FillEllipse(RectF(rm.Left, rm.Top, rm.Left + mw, rm.Top + mh), 1);
      ACanvas.DrawEllipse(RectF(rm.Left, rm.Top, rm.Left + mw, rm.Top + mh), 1);
      ACanvas.FillEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y, rm.CenterPoint.X - mw, rm.CenterPoint.Y + mh), 1);
      ACanvas.DrawEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y, rm.CenterPoint.X - mw, rm.CenterPoint.Y + mh), 1);
      ACanvas.FillEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y, rm.CenterPoint.X + mw, rm.CenterPoint.Y + mh), 1);
      ACanvas.DrawEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y, rm.CenterPoint.X + mw, rm.CenterPoint.Y + mh), 1);
      ACanvas.FillEllipse(RectF(rm.Right - mw, rm.Top + rm.Height, rm.Right, rm.Top + rm.Height - mh), 1);
      ACanvas.DrawEllipse(RectF(rm.Right - mw, rm.Top + rm.Height, rm.Right, rm.Top + rm.Height - mh), 1);
    end;
    ctMarker:
    begin
      mw := rm.Width / 4;
      mh := rm.Height / 4;
      ACanvas.FillEllipse(RectF(rm.Left, rm.Top, rm.Left + mw, rm.Top + mh), 1);
      ACanvas.DrawEllipse(RectF(rm.Left, rm.Top, rm.Left + mw, rm.Top + mh), 1);
      ACanvas.FillEllipse(RectF(rm.CenterPoint.X - mw * 1.5, rm.CenterPoint.Y, rm.CenterPoint.X - mw * 0.5, rm.CenterPoint.Y + mh), 1);
      ACanvas.DrawEllipse(RectF(rm.CenterPoint.X - mw * 1.5, rm.CenterPoint.Y, rm.CenterPoint.X - mw * 0.5, rm.CenterPoint.Y + mh), 1);
      ACanvas.FillEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y + mh, rm.CenterPoint.X + mw, rm.CenterPoint.Y + mh * 2), 1);
      ACanvas.DrawEllipse(RectF(rm.CenterPoint.X, rm.CenterPoint.Y + mh, rm.CenterPoint.X + mw, rm.CenterPoint.Y + mh * 2), 1);
      ACanvas.FillEllipse(RectF(rm.Right - mw, rm.Top + mh / 2, rm.Right, rm.Top + mh * 1.5), 1);
      ACanvas.DrawEllipse(RectF(rm.Right - mw, rm.Top + mh / 2, rm.Right, rm.Top + mh * 1.5), 1);
    end;
    ctStackedPercentageArea:
    begin
      pth := TPathData.Create;
      pth.MoveTo(PointF(rm.Left, rm.Top + rm.Height));
      pth.LineTo(PointF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 2));
      pth.LineTo(rm.CenterPoint);
      pth.LineTo(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 6));
      pth.LineTo(rm.BottomRight);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);

      c := ACanvas.Fill.Color;
      TAlphaColorRec(c).A := 150;
      ACanvas.Fill.Color := c;
      pth.MoveTo(pth.Points[0].Point);
      pth.LineTo(PointF(pth.Points[0].Point.X, rm.Top));
      pth.LineTo(PointF(pth.Points[1].Point.X, rm.Top));
      pth.LineTo(PointF(pth.Points[2].Point.X, rm.Top));
      pth.LineTo(PointF(pth.Points[3].Point.X, rm.Top));
      pth.LineTo(PointF(pth.Points[4].Point.X, rm.Top));
      pth.LineTo(pth.Points[4].Point);
      pth.LineTo(pth.Points[3].Point);
      pth.LineTo(pth.Points[2].Point);
      pth.LineTo(pth.Points[1].Point);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctStackedArea:
    begin
      pth := TPathData.Create;
      pth.MoveTo(PointF(rm.Left, rm.Top + rm.Height));
      pth.LineTo(PointF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 2));
      pth.LineTo(rm.CenterPoint);
      pth.LineTo(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 6));
      pth.LineTo(rm.BottomRight);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);

      c := ACanvas.Fill.Color;
      TAlphaColorRec(c).A := 150;
      ACanvas.Fill.Color := c;
      pth.MoveTo(pth.Points[0].Point);
      pth.LineTo(PointF(pth.Points[0].Point.X, rm.Top + rm.Height / 4));
      pth.LineTo(PointF(pth.Points[1].Point.X, rm.Top + rm.Height / 5));
      pth.LineTo(PointF(pth.Points[2].Point.X, rm.Top + rm.Height / 3));
      pth.LineTo(PointF(pth.Points[3].Point.X, rm.Top));
      pth.LineTo(PointF(pth.Points[4].Point.X, rm.Top + rm.Height / 3));
      pth.LineTo(pth.Points[4].Point);
      pth.LineTo(pth.Points[3].Point);
      pth.LineTo(pth.Points[2].Point);
      pth.LineTo(pth.Points[1].Point);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctBand:
    begin
      pth := TPathData.Create;
      pth.MoveTo(PointF(rm.Left, rm.Top + rm.Height));
      pth.LineTo(PointF(rm.Left + rm.Width / 5, rm.Top + rm.Height / 3));
      pth.LineTo(PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 3));
      pth.LineTo(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 3));
      pth.LineTo(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 5 * 4));
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctArea:
    begin
      pth := TPathData.Create;
      pth.MoveTo(PointF(rm.Left, rm.Top + rm.Height));
      pth.LineTo(PointF(rm.Left + rm.Width / 4, rm.Top));
      pth.LineTo(rm.CenterPoint);
      pth.LineTo(PointF(rm.Left + rm.Width / 5 * 4, rm.Top + rm.Height / 3));
      pth.LineTo(rm.BottomRight);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctStackedPercentageBar:
    begin
      rb := RectF(rm.Left, rm.Top + rm.Height / 3 * 2, rm.Left + rm.Width / 4, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 6 * 5, rm.Left + rm.Width / 4 * 2, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 4 * 3, rm.Left + rm.Width / 4 * 3, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 2, rm.Left + rm.Width, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);

      c := ACanvas.Fill.Color;
      TAlphaColorRec(c).A := 150;
      ACanvas.Fill.Color := c;
      rb := RectF(rm.Left, rm.Top, rm.Left + rm.Width / 4, rm.Top + rm.Height / 3 * 2);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4, rm.Top, rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 6 * 5);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 2, rm.Top, rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 4 * 3);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 3, rm.Top, rm.Left + rm.Width, rm.Top + rm.Height / 2);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
    end;
    ctVariableRadiusPie, ctSizedPie:
    begin
      pth := TPathData.Create;
      pth.MoveTo(rm.CenterPoint);
      pth.AddArc(rm.CenterPoint, PointF(rm.Height / 2, rm.Height / 2), 0, 75);
      pth.LineTo(rm.CenterPoint);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);

      pth.Clear;
      pth.MoveTo(rm.CenterPoint);
      pth.AddArc(rm.CenterPoint, PointF(rm.Height / 3, rm.Height / 3), 75, 90);
      pth.LineTo(rm.CenterPoint);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);

      pth.Clear;
      pth.MoveTo(rm.CenterPoint);
      pth.AddArc(rm.CenterPoint, PointF(rm.Height / 2, rm.Height / 2), 165, 90);
      pth.LineTo(rm.CenterPoint);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctSpider:
    begin
      pth := TPathData.Create;
      pth.MoveTo(PointF(rm.CenterPoint.X + rm.Height / 3, rm.CenterPoint.Y));
      pth.LineTo(PointF(rm.CenterPoint.X + rm.Height / 4, rm.CenterPoint.Y - rm.Height / 2));
      pth.LineTo(PointF(rm.CenterPoint.X, rm.CenterPoint.Y - rm.Height / 3));
      pth.LineTo(PointF(rm.CenterPoint.X - rm.Height / 2, rm.CenterPoint.Y + rm.Height / 2));
      pth.LineTo(PointF(rm.CenterPoint.X, rm.CenterPoint.Y + rm.Height / 3));
      pth.LineTo(PointF(rm.CenterPoint.X + rm.Height / 2, rm.CenterPoint.Y + rm.Height / 3));
      pth.ClosePath;
      ACanvas.FillPath(pth, 0.3);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctPie:
    begin
      pth := TPathData.Create;
      pth.MoveTo(rm.CenterPoint);
      pth.AddArc(rm.CenterPoint, PointF(rm.Height / 2, rm.Height / 2), 0, 235);
      pth.LineTo(rm.CenterPoint);
      pth.ClosePath;
      ACanvas.FillPath(pth, 1);
      ACanvas.DrawPath(pth, 1);
      pth.Free;
    end;
    ctStackedBar:
    begin
      rb := RectF(rm.Left, rm.Top + rm.Height / 3 * 2, rm.Left + rm.Width / 4, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 6 * 5, rm.Left + rm.Width / 4 * 2, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 4 * 3, rm.Left + rm.Width / 4 * 3, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 2, rm.Left + rm.Width, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);

      c := ACanvas.Fill.Color;
      TAlphaColorRec(c).A := 150;
      ACanvas.Fill.Color := c;
      rb := RectF(rm.Left, rm.Top + rm.Height / 3, rm.Left + rm.Width / 4, rm.Top + rm.Height / 3 * 2);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 5, rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 6 * 5);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 4, rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 4 * 3);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 5, rm.Left + rm.Width, rm.Top + rm.Height / 2);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
    end;
    ctOHLC, ctCandleStick, ctBoxPlot:
    begin
      ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 6), PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 6 * 5), 1);
      case AChartType of
        ctOHLC:
        begin
          ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2 - rm.Width / 3, rm.Top + rm.Height / 3), PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 3), 1);
          ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2 + rm.Width / 3, rm.Top + rm.Height / 5 * 3), PointF(rm.Left + rm.Width / 2, rm.Top + rm.Height / 5 * 3), 1);
        end;
        ctCandleStick, ctBoxPlot:
        begin
          ACanvas.FillRect(RectF(rm.Left + rm.Width / 2 - rm.Width / 3, rm.Top + rm.Height / 3, rm.Left + rm.Width / 2 + rm.Width / 3, rm.Top + rm.Height / 5 * 3), 0, 0, AllCorners, 1);
          case AChartType of
            ctBoxPlot:
            begin
              ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2 - rm.Width / 3, rm.Top + rm.Height / 3), PointF(rm.Left + rm.Width / 2 + rm.Width / 3, rm.Top + rm.Height / 3), 1);
              ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2 - rm.Width / 3, rm.Top + rm.Height / 2), PointF(rm.Left + rm.Width / 2 + rm.Width / 3, rm.Top + rm.Height / 2), 1);
              ACanvas.DrawLine(PointF(rm.Left + rm.Width / 2 - rm.Width / 3, rm.Top + rm.Height / 5 * 3), PointF(rm.Left + rm.Width / 2 + rm.Width / 3, rm.Top + rm.Height / 5 * 3), 1);
            end;
          end;
        end;
      end;
    end;
    ctBar:
    begin
      rb := RectF(rm.Left, rm.Top + rm.Height / 4, rm.Left + rm.Width / 4, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4, rm.Top + rm.Height / 5 * 4, rm.Left + rm.Width / 4 * 2, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 2, rm.Top + rm.Height / 2, rm.Left + rm.Width / 4 * 3, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
      rb := RectF(rm.Left + rm.Width / 4 * 3, rm.Top + rm.Height / 4, rm.Left + rm.Width, rm.Bottom);
      ACanvas.FillRect(rb, 0, 0, AllCorners, 1);
      ACanvas.DrawRect(rb, 0, 0, AllCorners, 1);
    end;
  end;
end;

{$IFDEF FREEWARE}
function Scramble(s:string): string;
var
  r:string;
  i: integer;
  c: char;
  b: byte;
begin
  r := '';
  {$IFDEF ZEROSTRINGINDEX}
  for i := 0 to length(s) - 1 do
  {$ELSE}
  for i := 1 to length(s) do
  {$ENDIF}
  begin
    b := ord(s[i]);
    b := (b and $E0) + ((b and $1F) xor 5);
    c := chr(b);
    r := r + c;
  end;
  Result := r;
end;
{$ENDIF}

function Hiword(L: DWORD): integer;
begin
  Result := L shr 16;
end;

function LoWord(L: DWORD): Integer;
begin
  Result := L AND $FFFF;
end;

function MakeWord(b1,b2: integer): integer;
begin
  Result := b1 or b2 shl 8;
end;

function MakeLong(i1,i2: integer): integer;
begin
  Result := i1 or i2 shl 16;
end;

function MinMaxToUnit(min, max, maxlabels: Double): Double;
var
  i: integer;
  delta, d, r: Double;
begin
  i := 0;
  r := 0;

  delta := max - min;

  if (delta > 0) and (maxlabels > 0) then
  begin

    while abs(delta) < 1 do
    begin
      delta := delta * 10;
      inc(i);
    end;

    while abs(delta) > 10 do
    begin
      delta := delta / 10;
      dec(i);
    end;

    d := (delta) / maxlabels;

    if (d < 0.1) then
      r := 0.1;

    if (d > 0.1) and (d <= 0.2) then
      r := 0.2;

    if (d > 0.2) and (d <= 0.5) then
      r := 0.5;

    if (d > 0.5) then
      r := 1;

    // denormalize unit
    if i < 0 then
    begin
      while i < 0 do
      begin
        r := r * 10;
        inc(i);
      end;
    end
    else
      while i > 0 do
      begin
        r := r / 10;
        dec(i);
      end;

    Result := r;
  end
  else
  begin
    r := 0;
    result := r;
  end;
end;

function AnimateDouble(var Start, Stop: Double; Delta: Double; Margin: Double): Boolean;
begin
  Result := true;
  if (Start > Stop - Margin) and (Start < Stop + Margin) then
  begin
    Start := Stop;
    Result := false;
  end
  else
  begin
    Delta := Max(Margin, Delta);
    if Start < Stop then
      Start := Start + Delta
    else
      Start := Start - Delta;
  end;
end;

{ TTMSFMXChart }

function TTMSFMXChart.AddBandSerie(AStartX, AEndX: integer; AStartValue1,
  AStartValue2, AEndValue1, AEndValue2: Double; AUseCommonYRange: Boolean;
  ASerieLineColor: TAlphaColor; ASerieLineWidth: integer;
  ASerieBandColor: TAlphaColor): TTMSFMXChartSerie;
var
  d1,d2: Double;
  f, t: integer;
  I, K: integer;
begin
  Result := nil;
  if AStartX > AEndX then
  begin
    t := AStartX;
    f := AEndX;
  end
  else
  begin
    t := AEndX;
    f := AStartX;
  end;

  if t - f > 0 then
  begin
    BeginUpdate;
    d1 := (AEndValue1 - AStartValue1) / (t - f);
    d2 := (AEndValue2 - AStartValue2) / (t - f);

    Result := Series.Add;
    Result.YValues.Positions := [];
    Result.XValues.Positions := [];
    Result.ChartType := ctBand;
    Result.Stroke.Color := ASerieLineColor;
    Result.StrokeThickness := ASerieLineWidth;
    Result.Fill.Color := ASerieBandColor;
    Result.ShowInLegend := False;

    K := 0;
    for I := f to t - 1 do
    begin
      Result.AddSecondPoint(AStartValue1 + (K * d1), AStartValue2 + (K * d2)).XValue := I;
      Inc(K);
    end;

    for I := 0 to Series.Count - 1 do
    begin
      if AUseCommonYRange then
        Series[I].AutoYRange := arCommon;

      Series[I].AutoXRange := arCommon;
    end;

    EndUpdate;
  end;
end;

function TTMSFMXChart.AddLineSerie(AStartX, AEndX: Integer; AStartValue,
  AEndValue: Double; AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: Integer): TTMSFMXChartSerie;
var
  d: Double;
  f, t: integer;
  I, K: integer;
begin
  Result := nil;
  BeginUpdate;

  if AStartX > AEndX then
  begin
    t := AStartX;
    f := AEndX;
  end
  else
  begin
    t := AEndX;
    f := AStartX;
  end;

  if t - f > 0 then
  begin
    d := (AEndValue - AStartValue) / (t - f);
    Result := Series.Add;

    Result.YValues.Positions := [];
    Result.XValues.Positions := [];
    Result.ChartType := ctXYLine;
    Result.Stroke.Color := ASerieLineColor;
    Result.StrokeThickness := ASerieLineWidth;
    Result.ShowInLegend := False;

    //values from startvalue to endvalue with delta
    K := 0;
    for I := f to t do
    begin
      Result.AddXYPoint(I, AStartValue + (k * d));
      Inc(K);
    end;

    for I := 0 to Series.Count - 1 do
    begin
      if AUseCommonYRange then
        Series[I].AutoYRange := arCommon;

      Series[I].AutoXRange := arCommon;
    end;
  end;

  EndUpdate;
end;

function TTMSFMXChart.AddMovingAverage(ASerieIndex, AStartX, AEndX,
  AWindow: Integer; AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: Integer): TTMSFMXChartSerie;
var
  i: integer;
  startsum: double;
  res: double;
begin
  if AWindow <= 0 then
    raise Exception.Create('Moving average window size must be 1 or higher');

  BeginUpdate;

  startsum := 0;
  for i := AStartX to AStartX + AWindow - 1 do
    startsum := startsum + Series[ASerieIndex].GetPoint(i).YValue;

  Result := Series.Add;
  Result.YValues.Positions := [];
  Result.XValues.Positions := [];
  Result.ChartType := ctXYLine;
  Result.Stroke.Color := ASerieLineColor;
  Result.StrokeThickness := ASerieLineWidth;
  Result.ShowInLegend := False;

  //Moving averages
  for i := AStartX + AWindow to AEndX do
  begin
    res := startsum / Awindow;
    Result.AddXYPoint(I, res);
    if i < AEndX then
    begin
      startsum := startsum - Series[ASerieIndex].GetPoint(i - AWindow).YValue;
      startsum := startsum + Series[ASerieIndex].GetPoint(i).YValue;
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    if AUseCommonYRange then
      Series[I].AutoYRange := arCommon;

    Series[I].AutoXRange := arCommon;
  end;

  EndUpdate;
end;

function TTMSFMXChart.AddParetoLine(ASerieIndex, AStartx, AEndx: Integer;
  AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: Integer): TTMSFMXChartSerie;
var
  i: integer;
  sum, sumtotal: Double;
  f, t: integer;
begin
  if AStartx >= AEndx then
    raise Exception.Create('End point should be larger than start point');

  BeginUpdate;

  if AStartx > AEndx then
  begin
    t := AStartx;
    f := AEndx;
  end
  else
  begin
    t := AEndx;
    f := AStartx;
  end;

  Result := Series.Add;
  Result.YValues.Positions := [];
  Result.XValues.Positions := [];
  Result.ChartType := ctLine;
  Result.Stroke.Color := ASerieLineColor;
  Result.StrokeThickness := ASerieLineWidth;
  Result.ShowInLegend := False;
  Result.Labels.Visible := True;
  Result.Labels.Format := '%.1f%%';

  sumtotal := 0;
  for I := f to t do
    sumtotal := sumtotal + Series[ASerieIndex].GetPoint(I).YValue;

  if sumtotal > 0 then
  begin
    sum := 0;
    for I := f to t do
    begin
      sum := sum + (Series[ASerieIndex].GetPoint(I).YValue / sumtotal * 100);
      Result.AddXYPoint(I, sum)
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    if AUseCommonYRange then
      Series[I].AutoYRange := arCommon;

    Series[I].AutoXRange := arCommon;
  end;

  EndUpdate;
end;

procedure TTMSFMXChart.AddTrendChannel(ASerieIndex, AStartX, AEndX: Integer;
  AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: Integer);
var
  i: integer;
  count: integer;
  yAxisValuesSum : double;
  xAxisValuesSum : double;
  xxSum : double;
  xySum : double;
  slope: double;
  intercept: double;
  mean: double;
  dev: double;
begin
  if AStartX >= AEndX then
    raise Exception.Create('End point should be larger than start point');

  count := AEndX - AStartX;
  if count > 0 then
  begin
    BeginUpdate;
    yAxisValuesSum := 0;
    xAxisValuesSum := 0;
    xxSum := 0;
    xySum := 0;

    for i := AStartX to AEndX do
      yAxisValuesSum := yAxisValuesSum + Series[ASerieIndex].GetPoint(i).YValue;

    mean := yAxisValuesSum / count;
    dev := 0;

    for i := AStartX to AEndX do
    begin
      xAxisValuesSum := xAxisValuesSum + i;
      xySum := xysum + i * Series[ASerieIndex].GetPoint(i).YValue;
      xxSum := xxsum + i * i;
      dev := dev + sqr((mean - Series[ASerieIndex].GetPoint(i).YValue));
    end;

    dev := sqrt(dev / count);

    slope := (count * xySum - xAxisValuesSum * yAxisValuesSum) / ( (count * xxSum) - (xAxisValuesSum * yAxisValuesSum));
    intercept := (yAxisValuesSum - slope * xAxisValuesSum) / count;

    AddLineSerie(AStartX, AEndX, intercept + dev,count * slope + (intercept + dev), AUseCommonYRange, ASerieLineColor, ASerieLineWidth);
    AddLineSerie(AStartX, AEndX, intercept - dev,count * slope + (intercept - dev), AUseCommonYRange, ASerieLineColor, ASerieLineWidth);
    EndUpdate;
  end;
end;

function TTMSFMXChart.AddTrendChannelBand(ASerieIndex, AStartX,
  AEndX: integer; AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: integer; ASerieBandColor: TAlphaColor): TTMSFMXChartSerie;
var
  i: integer;
  count: integer;
  yAxisValuesSum : double;
  xAxisValuesSum : double;
  xxSum : double;
  xySum : double;
  slope: double;
  intercept: double;
  mean: double;
  dev: double;
begin
  Result := nil;
  if Astartx >= Aendx then
    raise Exception.Create('End point should be larger than start point');

  count := AEndX - AStartX;
  if count > 0 then
  begin
    BeginUpdate;

    yAxisValuesSum := 0;
    xAxisValuesSum := 0;
    xxSum := 0;
    xySum := 0;

    for i := AStartX to AEndX do
      yAxisValuesSum := yAxisValuesSum + Series[ASerieIndex].GetPoint(i).YValue;

    mean := yAxisValuesSum / count;
    dev := 0;

    for i := AStartX to AEndX do
    begin
      xAxisValuesSum := xAxisValuesSum + i;
      xySum := xysum + i * Series[ASerieIndex].GetPoint(i).YValue;
      xxSum := xxsum + i * i;
      dev := dev + sqr((mean - Series[ASerieIndex].GetPoint(i).YValue));
    end;

    dev := sqrt(dev / count);

    slope := (count * xySum - xAxisValuesSum * yAxisValuesSum) / ( (count * xxSum) - (xAxisValuesSum * yAxisValuesSum));
    intercept := (yAxisValuesSum - slope * xAxisValuesSum) / count;

    Result := AddBandSerie(AStartX, AEndX, intercept - dev,intercept + dev,count * slope + (intercept - dev),count * slope + (intercept + dev), AUseCommonYRange, ASerieLineColor, ASerieLineWidth, ASerieBandColor);
    EndUpdate;
  end;
end;

function TTMSFMXChart.AddTrendLine(ASerieIndex, AStartX, AEndX: Integer;
  AUseCommonYRange: Boolean; ASerieLineColor: TAlphaColor;
  ASerieLineWidth: Integer): TTMSFMXChartSerie;
var
  i: integer;
  count: integer;
  yAxisValuesSum : double;
  xAxisValuesSum : double;
  xxSum : double;
  xySum : double;
  slope: double;
  intercept: double;
begin
  Result := nil;
  if AStartX >= AEndX then
    raise Exception.Create('End point should be larger than start point');

  count := AEndX - AStartX + 1;
  if count > 0 then
  begin
    BeginUpdate;
    yAxisValuesSum := 0;
    xAxisValuesSum := 0;
    xxSum := 0;
    xySum := 0;

    for i := AStartX to AEndX do
    begin
      yAxisValuesSum := yAxisValuesSum + Series[ASerieIndex].GetPoint(i).YValue;
      xAxisValuesSum := xAxisValuesSum + I;
      xySum := xysum + i * Series[ASerieIndex].GetPoint(i).YValue;
      xxSum := xxsum + i * i;
    end;

    slope := (count * xySum - xAxisValuesSum * yAxisValuesSum) / ( (count *  xxSum) - (xAxisValuesSum * xAxisValuesSum));
    intercept := (yAxisValuesSum - slope * xAxisValuesSum) / count;

    Result := AddLineSerie(AStartX, AEndX, AStartX * slope + intercept, AEndX * slope + intercept, AUseCommonYRange, ASerieLineColor, ASerieLineWidth);
    EndUpdate;
  end;
end;

procedure TTMSFMXChart.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChart then
  begin
    FInteractionOptions.Assign((Source as TTMSFMXChart).InteractionOptions);
    FSeries.Assign((Source as TTMSFMXChart).Series);
    FTitle.Assign((Source as TTMSFMXChart).Title);
    FXAxis.Assign((Source as TTMSFMXChart).XAxis);
    FYAxis.Assign((Source as TTMSFMXChart).YAxis);
    FLegend.Assign((Source as TTMSFMXChart).Legend);
    FSeriesMargins.Assign((Source as TTMSFMXChart).SeriesMargins);
    FFillOpacity := (Source as TTMSFMXChart).FillOpacity;
    FFill.Assign((Source as TTMSFMXChart).Fill);
    FStroke.Assign((Source as TTMSFMXChart).Stroke);
    FStrokeOpacity := (Source as TTMSFMXChart).StrokeOpacity;
    FStrokeJoin := (Source as TTMSFMXChart).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChart).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChart).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChart).StrokeDash;
    FInteraction := (Source as TTMSFMXChart).Interaction;
  end;
end;

procedure TTMSFMXChart.BeginUpdate;
var
  I: Integer;
begin
  {$IF DEFINED(DELPHIXE7_LVL) AND NOT DEFINED(DELPHIXE8_LVL)}
  inherited BeginUpdate;
  {$ELSE}
  inherited;
  {$ENDIF}
  Inc(FUpdateCount);

  for I := 0 to Series.Count - 1 do
  begin
    Series[I].FSequential := True;
    Series[I].FPrevXValue := -MaxDouble;
  end;
end;

procedure TTMSFMXChart.Calculate3DOffset;
var
  I: Integer;
  s: TTMSFMXChartSerie;
begin
  if HasPieChart then
    Exit;

  FTotalOffset3DX := GetTotal3DOffsetX;
  FTotalOffset3DY := GetTotal3DOffsetY;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    s.F3DXOffset := s.Get3DOffsetX;
    s.F3DYOffset := s.Get3DOffsetY;
  end;
end;

procedure TTMSFMXChart.CalculateCommonRangeX;
var
  s: TTMSFMXChartSerie;
  I: Integer;
  pt: TTMSFMXChartPointVirtual;
  totminx, totmaxx: Double;
  J: Integer;
  l, e, m, cnt: Integer;
  fnd: Boolean;
begin
  if HasPieChart then
    Exit;

  totmaxx := -MaxDouble;
  totminx := MaxDouble;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      case Series[I].AutoXRange of
        arCommon:
        begin
          totmaxx := Max(totmaxx, s.FXMax);
          totminx := Min(totminx, s.FXMin);
        end;
        arCommonZeroBased:
        begin
          totmaxx := Max(totmaxx, s.FXMax);
        end;
      end;
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      case s.AutoXRange of
        arCommon:
        begin
          s.FXMax := totmaxx;
          s.FXMin := totminx;
        end;
        arCommonZeroBased:
        begin
          s.FXMax := totmaxx;
        end;
      end;
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    s.FStartX := 0;
    s.FEndX := s.GetPointsCount - 1;

    if s.Visible then
    begin
      if s.FSequential then
      begin
        l  := s.FStartX;
        e := s.FEndX;
        fnd  := False;

        while (l < e) do
        begin
          m := (l + e) div 2;

          if s.GetPoint(m).XValue = s.FXMin then
          begin
            fnd := True;
            l := e;
            s.FStartX := m;
          end
          else if (s.GetPoint(m).XValue > s.FXMin) or (s.GetPoint(m + 1).XValue > s.FXMin) then
            e := m - 1
          else
            l := m + 1;
        end;

        if not fnd then
          s.FStartX := l;

        s.FStartX := s.FStartX - 1;

        l  := s.FStartX;
        e := s.FEndX;
        fnd  := False;

        while (l < e) do
        begin
          m := (l + e) div 2;

          if s.GetPoint(m).XValue = s.FXMax then
          begin
            fnd  := True;
            l := e;
            s.FEndX := m;
          end
          else if (s.GetPoint(m).XValue > s.FXMax) and (s.GetPoint(m - 1).XValue > s.FXMax) then
            e := m - 1
          else
            l := m + 1;
        end;

        if not fnd then
          s.FEndX := l;

        s.FEndX := s.FEndX + 1;
      end
      else
      begin
        cnt := s.GetPointsCount;
        for J := 0 to cnt - 1 do
        begin
          pt := s.GetPoint(J);
          if s.FXMin >= pt.XValue then
            s.FStartX := J;
        end;
        for J := cnt - 1 downto 0 do
        begin
          pt := s.GetPoint(J);
          if s.FXMax <= pt.XValue then
            s.FEndX := J;
        end;
      end;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateCommonRangeY;
var
  s, st: TTMSFMXChartSerie;
  I, J, K: Integer;
  totminy, totmaxy: Double;
  cmax, stmax: Double;
  sta, stp: Integer;
  offsymin, offsymax: Double;
begin
  if HasPieChart and not HasSpiderChart then
    Exit;

  totmaxy := -MaxDouble;
  totminy := MaxDouble;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      case s.AutoYRange of
        arCommonZeroBased, arCommon:
        begin
          if not s.IsStacked then
          begin
            totmaxy := Max(totmaxy, s.FYMax);
            if s.AutoYRange <> arCommonZeroBased then
              totminy := Min(totminy, s.FYMin);
          end;
        end;
      end;
    end;
  end;

  cmax := -MaxDouble;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.IsStacked then
    begin
      case s.AutoYRange of
        arCommonZeroBased, arCommon:
        begin
          sta := s.GetStartCountPoint;
          stp := s.GetEndCountPoint;
          for J := sta to stp do
          begin
            stmax := 0;
            for K := 0 to Series.Count - 1 do
            begin
              st := Series[K];
              if st.Visible and st.IsStacked then
              begin
                if (st.AutoYRange in [arCommon, arCommonZeroBased]) then
                  if (st.GroupIndex = s.GroupIndex) and (J >= 0) and (J <= st.GetPointsCount - 1) then
                    stMax := stMax + st.GetPoint(J).YValue;
              end;
            end;

            if stmax > cmax then
              cmax := stmax;
          end;
        end;
      end;
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.IsStacked then
    begin
      case s.AutoYRange of
        arCommonZeroBased, arCommon:
        begin
          if (s.ChartType = ctStackedPercentageArea) or (s.ChartType = ctStackedPercentageBar) then
            s.FYMax := 100
          else
            s.FYMax := cmax;

          s.FYMin := 0;

          offsymin := (s.FMinYOffsetPercentage * Abs(s.FYMax - s.FYMin)) / 100;
          offsymax := (s.FMaxYOffsetPercentage * Abs(s.FYMax - s.FYMin)) / 100;

          s.FYMin := s.FYMin - offsymin;
          s.FYMax := s.FYMax + offsymax;

          totmaxy := Max(totmaxy, s.FYMax);
          totminy := Min(totminy, s.FYMin);
        end;
      end;
    end;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      case s.AutoYRange of
        arCommon, arCommonZeroBased:
        begin
          s.FYMax := totmaxy;
          if s.AutoYRange <> arCommonZeroBased then
            s.FYMin := totminy;
        end;
      end;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateSeries;
var
  I: Integer;
begin
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Visible then
      Series[I].CalculateDrawPoints;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Visible then
      Series[I].CalculateAnimation;
  end;

  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Visible then
    begin
      Series[I].CalculateArea;
      Series[I].CalculateAnnotations;
      Series[I].CalculateLabels;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateMinMaxY;
var
  I: Integer;
begin
  if HasPieChart and not HasSpiderChart then
    Exit;

  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Visible then
      Series[I].CalculateMinMaxY;
  end;
end;

procedure TTMSFMXChart.CalculateMinMaxX;
var
  I: Integer;
begin
  if HasPieChart then
    Exit;

  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Visible then
      Series[I].CalculateMinMaxX;
  end;
end;

procedure TTMSFMXChart.CalculateXAxisHeight;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  bmp: TBitmap;
  majth, minth, maxh: Double;
  str: String;
  mn, mx: Double;
  mu, mi: Double;
  JMax, JMin: Double;
  domaj, domin: Boolean;
  domajv, dominv: Boolean;
  JMaxStr, JMinStr: String;
  J: Double;
  th: Double;
begin
  if HasPieChart then
    Exit;

  XAxis.FTopHeight := 0;
  XAxis.FCenterHeight := 0;
  XAxis.FBottomHeight := 0;

  bmp := TBitmap.Create(1, 1);
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and (s.XValues.Positions <> []) then
    begin
      maxh := 0;
      if s.XValues.Angle <> 0 then
      begin
        mn := s.FXMin;
        mx := s.FXMax;

        mu := s.XValues.MajorUnit;
        mi := s.XValues.MinorUnit;

        JMax := mn;
        JMin := mn;

        domaj := (mu > 0);
        domin := (mi > 0);

        if domin then
          Jmin := Round(mn / mi)  * mi;

        if domaj then
          Jmax := Round(mn / mu)  * mu;

        if (domaj) then
        begin
          while (JMin <= mx + mu) do
          begin
            domajv := JMax <= mx + mu;
            if domajv then
            begin
              bmp.Canvas.Font.Assign(s.XValues.MajorUnitFont);
              str := s.FindXAxisTextForValue(JMax);
              if str = '' then
              begin
                if s.XValues.MajorUnitFormat <> '' then
                begin
                  case s.XValues.MajorUnitFormatType of
                    vftNormal: str := Format(s.XValues.MajorUnitFormat,[Jmax]);
                    vftFloat: str := FormatFloat(s.XValues.MajorUnitFormat,Jmax);
                    vftDateTime: str := FormatDateTime(s.XValues.MajorUnitFormat, JMax);
                  end;
                end
                else
                  str := FloatToStr(JMax);
              end;

              JMaxStr := str;
              DoGetSerieXValue(s, -1, vkMajor, JMax, JMaxStr);
              th := bmp.Canvas.TextWidth(JMaxStr);
              th := th + s.XValues.MajorUnitTickMarkSize + s.XValues.MajorUnitSpacing + 4;
              if th > maxh then
                maxh := th;
            end;

            dominv := JMin <= mx + mu;
            if dominv and domin then
            begin
              while (CompareValue(JMin, JMax) <> EqualsValue) and (JMax > jmin) do
              begin
                j := JMin;
                bmp.Canvas.Font.Assign(s.XValues.MinorUnitFont);
                str := s.FindXAxisTextForValue(j);
                if str = '' then
                begin
                  if s.XValues.MinorUnitFormat <> '' then
                  begin
                    case s.XValues.MinorUnitFormatType of
                      vftNormal: str := Format(s.XValues.MinorUnitFormat,[j]);
                      vftFloat: str := FormatFloat(s.XValues.MinorUnitFormat, j);
                      vftDateTime: str := FormatDateTime(s.XValues.MinorUnitFormat, j);
                    end;
                  end
                  else
                    str := FloatToStr(j);
                end;

                if CompareValue(Jmin, JMax) <> EqualsValue then
                begin
                  JMinstr := str;
                  DoGetSerieXValue(s, -1, vkMinor, j, JMinStr);
                  th := bmp.Canvas.TextWidth(JMinStr);
                  th := th + s.XValues.MinorUnitTickMarkSize + s.XValues.MinorUnitSpacing + 4;
                  if th > maxh then
                    maxh := th;
                end;

                JMin := JMin + mi;

                if CompareValue(Jmin, Jmax) = 0 then
                  JMin := JMin + mi;
              end;
            end;

            if not domin then
            begin
              JMin := JMin + mu;
            end
            else
            begin
              if CompareValue(Jmin, Jmax) = 0 then
              begin
                JMin := JMin + mi;
              end;
            end;

            Jmax := JMax + mu;
            if (JMax > mx + mu) then
              Break;
          end;
        end;
      end
      else
      begin
        bmp.Canvas.Font.Assign(s.XValues.MajorUnitFont);
        majth := bmp.Canvas.TextHeight('gh');
        majth := majth + s.XValues.MajorUnitTickMarkSize + s.XValues.MajorUnitSpacing + 4;
        minth := 0;
        if s.XValues.MinorUnit > 0 then
        begin
          bmp.Canvas.Font.Assign(s.XValues.MinorUnitFont);
          minth := bmp.Canvas.TextHeight('gh');
          minth := minth + s.XValues.MinorUnitTickMarkSize + s.XValues.MinorUnitSpacing + 4;
        end;
        maxh := max(majth, minth);
      end;

      if s.XValues.Title.Text <> '' then
      begin
        bmp.Canvas.Font.Assign(s.XValues.Title.Font);
        maxh := maxh + bmp.Canvas.TextHeight(s.XValues.Title.Text) + 4 + s.XValues.Title.TextMargins.Top + s.XValues.Title.TextMargins.Bottom
      end;

      if (xpTop in XAxis.Positions) and (xpTop in s.XValues.Positions) then
      begin
        XAxis.FTopHeight := XAxis.FTopHeight + maxh;
        s.FXAxisTopHeight := XAxis.FTopHeight;
      end;

      if (xpCenter in XAxis.Positions) and (xpCenter in s.XValues.Positions) then
      begin
        XAxis.FCenterHeight := XAxis.FCenterHeight + maxh;
        s.FXAxisCenterHeight := XAxis.FCenterHeight;
      end;

      if (xpBottom in XAxis.Positions) and (xpBottom in s.XValues.Positions) then
      begin
        XAxis.FBottomHeight := XAxis.FBottomHeight + maxh;
        s.FXAxisBottomHeight := XAxis.FBottomHeight;
      end;
    end;
  end;
  bmp.Free;
end;

procedure TTMSFMXChart.CalculateXScale;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  c: Double;
  r: TRectF;
begin
  if HasPieChart then
    Exit;

  r := GetSeriesRectangle;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      s.FXScale := r.Width;
      c := Abs(s.FXMax - s.FXMin);
      if c > 0 then
      begin
        if s.CalculateOffset then
        begin
          if c > 1 then
            s.FXScale := r.Width / (c + 1)
          else
          begin
            if s.GetPointsCount > 1 then
              s.FXScale := r.Width / (c + (s.GetPoint(1).XValue - s.GetPoint(0).XValue))
          end;
        end
        else
          s.FXScale := r.Width / c;
      end;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateXAxis;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  va: Double;
  bmp: TBitmap;
  r: TRectF;
  str: string;
  mn, mx: Double;
  htpane: Double;
  httext: Double;
  mu, mi: Double;
  JMax, JMin: Double;
  domaj, domin, domajv, dominv: Boolean;
  JMaxStr, JMinStr: String;
  j: Double;
  dv: TTMSFMXChartDrawXYValue;
  dgl: TTMSFMXChartDrawXYGridLine;
  maxw, tw: Double;
  offsxt, offsxc, offsxb: Double;
  vl: Double;
  xrt, xrc, xrb: TRectF;
  K: Integer;
  strmax, strmin: String;
  a: Double;
  cy, ypos: Double;
  thMax, thMin: Double;
begin
  if HasPieChart then
    Exit;

  offsxt := 0;
  offsxc := 0;
  offsxb := 0;

  bmp := TBitmap.Create(1, 1);
  r := GetSeriesRectangle;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    s.FDrawXValues.Clear;
    s.FDrawXGridLines.Clear;
    if s.Visible and (s.XValues.Positions <> []) then
    begin
      mn := s.FXMin;
      mx := s.FXMax;

      bmp.Canvas.Font.Assign(s.XValues.MajorUnitFont);
      thMax := bmp.Canvas.TextHeight('gh');
      bmp.Canvas.Font.Assign(s.XValues.MinorUnitFont);
      thMin := bmp.Canvas.TextHeight('gh');

      if s.XValues.AutoUnits then
      begin
        htpane := r.Width;

        if s.XValues.MajorUnitFormat <> '' then
        begin
          case s.XValues.MajorUnitFormatType of
            vftNormal: strmin := Format(s.XValues.MajorUnitFormat,[mn]);
            vftFloat: strmin := FormatFloat(s.XValues.MajorUnitFormat,mn);
            vftDateTime: str := FormatDateTime(s.XValues.MajorUnitFormat, mn);
          end;
        end
        else
          strmin := FloatToStr(mn);

        if s.XValues.MajorUnitFormat <> '' then
        begin
          case s.XValues.MajorUnitFormatType of
            vftNormal: strmax := Format(s.XValues.MajorUnitFormat,[mx]);
            vftFloat: strmax := FormatFloat(s.XValues.MajorUnitFormat,mx);
            vftDateTime: str := FormatDateTime(s.XValues.MajorUnitFormat, mx);
          end;
        end
        else
          strmax := FloatToStr(mx);

        bmp.Canvas.Font.Assign(s.XValues.MajorUnitFont);
        httext := Max(bmp.Canvas.TextWidth(strmax), bmp.Canvas.TextWidth(strmin));
        mu := 0;
        if httext > 0 then
          mu := MinMaxToUnit(mn, mx, htpane / httext);

        bmp.Canvas.Font.Assign(s.XValues.MinorUnitFont);
        httext := Max(bmp.Canvas.TextWidth(strmax), bmp.Canvas.TextWidth(strmin));
        mi := 0;
        if httext > 0 then
          mi := MinMaxToUnit(mn, mx, htpane / httext);
      end
      else
      begin
        mu := s.XValues.MajorUnit;
        mi := s.XValues.MinorUnit;
      end;

      JMax := mn;
      JMin := mn;
      maxw := 0;

      domaj := (mu > 0);
      domin := (mi > 0);

      if domin then
        Jmin := Round(mn / mi)  * mi;

      if domaj then
        Jmax := Round(mn / mu)  * mu;

      if (domaj) then
      begin
        while (JMin <= mx + mu) do
        begin
          domajv := JMax <= mx + mu;
          if domajv then
          begin
            bmp.Canvas.Font.Assign(s.XValues.MajorUnitFont);
            str := s.FindXAxisTextForValue(JMax);
            if str = '' then
            begin
              if s.XValues.MajorUnitFormat <> '' then
              begin
                case s.XValues.MajorUnitFormatType of
                  vftNormal: str := Format(s.XValues.MajorUnitFormat,[Jmax]);
                  vftFloat: str := FormatFloat(s.XValues.MajorUnitFormat,Jmax);
                  vftDateTime: str := FormatDateTime(s.XValues.MajorUnitFormat, JMax);
                end;
              end
              else
                str := FloatToStr(JMax);
            end;

            JMaxStr := str;
            DoGetSerieXValue(s, s.FDrawXValues.Count, vkMajor, JMax, JMaxStr);
            dv.ValueString := JMaxStr;
            dv.Value := JMax;
            dv.Kind := vkMajor;
            vl := s.ValueToX(JMax) - s.F3DXOffset;
            dv.ValuePositionXTop.X := vl + FTotalOffset3DX;
            dv.ValuePositionXCenter.X := vl;
            dv.ValuePositionXBottom.X := vl;
            dv.ValuePositionXTop.Y := -offsxt;
            dv.ValuePositionXCenter.Y := offsxc;
            dv.ValuePositionXBottom.Y := offsxb;
            dgl.ValueGridLine := False;
            dgl.Value := jMax;
            dgl.Kind := vkMajor;
            dgl.XStartPoint.X := vl;
            dgl.XEndPoint.X := vl;

            dv.TextWidth := bmp.Canvas.TextWidth(JMaxStr) + 1;
            dv.TextHeight := thmax;

            if s.XValues.Angle <> 0 then
              tw := dv.TextWidth + s.XValues.MajorUnitTickMarkSize + s.XValues.MajorUnitSpacing + 3
            else
              tw := dv.TextHeight + s.XValues.MajorUnitTickMarkSize + s.XValues.MajorUnitSpacing + 3;

            if tw > maxw then
              maxw := tw;
            if (vl >= r.Left - 1) and (vl <= r.Right + 1) then
            begin
              dv.Index := s.FDrawXValues.Count;
              s.FDrawXValues.Add(dv);
              dgl.Index := s.FDrawXGridLines.Count;
              s.FDrawXGridLines.Add(dgl);
            end;
          end;

          dominv := JMin <= mx + mu;
          if dominv and domin then
          begin
            while (CompareValue(Jmin, JMax) <> EqualsValue) and (JMax > jmin) do
            begin
              j := JMin;
              bmp.Canvas.Font.Assign(s.XValues.MinorUnitFont);
              str := s.FindXAxisTextForValue(j);
              if str = '' then
              begin
                if s.XValues.MinorUnitFormat <> '' then
                begin
                  case s.XValues.MinorUnitFormatType of
                    vftNormal: str := Format(s.XValues.MinorUnitFormat,[j]);
                    vftFloat: str := FormatFloat(s.XValues.MinorUnitFormat, j);
                    vftDateTime: str := FormatDateTime(s.XValues.MinorUnitFormat, j);
                  end;
                end
                else
                  str := FloatToStr(j);
              end;

              if CompareValue(Jmin, JMax) <> EqualsValue then
              begin
                JMinstr := str;
                DoGetSerieXValue(s, s.FDrawXValues.Count, vkMinor, j, JMinStr);
                dv.ValueString := JMinStr;
                dv.Value := j;
                dv.Kind := vkMinor;
                vl := S.ValueToX(j) - s.F3DXOffset;
                dv.ValuePositionXTop.X := vl + FTotalOffset3DX;
                dv.ValuePositionXCenter.X := vl;
                dv.ValuePositionXBottom.X := vl;
                dv.ValuePositionXTop.Y := -offsxt;
                dv.ValuePositionXCenter.Y := offsxc;
                dv.ValuePositionXBottom.Y := offsxb;
                dgl.ValueGridLine := False;
                dgl.Value := j;
                dgl.Kind := vkMinor;
                dgl.XStartPoint.X := vl;
                dgl.XEndPoint.X := vl;
                dv.TextWidth := bmp.Canvas.TextWidth(JMinStr) + 1;
                dv.TextHeight := thMin;
                if s.XValues.Angle <> 0 then
                  tw := dv.TextWidth + s.XValues.MinorUnitTickMarkSize + s.XValues.MinorUnitSpacing + 3
                else
                  tw := dv.TextHeight + s.XValues.MinorUnitTickMarkSize + s.XValues.MinorUnitSpacing + 3;

                if tw > maxw then
                  maxw := tw;

                if (vl >= R.Left - 1) and (vl <= r.Right + 1) then
                begin
                  dv.Index := s.FDrawXValues.Count;
                  s.FDrawXValues.Add(dv);
                  dgl.Index := s.FDrawXGridLines.Count;
                  s.FDrawXGridLines.Add(dgl);
                end;
              end;

              JMin := JMin + mi;

              if CompareValue(Jmin, Jmax) = 0 then
                JMin := JMin + mi;
            end;
          end;

          if not domin then
          begin
            JMin := JMin + mu;
          end
          else
          begin
            if CompareValue(Jmin, Jmax) = 0 then
            begin
              JMin := JMin + mi;
            end;
          end;

          Jmax := JMax + mu;
          if (JMax > mx + mu) then
            Break;
        end;
      end;

      if s.XValues.Title.Text <> '' then
      begin
        bmp.Canvas.Font.Assign(s.XValues.Title.Font);
        maxw := maxw + bmp.Canvas.TextHeight(s.XValues.Title.Text) + 4 + s.XValues.Title.TextMargins.Top + s.XValues.Title.TextMargins.Bottom
      end;

      if (xpTop in XAxis.Positions) and (xpTop in s.XValues.Positions) then
        offsxt := offsxt + maxw;

      if (xpCenter in XAxis.Positions) and (xpCenter in s.XValues.Positions) then
        offsxc := offsxc + maxw;

      if (xpBottom in XAxis.Positions) and (xpBottom in s.XValues.Positions) then
        offsxb := offsxb + maxw;
    end;
  end;

  bmp.Free;

  xrt := XAxis.GetRectangle(xpTop);
  xrc := XAxis.GetRectangle(xpCenter);
  xrb := XAxis.GetRectangle(xpBottom);
  r := GetSeriesRectangle;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      a := DegToRad(s.XValues.Angle);
      va := Sin(a);
      for K := 0 to s.FDrawXValues.Count - 1 do
      begin
        dv := s.FDrawXValues[K];
        dv.ValuePositionXTop.Y := dv.ValuePositionXTop.Y + xrt.Bottom;
        dv.ValuePositionXCenter.Y := dv.ValuePositionXCenter.Y + xrc.CenterPoint.Y;
        dv.ValuePositionXBottom.Y := dv.ValuePositionXBottom.Y + xrb.Top;

        dv.TextPositionXTop := dv.ValuePositionXTop;
        dv.TextPositionXCenter := dv.ValuePositionXCenter;
        dv.TextPositionXBottom := dv.ValuePositionXBottom;

        cy := dv.TextPositionXBottom.Y + dv.TextHeight / 2;
        ypos := cy + (dv.TextWidth / 2) * va;
        dv.TextPositionXBottom.Y := dv.TextPositionXBottom.Y + Abs(ypos - cy) - Abs((dv.TextHeight / 2) * va);

        cy := dv.TextPositionXCenter.Y + dv.TextHeight / 2;
        ypos := cy + (dv.TextWidth / 2) * sin(a);
        dv.TextPositionXCenter.Y := dv.TextPositionXCenter.Y + Abs(ypos - cy) - Abs((dv.TextHeight / 2) * va);

        cy := dv.TextPositionXTop.Y + dv.TextHeight / 2;
        ypos := cy + (dv.TextWidth / 2) * va;
        dv.TextPositionXTop.Y := dv.TextPositionXTop.Y - Abs(ypos - cy) + Abs((dv.TextHeight / 2) * va);

        s.FDrawXValues[K] := dv;
      end;

      for K := 0 to s.FDrawXGridLines.Count - 1 do
      begin
        dgl := s.FDrawXGridLines[K];
        if s.XGrid.Extended then
        begin
          dgl.XStartPoint.Y := xrt.Bottom;
          dgl.XEndPoint.Y := xrb.Top;
        end
        else
        begin
          dgl.XStartPoint.Y := r.Top;
          dgl.XEndPoint.Y := r.Bottom;
        end;

        dgl.XStartPoint.X := dgl.XStartPoint.X + FTotalOffset3DX;
        dgl.XEndPoint.X := dgl.XEndPoint.X + FTotalOffset3DX;
        dgl.XEndPoint.Y := dgl.XEndPoint.Y - FTotalOffset3DY;
        s.FDrawXGridLines[K] := dgl;
      end;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateYAxis;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  bmp: TBitmap;
  r: TRectF;
  str: string;
  min, max: Double;
  htpane: Double;
  mu, mi: Double;
  JMax, JMin: Double;
  domaj, domin, domajv, dominv: Boolean;
  JMaxStr, JMinStr: String;
  j: Double;
  dv: TTMSFMXChartDrawXYValue;
  dgl: TTMSFMXChartDrawXYGridLine;
  maxw, tw: Double;
  offsyl, offsyc, offsyr: Double;
  vl: Double;
  yrl, yrc, yrr, pr: TRectF;
  K: Integer;
  thMax, thMin: Double;
begin
  if HasPieChart and not HasSpiderChart then
    Exit;

  YAxis.FLeftWidth := 0;
  YAxis.FCenterWidth := 0;
  YAxis.FRightWidth := 0;
  offsyl := 0;
  offsyc := 0;
  offsyr := 0;

  bmp := TBitmap.Create(1, 1);
  r := GetSeriesRectangle;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    s.FDrawYValues.Clear;
    s.FDrawYGridLines.Clear;
    if s.Visible and ((s.YValues.Positions <> []) or s.IsSpider) then
    begin
      if s.IsSpider then
        min := 0
      else
        min := s.FYMin;

      max := s.FYMax;

      bmp.Canvas.Font.Assign(s.YValues.MajorUnitFont);
      thMax := bmp.Canvas.TextHeight('gh');
      bmp.Canvas.Font.Assign(s.YValues.MinorUnitFont);
      thMin := bmp.Canvas.TextHeight('gh');
      pr := s.GetPieRect(False);

      if s.YValues.AutoUnits then
      begin
        if s.IsSpider then
          htpane := pr.Height
        else
          htpane := r.Height;

        mu := 0;
        if thMax > 0 then
          mu := MinMaxToUnit(min, max, htpane / thMax);

        mi := 0;
        if thMin > 0 then
          mi := MinMaxToUnit(min, max, htpane / thMin);
      end
      else
      begin
        mu := s.YValues.MajorUnit;
        mi := s.YValues.MinorUnit;
      end;

      JMax := min;
      JMin := min;
      maxw := 0;

      domaj := (mu > 0);
      domin := (mi > 0);

      if not s.IsSpider then
      begin
        if domin then
          Jmin := Round(min / mi)  * mi;

        if domaj then
          Jmax := Round(min / mu)  * mu;
      end;

      if (domaj) then
      begin
        while (JMin <= max + mu) do
        begin
          domajv := (JMax <= max + mu) or s.IsSpider;
          if domajv then
          begin
            bmp.Canvas.Font.Assign(s.YValues.MajorUnitFont);
            if s.YValues.MajorUnitFormat <> '' then
            begin
              case s.YValues.MajorUnitFormatType of
                vftNormal: str := Format(s.YValues.MajorUnitFormat,[Jmax]);
                vftFloat: str := FormatFloat(s.YValues.MajorUnitFormat,Jmax);
                vftDateTime: str := FormatDateTime(s.YValues.MajorUnitFormat, JMax);
              end;
            end
            else
              str := FloatToStr(JMax);
         
            JMaxStr := str;
            DoGetSerieYValue(s, s.FDrawYValues.Count, vkMajor, JMax, JMaxStr);
            dv.ValueString := JMaxStr;
            dv.Value := JMax;
            dv.Kind := vkMajor;
            vl := s.ValueToY(JMax) + s.F3DYOffset;
            dv.ValuePositionYLeft.Y := vl;
            dv.ValuePositionYCenter.Y := vl;
            dv.ValuePositionYRight.Y := vl - FTotalOffset3DY;
            dv.ValuePositionYLeft.X := -offsyl;
            dv.ValuePositionYCenter.X := offsyc;
            dv.ValuePositionYRight.X := offsyr;
            dv.ValueRadius := PointF(pr.Height / max * JMax / 2, pr.Height / max * JMax / 2);
            dgl.ValueGridLine := False;
            dgl.Value := JMax;
            dgl.Kind := vkMajor;
            dgl.YStartPoint.Y := vl;
            dgl.YEndPoint.Y := vl;
            dgl.ValueRadius := dv.ValueRadius;

            dv.TextWidth := bmp.Canvas.TextWidth(JMaxStr) + 1;
            dv.TextHeight := thMax;
            tw := dv.TextWidth + s.YValues.MajorUnitTickMarkSize + s.YValues.MajorUnitSpacing + 3;
            if tw > maxw then
              maxw := tw;

            if ((vl >= r.Top - 1) and (vl <= r.Bottom + 1)) or s.IsSpider then
            begin
              dv.Index := s.FDrawYValues.Count;
              s.FDrawYValues.Add(dv);
              dgl.Index := s.FDrawYGridLines.Count;
              s.FDrawYGridLines.Add(dgl);
            end;
          end;

          dominv := (JMin <= max + mu) or s.IsSpider;
          if dominv and domin then
          begin
            while (CompareValue(JMin, JMax) <> EqualsValue) and (JMax > jmin) do
            begin
              j := JMin;
              bmp.Canvas.Font.Assign(s.YValues.MinorUnitFont);
              if s.YValues.MinorUnitFormat <> '' then
              begin
                case s.YValues.MinorUnitFormatType of
                  vftNormal: str := Format(s.YValues.MinorUnitFormat,[j]);
                  vftFloat: str := FormatFloat(s.YValues.MinorUnitFormat, j);
                  vftDateTime: str := FormatDateTime(s.YValues.MinorUnitFormat, j);
                end;
              end
              else
                str := FloatToStr(j);

              if CompareValue(Jmin, JMax) <> EqualsValue then
              begin
                JMinstr := str;
                DoGetSerieYValue(s, s.FDrawYValues.Count, vkMinor, j, JMinStr);
                dv.ValueString := JMinStr;
                dv.Value := j;
                dv.Kind := vkMinor;
                vl := S.ValueToY(j) + s.F3DYOffset;
                dv.ValuePositionYLeft.Y := vl;
                dv.ValuePositionYCenter.Y := vl;
                dv.ValuePositionYRight.Y := vl - FTotalOffset3DY;
                dv.ValuePositionYLeft.X := -offsyl;
                dv.ValuePositionYCenter.X := offsyc;
                dv.ValuePositionYRight.X := offsyr;
                dv.ValueRadius := PointF(pr.Height / max * j / 2, pr.Height / max * j / 2);
                dgl.ValueGridLine := False;
                dgl.Value := j;
                dgl.Kind := vkMinor;
                dgl.YStartPoint.Y := vl;
                dgl.YEndPoint.Y := vl;
                dgl.ValueRadius := dv.ValueRadius;

                dv.TextWidth := bmp.Canvas.TextWidth(JMinStr) + 1;
                dv.TextHeight := thMin;
                tw := dv.TextWidth + s.YValues.MinorUnitTickMarkSize + s.YValues.MinorUnitSpacing + 3;
                if tw > maxw then
                  maxw := tw;

                if ((vl >= R.Top - 1) and (vl <= r.Bottom + 1)) or s.IsSpider then
                begin
                  dv.Index := s.FDrawYValues.Count;
                  s.FDrawYValues.Add(dv);
                  dgl.Index := s.FDrawYGridLines.Count;
                  s.FDrawYGridLines.Add(dgl);
                end;
              end;

              JMin := JMin + mi;

              if CompareValue(Jmin, Jmax) = 0 then
                JMin := JMin + mi;
            end;
          end;

          if not domin then
          begin
            JMin := JMin + mu;
          end
          else
          begin
            if CompareValue(Jmin, Jmax) = 0 then
              JMin := JMin + mi;
          end;

          Jmax := JMax + mu;
          if (JMax > max + mu) then
            Break;
        end;
      end;

      if s.YValues.Title.Text <> '' then
      begin
        bmp.Canvas.Font.Assign(s.YValues.Title.Font);
        maxw := maxw + bmp.Canvas.TextHeight(s.YValues.Title.Text) + 4 + s.YValues.Title.TextMargins.Left + s.YValues.Title.TextMargins.Right
      end;

      if (ypLeft in YAxis.Positions) and (ypLeft in s.YValues.Positions) then
      begin
        YAxis.FLeftWidth := YAxis.FLeftWidth + maxw;
        offsyl := offsyl + maxw;
        s.FYAxisLeftWidth := YAxis.FLeftWidth;
      end;

      if (ypCenter in YAxis.Positions) and (ypCenter in s.YValues.Positions) then
      begin
        YAxis.FCenterWidth := YAxis.FCenterWidth + maxw;
        offsyc := offsyc + maxw;
        s.FYAxisCenterWidth := YAxis.FCenterWidth;
      end;

      if (ypRight in YAxis.Positions) and (ypRight in s.YValues.Positions) then
      begin
        YAxis.FRightWidth := YAxis.FRightWidth + maxw;
        offsyr := offsyr + maxw;
        s.FYAxisRightWidth := YAxis.FRightWidth;
      end;
    end;
  end;

  bmp.Free;

  yrl := YAxis.GetRectangle(ypLeft);
  yrc := YAxis.GetRectangle(ypCenter);
  yrr := YAxis.GetRectangle(ypRight);
  r := GetSeriesRectangle;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      for K := 0 to s.FDrawYValues.Count - 1 do
      begin
        dv := s.FDrawYValues[K];
        dv.ValuePositionYLeft.X := dv.ValuePositionYLeft.X + yrl.Right;
        dv.ValuePositionYCenter.X := dv.ValuePositionYCenter.X + yrc.CenterPoint.X;
        dv.ValuePositionYRight.X := dv.ValuePositionYRight.X + yrr.Left;
        dv.TextPositionYLeft := dv.ValuePositionYLeft;
        dv.TextPositionYCenter := dv.ValuePositionYCenter;
        dv.TextPositionYRight := dv.ValuePositionYRight;
        s.FDrawYValues[K] := dv;
      end;

      for K := 0 to s.FDrawYGridLines.Count - 1 do
      begin
        dgl := s.FDrawYGridLines[K];
        if s.YGrid.Extended then
        begin
          dgl.YStartPoint.X := yrl.Right;
          dgl.YEndPoint.X := yrr.Left;
        end
        else
        begin
          dgl.YStartPoint.X := r.Left;
          dgl.YEndPoint.X := r.Right;
        end;

        dgl.YStartPoint.Y := dgl.YStartPoint.Y - FTotalOffset3DY;
        dgl.YEndPoint.Y := dgl.YEndPoint.Y - FTotalOffset3DY;
        dgl.YStartPoint.X := dgl.YStartPoint.X + FTotalOffset3DX;

        s.FDrawYGridLines[K] := dgl;
      end;
    end;
  end;
end;

procedure TTMSFMXChart.CalculateYScale;
var
  s: TTMSFMXChartSerie;
  r: TRectF;
  I: Integer;
  c: Double;
begin
  if HasPieChart then
    Exit;

  r := GetSeriesRectangle;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      c := (s.FYMax - s.FYMin);
      if c > 0 then
        s.FYScale := r.Height / c
      else
        s.FYScale := r.Height;
    end;
  end;
end;

function TTMSFMXChart.CanLoadProperty(AObject: TObject; APropertyName: string;
  APropertyType: TTypeKind): Boolean;
begin
  if AObject = Self then
    Result := IndexText(APropertyName, ExcludePropertyList) = -1
  else
    Result := True;

  DoCanLoadProperty(AObject, APropertyName, APropertyType, Result);
end;

function TTMSFMXChart.CanSaveProperty(AObject: TObject; APropertyName: string;
  APropertyType: TTypeKind): Boolean;
begin
  if AObject = Self then
    Result := IndexText(APropertyName, ExcludePropertyList) = -1
  else
    Result := True;

  DoCanSaveProperty(AObject, APropertyName, APropertyType, Result);
end;

procedure TTMSFMXChart.DoCanSaveProperty(AObject: TObject; APropertyName: string; APropertyType: TTypeKind; var ACanSave: Boolean);
begin
  if Assigned(OnCanSaveProperty) then
    OnCanSaveProperty(Self, AObject, APropertyName, APropertyType, ACanSave);
end;

procedure TTMSFMXChart.DoCanLoadProperty(AObject: TObject; APropertyName: string; APropertyType: TTypeKind; var ACanLoad: Boolean);
begin
  if Assigned(OnCanLoadProperty) then
    OnCanLoadProperty(Self, AObject, APropertyName, APropertyType, ACanLoad);
end;

procedure TTMSFMXChart.Clear;
begin
  Series.Clear;
end;

procedure TTMSFMXChart.CMGesture(var EventInfo: TGestureEventInfo);
var
  p: Double;
  {$IFDEF MSWINDOWS}
  px, py: Double;
  {$ENDIF}
begin
  if not Interaction then
    Exit;

  if EventInfo.GestureID = igiZoom then
  begin
    if TInteractiveGestureFlag.gfBegin in EventInfo.Flags then
    begin
      FTouchDistance := EventInfo.Distance;
    end
    else
    begin
      if EventInfo.Flags = [] then
      begin
        p := (EventInfo.Distance - FTouchDistance);
        UpdateRange(p, p, True);
      end;
      FTouchDistance := EventInfo.Distance;
    end;
  end
  {$IFDEF MSWINDOWS}
  else if EventInfo.GestureID = igiPan then
  begin
    if TInteractiveGestureFlag.gfBegin in EventInfo.Flags then
    begin
      FTouchPoint := EventInfo.Location;
    end
    else
    begin
      if EventInfo.Flags = [] then
      begin
        px := FTouchPoint.X - EventInfo.Location.X;
        py := FTouchPoint.Y - EventInfo.Location.Y;
        UpdateRange(px, py, False);
      end;
      FTouchPoint := EventInfo.Location;
    end;
  end
  {$ENDIF}
  else
    inherited;
end;

constructor TTMSFMXChart.Create(AOwner: TComponent);
begin
  inherited;
  ClipChildren := True;
  FFill := TBrush.Create(TBrushKind.Solid, claWhite);
  FFill.OnChanged := FillChanged;
  FFillOpacity := 1;
  FStroke := TBrush.Create(TBrushKind.Solid, claGray);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeOpacity := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
  FCachingQuality := TCanvasQuality.HighQuality;
  FInteraction := True;
  FClickMargin := 10;
  FCache := False;
  FCaching := False;

  FInteractionOptions := TTMSFMXChartInteractionOptions.Create(Self);

  FSeries := TTMSFMXChartSeries.Create(Self);
  FTitle := TTMSFMXChartTitle.Create(Self);
  FXAxis := TTMSFMXChartXAxis.Create(Self);
  FYAxis := TTMSFMXChartYAxis.Create(Self);
  FLegend := TTMSFMXChartLegend.Create(Self);
  FSeriesMargins := TBounds.Create(TRectF.Empty);
  FSeriesMargins.OnChange := SeriesMarginsChanged;

  FDrawOrder := csdoBarsFirst;

  Width := 550;
  Height := 350;
  if (csDesigning in ComponentState) and not ((csReading in Owner.ComponentState) or (csLoading in Owner.ComponentState)) then
    InitSample;

  Touch.DefaultInteractiveGestures := Touch.DefaultInteractiveGestures + [TInteractiveGesture.Pan, TInteractiveGesture.Zoom];
  Touch.InteractiveGestures := Touch.InteractiveGestures + [TInteractiveGesture.Pan, TInteractiveGesture.Zoom];

  if not (csDesigning in ComponentState) then
    RegisterRuntimeClasses;
end;

destructor TTMSFMXChart.Destroy;
begin
  if Assigned(FCachedBitmap) then
    FCachedBitmap.Free;
  FInteractionOptions.Free;
  FSeriesMargins.Free;
  FLegend.Free;
  FYAxis.Free;
  FXAxis.Free;
  FTitle.Free;
  FSeries.Free;
  FStroke.Free;
  FFill.Free;
  inherited;
end;

procedure TTMSFMXChart.DoAfterDrawBackground(ACanvas: TCanvas; ARect: TRectF);
begin
  if Assigned(OnAfterDrawBackground) then
    OnAfterDrawBackground(Self, ACanvas, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawChart(ACanvas: TCanvas; ARect: TRectF);
begin
  if Assigned(OnAfterDrawChart) then
    OnAfterDrawChart(Self, ACanvas, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawLegend(ACanvas: TCanvas; ARect: TRectF);
begin
  if Assigned(OnAfterDrawLegend) then
    OnAfterDrawLegend(Self, ACanvas, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawLegendIcon(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ARect: TRectF);
begin
  if Assigned(OnAfterDrawLegendIcon) then
    OnAfterDrawLegendIcon(Self, ACanvas, ASerie, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieMarker(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawPoint: TTMSFMXChartDrawPoint);
begin
  if Assigned(OnAfterDrawSerieMarker) then
    OnAfterDrawSerieMarker(Self, ACanvas, ASerie, ADrawPoint);
end;

procedure TTMSFMXChart.DoAfterDrawSerieMultiPoint(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawMultiPoint: TTMSFMXChartDrawMultiPoint);
begin
  if Assigned(OnAfterDrawSerieMultiPoint) then
    OnAfterDrawSerieMultiPoint(Self, ACanvas, ASerie, ADrawMultiPoint);
end;

procedure TTMSFMXChart.DoAfterDrawSeries(ACanvas: TCanvas; ARect: TRectF);
begin
  if Assigned(OnAfterDrawSeries) then
    OnAfterDrawSeries(Self, ACanvas, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieSlice(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawSlice: TTMSFMXChartDrawSlice);
begin
  if Assigned(OnAfterDrawSerieSlice) then
    OnAfterDrawSerieSlice(Self, ACanvas, ASerie, ADrawSlice);
end;

procedure TTMSFMXChart.DoAfterDrawSerieAnnotation(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawAnnotation: TTMSFMXChartDrawAnnotation);
begin
  if Assigned(OnAfterDrawSerieAnnotation) then
    OnAfterDrawSerieAnnotation(Self, ACanvas, ASerie, ADrawAnnotation);
end;

procedure TTMSFMXChart.DoAfterDrawSerieBar(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawRect: TTMSFMXChartDrawRect);
begin
  if Assigned(OnAfterDrawSerieBar) then
    OnAfterDrawSerieBar(Self, ACanvas, ASerie, ADrawRect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieGridLegendText(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawGridLegendText: TTMSFMXChartDrawLabel);
begin
  if Assigned(OnAfterDrawSerieGridLegendText) then
    OnAfterDrawSerieGridLegendText(Self, ACanvas, ASerie, ADrawGridLegendText);
end;

procedure TTMSFMXChart.DoAfterDrawSerieLabel(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawLabel: TTMSFMXChartDrawLabel);
begin
  if Assigned(OnAfterDrawSerieLabel) then
    OnAfterDrawSerieLabel(Self, ACanvas, ASerie, ADrawLabel);
end;

procedure TTMSFMXChart.DoAfterDrawSerieLegend(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF);
begin
  if Assigned(OnAfterDrawSerieLegend) then
    OnAfterDrawSerieLegend(Self, ACanvas, ASerie, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieLegendIcon(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF);
begin
  if Assigned(OnAfterDrawSerieLegendIcon) then
    OnAfterDrawSerieLegendIcon(Self, ACanvas, ASerie, APoint, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieLegendIconVirtual(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF);
begin
  if Assigned(OnAfterDrawSerieLegendIconVirtual) then
    OnAfterDrawSerieLegendIconVirtual(Self, ACanvas, ASerie, APoint, ARect);
end;

procedure TTMSFMXChart.DoAfterDrawSerieLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawLine: TTMSFMXChartDrawLine);
begin
  if Assigned(OnAfterDrawSerieLine) then
    OnAfterDrawSerieLine(Self, ACanvas, ASerie, ADrawLine);
end;

procedure TTMSFMXChart.DoAfterDrawSerieXGridLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine);
begin
  if Assigned(OnAfterDrawSerieXGridLine) then
    OnAfterDrawSerieXGridLine(Self, ACanvas, ASerie, ADrawGridLine);
end;

procedure TTMSFMXChart.DoAfterDrawSerieXValue(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue);
begin
  if Assigned(OnAfterDrawSerieXValue) then
    OnAfterDrawSerieXValue(Self, ACanvas, ASerie, APosition, ADrawValue);
end;

procedure TTMSFMXChart.DoAfterDrawSerieYGridLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; ADrawGridLine: TTMSFMXChartDrawXYGridLine);
begin
  if Assigned(OnAfterDrawSerieYGridLine) then
    OnAfterDrawSerieYGridLine(Self, ACanvas, ASerie, ADrawGridLine);
end;

procedure TTMSFMXChart.DoAfterDrawSerieYValue(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; ADrawValue: TTMSFMXChartDrawXYValue);
begin
  if Assigned(OnAfterDrawSerieYValue) then
    OnAfterDrawSerieYValue(Self, ACanvas, ASerie, APosition, ADrawValue);
end;

procedure TTMSFMXChart.DoAfterDrawTitle(ACanvas: TCanvas; ARect: TRectF;
  APosition: TTMSFMXChartTitlePosition);
begin
  if Assigned(OnAfterDrawTitle) then
    OnAfterDrawTitle(Self, ACanvas, ARect, APosition);
end;

procedure TTMSFMXChart.DoAfterDrawXAxis(ACanvas: TCanvas;
  ARect: TRectF; APosition: TTMSFMXChartXAxisPosition);
begin
  if Assigned(OnAfterDrawXAxis) then
    OnAfterDrawXAxis(Self, ACanvas, ARect, APosition);
end;

procedure TTMSFMXChart.DoAfterDrawXValuesTitle(ACanvas: TCanvas; ARect,
  ATextRect: TRectF; APosition: TTMSFMXChartXAxisPosition;
  ASerie: TTMSFMXChartSerie);
begin
  if Assigned(OnAfterDrawXValuesTitle) then
    OnAfterDrawXValuesTitle(Self, ACanvas, ARect, APosition, ASerie, ATextRect);
end;

procedure TTMSFMXChart.DoAfterDrawYAxis(ACanvas: TCanvas;
  ARect: TRectF; APosition: TTMSFMXChartYAxisPosition);
begin
  if Assigned(OnAfterDrawYAxis) then
    OnAfterDrawYAxis(Self, ACanvas, ARect, APosition);
end;

procedure TTMSFMXChart.DoAfterDrawYValuesTitle(ACanvas: TCanvas; ARect,
  ATextRect: TRectF; APosition: TTMSFMXChartYAxisPosition;
  ASerie: TTMSFMXChartSerie);
begin
  if Assigned(OnAfterDrawYValuesTitle) then
    OnAfterDrawYValuesTitle(Self, ACanvas, ARect, APosition, ASerie, ATextRect);
end;

procedure TTMSFMXChart.DoAnimateSerieFinished(ASerie: TTMSFMXChartSerie);
begin
  if Assigned(OnAnimateSerieFinished) then
    OnAnimateSerieFinished(Self, ASerie);
end;

procedure TTMSFMXChart.DoAnimateSerieStarted(ASerie: TTMSFMXChartSerie);
begin
  if Assigned(OnAnimateSerieStarted) then
    OnAnimateSerieStarted(Self, ASerie);
end;

procedure TTMSFMXChart.DoBeforeDrawBackground(ACanvas: TCanvas; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawBackground) then
    OnBeforeDrawBackground(Self, ACanvas, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawChart(ACanvas: TCanvas; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawChart) then
    OnBeforeDrawChart(Self, ACanvas, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawLegend(ACanvas: TCanvas; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawLegend) then
    OnBeforeDrawLegend(Self, ACanvas, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawLegendIcon(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawLegendIcon) then
    OnBeforeDrawLegendIcon(Self, ACanvas, ASerie, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieMarker(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawPoint: TTMSFMXChartDrawPoint; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieMarker) then
    OnBeforeDrawSerieMarker(Self, ACanvas, ASerie, ADrawPoint, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSeries(ACanvas: TCanvas; ARect: TRectF; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSeries) then
    OnBeforeDrawSeries(Self, ACanvas, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieSlice(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawSlice: TTMSFMXChartDrawSlice;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieSlice) then
    OnBeforeDrawSerieSlice(Self, ACanvas, ASerie, ADrawSlice, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieAnnotation(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawAnnotation: TTMSFMXChartDrawAnnotation;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieAnnotation) then
    OnBeforeDrawSerieAnnotation(Self, ACanvas, ASerie, ADrawAnnotation, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieBar(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawRect: TTMSFMXChartDrawRect;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieBar) then
    OnBeforeDrawSerieBar(Self, ACanvas, ASerie, ADrawRect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieGridLegendText(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawGridLegendText: TTMSFMXChartDrawLabel;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieGridLegendText) then
    OnBeforeDrawSerieGridLegendText(Self, ACanvas, ASerie, ADrawGridLegendText, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieLabel(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawLabel: TTMSFMXChartDrawLabel;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieLabel) then
    OnBeforeDrawSerieLabel(Self, ACanvas, ASerie, ADrawLabel, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieLegend(ACanvas: TCanvas; ASerie: TTMSFMXChartSerie; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieLegend) then
    OnBeforeDrawSerieLegend(Self, ACanvas, ASerie, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieLegendIcon(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieLegendIcon) then
    OnBeforeDrawSerieLegendIcon(Self, ACanvas, ASerie, APoint, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieLegendIconVirtual(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual; ARect: TRectF;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieLegendIconVirtual) then
    OnBeforeDrawSerieLegendIconVirtual(Self, ACanvas, ASerie, APoint, ARect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawLine: TTMSFMXChartDrawLine;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieLine) then
    OnBeforeDrawSerieLine(Self, ACanvas, ASerie, ADrawLine, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieMultiPoint(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawMultiPoint: TTMSFMXChartDrawMultiPoint;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieMultiPoint) then
    OnBeforeDrawSerieMultiPoint(Self, ACanvas, ASerie, ADrawMultiPoint, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieXGridLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieXGridLine) then
    OnBeforeDrawSerieXGridLine(Self, ACanvas, ASerie, ADrawGridLine, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieXValue(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartXAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieXValue) then
    OnBeforeDrawSerieXValue(Self, ACanvas, ASerie, APosition, ADrawValue, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieYGridLine(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; var ADrawGridLine: TTMSFMXChartDrawXYGridLine; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieYGridLine) then
    OnBeforeDrawSerieYGridLine(Self, ACanvas, ASerie, ADrawGridLine, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawSerieYValue(ACanvas: TCanvas;
  ASerie: TTMSFMXChartSerie; APosition: TTMSFMXChartYAxisPosition; var ADrawValue: TTMSFMXChartDrawXYValue;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawSerieYValue) then
    OnBeforeDrawSerieYValue(Self, ACanvas, ASerie, APosition, ADrawValue, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawTitle(ACanvas: TCanvas; ARect: TRectF;
  APosition: TTMSFMXChartTitlePosition; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawTitle) then
    OnBeforeDrawTitle(Self, ACanvas, ARect, APosition, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawXAxis(ACanvas: TCanvas;
  ARect: TRectF; APosition: TTMSFMXChartXAxisPosition;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawXAxis) then
    OnBeforeDrawXAxis(Self, ACanvas, ARect, APosition, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawXValuesTitle(ACanvas: TCanvas; ARect,
  ATextRect: TRectF; APosition: TTMSFMXChartXAxisPosition;
  ASerie: TTMSFMXChartSerie; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawXValuesTitle) then
    OnBeforeDrawXValuesTitle(Self, ACanvas, ARect, APosition, ASerie, ATextRect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawYAxis(ACanvas: TCanvas;
  ARect: TRectF; APosition: TTMSFMXChartYAxisPosition;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawYAxis) then
    OnBeforeDrawYAxis(Self, ACanvas, ARect, APosition, ADefaultDraw);
end;

procedure TTMSFMXChart.DoBeforeDrawYValuesTitle(ACanvas: TCanvas; ARect,
  ATextRect: TRectF; APosition: TTMSFMXChartYAxisPosition;
  ASerie: TTMSFMXChartSerie; var ADefaultDraw: Boolean);
begin
  if Assigned(OnBeforeDrawYValuesTitle) then
    OnBeforeDrawYValuesTitle(Self, ACanvas, ARect, APosition, ASerie, ATextRect, ADefaultDraw);
end;

procedure TTMSFMXChart.DoCustomizeAnnotationFill(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual;
  AFill: TBrush);
begin
  if Assigned(OnCustomizeAnnotationFill) then
    OnCustomizeAnnotationFill(Self, ASerie, APoint, AAnnotation, AFill);
end;

procedure TTMSFMXChart.DoCustomizeAnnotationFont(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual;
  AFont: TFont);
begin
  if Assigned(OnCustomizeAnnotationFont) then
    OnCustomizeAnnotationFont(Self, ASerie, APoint, AAnnotation, AFont);
end;

procedure TTMSFMXChart.DoCustomizeAnnotationStroke(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; AAnnotation: TTMSFMXChartAnnotationVirtual;
  AStroke: TStrokeBrush);
begin
  if Assigned(OnCustomizeAnnotationStroke) then
    OnCustomizeAnnotationStroke(Self, ASerie, APoint, AAnnotation, AStroke);
end;

procedure TTMSFMXChart.DoDrawTitleText(ACanvas: TCanvas; ARect: TRectF;
  APosition: TTMSFMXChartTitlePosition; var AText: String;
  var ADefaultDraw: Boolean);
begin
  if Assigned(OnDrawTitleText) then
    OnDrawTitleText(Self, ACanvas, ARect, APosition, AText, ADefaultDraw);
end;

procedure TTMSFMXChart.DoGetAnnotation(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; AIndex: Integer;
  var AAnnotation: TTMSFMXChartAnnotationVirtual);
begin
  if Assigned(OnGetAnnotation) then
    OnGetAnnotation(Self, ASerie, APoint, AIndex, AAnnotation);
end;

procedure TTMSFMXChart.DoGetNumberOfAnnotations(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; var ANumberOfAnnotations: Integer);
begin
  if Assigned(OnGetNumberOfAnnotations) then
    OnGetNumberOfAnnotations(Self, ASerie, APoint, ANumberOfAnnotations);
end;

procedure TTMSFMXChart.DoGetNumberOfPoints(ASerie: TTMSFMXChartSerie;
  var ANumberOfPoints: Integer);
begin
  if Assigned(OnGetNumberOfPoints) then
  begin
    if Assigned(ASerie) then
      ASerie.Points.Clear;

    OnGetNumberOfPoints(Self, ASerie, ANumberOfPoints);
    FVirtualMode := ANumberOfPoints > 0;
  end;
end;

procedure TTMSFMXChart.DoGetPoint(ASerie: TTMSFMXChartSerie; AIndex: Integer; var APoint: TTMSFMXChartPointVirtual);
begin
  if Assigned(OnGetPoint) then
    OnGetPoint(Self, ASerie, AIndex, APoint);
end;

procedure TTMSFMXChart.DoGetSerieLabel(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var AValueString: String);
begin
  if Assigned(OnGetSerieLabel) then
    OnGetSerieLabel(Self, ASerie, APoint, AValueString);
end;

procedure TTMSFMXChart.DoGetSerieLabelVirtual(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; var AValueString: String);
begin
  if Assigned(OnGetSerieLabelVirtual) then
    OnGetSerieLabelVirtual(Self, ASerie, APoint, AValueString);
end;

procedure TTMSFMXChart.DoGetSerieLegendText(ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPoint; var ALegendText: String);
begin
  if Assigned(OnGetSerieLegendText) then
    OnGetSerieLegendText(Self, ASerie, APoint, ALegendText);
end;

procedure TTMSFMXChart.DoGetSerieLegendTextVirtual(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPointVirtual; var ALegendText: String);
begin
  if Assigned(OnGetSerieLegendTextVirtual) then
    OnGetSerieLegendTextVirtual(Self, ASerie, APoint, ALegendText);
end;

procedure TTMSFMXChart.DoGetSerieSpiderLegendText(ASerie: TTMSFMXChartSerie;
  APoint: TTMSFMXChartPoint; var ALegendText: String);
begin
  if Assigned(OnGetSerieSpiderLegendText) then
    OnGetSerieSpiderLegendText(Self, ASerie, APoint, ALegendText);
end;

procedure TTMSFMXChart.DoGetSerieSpiderLegendTextVirtual(
  ASerie: TTMSFMXChartSerie; APoint: TTMSFMXChartPointVirtual;
  var ALegendText: String);
begin
  if Assigned(OnGetSerieSpiderLegendTextVirtual) then
    OnGetSerieSpiderLegendTextVirtual(Self, ASerie, APoint, ALegendText);
end;

procedure TTMSFMXChart.DoGetSerieXValue(ASerie: TTMSFMXChartSerie; AIndex: Integer; AKind: TTMSFMXChartDrawXYValueKind;
  AValue: Double; var AValueString: String);
begin
  if Assigned(OnGetSerieXValue) then
    OnGetSerieXValue(Self, ASerie, AIndex, AKind, AValue, AValueString);
end;

procedure TTMSFMXChart.DoGetSerieYValue(ASerie: TTMSFMXChartSerie; AIndex: Integer; AKind: TTMSFMXChartDrawXYValueKind;
  AValue: Double; var AValueString: String);
begin
  if Assigned(OnGetSerieYValue) then
    OnGetSerieYValue(Self, ASerie, AIndex, AKind, AValue, AValueString);
end;

procedure TTMSFMXChart.DoLegendItemClick(AIndex: Integer);
begin
  if Assigned(OnLegendItemClick) then
    OnLegendItemClick(Self, AIndex);
end;

procedure TTMSFMXChart.DoSerieBarClick(APoint: TTMSFMXChartPoint);
begin
  if Assigned(OnSerieBarClick) then
    OnSerieBarClick(Self, APoint);
end;

procedure TTMSFMXChart.DoSerieLegendItemClick(ASerie: TTMSFMXChartSerie; AIndex: Integer);
begin
  if Assigned(OnSerieLegendItemClick) then
    OnSerieLegendItemClick(Self, ASerie, AIndex);
end;

procedure TTMSFMXChart.DoSerieBarClickVirtual(APoint: TTMSFMXChartPointVirtual);
begin
  if Assigned(OnSerieBarClickVirtual) then
    OnSerieBarClickVirtual(Self, APoint);
end;

procedure TTMSFMXChart.DoSeriePointClick(APoint: TTMSFMXChartPoint);
begin
  if Assigned(OnSeriePointClick) then
    OnSeriePointClick(Self, APoint);
end;

procedure TTMSFMXChart.DoSeriePointClickVirtual(APoint: TTMSFMXChartPointVirtual);
begin
  if Assigned(OnSeriePointClickVirtual) then
    OnSeriePointClickVirtual(Self, APoint);
end;

procedure TTMSFMXChart.DoSerieSliceClick(APoint: TTMSFMXChartPoint);
begin
  if Assigned(OnSerieSliceClick) then
    OnSerieSliceClick(Self, APoint);
end;

procedure TTMSFMXChart.DoSerieSliceClickVirtual(APoint: TTMSFMXChartPointVirtual);
begin
  if Assigned(OnSerieSliceClickVirtual) then
    OnSerieSliceClickVirtual(Self, APoint);
end;

procedure TTMSFMXChart.DrawBackGround(ACanvas: TCanvas);
var
  dr: Boolean;
  r: TRectF;
  st: TCanvasSaveState;
begin
  ACanvas.Fill.Assign(Fill);
  ACanvas.Stroke.Assign(Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := StrokeThickness;
  ACanvas.Stroke.Cap := StrokeCap;
  ACanvas.Stroke.Dash := StrokeDash;
  ACanvas.Stroke.Join := StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := StrokeThickness;
  ACanvas.StrokeCap := StrokeCap;
  ACanvas.StrokeDash := StrokeDash;
  ACanvas.StrokeJoin := StrokeJoin;
  {$ENDIF}

  r := GetChartRectangle;
  r := ModifyRect(r);
  dr := True;
  st := ACanvas.SaveState;
  DoBeforeDrawBackground(ACanvas, r, dr);
  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, FillOpacity);
    ACanvas.DrawRect(r, 0, 0, AllCorners, StrokeOpacity);
    DoAfterDrawBackground(ACanvas, r);
  end;
  ACanvas.RestoreState(st);
end;

procedure TTMSFMXChart.DrawChart(ACanvas: TCanvas);
var
  r: TRectF;
  dr: Boolean;
  st: TCanvasSaveState;
  {$IFDEF FREEWARE}
  str: String;
  {$ENDIF}
begin
  r := GetChartRectangle;
  dr := True;
  st := ACanvas.SaveState;
  DoBeforeDrawChart(ACanvas, r, dr);
  if dr then
  begin
    DrawBackGround(ACanvas);
    Title.Draw(ACanvas);
    DrawXYGrid(ACanvas);
    XAxis.Draw(ACanvas);
    YAxis.Draw(ACanvas);
    Series.Draw(ACanvas);
    Legend.Draw(ACanvas);
    DoAfterDrawChart(ACanvas, r);
  end;
  ACanvas.RestoreState(st);

  {$IFDEF FREEWARE}
  ACanvas.Fill.Color := TAlphaColorRec.Red;
  str := Scramble('Duuilfdqljk%pv`v%qwldi%s`wvljk%jc%QHV%vjcqrdw`%fjhujk`kqv+')+#13#10+
    Scramble('Fjkqdfq%QHV%vjcqrdw`%mqqu?**rrr+qhvvjcqrdw`+fjh%cjw%sdila%ilf`kvlkb+');
  ACanvas.FillText(GetChartRectangle, str, True, 1, [], TTextAlign.Leading, TTextAlign.Leading);
  {$ENDIF}
end;

procedure TTMSFMXChart.DrawXYGrid(ACanvas: TCanvas);
var
  I: Integer;
  s: TTMSFMXChartSerie;
begin
  if HasPieChart then
    Exit;

  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible then
    begin             
      s.XGrid.Draw(ACanvas);
      s.YGrid.Draw(ACanvas, False);  
    end;
  end;
end;

procedure TTMSFMXChart.EndUpdate;
begin
  {$IF DEFINED(DELPHIXE7_LVL) AND NOT DEFINED(DELPHIXE8_LVL)}
  inherited EndUpdate;
  {$ELSE}
  inherited;
  {$ENDIF}
  Dec(FUpdateCount);
  if FUpdateCount = 0 then
    UpdateChart;
end;

procedure TTMSFMXChart.SaveSettingsToFile(AFileName: string);
begin
  TTMSFMXChartPersistence.SaveSettingsToFile(Self, AFileName);
end;

procedure TTMSFMXChart.SaveSettingsToStream(AStream: TStream);
begin
  TTMSFMXChartPersistence.SaveSettingsToStream(Self, AStream);
end;

procedure TTMSFMXChart.SaveToImage(AFileName: String);
var
  bmp: TBitmap;
begin
  bmp := nil;
  try
    bmp := MakeScreenshot;
    if Assigned(bmp) then
      bmp.SaveToFile(AFileName);
  finally
    if Assigned(bmp) then
      bmp.Free;
  end;
end;

procedure TTMSFMXChart.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChart.GetChartRectangle: TRectF;
begin
  Result := RectF(1, 1, Width - 1, Height - 1);
end;

function TTMSFMXChart.GetSerieByName(AName: string): TTMSFMXChartSerie;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].LegendText = AName then
    begin
      Result := Series[I];
      Exit;
    end;
  end;
end;

function TTMSFMXChart.GetSeriesRectangle: TRectF;
var
  r: TRectF;
  offx, offy: Double;
begin
  r := GetChartRectangle;
  if XAxis.Visible then
  begin
    if xpTop in XAxis.Positions then
      R.Top := R.Top + XAxis.GetHeight(xpTop);

    if xpBottom in XAxis.Positions then
      R.Bottom := R.Bottom - XAxis.GetHeight(xpBottom);
  end;

  if Title.Visible then
  begin
    if tipTop in Title.Positions then
      R.Top := R.Top + Title.Height;
    if tipBottom in Title.Positions then
      R.Bottom := R.Bottom - Title.Height;
  end;

  if YAxis.Visible then
  begin
    if ypLeft in YAxis.Positions then
      R.Left := R.Left + YAxis.GetWidth(ypLeft);

    if ypRight in YAxis.Positions then
      R.Right := R.Right - YAxis.GetWidth(ypRight);
  end;

  Result := RectF(R.Left + SeriesMargins.Left, R.Top + SeriesMargins.Top, R.Right - SeriesMargins.Right,
    R.Bottom - SeriesMargins.Bottom);

  offx := FTotalOffset3DX;
  offy := FTotalOffset3DY;

  Result.Right := Result.Right - offx;
  Result.Top := Result.Top + offy;
end;

function TTMSFMXChart.GetTotal3DOffsetX: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  gp, l: Integer;
  lgp: TArray<Integer>;
  J: Integer;
  maxgp: Double;
begin
  Result := 0;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and not s.IsBar and not s.IsStacked then
      Result := Result + s.Offset3DX;
  end;

  gp := -1;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and (s.GroupIndex > gp) and (s.IsBar or s.IsArea) then
    begin
      gp := s.GroupIndex;
      l := Length(lgp);
      SetLength(lgp, l + 1);
      lgp[l] := gp;
    end;
  end;

  maxgp := 0;
  for I := 0 to Length(lgp) - 1 do
  begin
    for J := 0 to Series.Count - 1 do
    begin
      s := Series[J];
      if s.Visible and s.Is3DEnabled and (s.GroupIndex = lgp[I]) and ((s.IsBar and not s.IsStacked) or (s.IsArea and s.IsStacked)) then
        maxgp := Max(maxgp, s.Offset3DX);
    end;
    Result := Result + maxgp;
  end;

  maxgp := 0;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and s.IsBar and s.IsStacked then
      maxgp := Max(maxgp, s.Offset3DX);
  end;
  Result := Result + maxgp;
end;

function TTMSFMXChart.GetTotal3DOffsetY: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  gp, l: Integer;
  lgp: TArray<Integer>;
  J: Integer;
  maxgp: Double;
begin
  Result := 0;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and not s.IsBar and not s.IsStacked then
      Result := Result + s.Offset3DY;
  end;

  gp := -1;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and (s.GroupIndex > gp) and (s.IsBar or s.IsArea) then
    begin
      gp := s.GroupIndex;
      l := Length(lgp);
      SetLength(lgp, l + 1);
      lgp[l] := gp;
    end;
  end;

  maxgp := 0;
  for I := 0 to Length(lgp) - 1 do
  begin
    for J := 0 to Series.Count - 1 do
    begin
      s := Series[J];
      if s.Visible and s.Is3DEnabled and (s.GroupIndex = lgp[I]) and ((s.IsBar and not s.IsStacked) or (s.IsArea and s.IsStacked)) then
        maxgp := Max(maxgp, s.Offset3DY);
    end;
    Result := Result + maxgp;
  end;

  maxgp := 0;
  for I := 0 to Series.Count - 1 do
  begin
    s := Series[I];
    if s.Visible and s.Is3DEnabled and s.IsBar and s.IsStacked then
      maxgp := Max(maxgp, s.Offset3DY);
  end;
  Result := Result + maxgp;
end;

function TTMSFMXChart.GetVersion: String;
var
  vn: Integer;
begin
  vn := GetVersionNr;
  Result := IntToStr(Hi(Hiword(vn)))+'.'+IntToStr(Lo(Hiword(vn)))+'.'+IntToStr(Hi(Loword(vn)))+'.'+IntToStr(Lo(Loword(vn)));
end;

function TTMSFMXChart.GetVersionNr: integer;
begin
  Result := MakeLong(MakeWord(BLD_VER,REL_VER),MakeWord(MIN_VER,MAJ_VER));
end;

function TTMSFMXChart.HasPieChart: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].IsPie then
    begin
      Result := True;
      Break;
    end;
  end;
end;

function TTMSFMXChart.HasSpiderChart: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].IsSpider then
    begin
      Result := True;
      Break;
    end;
  end;
end;

function TTMSFMXChart.HasStackedPieChart: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].Pie.Stacked then
    begin
      Result := True;
      Break;
    end;
  end;
end;

procedure TTMSFMXChart.InitSample;
var
  s: TTMSFMXChartSerie;
  I: Integer;
  J: Integer;
  pt: TTMSFMXChartPoint;
begin
  BeginUpdate;
  Series.Clear;
  s := Series.Add;
  s.Stroke.Color := claDarkslategray;
  s.Fill.Color := claLightslategray;
  s.YValues.Positions := [ypLeft];
  s.XValues.Positions := [xpBottom];
  s.XGrid.Visible := True;
  s.YGrid.Visible := True;

  s := Series.Add;
  s.Stroke.Color := claDarkgreen;
  s.Fill.Color := claLightseagreen;
  s.YValues.Positions := [];
  s.XValues.Positions := [];

  s := Series.Add;
  s.Stroke.Color := claDarkred;
  s.Fill.Color := claOrange;
  s.YValues.Positions := [];
  s.XValues.Positions := [];

  Title.Text := 'TMS FMX Chart';

  for I := 0 to Series.Count - 1 do
  begin
    Series[I].YValues.Title.Text := 'Y-Axis Serie ' + inttostr(I + 1);
    Series[I].XValues.Title.Text := 'X-Axis Serie ' + inttostr(I + 1);
    Series[I].LegendText := 'Serie ' + inttostr(I + 1);
    Series[I].AutoYRange := arDisabled;
    Series[I].MinY := 0;
    Series[I].MaxY := 150;
    Series[I].Mode := smMathematical;
    Series[I].Markers.Visible := True;
    Series[I].Markers.Fill.Assign(Series[I].Fill);
    Series[I].Markers.Stroke.Assign(Series[I].Stroke);
    for J := 0 to 10 do
    begin
      pt := Series[I].Points.Add;
      pt.YValue := RandomRange(0, 130) + 5;
      pt.YValueSecond := pt.YValue - 20;
      pt.YValueVariable := pt.YValue;
      pt.YValueClose := pt.YValue - 20;
      pt.YValueHigh := pt.YValue + 20;
      pt.YValueOpen := pt.YValue + RandomRange(-10, 10);
      pt.YValueClose := pt.YValue + RandomRange(-10, 10);
      pt.YValueMedian := pt.YValueOpen + RandomRange(-5, 5);
    end;
  end;

  EndUpdate;
end;

procedure TTMSFMXChart.Loaded;
begin
  inherited;
  UpdateChart;
end;

procedure TTMSFMXChart.LoadSettingsFromFile(AFileName: string);
begin
  TTMSFMXChartPersistence.LoadSettingsFromFile(Self, AFileName);
end;

procedure TTMSFMXChart.LoadSettingsFromStream(AStream: TStream);
begin
  TTMSFMXChartPersistence.LoadSettingsFromStream(Self, AStream);
end;

procedure TTMSFMXChart.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Single);
begin
  inherited;
  if not Interaction then
    Exit;

  Capture;
  FMouseDown := True;
  FDownPt := PointF(X, Y);
  FMovePt := FDownPt;
end;

procedure TTMSFMXChart.MouseMove(Shift: TShiftState; X, Y: Single);
begin
  inherited;
  if not Interaction then
    Exit;

  if (InteractionOptions.ScaleMode = smNone) and (InteractionOptions.Panning = False) then
    Exit;

  if FMouseDown then
  begin
    FMouseMoved := (Abs(FDownPt.X - X) > INTERACTIONMARGIN) or (Abs(FDownPt.Y - Y) > INTERACTIONMARGIN);
    UpdateRange(FMovePt.X - X, FMovePt.Y - Y, ssCtrl in Shift);
    FMovePt := PointF(X, Y);
  end;
end;

procedure TTMSFMXChart.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
  Y: Single);
var
  ptPoint, ptBar, ptSlice: TTMSFMXChartPoint;
  ptPointv, ptBarv, ptSlicev: TTMSFMXChartPointVirtual;
  b: Boolean;
  i: Integer;
  s: TTMSFMXChartSerie;
begin
  inherited;
  if not Interaction then
    Exit;

  ReleaseCapture;

  FMouseDown := False;
  if FMouseMoved then
  begin
    FMouseMoved := False;
    Exit;
  end;

  ptPointv := GetDefaultVirtualPoint;
  ptBarv := GetDefaultVirtualPoint;
  ptSlicev := GetDefaultVirtualPoint;

  b := False;

  if FVirtualMode then
  begin
    if XYToPointVirtual(X, Y, ptPointv) then
    begin
      DoSeriePointClickVirtual(ptPointv);
      b := True;
    end
    else
    begin
      if XYToBarVirtual(X, Y, ptBarv) then
      begin
        DoSerieBarClickVirtual(ptBarv);
        b := True;
      end;
    end;

    if XYToSliceVirtual(X, Y, ptSlicev) then
    begin
      DoSerieSliceClickVirtual(ptSlicev);
      b := True;
    end;
  end
  else
  begin
    ptPoint := XYToPoint(X, Y);
    if Assigned(ptPoint) then
    begin
      DoSeriePointClick(ptPoint);
      b := True;
    end
    else
    begin
      ptBar := XYToBar(X, Y);
      if Assigned(ptBar) then
      begin
        DoSerieBarClick(ptBar);
        b := True;
      end;
    end;

    ptSlice := XYToSlice(X, Y);
    if Assigned(ptSlice) then
    begin
      DoSerieSliceClick(ptSlice);
      b := True;
    end;
  end;

  if not b then
  begin
    s := nil;
    i := XYToSerieLegendItem(X, Y, s);
    if (i <> -1) and Assigned(s) then
    begin
      DoSerieLegendItemClick(s, i);
      b := True;
    end;

    if not b then
    begin
      i := XYToLegendItem(X, Y);
      if (i <> -1) then
        DoLegendItemClick(i);
    end;
  end;
end;

procedure TTMSFMXChart.MouseWheel(Shift: TShiftState; WheelDelta: Integer;
  var Handled: Boolean);
begin
  inherited;
  if not Interaction then
    Exit;

  UpdateRange(0, WheelDelta, ssCtrl in Shift)
end;

procedure TTMSFMXChart.Paint;
var
  ca: TCanvas;
begin
  inherited;

  if Caching then
  begin
    if not FCache then
    begin
      if Assigned(FCachedBitmap) then
        FCachedBitmap.Free;

      FCachedBitmap := TBitmap.Create(Round(Width), Round(Height));
      ca := TCanvasManager.CreateFromBitmap(FCachedBitmap, CachingQuality);
      ca.BeginScene;
      ca.Clear(TAlphaColorRec.White);
      DrawChart(ca);
      ca.EndScene;
      ca.Free;
      FCache := True;
    end;
    Canvas.DrawBitmap(FCachedBitmap, RectF(0, 0, FCachedBitmap.Width, FCachedBitmap.Height), RectF(0, 0, FCachedBitmap.Width, FCachedBitmap.Height), AbsoluteOpacity);
  end
  else
    DrawChart(Canvas);
end;

procedure TTMSFMXChart.RegisterRuntimeClasses;
begin
  inherited;
  RegisterClasses([TTMSFMXChart, TTMSFMXChartSerie, TTMSFMXChartPoint, TTMSFMXChartAnnotation, TGradientPoint]);
end;

procedure TTMSFMXChart.RepaintChart;
begin
  if FUpdateCount > 0 then
    Exit;

  FCache := False;
  Repaint;
end;

procedure TTMSFMXChart.Resize;
begin
  inherited;
  UpdateChart;
end;

procedure TTMSFMXChart.SeriesMarginsChanged(Sender: TObject);
begin
  UpdateChart;
end;

procedure TTMSFMXChart.SetCaching(const Value: Boolean);
begin
  if FCaching <> Value then
  begin
    FCaching := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetCachingQuality(const Value: TCanvasQuality);
begin
  if FCachingQuality <> Value then
  begin
    FCachingQuality := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetDrawOrder(const Value: TTMSFMXChartSeriesDrawOrder);
begin
  if FDrawOrder <> Value then
  begin
    FDrawOrder := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetLegend(const Value: TTMSFMXChartLegend);
begin
  if FLegend <> Value then
  begin
    FLegend.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetSeries(const Value: TTMSFMXChartSeries);
begin
  if FSeries <> Value then
  begin
    FSeries.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChart.SetSeriesMargins(const Value: TBounds);
begin
  if FSeriesMargins <> Value then
  begin
    FSeriesMargins.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChart.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChart.SetTitle(const Value: TTMSFMXChartTitle);
begin
  if FTitle <> Value then
  begin
    FTitle.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChart.SetInteractionOptions(const Value: TTMSFMXChartInteractionOptions);
begin
  FInteractionOptions.Assign(Value);
end;

procedure TTMSFMXChart.SetVersion(const Value: String);
begin

end;

procedure TTMSFMXChart.SetXAxis(const Value: TTMSFMXChartXAxis);
begin
  if FXAxis <> Value then
  begin
    FXAxis.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChart.SetYAxis(const Value: TTMSFMXChartYAxis);
begin
  if FYAxis <> Value then
  begin
    FYAxis.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChart.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChart.UpdateChart;
begin
  if (FUpdateCount > 0) or (csDestroying in ComponentState) then
    Exit;

  Calculate3DOffset;
  CalculateMinMaxX;
  CalculateCommonRangeX;
  CalculateMinMaxY;
  CalculateCommonRangeY;
  CalculateXAxisHeight;
  CalculateYScale;
  CalculateYAxis;
  CalculateXScale;
  CalculateXAxis;
  CalculateSeries;
  if YAxis.DisplayAtReferenceValue then
    CalculateYAxis;
  if XAxis.DisplayAtReferenceValue then
    CalculateXAxis;
  RepaintChart;
end;

procedure TTMSFMXChart.UpdateRange(AXRange, AYRange: Double; AScale: Boolean);
var
  I: Integer;
  vx, vy: Double;
begin
  BeginUpdate;
  for I := 0 to Series.Count - 1 do
  begin
    if Series[I].AutoXRange = arDisabled then
    begin
      vx := Series[I].XToValue(AXRange, False);
      if AScale and (InteractionOptions.ScaleMode = smHorizontal) then
      begin
        Series[I].MinX := Series[I].MinX + vx;
        Series[I].MaxX := Series[I].MaxX - vx;
      end
      else if not AScale and InteractionOptions.Panning then
      begin
        Series[I].MinX := Series[I].MinX + vx;
        Series[I].MaxX := Series[I].MaxX + vx;
      end;
    end;

    if Series[I].AutoYRange = arDisabled then
    begin
      vy := Series[I].YToValue(AYRange, False);
      if AScale and (InteractionOptions.ScaleMode = smVertical) then
      begin
        Series[I].MinY := Series[I].MinY + vy;
        Series[I].MaxY := Series[I].MaxY - vy;
      end
      else if not AScale and InteractionOptions.Panning then
      begin
        Series[I].MinY := Series[I].MinY - vy;
        Series[I].MaxY := Series[I].MaxY - vy;
      end;
    end;
  end;
  EndUpdate;
end;

function TTMSFMXChart.XYToBar(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPoint;
begin
  Result := nil;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      res := s.XYToBar(X, Y);
      if Assigned(res) then
      begin
        Result := res;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToBarVirtual(X, Y: Double;
  var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPointVirtual;
begin
  Result := False;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      if s.XYToBarVirtual(X, Y, res) then
      begin
        APoint := res;
        Result := True;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToLegendItem(X, Y: Double): Integer;
begin
  Result := Legend.XYToItem(X, Y);
end;

function TTMSFMXChart.XYToPoint(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPoint;
begin
  Result := nil;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      res := s.XYToPoint(X, Y);
      if Assigned(res) then
      begin
        Result := res;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToPointVirtual(X, Y: Double;
  var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPointVirtual;
begin
  Result := False;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      if s.XYToPointVirtual(X, Y, res) then
      begin
        APoint := res;
        Result := True;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToSerieLegendItem(X,
  Y: Double; var ASerie: TTMSFMXChartSerie): Integer;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: Integer;
begin
  ASerie := nil;
  Result := -1;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      res := s.XYToLegendItem(X, Y);
      if res <> -1 then
      begin
        Result := res;
        ASerie := s;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToSlice(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPoint;
begin
  Result := nil;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      res := s.XYToSlice(X, Y);
      if Assigned(res) then
      begin
        Result := res;
        Break;
      end;
    end;
  end;
end;

function TTMSFMXChart.XYToSliceVirtual(X, Y: Double;
  var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  res: TTMSFMXChartPointVirtual;
begin
  Result := False;
  for I := Series.Count - 1 downto 0 do
  begin
    s := Series[I];
    if s.Visible then
    begin
      if s.XYToSliceVirtual(X, Y, res) then
      begin
        APoint := res;
        Result := True;
        Break;
      end;
    end;
  end;
end;

{ TTMSFMXChartSerie }

procedure TTMSFMXChartSerie.AnimateProcess(Sender: TObject);
var
  dt, db, post, posb, posh, posl, posc, poso, dmh, dml, dmc, dmo: Double;
  rest, resb, resh, resl, resc, reso: Boolean;
  I, K: Integer;
  iCount: Integer;
  dr, drsave: TTMSFMXChartDrawRect;
  dl, dlsave: TTMSFMXChartDrawLine;
  dp, dpsave: TTMSFMXChartDrawPoint;
  ds, dssave: TTMSFMXChartDrawSlice;
  dm, dmsave: TTMSFMXChartDrawMultiPoint;
  drt, drh, drl, drc, dro: Double;
  c: TTMSFMXChart;
begin
  for I := 0 to FDrawRects.Count - 1 do
  begin
    dr := FDrawRects[I];
    if FRevert then
      drSave := FAnimateDrawRectsReverse[I]
    else
      drSave := FAnimateDrawRects[I];
    if dr.DoAnimation then
    begin
      dt := Abs(drsave.Rect.Top - dr.Rect.Top) / AnimationFactor;
      post := dr.Rect.Top;
      drt := drSave.Rect.Top;
      rest := AnimateDouble(post, drt, dt, 0.1);

      db := Abs(drsave.Rect.Bottom - dr.Rect.Bottom) / AnimationFactor;
      posb := dr.Rect.Bottom;
      drt := drSave.Rect.Bottom;
      resb := AnimateDouble(posb, drt, db, 0.1);
      if rest or resb then
      begin
        dr.Rect.Top := post;
        dr.Rect.Bottom := posb;
        FDrawRects[I] := dr;
        RepaintChart;
      end
      else
      begin
        dr.DoAnimation := False;
        dr.Rect.Top := drsave.Rect.Top;
        dr.Rect.Bottom := drsave.Rect.Bottom;
        FDrawRects[I] := dr;
        RepaintChart;
      end;

      if (I < FDrawRects.Count - 1) and ((AnimationFlow and (dt < 7.5)) or not FAnimationFlow) then
      begin
        dr := FDrawRects[I + 1];
        dr.DoAnimation := True;
        FDrawRects[I + 1] := dr;
      end;
    end;
  end;

  for I := 0 to FDrawLines.Count - 1 do
  begin
    dl := FDrawLines[I];
    if FRevert then
      dlSave := FAnimateDrawLinesReverse[I]
    else
      dlSave := FAnimateDrawLines[I];

    if dl.DoAnimation then
    begin
      dt := Abs(dlsave.StartPoint.Y - dl.StartPoint.Y) / AnimationFactor;
      post := dl.StartPoint.Y;
      drt := dlSave.StartPoint.Y;
      rest := AnimateDouble(post, drt, dt, 0.1);

      db := Abs(dlsave.EndPoint.Y - dl.EndPoint.Y) / AnimationFactor;
      posb := dl.EndPoint.Y;
      drt := dlSave.EndPoint.Y;
      resb := AnimateDouble(posb, drt, db, 0.1);
      if rest or resb then
      begin
        dl.StartPoint.Y := post;
        dl.EndPoint.Y := posb;
        FDrawLines[I] := dl;
        RepaintChart;
      end
      else
      begin
        dl.DoAnimation := False;
        dl.StartPoint.Y := dlsave.StartPoint.Y;
        dl.EndPoint.Y := dlsave.EndPoint.Y;
        FDrawLines[I] := dl;
        RepaintChart;
      end;

      if IsArea and (FDrawLines.Count > 0) then
      begin
        FDrawPath.Clear;
        FDrawPath.MoveTo(FDrawLines[0].StartPoint);
        for K := 0 to FDrawLines.Count - 1 do
          FDrawPath.LineTo(FDrawLines[K].EndPoint);
        FDrawPath.ClosePath;
      end;

      if (I < FDrawLines.Count - 1) and ((AnimationFlow and (dt < 7.5)) or not FAnimationFlow) then
      begin
        dl := FDrawLines[I + 1];
        dl.DoAnimation := True;
        FDrawLines[I + 1] := dl;
      end;
    end;
  end;

  for I := 0 to FDrawMultiPoints.Count - 1 do
  begin
    dm := FDrawMultiPoints[I];
    if FRevert then
      dmSave := FAnimateDrawMultiPointsReverse[I]
    else
      dmSave := FAnimateDrawMultiPoints[I];

    if dm.DoAnimation then
    begin
      dmh := Abs(dmsave.PointHigh.Y - dm.PointHigh.Y) / AnimationFactor;
      posh := dm.PointHigh.Y;
      drh := dmSave.PointHigh.Y;
      resh := AnimateDouble(posh, drh, dmh, 0.1);

      dmo := Abs(dmsave.PointOpen.Y - dm.PointOpen.Y) / AnimationFactor;
      poso := dm.PointOpen.Y;
      dro := dmSave.PointOpen.Y;
      reso := AnimateDouble(poso, dro, dmo, 0.1);

      dmc := Abs(dmsave.PointClose.Y - dm.PointClose.Y) / AnimationFactor;
      posc := dm.PointClose.Y;
      drc := dmSave.PointClose.Y;
      resc := AnimateDouble(posc, drc, dmc, 0.1);

      dml := Abs(dmsave.PointLow.Y - dm.PointLow.Y) / AnimationFactor;
      posl := dm.PointLow.Y;
      drl := dmSave.PointLow.Y;
      resl := AnimateDouble(posl, drl, dml, 0.1);

      if resh or reso or resc or resl then
      begin
        dm.PointHigh.Y := posh;
        dm.PointLow.Y := posl;
        dm.PointOpen.Y := poso;
        dm.PointClose.Y := posc;
        FDrawMultiPoints[I] := dm;
        RepaintChart;
      end
      else
      begin
        dm.DoAnimation := False;
        dm.PointHigh.Y := dmsave.PointHigh.Y;
        dm.PointLow.Y := dmsave.PointLow.Y;
        dm.PointClose.Y := dmsave.PointLow.Y;
        dm.PointOpen.Y := dmsave.PointOpen.Y;
        FDrawMultiPoints[I] := dm;
        RepaintChart;
      end;

      if (I < FDrawMultiPoints.Count - 1) and ((AnimationFlow and (dmh < 7.5)) or not FAnimationFlow) then
      begin
        dm := FDrawMultiPoints[I + 1];
        dm.DoAnimation := True;
        FDrawMultiPoints[I + 1] := dm;
      end;
    end;
  end;


  for I := 0 to FDrawPoints.Count - 1 do
  begin
    dp := FDrawPoints[I];
    if FRevert then
      dpSave := FAnimateDrawPointsReverse[I]
    else
      dpSave := FAnimateDrawPoints[I];

    if dp.DoAnimation then
    begin
      dt := Abs(dpsave.Point.Y - dp.Point.Y) / AnimationFactor;
      post := dp.Point.Y;
      drt := dpSave.Point.Y;
      rest := AnimateDouble(post, drt, dt, 0.1);

      if rest then
      begin
        dp.Point.Y := post;
        FDrawPoints[I] := dp;
        RepaintChart;
      end
      else
      begin
        dp.DoAnimation := False;
        dp.Point.Y := dpsave.Point.Y;
        FDrawPoints[I] := dp;
        RepaintChart;
      end;

      if (I < FDrawPoints.Count - 1) and ((AnimationFlow and (dt < 7.5)) or not FAnimationFlow) then
      begin
        dp := FDrawPoints[I + 1];
        dp.DoAnimation := True;
        FDrawPoints[I + 1] := dp;
      end;
    end;
  end;

  for I := 0 to FDrawSlices.Count - 1 do
  begin
    ds := FDrawSlices[I];
    if FRevert then
      dsSave := FAnimateDrawSlicesReverse[I]
    else
      dsSave := FAnimateDrawSlices[I];

    if ds.DoAnimation then
    begin
      if ds.AnimateEndPoint then
      begin
        dt := Abs(dssave.EndPoint.Y - ds.EndPoint.Y) / AnimationFactor;
        post := ds.EndPoint.Y;
        drt := dssave.EndPoint.Y;
        rest := AnimateDouble(post, drt, dt, 0.1);
      end
      else
      begin
        dt := Abs(dssave.SweepAngle - ds.SweepAngle) / AnimationFactor;
        post := ds.SweepAngle;
        drt := dssave.SweepAngle;
        rest := AnimateDouble(post, drt, dt, 0.1);
      end;

      if rest then
      begin
        if ds.AnimateEndPoint then
          ds.EndPoint.Y := post
        else
          ds.SweepAngle := post;

        FDrawSlices[I] := ds;
        RepaintChart;
      end
      else
      begin
        ds.DoAnimation := False;
        if ds.AnimateEndPoint then
          ds.EndPoint.Y := dssave.EndPoint.Y
        else
          ds.SweepAngle := dssave.SweepAngle;

        FDrawSlices[I] := ds;
        RepaintChart;
      end;

      if (I < FDrawSlices.Count - 1) and ((AnimationFlow and (dt < 0.1)) or not FAnimationFlow) then
      begin
        ds := FDrawSlices[I + 1];
        ds.DoAnimation := True;
        ds.AnimateEndPoint := ChartType = ctSpider;
        FDrawSlices[I + 1] := ds;
      end;
    end;
  end;

  iCount := 0;
  for I := 0 to FDrawRects.Count - 1 do
  begin
    if FDrawRects[I].DoAnimation then
      Inc(iCount);
  end;

  for I := 0 to FDrawLines.Count - 1 do
  begin
    if FDrawLines[I].DoAnimation then
      Inc(iCount);
  end;

  for I := 0 to FDrawPoints.Count - 1 do
  begin
    if FDrawPoints[I].DoAnimation then
      Inc(iCount);
  end;

  for I := 0 to FDrawSlices.Count - 1 do
  begin
    if FDrawSlices[I].DoAnimation then
      Inc(iCount);
  end;

  for I := 0 to FDrawMultiPoints.Count - 1 do
  begin
    if FDrawMultiPoints[I].DoAnimation then
      Inc(iCount);
  end;

  if iCount = 0 then
  begin
    c := Chart;
    if Assigned(c) then
      c.DoAnimateSerieFinished(Self);
    FAnimateTimer.Enabled := False;
  end;
end;

function TTMSFMXChartSerie.AddPoint(AYValue: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValue;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddMultiPoint(AYValueOpen, AYValueHigh, AYValueLow,
  AYValueClose: Double; AColor: TAlphaColor;
  ALegendText: String): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValueOpen;
  Result.YValueVariable := AYValueOpen;
  Result.YValueOpen := AYValueOpen;
  Result.YValueLow := AYValueLow;
  Result.YValueClose := AYValueClose;
  Result.YValueHigh := AYValueHigh;
  Result.YValueMedian := AYValueOpen;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddMultiPoint(AYValueOpen, AYValueHigh, AYValueLow,
  AYValueClose: Double; AXValueText: String; AColor: TAlphaColor;
  ALegendText: String): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValueOpen;
  Result.YValueVariable := AYValueOpen;
  Result.YValueOpen := AYValueOpen;
  Result.YValueLow := AYValueLow;
  Result.YValueClose := AYValueClose;
  Result.YValueHigh := AYValueHigh;
  Result.XValueText := AXValueText;
  Result.YValueMedian := AYValueOpen;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddMultiPoint(AYValueMinimum, AYValueQ1,
  AYValueMedian, AYValueQ3, AYValueMaximum: Double;
  AColor: TAlphaColor; ALegendText: String): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValueQ3;
  Result.YValueVariable := AYValueQ3;
  Result.YValueOpen := AYValueQ3;
  Result.YValueLow := AYValueMinimum;
  Result.YValueClose := AYValueQ1;
  Result.YValueHigh := AYValueMaximum;
  Result.YValueMedian := AYValueMedian;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddMultiPoint(AYValueMinimum, AYValueQ1,
  AYValueMedian, AYValueQ3, AYValueMaximum: Double; AXValueText: String;
  AColor: TAlphaColor; ALegendText: String): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValueQ3;
  Result.YValueVariable := AYValueQ3;
  Result.YValueOpen := AYValueQ3;
  Result.YValueLow := AYValueMinimum;
  Result.YValueClose := AYValueQ1;
  Result.YValueHigh := AYValueMaximum;
  Result.YValueMedian := AYValueMedian;
  Result.XValueText := AXValueText;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddPoint(AYValue: Double;
  AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValue;
  Result.XValueText := AXValueText;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddSecondPoint(AYValue, AYValueSecond: Double;
  AColor: TAlphaColor; ALegendText: String): TTMSFMXChartPoint;
begin
  Result := AddPoint(AYValue, AColor, ALegendText);
  Result.YValueSecond := AYValueSecond;
end;

function TTMSFMXChartSerie.AddSecondPoint(AYValue, AYValueSecond: Double;
  AXValueText: String; AColor: TAlphaColor;
  ALegendText: String): TTMSFMXChartPoint;
begin
  Result := AddPoint(AYValue, AXValueText, AColor, ALegendText);
  Result.YValueSecond := AYValueSecond;
end;

function TTMSFMXChartSerie.AddVariablePoint(AYValue, AYValueVariable: Double;
  AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValueVariable;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddVariablePoint(AYValue, AYValueVariable: Double;
  AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValueVariable;
  Result.XValueText := AXValueText;
  Result.Color := AColor;
  Result.LegendText := ALegendText;
end;

function TTMSFMXChartSerie.AddXYPoint(AXValue,
  AYValue: Double; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.XValue := AXValue;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValue;
  Result.LegendText := ALegendText;
  Result.Color := AColor;
end;

function TTMSFMXChartSerie.AddXYPoint(AXValue, AYValue: Double;
  AXValueText: String; AColor: TAlphaColor = claNull; ALegendText: String = ''): TTMSFMXChartPoint;
begin
  Result := Points.Add;
  Result.XValue := AXValue;
  Result.YValue := AYValue;
  Result.YValueVariable := AYValue;
  Result.XValueText := AXValueText;
  Result.LegendText := ALegendText;
  Result.Color := AColor;
end;

procedure TTMSFMXChartSerie.Animate;
var
  dr: TTMSFMXChartDrawRect;
  dl: TTMSFMXChartDrawLine;
  dp: TTMSFMXChartDrawPoint;
  dm: TTMSFMXChartDrawMultiPoint;
  ds: TTMSFMXChartDrawSlice;
  c: TTMSFMXChart;
begin
  if FDrawRects.Count > 0 then
  begin
    dr := FDrawRects[0];
    dr.DoAnimation := True;
    FDrawRects[0] := dr;
  end;

  if FDrawLines.Count > 0 then
  begin
    dl := FDrawLines[0];
    dl.DoAnimation := True;
    FDrawLines[0] := dl;
  end;

  if FDrawPoints.Count > 0 then
  begin
    dp := FDrawPoints[0];
    dp.DoAnimation := True;
    FDrawPoints[0] := dp;
  end;

  if FDrawSlices.Count > 0 then
  begin
    ds := FDrawSlices[0];
    ds.DoAnimation := True;
    ds.AnimateEndPoint := ChartType = ctSpider;
    FDrawSlices[0] := ds;
  end;

  if FDrawMultiPoints.Count > 0 then
  begin
    dm := FDrawMultiPoints[0];
    dm.DoAnimation := True;
    FDrawMultiPoints[0] := dm;
  end;

  FAnimateTimer.Enabled := True;
  c := Chart;
  if Assigned(c) then
    c.DoAnimateSerieStarted(Self);  
end;

procedure TTMSFMXChartSerie.Assign(Source: TPersistent);
begin
  if (Source is TTMSFMXChartSerie) then
  begin
    FGroupIndex := (Source as TTMSFMXChartSerie).GroupIndex;
    FEnable3D := (Source as TTMSFMXChartSerie).Enable3D;
    FOffset3DX := (Source as TTMSFMXChartSerie).Offset3DX;
    FOffset3DY := (Source as TTMSFMXChartSerie).Offset3DY;
    FMinXOffsetPercentage := (Source as TTMSFMXChartSerie).MinXOffsetPercentage;
    FMinYOffsetPercentage := (Source as TTMSFMXChartSerie).MinYOffsetPercentage;
    FMaxXOffsetPercentage := (Source as TTMSFMXChartSerie).MaxXOffsetPercentage;
    FMaxYOffsetPercentage := (Source as TTMSFMXChartSerie).MaxYOffsetPercentage;
    FXGrid.Assign((Source as TTMSFMXChartSerie).XGrid);
    FXValues.Assign((Source as TTMSFMXChartSerie).XValues);
    FYGrid.Assign((Source as TTMSFMXChartSerie).YGrid);
    FYValues.Assign((Source as TTMSFMXChartSerie).YValues);
    FPoints.Assign((Source as TTMSFMXChartSerie).Points);
    FBar.Assign((Source as TTMSFMXChartSerie).Bar);
    FMultiPoint.Assign((Source as TTMSFMXChartSerie).MultiPoint);
    FLegend.Assign((Source as TTMSFMXChartSerie).Legend);
    FPie.Assign((Source as TTMSFMXChartSerie).Pie);
    FMarkers.Assign((Source as TTMSFMXChartSerie).Markers);
    FLabels.Assign((Source as TTMSFMXChartSerie).Labels);
    FMode := (Source as TTMSFMXChartSerie).Mode;
    FTag := (Source as TTMSFMXChartSerie).Tag;
    FLegendText := (Source as TTMSFMXChartSerie).LegendText;
    FShowInLegend := (Source as TTMSFMXChartSerie).ShowInLegend;
    FVisible := (Source as TTMSFMXChartSerie).Visible;
    FChartType := (Source as TTMSFMXChartSerie).ChartType;
    FAutoXRange := (Source as TTMSFMXChartSerie).AutoXRange;
    FAutoYRange := (Source as TTMSFMXChartSerie).AutoYRange;
    FMinY := (Source as TTMSFMXChartSerie).MinY;
    FMaxY := (Source as TTMSFMXChartSerie).MaxY;
    FMinX := (Source as TTMSFMXChartSerie).MinX;
    FMaxX := (Source as TTMSFMXChartSerie).MaxX;
    FFillOpacity := (Source as TTMSFMXChartSerie).FillOpacity;
    FFill.Assign((Source as TTMSFMXChartSerie).Fill);
    FStroke.Assign((Source as TTMSFMXChartSerie).Stroke);
    FStrokeOpacity := (Source as TTMSFMXChartSerie).StrokeOpacity;
    FStrokeJoin := (Source as TTMSFMXChartSerie).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChartSerie).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartSerie).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartSerie).StrokeDash;
  end;
end;

procedure TTMSFMXChartSerie.CalculateAnimation;
var
  I: Integer;
  dr: TTMSFMXChartDrawRect;
  dl: TTMSFMXChartDrawLine;
  dp: TTMSFMXChartDrawPoint;
  dm: TTMSFMXChartDrawMultiPoint;
  ds: TTMSFMXChartDrawSlice;
  zrv: Double;
begin
  zrv := ValueToY(ZeroReferenceValue);

  for I := 0 to FDrawRects.Count - 1 do
  begin
    dr := FDrawRects[I];
    dr.Rect.Top := zrv;
    dr.Rect.Bottom := zrv;
    FAnimateDrawRects.Add(dr);
    FAnimateDrawRectsReverse.Add(FDrawRects[I]);
  end;

  for I := 0 to FDrawLines.Count - 1 do
  begin
    dl := FDrawLines[I];
    dl.StartPoint.Y := zrv;
    dl.EndPoint.Y := zrv;
    FAnimateDrawLines.Add(dl);
    FAnimateDrawLinesReverse.Add(FDrawLines[I]);
  end;

  for I := 0 to FDrawPoints.Count - 1 do
  begin
    dp := FDrawPoints[I];
    dp.Point.Y := zrv;
    FAnimateDrawPoints.Add(dp);
    FAnimateDrawPointsReverse.Add(FDrawPoints[I]);
  end;

  for I := 0 to FDrawMultiPoints.Count - 1 do
  begin
    dm := FDrawMultiPoints[I];
    dm.PointHigh.Y := zrv;
    dm.PointLow.Y := zrv;
    dm.PointOpen.Y := zrv;
    dm.PointClose.Y := zrv;
    FAnimateDrawMultiPoints.Add(dm);
    FAnimateDrawMultiPointsReverse.Add(FDrawMultiPoints[I]);
  end;

  for I := 0 to FDrawSlices.Count - 1 do
  begin
    ds := FDrawSlices[I];
    ds.SweepAngle := 0;
    ds.EndPoint.Y := ds.CenterPoint.Y;
    FAnimateDrawSlices.Add(ds);
    FAnimateDrawSlicesReverse.Add(FDrawSlices[I]);
  end;

  if FRevert then
  begin
    FDrawLines.Clear;
    FDrawLines.AddRange(FAnimateDrawLines);
    FDrawRects.Clear;
    FDrawRects.AddRange(FAnimateDrawRects);
    FDrawPoints.Clear;
    FDrawPoints.AddRange(FAnimateDrawPoints);
    FDrawSlices.Clear;
    FDrawSlices.AddRange(FAnimateDrawSlices);
    FDrawMultiPoints.Clear;
    FDrawMultiPoints.AddRange(FAnimateDrawMultiPoints);
  end;
end;

procedure TTMSFMXChartSerie.CalculateAnnotations;
var
  bmp: TBitmap;
  dann: TTMSFMXChartDrawAnnotation;
  I: Integer;
  ann: TTMSFMXChartAnnotationVirtual;
  th: Double;
  ft: TFont;
  c: TTMSFMXChart;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if FDrawAnnotations.Count > 0 then
  begin
    ft := TFont.Create;
    bmp := TBitmap.Create(1, 1);
    for I := 0 to FDrawAnnotations.Count - 1 do
    begin
      dann := FDrawAnnotations[I];
      ann := dann.VirtualAnnotationReference;

      if ann.AutoSize and (ann.Text <> '') then
      begin
        if Assigned(ann.Annotation) then
          ft.Assign(ann.Annotation.Font);

        c.DoCustomizeAnnotationFont(Self, dann.VirtualPointReference, ann, ft);

        bmp.Canvas.Font.Assign(ft);
        th := bmp.Canvas.TextHeight(ann.Text) + 2;
        dann.TextHeight := th;
        dann.TextWidth := bmp.Canvas.TextWidth(ann.Text) + 4;
      end
      else
      begin
        dann.TextHeight := ann.Height;
        dann.TextWidth := ann.Width;
      end;
      dann.AnnotationPoint.X := dann.AnnotationPoint.X + ann.OffsetX;
      if IsPie then
        dann.AnnotationPoint.Y := dann.AnnotationPoint.Y + ann.OffsetY + dann.TextHeight / 2
      else
        dann.AnnotationPoint.Y := dann.AnnotationPoint.Y + ann.OffsetY;

      dann.Rect := RectF(dann.AnnotationPoint.X - dann.TextWidth / 2, dann.AnnotationPoint.Y - dann.TextHeight, dann.AnnotationPoint.X + dann.TextWidth / 2, dann.AnnotationPoint.Y);
      FDrawAnnotations[I] := dann;
    end;
    ft.Free;
    bmp.Free;
  end;
end;

procedure TTMSFMXChartSerie.CalculateArea;
var
  I: Integer;
begin
  if IsArea and (FDrawLines.Count > 0) then
  begin
    FDrawPath.MoveTo(FDrawLines[0].StartPoint);
    for I := 0 to FDrawLines.Count - 1 do
      FDrawPath.LineTo(FDrawLines[I].EndPoint);
    FDrawPath.ClosePath;
  end;
end;

procedure TTMSFMXChartSerie.CalculateDrawPoints;
var
  I: Integer;
  dr, drPrev: TTMSFMXChartDrawRect;
  dl, dlPrev, dlPrevn: TTMSFMXChartDrawLine;
  ds: TTMSFMXChartDrawSlice;
  dp: TTMSFMXChartDrawPoint;
  dm: TTMSFMXChartDrawMultiPoint;
  ohw: Single;
  dlbl: TTMSFMXChartDrawLabel;
  dann: TTMSFMXChartDrawAnnotation;
  pt, ptNext: TTMSFMXChartPointVirtual;
  xVal, yVal, xValNext, yValNext, yValPrev, yValLow, yValOpen, yValMedian, yValClose, yValHigh: Double;
  r: TRectF;
  c: TTMSFMXChart;
  bwpos, bw, bwtot: Double;
  zrv: Double;
  start, stop: Integer;
  ann: TTMSFMXChartAnnotationVirtual;
  a: Integer;
  vl, vlNext, vlstmax, vlstmaxnext, pttot, ptmax, vlHigh, vlLow, vlClose, vlOpen, vlMedian: Double;
  pir: TRectF;
  st, swp: Double;
  cnt: Integer;
  cnta: Integer;
  {$IFDEF DELPHIXE8_LVL}
  tmp: TTMSFMXChartDrawLines;
  {$ENDIF}
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  FDrawLabels.Clear;
  FDrawAnnotations.Clear;
  FDrawPoints.Clear;
  FAnimateDrawPoints.Clear;
  FAnimateDrawPointsReverse.Clear;
  FDrawMultiPoints.Clear;
  FAnimateDrawMultiPoints.Clear;
  FAnimateDrawMultiPointsReverse.Clear;
  FDrawSlices.Clear;
  FAnimateDrawSlices.Clear;
  FAnimateDrawSlicesReverse.Clear;
  FDrawLines.Clear;
  FAnimateDrawLines.Clear;
  FAnimateDrawLinesReverse.Clear;
  FDrawRects.Clear;
  FAnimateDrawRects.Clear;
  FAnimateDrawRectsReverse.Clear;
  FDrawPath.Clear;

  r := c.GetSeriesRectangle;
  zrv := ValueToY(ZeroReferenceValue);

  start := GetStartPoint;
  stop := GetEndPoint;

  case ChartType of
    ctPie, ctVariableRadiusPie, ctSizedPie, ctSpider:
    begin
      pttot := GetPointsTotal;
      if IsSpider then
        ptmax := YMax
      else
        ptmax := GetPointsMax;

      if pttot > 0 then
      begin
        pir := GetPieRect;
        st := Pie.StartAngle;
        cnt := GetPointsCount;
        for I := start to stop do
        begin
          pt := GetPoint(i);
          if not (pt.YValue > 0) then
            Continue;

          ds.Radius := PointF(pir.Width / 2, pir.Height / 2);
          swp := Pie.SweepAngle / pttot * pt.YValue;

          if (ChartType = ctSizedPie) or (ChartType = ctSpider) then
          begin
            swp := Pie.SweepAngle / Max(1, cnt);
            ds.Radius.X := (ds.Radius.X / Max(1, ptmax)) * pt.YValue;
            ds.Radius.Y := (ds.Radius.Y / Max(1, ptmax)) * pt.YValue;
          end
          else if (ChartType = ctVariableRadiusPie) then
          begin
            ds.Radius.X := (ds.Radius.X / Max(1, ptmax)) * pt.YValueVariable;
            ds.Radius.Y := (ds.Radius.Y / Max(1, ptmax)) * pt.YValueVariable;
          end;

          ds.CenterPoint := pir.CenterPoint;
          ds.CenterPoint.X := ds.CenterPoint.X + pt.Explode * Cos(DegToRad(st + swp / 2));
          ds.CenterPoint.Y := ds.CenterPoint.Y + pt.Explode * Sin(DegToRad(st + swp / 2));
          ds.InnerRadius := PointF(Pie.InnerSize / 2, Pie.InnerSize / 2);
          ds.StartPoint.X := ds.CenterPoint.X + ds.InnerRadius.X * Cos(DegToRad(st));
          ds.StartPoint.Y := ds.CenterPoint.Y + ds.InnerRadius.Y * Sin(DegToRad(st));

          xval := ds.Radius.X * Cos(DegToRad(st + swp / 2));
          yval := ds.Radius.Y * Sin(DegToRad(st + swp / 2));

          ds.EndPoint.X := ds.CenterPoint.X + xval;
          ds.EndPoint.Y := ds.CenterPoint.Y + yval;
          ds.StartAngle := st;
          ds.SweepAngle := swp;
          ds.Reference := pt.Point;
          ds.VirtualReference := pt;
          FDrawSlices.Add(ds);

          dlbl.Reference := pt.Point;
          dlbl.VirtualReference := pt;
          if ChartType = ctSpider then
          begin
            dlbl.Point.X := ds.CenterPoint.X + xval;
            dlbl.Point.Y := ds.CenterPoint.Y + yval;
          end
          else
          begin
            dlbl.Point.X := ds.CenterPoint.X + xval / 4 * 3;
            dlbl.Point.Y := ds.CenterPoint.Y + yval / 4 * 3;
          end;

          dlbl.Value := pt.YValue;
          FDrawLabels.Add(dlbl);

          cnta := 0;
          if Assigned(pt.Point) then
            cnta := pt.Point.Annotations.Count;

          c.DoGetNumberOfAnnotations(Self, pt, cnta);
          for a := 0 to cnta - 1 do
          begin
            if Assigned(pt.Point) and (a >= 0) and (a <= pt.Point.Annotations.Count - 1) then
              ann := pt.Point.Annotations[a].GetValue
            else
              ann := GetDefaultVirtualAnnotation;

            ann.Index := a;
            c.DoGetAnnotation(Self, pt, a, ann);
            if ann.Visible then
            begin
              dann.PointReference := pt.Point;
              dann.VirtualPointReference := pt;
              dann.AnnotationReference := ann.Annotation;
              dann.VirtualAnnotationReference := ann;
              dann.AnnotationPoint := PointF(ds.CenterPoint.X + xval, ds.CenterPoint.Y + yval);
              dann.StartPoint := dann.AnnotationPoint;
              FDrawAnnotations.Add(dann);
            end;
          end;

          st := st + swp;
        end;
      end;
    end;
    ctLine, ctArea, ctStackedArea, ctStackedPercentageArea, ctDigitalLine, ctXYLine, ctBand, ctOHLC, ctCandleStick, ctBoxPlot:
    begin
      for I := start to stop do
      begin
        pt := GetPoint(i);
        ptNext := GetPoint(i + 1);
        xVal := ValueToX(pt.XValue);
        xValNext := ValueToX(ptNext.XValue);
        vl := pt.YValue;
        vlHigh := pt.YValueHigh;
        vlLow := pt.YValueLow;
        vlClose := pt.YValueClose;
        vlOpen := pt.YValueOpen;
        vlMedian := pt.YValueMedian;
        vlNext := ptNext.YValue;

        vlstmax := GetStackedMax(i);
        vlstmaxnext := GetStackedMax(i + 1);

        dl.Bottom := False;
        dl.StackedValueStart := 0;
        dl.StackedValueEnd := 0;

        if (ChartType = ctStackedArea) or (ChartType = ctStackedPercentageArea) then
        begin
          if ChartType = ctStackedPercentageArea then
          begin
            if vlstmax > 0 then
              vl := vl / vlstmax * 100;

            if vlstmaxnext > 0 then
              vlNext := vlNext / vlstmaxnext * 100;
          end;

          dlPrev := GetDefaultDrawLine;
          if GetStackedDrawLine(i + 1 - start, dlPrev) then
          begin
            yVal := ValueToY(vl + dlPrev.StackedValueStart);
            yValNext := ValueToY(vlNext + dlPrev.StackedValueEnd);
            dl.StackedValueStart := dlPrev.StackedValueStart + vl;
            dl.StackedValueEnd := dlPrev.StackedValueEnd + vlNext;
          end
          else
          begin
            yVal := ValueToY(vl);
            yValNext := ValueToY(vlNext);
            dl.StackedValueStart := dl.StackedValueStart + vl;
            dl.StackedValueEnd := dl.StackedValueEnd + vlNext;
          end;
        end
        else
        begin
          yVal := ValueToY(vl);
          yValNext := ValueToY(vlNext);
        end;

        yValLow := ValueToY(vlLow);
        yValOpen := ValueToY(vlOpen);
        yValMedian := ValueToY(vlMedian);
        yValClose := ValueToY(vlClose);
        yValHigh := ValueToY(vlHigh);

        if ChartType = ctDigitalLine then
        begin
          dl.StartPoint := PointF(xVal, yVal);
          dl.EndPoint := PointF(xValNext, yVal);
          dl.StartReference := pt.Point;
          dl.EndReference := pt.Point;
          dl.StartVirtualReference := pt;
          dl.EndVirtualReference := pt;

          FDrawLines.Add(dl);
          dl.StartPoint := PointF(xValNext, yVal);
          dl.EndPoint := PointF(xValNext, yValNext);
          dl.StartReference := pt.Point;
          dl.StartVirtualReference := pt;
          dl.EndReference := pt.Point;
          dl.EndVirtualReference := pt;
          FDrawLines.Add(dl);
        end
        else
        begin
          dl.StartPoint := PointF(xVal, yVal);
          dl.EndPoint := PointF(xValNext, yValNext);
          dl.StartReference := pt.Point;
          dl.StartVirtualReference := pt;
          dl.EndReference := pt.Point;
          dl.EndVirtualReference := pt;
          FDrawLines.Add(dl);
        end;

        dp.Point := PointF(xVal, yVal);
        dp.Reference := pt.Point;
        dp.VirtualReference := pt;
        FDrawPoints.Add(dp);

        if I = stop then
        begin
          dp.Point := PointF(xValNext, yValNext);
          dp.Reference := ptNext.Point;
          dp.VirtualReference := ptNext;
          FDrawPoints.Add(dp);
        end;

        dlbl.Point := PointF(xVal, yVal);
        dlbl.Reference := pt.Point;
        dlbl.VirtualReference := pt;
        dlbl.Value := pt.YValue;
        if IsStacked and (Labels.Mode = lmStacked) then
          dlbl.Value := dl.StackedValueStart;

        FDrawLabels.Add(dlbl);

        dm.PointHigh := PointF(xVal, yValHigh);
        dm.PointLow := PointF(xVal, yValLow);
        dm.PointClose := PointF(xVal, yValClose);
        dm.PointOpen := PointF(xVal, yValOpen);
        dm.PointMedian := PointF(xVal, yValMedian);
        dm.Reference := pt.Point;
        ohw := GetMultiPointWidth;
        if yValOpen > yValClose then
          dm.Rect := RectF(xVal - ohw / 2, yValClose, xVal + ohw / 2, yValOpen)
        else
          dm.Rect := RectF(xVal - ohw / 2, yValOpen, xVal + ohw / 2, yValClose);
        dm.VirtualReference := pt;

        FDrawMultiPoints.Add(dm);

        cnta := 0;
        if Assigned(pt.Point) then
          cnta := pt.Point.Annotations.Count;

        c.DoGetNumberOfAnnotations(Self, pt, cnta);
        for a := 0 to cnta - 1 do
        begin
          if Assigned(pt.Point) and (a >= 0) and (a <= pt.Point.Annotations.Count - 1) then
            ann := pt.Point.Annotations[a].GetValue
          else
            ann := GetDefaultVirtualAnnotation;

          ann.Index := a;
          c.DoGetAnnotation(Self, pt, a, ann);
          if ann.Visible then
          begin
            dann.PointReference := pt.Point;
            dann.VirtualPointReference := pt;
            dann.AnnotationReference := ann.Annotation;
            dann.VirtualAnnotationReference := ann;
            dann.AnnotationPoint := PointF(xVal, yVal);
            dann.StartPoint := dann.AnnotationPoint;
            FDrawAnnotations.Add(dann);
          end;
        end;

        if I = stop then
        begin
          dlbl.Point := PointF(xValNext, yValNext);
          dlbl.Reference := ptNext.Point;
          dlbl.VirtualReference := ptNext;
          dlbl.Value := ptNext.YValue;
          if IsStacked and (Labels.Mode = lmStacked) then
            dlbl.Value := dl.StackedValueEnd;
          FDrawLabels.Add(dlbl);

          cnta := 0;
          if Assigned(ptNext.Point) then
            cnta := ptNext.Point.Annotations.Count;

          c.DoGetNumberOfAnnotations(Self, ptNext, cnta);
          for a := 0 to cnta - 1 do
          begin
            if Assigned(ptNext.Point) and (a >= 0) and (a <= ptNext.Point.Annotations.Count - 1) then
              ann := ptNext.Point.Annotations[a].GetValue
            else
              ann := GetDefaultVirtualAnnotation;

            ann.Index := a;
            c.DoGetAnnotation(Self, ptNext, a, ann);
            if ann.Visible then
            begin
              dann.PointReference := ptNext.Point;
              dann.VirtualPointReference := ptNext;
              dann.AnnotationReference := ann.Annotation;
              dann.VirtualAnnotationReference := ann;
              dann.AnnotationPoint := PointF(xValNext, yValNext);
              dann.StartPoint := dann.AnnotationPoint;
              FDrawAnnotations.Add(dann);
            end;
          end;
        end;
      end;

      if ChartType = ctBand then
      begin
        pt := GetPoint(stop + 1);
        pt.SecondValue := True;
        xVal := ValueToX(pt.XValue);
        vl := pt.YValue;
        vlNext := pt.YValueSecond;

        yVal := ValueToY(vl);
        yValNext := ValueToY(vlNext);

        dl.StartPoint := PointF(xVal, yVal);
        dl.EndPoint := PointF(xVal, yValNext);
        dl.StartReference := pt.Point;
        dl.StartVirtualReference := pt;
        dl.EndReference := pt.Point;
        dl.EndVirtualReference := pt;
        FDrawLines.Add(dl);

        dp.Point := PointF(xVal, yVal);
        dp.Reference := pt.Point;
        dp.VirtualReference := pt;
        FDrawPoints.Add(dp);

        for I := stop + 1 downto start + 1 do
        begin
          pt := GetPoint(i);
          pt.SecondValue := True;
          ptNext := GetPoint(i - 1);
          ptNext.SecondValue := True;

          xVal := ValueToX(pt.XValue);
          xValNext := ValueToX(ptNext.XValue);
          vl := pt.YValueSecond;
          vlNext := ptNext.YValueSecond;

          yVal := ValueToY(vl);
          yValNext := ValueToY(vlNext);

          dl.StartPoint := PointF(xVal, yVal);
          dl.EndPoint := PointF(xValNext, yValNext);
          dl.StartReference := pt.Point;
          dl.StartVirtualReference := pt;
          dl.EndReference := pt.Point;
          dl.EndVirtualReference := pt;
          FDrawLines.Add(dl);

          dp.Point := PointF(xVal, yVal);
          dp.Reference := pt.Point;
          dp.VirtualReference := pt;
          FDrawPoints.Add(dp);

          if I = start + 1 then
          begin
            dp.Point := PointF(xValNext, yValNext);
            dp.Reference := ptNext.Point;
            dp.VirtualReference := ptNext;
            FDrawPoints.Add(dp);
          end;

          dlbl.Point := PointF(xVal, yVal);
          dlbl.Reference := pt.Point;
          dlbl.VirtualReference := pt;
          dlbl.Value := pt.YValueSecond;

          FDrawLabels.Add(dlbl);

          if I = start + 1 then
          begin
            dlbl.Point := PointF(xValNext, yValNext);
            dlbl.Reference := ptNext.Point;
            dlbl.VirtualReference := ptNext;
            dlbl.Value := ptNext.YValueSecond;
            FDrawLabels.Add(dlbl);
          end;
        end;
      end;

      if IsArea and (FDrawLines.Count > 0) and not (ChartType = ctBand) then
      begin
        dlPrev := GetDefaultDrawLine;
        dlPrevn := GetDefaultDrawLine;
        if GetStackedDrawLine(1, dlPrev) and GetStackedDrawLine(FDrawLines.Count, dlPrevn) then
        begin
          dl.StartPoint := FDrawLines[FDrawLines.Count - 1].EndPoint;
          dl.EndPoint := PointF(FDrawLines[FDrawLines.Count - 1].EndPoint.X, dlPrevn.EndPoint.Y);
          dl.StartReference := nil;
          dl.EndReference := nil;
          dl.StartVirtualReference := GetDefaultVirtualPoint;
          dl.EndVirtualReference := GetDefaultVirtualPoint;
          dl.Bottom := False;
          FDrawLines.Add(dl);

          dl.StartPoint := PointF(FDrawLines[0].StartPoint.X, dlPrev.StartPoint.Y);
          dl.EndPoint := FDrawLines[0].StartPoint;
          dl.StartReference := nil;
          dl.EndReference := nil;
          dl.StartVirtualReference := GetDefaultVirtualPoint;
          dl.EndVirtualReference := GetDefaultVirtualPoint;
          dl.Bottom := False;
          {$IFDEF DELPHIXE8_LVL}
          tmp := TTMSFMXChartDrawLines.Create;
          tmp.AddRange(FDrawLines);
          FDrawLines.Clear;
          FDrawLines.Add(dl);
          FDrawLines.AddRange(tmp);
          tmp.Free;
          {$ELSE}
          FDrawLines.Insert(0, dl);
          {$ENDIF}

          for I := FDrawLines.Count - 3 downto 1 do
          begin
            if GetStackedDrawLine(I, dlPrev) then
            begin
              dl := dlPrev;
              dl.StartReference := nil;
              dl.EndReference := nil;
              dl.StartVirtualReference := GetDefaultVirtualPoint;
              dl.EndVirtualReference := GetDefaultVirtualPoint;
              dl.Bottom := True;
              FDrawLines.Add(dl);
            end;
          end;
        end
        else
        begin
          dl.StartPoint := FDrawLines[FDrawLines.Count - 1].EndPoint;
          dl.EndPoint := PointF(FDrawLines[FDrawLines.Count - 1].EndPoint.X, zrv);
          dl.StartReference := nil;
          dl.EndReference := nil;
          dl.StartVirtualReference := GetDefaultVirtualPoint;
          dl.EndVirtualReference := GetDefaultVirtualPoint;
          FDrawLines.Add(dl);

          dl.StartPoint := PointF(FDrawLines[0].StartPoint.X, zrv);
          dl.EndPoint := FDrawLines[0].StartPoint;
          dl.StartReference := nil;
          dl.EndReference := nil;
          dl.StartVirtualReference := GetDefaultVirtualPoint;
          dl.EndVirtualReference := GetDefaultVirtualPoint;
          {$IFDEF DELPHIXE8_LVL}
          tmp := TTMSFMXChartDrawLines.Create;
          tmp.AddRange(FDrawLines);
          FDrawLines.Clear;
          FDrawLines.Add(dl);
          FDrawLines.AddRange(tmp);
          tmp.Free;
          {$ELSE}
          FDrawLines.Insert(0, dl);
          {$ENDIF}
        end;
      end;
    end;
    ctMarker, ctXYMarker:
    begin
      for I := start to stop do
      begin
        pt := GetPoint(i);
        xVal := ValueToX(pt.XValue);
        yVal := ValueToY(pt.YValue);
        dp.Reference := pt.Point;
        dp.VirtualReference := pt;
        dp.Point := PointF(xVal, yVal);
        FDrawPoints.Add(dp);
        dlbl.Reference := pt.Point;
        dlbl.VirtualReference := pt;
        dlbl.Point := PointF(xVal, yVal);
        dlbl.Value := pt.YValue;
        FDrawLabels.Add(dlbl);

        cnta := 0;
        if Assigned(pt.Point) then
          cnta := pt.Point.Annotations.Count;

        c.DoGetNumberOfAnnotations(Self, pt, cnta);
        for a := 0 to cnta - 1 do
        begin
          if Assigned(pt.Point) and (a >= 0) and (a <= pt.Point.Annotations.Count - 1) then
            ann := pt.Point.Annotations[a].GetValue
          else
            ann := GetDefaultVirtualAnnotation;

          ann.Index := a;
          c.DoGetAnnotation(Self, pt, a, ann);
          if ann.Visible then
          begin
            dann.PointReference := pt.Point;
            dann.VirtualPointReference := pt;
            dann.AnnotationReference := ann.Annotation;
            dann.VirtualAnnotationReference := ann;
            dann.AnnotationPoint := PointF(xVal, yVal);
            dann.StartPoint := dann.AnnotationPoint;
            FDrawAnnotations.Add(dann);
          end;
        end;
      end;
    end;
    ctBar, ctStackedBar, ctStackedPercentageBar:
    begin
      bwtot := GetTotalBarWidth;
      bwpos := GetBarOffset;
      if Offset then
        bwpos := bwpos + (FXScale - bwtot) / 2 - (FXScale / 2);

      bw := GetBarWidth;
      for I := start to stop do
      begin
        pt := GetPoint(i);
        dr.Reference := pt.Point;
        dr.VirtualReference := pt;
        dr.StackedValue := 0;
        xVal := ValueToX(pt.XValue);
        vlstmax := GetStackedMax(i);
        vl := pt.YValue;

        if (ChartType = ctStackedBar) or (ChartType = ctStackedPercentageBar) then
        begin
          if ChartType = ctStackedPercentageBar then
          begin
            if vlstmax > 0 then
              vl := vl / vlstmax * 100;
          end;

          drPrev := GetDefaultDrawRect;
          if GetStackedDrawRect(i - start, drPrev) then
          begin
            yVal := ValueToY(vl + drPrev.StackedValue);
            yValPrev := ValueToY(drPrev.StackedValue);
            dr.StackedValue := drPrev.StackedValue + vl;
          end
          else
          begin
            yVal := ValueToY(vl);
            yValPrev := zrv;
            dr.StackedValue := dr.StackedValue + vl;
          end;
        end
        else
        begin
          yVal := ValueToY(vl);
          yValPrev := zrv;
        end;

        if yval > yValPrev then
          dr.Rect := RectF(xVal + bwpos, yValPrev, xVal + bwpos + bw, yval)
        else
          dr.Rect := RectF(xVal + bwpos, yVal, xVal + bwpos + bw, yValPrev);

        FDrawRects.Add(dr);
        dp.Point.Y := dr.Rect.Top;
        dp.Point.X := dr.Rect.CenterPoint.X;
        dp.Reference := pt.Point;
        dp.VirtualReference := pt;
        FDrawPoints.Add(dp);

        dlbl.Point.Y := dr.Rect.Top;
        dlbl.Point.X := dr.Rect.CenterPoint.X;
        dlbl.Reference := pt.Point;
        dlbl.VirtualReference := pt;
        dlbl.Value := pt.YValue;
        if IsStacked and (Labels.Mode = lmStacked) then
          dlbl.Value := dr.StackedValue;

        FDrawLabels.Add(dlbl);

        cnta := 0;
        if Assigned(pt.Point) then
          cnta := pt.Point.Annotations.Count;

        c.DoGetNumberOfAnnotations(Self, pt, cnta);
        for a := 0 to cnta - 1 do
        begin
          if Assigned(pt.Point) and (a >= 0) and (a <= pt.Point.Annotations.Count - 1) then
            ann := pt.Point.Annotations[a].GetValue
          else
            ann := GetDefaultVirtualAnnotation;

          ann.Index := a;
          c.DoGetAnnotation(Self, pt, a, ann);
          if ann.Visible then
          begin
            dann.PointReference := pt.Point;
            dann.VirtualPointReference := pt;
            dann.AnnotationReference := ann.Annotation;
            dann.VirtualAnnotationReference := ann;
            dann.AnnotationPoint.Y := dr.Rect.Top;
            dann.AnnotationPoint.X := dr.Rect.CenterPoint.X;
            dann.StartPoint := dann.AnnotationPoint;
            FDrawAnnotations.Add(dann);
          end;
        end;
      end;
    end;
  end;
end;

procedure TTMSFMXChartSerie.CalculateLabels;
var
  bmp: TBitmap;
  th: Double;
  dlbl: TTMSFMXChartDrawLabel;
  I: Integer;
  c: TTMSFMXChart;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if Labels.Visible then
  begin
    bmp := TBitmap.Create(1, 1);
    bmp.Canvas.Font.Assign(Labels.Font);
    th := bmp.Canvas.TextHeight('gh') + 2;
    for I := 0 to FDrawLabels.Count - 1 do
    begin
      dlbl := FDrawLabels[I];
      if Labels.Format <> '' then
      begin
        case Labels.FormatType of
          vftNormal: dlbl.ValueString := Format(Labels.Format,[dlbl.Value]);
          vftFloat: dlbl.ValueString := FormatFloat(Labels.Format,dlbl.Value);
          vftDateTime: dlbl.ValueString := FormatDateTime(Labels.Format, dlbl.Value);
        end;
      end
      else
        dlbl.ValueString := FloatToStr(dlbl.Value);

      if c.FVirtualMode then
        c.DoGetSerieLabelVirtual(Self, dlbl.VirtualReference, dlbl.ValueString)
      else
        c.DoGetSerieLabel(Self, dlbl.Reference, dlbl.ValueString);

      dlbl.TextHeight := th;
      dlbl.TextWidth := bmp.Canvas.TextWidth(dlbl.ValueString) + 4;
      dlbl.Point.X := dlbl.Point.X + Labels.OffsetX;
      if IsPie then
        dlbl.Point.Y := dlbl.Point.Y + Labels.OffsetY + dlbl.TextHeight / 2
      else
        dlbl.Point.Y := dlbl.Point.Y + Labels.OffsetY;

      dlbl.Rect := RectF(dlbl.Point.X - dlbl.TextWidth / 2, dlbl.Point.Y - dlbl.TextHeight, dlbl.Point.X + dlbl.TextWidth / 2, dlbl.Point.Y);

      FDrawLabels[I] := dlbl;
    end;
    bmp.Free;
  end;
end;

procedure TTMSFMXChartSerie.CalculateMinMaxX;
var
  I: Integer;
  pt: TTMSFMXChartPointVirtual;
  c: TTMSFMXChart;
  offsxmin, offsxmax: Double;
  sta, ste, cnt: Integer;
begin
  cnt := GetPointsCount;
  if cnt = 0 then
  begin
    FXMax := 0;
    FXMin := 0;
    Exit;
  end;

  c := Chart;
  if Assigned(c) then
  begin
    sta := 0;
    ste := cnt - 1;

    case AutoXRange of
      arDisabled:
      begin
        FXMax := MaxX;
        FXMin := MinX;
      end;
      arEnabled, arCommon:
      begin
        FXMax := -MaxDouble;
        FXMin := MaxDouble;

        for I := sta to ste do
        begin
          pt := GetPoint(I);
          FXMax := Max(FXMax, pt.XValue);
          FXMin := Min(FXMin, pt.XValue);
        end;
      end;
      arEnabledZeroBased, arCommonZeroBased:
      begin
        FXMin := 0;
        FXMax := -MaxDouble;

        for I := sta to ste do
        begin
          pt := GetPoint(I);
          FXMax := Max(FXMax, pt.XValue);
        end;
      end;
    end;
  end;

  offsxmin := (FMinXOffsetPercentage * Abs(FXMax - FXMin)) / 100;
  offsxmax := (FMaxXOffsetPercentage * Abs(FXMax - FXMin)) / 100;
  FXMin := FXMin - offsxmin;
  FXMax := FXMax + offsxmax;
end;

procedure TTMSFMXChartSerie.CalculateMinMaxY;
var
  I: Integer;
  pt: TTMSFMXChartPointVirtual;
  c: TTMSFMXChart;
  offsymin, offsymax: Double;
  sta, ste, cnt: Integer;
begin
  cnt := GetPointsCount;
  if cnt = 0 then
  begin
    FYMax := 0;
    FYMin := 0;
    Exit;
  end;

  c := Chart;
  if Assigned(c) then
  begin
    sta := GetStartCountPoint;
    ste := GetEndCountPoint;

    case AutoYRange of
      arDisabled:
      begin
        FYMax := MaxY;
        FYMin := MinY;
      end;
      arEnabled, arCommon:
      begin
        FYMax := -MaxDouble;
        FYMin := MaxDouble;

        for I := sta to ste do
        begin
          pt := GetPoint(I);
          if ChartType in [ctOHLC, ctCandleStick, ctBoxPlot] then
            FYMax := Max(FYMax, pt.YValueHigh)
          else
            FYMax := Max(FYMax, pt.YValue);

          if ChartType = ctBand then
            FYMin := Min(FYMin, pt.YValueSecond)
          else if ChartType in [ctCandleStick, ctOHLC, ctBoxPlot] then
            FYMin := Min(FYMin, pt.YValueLow)
          else
            FYMin := Min(FYMin, pt.YValue);
        end;
      end;
      arEnabledZeroBased, arCommonZeroBased:
      begin
        FYMin := 0;
        FYMax := -MaxDouble;

        for I := sta to ste do
        begin
          pt := GetPoint(I);
          if ChartType in [ctOHLC, ctCandleStick, ctBoxPlot] then
            FYMax := Max(FYMax, pt.YValueHigh)
          else
            FYMax := Max(FYMax, pt.YValue);
        end;
      end;
    end;
  end;

  if IsSpider and not (AutoYRange = arDisabled) then
  begin
    offsymin := (Max(10, FMinYOffsetPercentage) * Abs(FYMax - FYMin)) / 100;
    offsymax := (Max(10, FMaxYOffsetPercentage) * Abs(FYMax - FYMin)) / 100;
  end
  else
  begin
    offsymin := (FMinYOffsetPercentage * Abs(FYMax - FYMin)) / 100;
    offsymax := (FMaxYOffsetPercentage * Abs(FYMax - FYMin)) / 100;
  end;

  FYMin := FYMin - offsymin;
  FYMax := FYMax + offsymax;
end;

procedure TTMSFMXChartSerie.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

function TTMSFMXChartSerie.Chart: TTMSFMXChart;
begin
  Result := FChart;
end;

constructor TTMSFMXChartSerie.Create(Collection: TCollection);
begin
  inherited;
  if Assigned(Collection) then
    FChart := (Collection as TTMSFMXChartSeries).Chart;

  FGroupIndex := 0;
  FSequential := True;
  FPrevXValue := -MaxDouble;
  FPoints := TTMSFMXChartPoints.Create(Self);
  FBar := TTMSFMXChartSerieBar.Create(Self);
  FMultiPoint := TTMSFMXChartSerieMultiPoint.Create(Self);
  FPie := TTMSFMXChartSeriePie.Create(Self);
  FMarkers := TTMSFMXChartSerieMarkers.Create(Self);
  FXGrid := TTMSFMXChartSerieXYGrid.Create(Self);
  FYGrid := TTMSFMXChartSerieXYGrid.Create(Self);
  FXValues := TTMSFMXChartSerieXValues.Create(Self);
  FYValues := TTMSFMXChartSerieYValues.Create(Self);
  FLabels := TTMSFMXChartSerieLabels.Create(Self);
  FLegend := TTMSFMXChartSerieLegend.Create(Self);

  FDrawLines := TTMSFMXChartDrawLines.Create;
  FAnimateDrawLines := TTMSFMXChartDrawLines.Create;
  FAnimateDrawLinesReverse := TTMSFMXChartDrawLines.Create;
  FDrawSlices := TTMSFMXChartDrawSlices.Create;
  FAnimateDrawSlices := TTMSFMXChartDrawSlices.Create;
  FAnimateDrawSlicesReverse := TTMSFMXChartDrawSlices.Create;
  FDrawMultiPoints := TTMSFMXChartDrawMultiPoints.Create;
  FAnimateDrawMultiPoints := TTMSFMXChartDrawMultiPoints.Create;
  FAnimateDrawMultiPointsReverse := TTMSFMXChartDrawMultiPoints.Create;
  FDrawPoints := TTMSFMXChartDrawPoints.Create;
  FAnimateDrawPoints := TTMSFMXChartDrawPoints.Create;
  FAnimateDrawPointsReverse := TTMSFMXChartDrawPoints.Create;
  FDrawXValues := TTMSFMXChartDrawXYValues.Create;
  FDrawYValues := TTMSFMXChartDrawXYValues.Create;
  FDrawXGridLines := TTMSFMXChartDrawXYGridLines.Create;
  FDrawYGridLines := TTMSFMXChartDrawXYGridLines.Create;
  FDrawLabels := TTMSFMXChartDrawLabels.Create;
  FDrawAnnotations := TTMSFMXChartDrawAnnotations.Create;
  
  FDrawRects := TTMSFMXChartDrawRects.Create;
  FAnimateDrawRects := TTMSFMXChartDrawRects.Create;
  FAnimateDrawRectsReverse := TTMSFMXChartDrawRects.Create;
  FDrawPath := TPathData.Create;
  FAnimationFactor := 4.0;
  FOffset3DX := 15;
  FOffset3DY := 15;
  FEnable3D := False;
  FAnimationFlow := True;
  FVisible := True;
  FShowInLegend := True;
  FChartType := ctLine;
  FAutoYRange := arEnabled;
  FAutoXRange := arDisabled;
  FZeroReferenceValue := 0;
  FMinX := 0;
  FMaxX := 10;
  FMinY := 0;
  FMaxY := 10;
  FTag := 0;
  FFillOpacity := 1;
  FFill := TBrush.Create(TBrushKind.Solid, claYellowgreen);
  FFill.OnChanged := FillChanged;
  FStrokeOpacity := 1;
  FStroke := TBrush.Create(TbrushKind.Solid, claDarkgreen);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;

  FAnimateTimer := TTimer.Create(nil);
  FAnimateTimer.Enabled := False;
  FAnimateTimer.Interval := 10;
  FAnimateTimer.OnTimer := AnimateProcess;

  FLegendText := 'Serie ' + inttostr(Index);
  UpdateChart;
end;

destructor TTMSFMXChartSerie.Destroy;
begin
  FAnimateTimer.Free;
  FFill.Free;
  FStroke.Free;
  FPoints.Free;
  FDrawXValues.Free;
  FDrawYValues.Free;
  FDrawXGridLines.Free;
  FDrawYGridLines.Free;
  FAnimateDrawRectsReverse.Free;
  FAnimateDrawRects.Free;
  FDrawRects.Free;
  FDrawLabels.Free;
  FDrawAnnotations.Free;
  FAnimateDrawLines.Free;
  FAnimateDrawLinesReverse.Free;
  FDrawLines.Free;
  FAnimateDrawSlices.Free;
  FAnimateDrawSlicesReverse.Free;
  FDrawslices.Free;
  FAnimateDrawPoints.Free;
  FAnimateDrawPointsReverse.Free;
  FDrawPoints.Free;
  FAnimateDrawMultiPoints.Free;
  FAnimateDrawMultiPointsReverse.Free;
  FDrawMultiPoints.Free;
  FDrawPath.Free;
  FLabels.Free;
  FXValues.Free;
  FYValues.Free;
  FXGrid.Free;
  FYGrid.Free;
  FMarkers.Free;
  FLegend.Free;
  FBar.Free;
  FMultiPoint.Free;
  FPie.Free;
  inherited;
  UpdateChart;
end;

procedure TTMSFMXChartSerie.Draw(ACanvas: TCanvas; ADrawPoint: Integer = -1);
var
  I: Integer;
  c: TTMSFMXChart;
  dl: TTMSFMXChartDrawLine;
  dm: TTMSFMXChartDrawMultiPoint;
  dr: TTMSFMXChartDrawRect;
  ds, dsStart: TTMSFMXChartDrawSlice;
  df: Boolean;
  st: TCanvasSaveState;
  pth, pp: TPathData;
  br: TRectF;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  ACanvas.Fill.Assign(Fill);
  ACanvas.Stroke.Assign(Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := StrokeThickness;
  ACanvas.Stroke.Cap := StrokeCap;
  ACanvas.Stroke.Dash := StrokeDash;
  ACanvas.Stroke.Join := StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := StrokeThickness;
  ACanvas.StrokeCap := StrokeCap;
  ACanvas.StrokeDash := StrokeDash;
  ACanvas.StrokeJoin := StrokeJoin;
  {$ENDIF}

  pth := nil;
  if Is3DEnabled then
    pth := TPathData.Create;

  case ChartType of
    ctSpider:
    begin
      if FDrawSlices.Count > 0 then
      begin
        pp := TPathData.Create;
        dsStart := FDrawSlices[0];
        pp.MoveTo(dsStart.EndPoint);
        for I := 1 to FDrawSlices.Count - 1 do
        begin
          ds := FDrawSlices[I];
          pp.LineTo(ds.EndPoint);
        end;
        pp.ClosePath;
        st := ACanvas.SaveState;
        ACanvas.FillPath(pp, FillOpacity);
        ACanvas.DrawPath(pp, StrokeOpacity);
        ACanvas.RestoreState(st);
        pp.Free;
      end;
    end;
    ctPie, ctVariableRadiusPie, ctSizedPie:
    begin
      pp := TPathData.Create;
      for I := 0 to FDrawSlices.Count - 1 do
      begin
        ds := FDrawSlices[I];
        df := True;
        st := ACanvas.SaveState;

        if Assigned(ds.Reference) and (ds.Reference.Color <> claNull) then
          ACanvas.Fill.Color := ds.Reference.Color
        else if ds.VirtualReference.Color <> claNull then
          ACanvas.Fill.Color := ds.VirtualReference.Color;

        c.DoBeforeDrawSerieSlice(ACanvas, Self, ds, df);
        if df then
        begin
          pp.Clear;
          pp.MoveTo(ds.StartPoint);
          pp.AddArc(ds.CenterPoint, ds.Radius, ds.StartAngle, ds.SweepAngle);
          pp.AddArc(ds.CenterPoint, ds.InnerRadius, ds.StartAngle + ds.SweepAngle, -ds.SweepAngle);
          pp.ClosePath;
          ACanvas.FillPath(pp, FillOpacity);
          ACanvas.DrawPath(pp, StrokeOpacity);
          c.DoAfterDrawSerieSlice(ACanvas, Self, ds);
        end;
        ACanvas.RestoreState(st);
      end;
      pp.Free;
    end;
    ctOHLC, ctCandleStick, ctBoxPlot:
    begin
      for I := 0 to FDrawMultiPoints.Count - 1 do
      begin
        dm := FDrawMultiPoints[I];
        df := True;
        st := ACanvas.SaveState;
        if Assigned(dm.Reference) then
        begin
          if (dm.Reference.Color <> claNull) then
          begin
            ACanvas.Fill.Color := dm.Reference.Color;
            ACanvas.Stroke.Color := dm.Reference.Color;
          end
          else if dm.VirtualReference.Color <> claNull then
          begin
            ACanvas.Fill.Color := dm.VirtualReference.Color;
            ACanvas.Stroke.Color := dm.VirtualReference.Color;
          end
          else
          begin
            if dm.Reference.YValueClose < dm.Reference.YValueOpen then
            begin
              if MultiPoint.DecreaseStrokeColor <> claNull then
                ACanvas.Stroke.Color := MultiPoint.DecreaseStrokeColor;

              if MultiPoint.DecreaseFillColor <> claNull then
                ACanvas.Fill.Color := MultiPoint.DecreaseFillColor;
            end
            else
            begin
              if MultiPoint.IncreaseStrokeColor <> claNull then
                ACanvas.Stroke.Color := MultiPoint.IncreaseStrokeColor;

              if MultiPoint.IncreaseFillColor <> claNull then
                ACanvas.Fill.Color := MultiPoint.IncreaseFillColor;
            end;
          end;
        end;

        c.DoBeforeDrawSerieMultiPoint(ACanvas, Self, dm, df);
        if df then
        begin
          ACanvas.DrawLine(dm.PointHigh, dm.PointLow, 1);
          case ChartType of
            ctOHLC:
            begin
              ACanvas.DrawLine(PointF(dm.Rect.Left, dm.PointOpen.Y), dm.PointOpen, 1);
              ACanvas.DrawLine(dm.PointClose, PointF(dm.Rect.Right, dm.PointClose.Y), 1);
            end;
            ctCandleStick, ctBoxPlot:
            begin
              ACanvas.FillRect(dm.Rect, 0, 0, AllCorners, 1);
              case ChartType of
                ctBoxPlot:
                begin
                  ACanvas.DrawLine(PointF(dm.Rect.Left, dm.PointHigh.Y), PointF(dm.Rect.Right, dm.PointHigh.Y), 1);
                  ACanvas.DrawLine(PointF(dm.Rect.Left, dm.PointLow.Y), PointF(dm.Rect.Right, dm.PointLow.Y), 1);
                  ACanvas.DrawLine(PointF(dm.Rect.Left, dm.PointMedian.Y), PointF(dm.Rect.Right, dm.PointMedian.Y), 1);
                end;
              end;
            end;
          end;
          c.DoAfterDrawSerieMultiPoint(ACanvas, Self, dm);
        end;
        ACanvas.RestoreState(st);
      end;
    end;
    ctLine, ctDigitalLine, ctXYLine:
    begin
      for I := 0 to FDrawLines.Count - 1 do
      begin
        dl := FDrawLines[I];
        df := True;
        st := ACanvas.SaveState;
        c.DoBeforeDrawSerieLine(ACanvas, Self, dl, df);
        if df then
        begin
          if Is3DEnabled then
          begin
            if Assigned(pth) then
            begin
              pth.Clear;
              pth.MoveTo(dl.StartPoint);
              pth.LineTo(dl.EndPoint);
              pth.LineTo(PointF(dl.EndPoint.X + Offset3DX, dl.EndPoint.Y - Offset3DY));
              pth.LineTo(PointF(dl.StartPoint.X + Offset3DX, dl.StartPoint.Y - Offset3DY));
              pth.ClosePath;
              ACanvas.FillPath(pth, FillOpacity);
              ACanvas.DrawPath(pth, StrokeOpacity);
            end;
          end
          else
            ACanvas.DrawLine(dl.StartPoint, dl.EndPoint, StrokeOpacity);

          c.DoAfterDrawSerieLine(ACanvas, Self, dl);
        end;
        ACanvas.RestoreState(st);
      end;
    end;
    ctArea, ctStackedArea, ctStackedPercentageArea, ctBand:
    begin
      if Is3DEnabled and Assigned(pth) then
      begin
        if ChartType = ctBand then
        begin
          for I := FDrawLines.Count - 1 downto FDrawLines.Count div 2 do
          begin
            dl := FDrawLines[I];
            pth.Clear;
            pth.MoveTo(dl.StartPoint);
            pth.LineTo(dl.EndPoint);
            pth.LineTo(PointF(dl.EndPoint.X + Offset3DX, dl.EndPoint.Y - Offset3DY));
            pth.LineTo(PointF(dl.StartPoint.X + Offset3DX, dl.StartPoint.Y - Offset3DY));
            pth.ClosePath;
            ACanvas.FillPath(pth, FillOpacity);
            ACanvas.DrawPath(pth, StrokeOpacity);
          end;

          for I := 0 to FDrawLines.Count div 2 do
          begin
            dl := FDrawLines[I];
            pth.Clear;
            pth.MoveTo(dl.StartPoint);
            pth.LineTo(dl.EndPoint);
            pth.LineTo(PointF(dl.EndPoint.X + Offset3DX, dl.EndPoint.Y - Offset3DY));
            pth.LineTo(PointF(dl.StartPoint.X + Offset3DX, dl.StartPoint.Y - Offset3DY));
            pth.ClosePath;
            ACanvas.FillPath(pth, FillOpacity);
            ACanvas.DrawPath(pth, StrokeOpacity);
          end;
        end
        else
        begin
          for I := 0 to FDrawLines.Count - 1 do
          begin
            dl := FDrawLines[I];
            if dl.Bottom then
              Continue;
            pth.Clear;
            pth.MoveTo(dl.StartPoint);
            pth.LineTo(dl.EndPoint);
            pth.LineTo(PointF(dl.EndPoint.X + Offset3DX, dl.EndPoint.Y - Offset3DY));
            pth.LineTo(PointF(dl.StartPoint.X + Offset3DX, dl.StartPoint.Y - Offset3DY));
            pth.ClosePath;
            ACanvas.FillPath(pth, FillOpacity);
            ACanvas.DrawPath(pth, StrokeOpacity);
          end;
        end;
      end;
      ACanvas.FillPath(FDrawPath, FillOpacity);
      ACanvas.DrawPath(FDrawPath, StrokeOpacity);
    end;
    ctBar, ctStackedBar, ctStackedPercentageBar:
    begin
      dr := FDrawRects[ADrawPoint];
      br := dr.Rect;
      br := RectF(int(br.Left) + 0.5, int(br.Top) + 0.5, int(br.Right) - 0.5, int(br.Bottom) - 0.5);
      st := ACanvas.SaveState;
      if Assigned(dr.Reference) and (dr.Reference.Color <> claNull) then
        ACanvas.Fill.Color := dr.Reference.Color
      else if dr.VirtualReference.Color <> claNull then
        ACanvas.Fill.Color := dr.VirtualReference.Color;

      df := True;
      c.DoBeforeDrawSerieBar(ACanvas, Self, dr, df);
      if df then
      begin
        if Is3DEnabled and Assigned(pth) then
        begin
          pth.Clear;
          pth.MoveTo(br.TopLeft);
          pth.LineTo(PointF(br.Right, br.Top));
          pth.LineTo(PointF(br.Right + Offset3DX, br.Top - Offset3DY));
          pth.LineTo(PointF(br.Left + Offset3DX, br.Top - Offset3DY));
          pth.ClosePath;
          ACanvas.FillPath(pth, FillOpacity);
          ACanvas.DrawPath(pth, StrokeOpacity);
          pth.Clear;
          pth.MoveTo(PointF(br.Right, br.Top));
          pth.LineTo(PointF(br.Right, br.Bottom));
          pth.LineTo(PointF(br.Right + Offset3DX, br.Bottom - Offset3DY));
          pth.LineTo(PointF(br.Right + Offset3DX, br.Top - Offset3DY));
          pth.ClosePath;
          ACanvas.FillPath(pth, FillOpacity);
          ACanvas.DrawPath(pth, StrokeOpacity);
        end;

        ACanvas.FillRect(br, 0, 0, AllCorners, FillOpacity);
        ACanvas.DrawRect(br, 0, 0, AllCorners, StrokeOpacity);
        c.DoAfterDrawSerieBar(ACanvas, Self, dr);
      end;
      ACanvas.RestoreState(st);
    end;
  end;

  if Assigned(pth) then
    pth.Free;
end;

procedure TTMSFMXChartSerie.DrawAnnotations(ACanvas: TCanvas);
var
  c: TTMSFMXChart;
  I: Integer;
  dann: TTMSFMXChartDrawAnnotation;
  dr: Boolean;
  st, sts, strs, ars: TCanvasSaveState;
  rt: TRectF;
  an: TTMSFMXChartAnnotationVirtual;
  bsx, bsy: Double;
  pth: TPathData;
  epd: TPointF;
  fl: TBrush;
  stb: TStrokeBrush;
  ft: TFont;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if not (FDrawAnnotations.Count > 0) or FAnimateTimer.Enabled or Revert then
    Exit;

  fl := TBrush.Create(TBrushKind.Solid, claGhostwhite);
  stb := TStrokeBrush.Create(TBrushKind.Solid, claGray);

  ft := TFont.Create;

  for I := 0 to FDrawAnnotations.Count - 1 do
  begin
    dann := FDrawAnnotations[I];
    an := dann.VirtualAnnotationReference;
    bsx := an.BalloonArrowSize;
    bsy := an.BalloonArrowSize;

    if Assigned(an.Annotation) then
    begin
      fl.Assign(an.Annotation.Fill);
      ft.Assign(an.Annotation.Font);
      stb.Assign(an.Annotation.Stroke);
      stb.Thickness := an.Annotation.StrokeThickness;
      stb.Cap := an.Annotation.StrokeCap;
      stb.Dash := an.Annotation.StrokeDash;
      stb.Join := an.Annotation.StrokeJoin;
    end;

    c.DoCustomizeAnnotationFill(Self, dann.VirtualPointReference, an, fl);
    c.DoCustomizeAnnotationStroke(Self, dann.VirtualPointReference, an, stb);
    c.DoCustomizeAnnotationFont(Self, dann.VirtualPointReference, an, ft);

    ACanvas.Fill.Assign(fl);
    ACanvas.Font.Assign(ft);
    ACanvas.Stroke.Assign(stb);

    st := ACanvas.SaveState;
    dr := True;
    c.DoBeforeDrawSerieAnnotation(ACanvas, Self, dann, dr);
    if dr then
    begin
      rt := RectF(dann.AnnotationPoint.X - dann.TextWidth / 2, dann.AnnotationPoint.Y - dann.TextHeight,
        dann.AnnotationPoint.X + dann.TextWidth / 2, dann.AnnotationPoint.Y);

      rt := ModifyRect(rt);

      if an.Shape = asBalloon then
      begin
        case an.BalloonDirection of
          bdUpRight: epd := PointF(rt.Right - an.CornerRadius, rt.Top - bsy);
          bdUpLeft: epd := PointF(rt.Left + an.CornerRadius, rt.Top - bsy);
          bdDownRight: epd := PointF(rt.Right - an.CornerRadius, rt.Bottom + bsy);
          bdDownLeft: epd := PointF(rt.Left + an.CornerRadius, rt.Bottom + bsy);
        end;
      end
      else
        epd := PointF(rt.Left + rt.Width / 2, rt.Bottom);

      if an.LineColor <> TAlphaColorRec.Null then
      begin
        strs := ACanvas.SaveState;
        {$IFDEF DELPHIXE11_LVL}
        ACanvas.Stroke.Thickness := an.LineThickness;
        {$ELSE}
        ACanvas.StrokeThickness := an.LineThickness;
        {$ENDIF}
        ACanvas.Stroke.Color := an.LineColor;
        ACanvas.DrawLine(dann.StartPoint, epd, an.LineOpacity);
        ACanvas.RestoreState(strs);
      end;

      if (an.ArrowColor <> TAlphaColorRec.Null) and (an.Arrow <> arLine) then
      begin
        ars := ACanvas.SaveState;
        ACanvas.Fill.Kind := TBrushKind.Solid;
        case an.Arrow of
          arEndArrow: DrawArrow(ACanvas, an.ArrowColor, an.ArrowSize, an.ArrowOpacity, dann.StartPoint, epd);
          arStartArrow: DrawArrow(ACanvas, an.ArrowColor, an.ArrowSize, an.ArrowOpacity, epd, dann.StartPoint);
          arDoubleArrow:
          begin
            DrawArrow(ACanvas, an.ArrowColor, an.ArrowSize, an.ArrowOpacity, dann.StartPoint, epd);
            DrawArrow(ACanvas, an.ArrowColor, an.ArrowSize, an.ArrowOpacity, epd, dann.StartPoint);
          end;
        end;
        ACanvas.RestoreState(ars);
      end;

      case an.Shape of
        asRectangle:
        begin
          ACanvas.FillRect(rt, an.CornerRadius, an.CornerRadius, AllCorners, an.FillOpacity);
          Acanvas.DrawRect(rt, an.CornerRadius, an.CornerRadius, AllCorners, an.StrokeOpacity);
        end;
        asCircle:
        begin
          ACanvas.FillEllipse(rt, an.FillOpacity);
          ACanvas.DrawEllipse(rt, an.StrokeOpacity);
        end;
        asBalloon:
        begin
          ACanvas.FillRect(rt, an.CornerRadius, an.CornerRadius, AllCorners, an.FillOpacity);
          ACanvas.DrawRect(rt, an.CornerRadius, an.CornerRadius, AllCorners, an.StrokeOpacity);
          pth := TPathData.Create;
          case an.BalloonDirection of
            bdUpRight:
            begin
              pth.MoveTo(PointF(rt.Right - an.CornerRadius, rt.Top + stb.Thickness));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Top));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Top - bsy));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Top));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Top + stb.Thickness));
              pth.ClosePath;
              ACanvas.FillPath(pth, an.FillOpacity);
              pth.Clear;
              pth.MoveTo(PointF(rt.Right - an.CornerRadius, rt.Top));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Top - bsy));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Top));
              ACanvas.DrawPath(pth, an.StrokeOpacity);
            end;
            bdUpLeft:
            begin
              pth.MoveTo(PointF(rt.Left + an.CornerRadius, rt.Top + stb.Thickness));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Top));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Top - bsy));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Top));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Top + stb.Thickness));
              pth.ClosePath;
              ACanvas.FillPath(pth, an.FillOpacity);
              pth.Clear;
              pth.MoveTo(PointF(rt.Left + an.CornerRadius, rt.Top));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Top - bsy));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Top));
              ACanvas.DrawPath(pth, an.StrokeOpacity);
            end;
            bdDownRight:
            begin
              pth.MoveTo(PointF(rt.Right - an.CornerRadius, rt.Bottom - stb.Thickness));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Bottom));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Bottom + bsy));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Bottom));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Bottom - stb.Thickness));
              pth.ClosePath;
              ACanvas.FillPath(pth, an.FillOpacity);
              pth.Clear;
              pth.MoveTo(PointF(rt.Right - an.CornerRadius, rt.Bottom));
              pth.LineTo(PointF(rt.Right - an.CornerRadius, rt.Bottom + bsy));
              pth.LineTo(PointF(rt.Right - an.CornerRadius - bsx, rt.Bottom));
              ACanvas.DrawPath(pth, an.StrokeOpacity);
            end;
            bdDownLeft:
            begin
              pth.MoveTo(PointF(rt.Left + an.CornerRadius, rt.Bottom - stb.Thickness));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Bottom));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Bottom + bsy));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Bottom));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Bottom - stb.Thickness));
              pth.ClosePath;
              ACanvas.FillPath(pth, an.FillOpacity);
              pth.Clear;
              pth.MoveTo(PointF(rt.Left + an.CornerRadius, rt.Bottom));
              pth.LineTo(PointF(rt.Left + an.CornerRadius, rt.Bottom + bsy));
              pth.LineTo(PointF(rt.Left + an.CornerRadius + bsx, rt.Bottom));
              ACanvas.DrawPath(pth, an.StrokeOpacity);
            end;
          end;
          pth.Free;
        end;
      end;
      sts := ACanvas.SaveState;
      ACanvas.Fill.Color := an.FontColor;
      ACanvas.FillText(rt, an.Text, an.WordWrap, 1, [], an.TextHorizontalAlignment, an.TextVerticalAlignment);
      ACanvas.RestoreState(sts);
      c.DoAfterDrawSerieAnnotation(ACanvas, Self, dann);
      ACanvas.RestoreState(st);
    end;
  end;

  fl.Free;
  stb.Free;
  ft.Free;
end;

procedure TTMSFMXChartSerie.DrawArrow(ACanvas: TCanvas; ArrowColor: TAlphaColor;
  ArrowSize, ArrowOpacity: Double; origin, target: TPointF);
var
  quarter: Integer;
  fx, px: Double;
  fy, py: Double;
  x, y: Double;
  arrowpts: TPolygon;
  p: TPointF;
  ar: TPointF;
  h: Double;
  arx, ary: Double;
begin
  SetLength(arrowpts, 4);

  arx := ArrowSize;
  ary := ArrowSize;

  arrowpts[0] := target;

  x := target.x - origin.x;
  y := target.y - origin.y;
  h := sqrt(sqr(x) + sqr(y));

  if h = 0 then
    h := 1;

  if origin.x < target.x then
  begin
    if origin.y < target.y then
      quarter := 1
    else
      quarter := 3;
  end
  else
  begin
    if origin.y < target.y then
      quarter := 2
    else
      quarter := 4;
  end;

  px := x * arx / h;
  py := y * ary / h;
  case quarter of
    1 :
      begin
        p.x := target.x - px;
        p.y := target.y - py;
        ar.x := target.x - (x * arx / h);
        ar.y := target.y - (y * ary / h);
      end;
    2 :
      begin
        p.x := target.x - px;
        p.y := target.y - py;
        ar.x := target.x - (x * arx / h);
        ar.y := target.y - (y * ary / h);
      end;
    3 :
      begin
        p.x := target.x - px;
        p.y := target.y - py;
        ar.x := target.x - (x * arx / h);
        ar.y := target.y - (y * ary / h);
      end;
    4 :
      begin
        p.x := Target.x - px;
        p.y := Target.y - py;
        ar.x := target.x - (x * arx / h);
        ar.y := target.y - (y * ary / h);
      end;
  end;

  fx := y * (arx / 2) / h;
  fy := x * (ary / 2) / h;
  case quarter of
    1 :
      begin
        arrowpts[1].x := p.x - fx;
        arrowpts[1].y := p.y + fy;
        arrowpts[3].x := p.x + fx;
        arrowpts[3].y := p.y - fy;
      end;
    2 :
      begin
        arrowpts[1].x := p.x + fx;
        arrowpts[1].y := p.y - fy;
        arrowpts[3].x := p.x - fx;
        arrowpts[3].y := p.y + fy;
      end;
    3 :
      begin
        arrowpts[1].x := p.x + fx;
        arrowpts[1].y := p.y - fy;
        arrowpts[3].x := p.x - fx;
        arrowpts[3].y := p.y + fy;
      end;
    4 :
      begin
        arrowpts[1].x := p.x + fx;
        arrowpts[1].y := p.y - fy;
        arrowpts[3].x := p.x - fx;
        arrowpts[3].y := p.y + fy;
      end;
  end;

  arrowpts[2] := ar;
  ACanvas.Fill.Color := ArrowColor;
  ACanvas.FillPolygon(arrowpts, ArrowOpacity);
end;

procedure TTMSFMXChartSerie.DrawLabels(ACanvas: TCanvas);
var
  c: TTMSFMXChart;
  I: Integer;
  dlbl: TTMSFMXChartDrawLabel;
  dr: Boolean;
  st, sts: TCanvasSaveState;
  rt: TRectF;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if not Labels.Visible or FAnimateTimer.Enabled or Revert then
    Exit;

  ACanvas.Fill.Assign(Labels.Fill);
  ACanvas.Font.Assign(Labels.Font);
  ACanvas.Stroke.Assign(Labels.Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := Labels.StrokeThickness;
  ACanvas.Stroke.Cap := Labels.StrokeCap;
  ACanvas.Stroke.Dash := Labels.StrokeDash;
  ACanvas.Stroke.Join := Labels.StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := Labels.StrokeThickness;
  ACanvas.StrokeCap := Labels.StrokeCap;
  ACanvas.StrokeDash := Labels.StrokeDash;
  ACanvas.StrokeJoin := Labels.StrokeJoin;
  {$ENDIF}

  for I := 0 to FDrawLabels.Count - 1 do
  begin
    dlbl := FDrawLabels[I];
    st := ACanvas.SaveState;
    dr := True;
    c.DoBeforeDrawSerieLabel(ACanvas, Self, dlbl, dr);
    if dr then
    begin
      rt := RectF(dlbl.Point.X - dlbl.TextWidth / 2, dlbl.Point.Y - dlbl.TextHeight, dlbl.Point.X + dlbl.TextWidth / 2, dlbl.Point.Y);
      rt := ModifyRect(rt);
      ACanvas.FillRect(rt, 0, 0, AllCorners, Labels.FillOpacity);
      Acanvas.DrawRect(rt, 0, 0, AllCorners, Labels.StrokeOpacity);
      sts := ACanvas.SaveState;
      ACanvas.Fill.Color := Labels.FontColor;
      ACanvas.FillText(rt, dlbl.ValueString, False, 1, [], TTextAlign.Center);
      ACanvas.RestoreState(sts);
      c.DoAfterDrawSerieLabel(ACanvas, Self, dlbl);
    end;
    ACanvas.RestoreState(st);
  end;
end;

procedure TTMSFMXChartSerie.DrawLegend(ACanvas: TCanvas);
var
  c: TTMSFMXChart;
  r, rm, rt: TRectF;
  I, cnt: Integer;
  pt: TTMSFMXChartPointVirtual;
  th, x, y: Double;
  dr: Boolean;
  st: TCanvasSaveState;
  str: String;
  li: Boolean;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  cnt := GetPointsCount;
  if cnt = 0 then
    Exit;

  if not Legend.Visible then
    Exit;

  r := Legend.GetRectangle;
  r := ModifyRect(r);
  if r = TRectF.Empty then
    Exit;


  ACanvas.Fill.Assign(Legend.Fill);
  ACanvas.Stroke.Assign(Legend.Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := Legend.StrokeThickness;
  ACanvas.Stroke.Cap := Legend.StrokeCap;
  ACanvas.Stroke.Dash := Legend.StrokeDash;
  ACanvas.Stroke.Join := Legend.StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := Legend.StrokeThickness;
  ACanvas.StrokeCap := Legend.StrokeCap;
  ACanvas.StrokeDash := Legend.StrokeDash;
  ACanvas.StrokeJoin := Legend.StrokeJoin;
  {$ENDIF}

  dr := True;
  st := ACanvas.SaveState;
  c.DoBeforeDrawSerieLegend(ACanvas, Self, r, dr);
  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, Legend.FillOpacity);
    ACanvas.DrawRect(r, 0, 0, AllCorners, Legend.StrokeOpacity);
    ACanvas.Font.Assign(Legend.Font);
    ACanvas.Fill.Color := Legend.FontColor;

    x := r.Left + 5;
    y := r.Top + 5;

    if LegendText <> '' then
    begin
      th := ACanvas.TextHeight(LegendText);
      rt := RectF(r.Left, r.Top, r.Right, r.Top + th + 5);
      InflateRect(rt, -2, -2);
      ACanvas.FillText(rt, LegendText, False, 1, [], TTextAlign.Center);
      y := y + th;
      ACanvas.DrawLine(PointF(r.Left, y), PointF(r.Right, y), Legend.StrokeOpacity);
      y := y + 5;
    end;

    for I := 0 to cnt - 1 do
    begin
      pt := GetPoint(I);
      str := GetLegendText(pt);
      if str <> '' then
      begin
        th := ACanvas.TextHeight(str);
        rt := RectF(x + th + 5, y, x + r.Width - 10, y + th);
        rm := RectF(x, y, x + th, y + th);
        y := y + th + 5;

        ACanvas.Fill.Color := Legend.FontColor;

        ACanvas.FillText(rt, str, False, 1, [], TTextAlign.Leading, TTextAlign.Leading);

        ACanvas.Fill.Assign(Fill);
        ACanvas.Stroke.Assign(Stroke);
        {$IFDEF DELPHIXE11_LVL}
        ACanvas.Stroke.Thickness := StrokeThickness;
        ACanvas.Stroke.Cap := StrokeCap;
        ACanvas.Stroke.Dash := StrokeDash;
        ACanvas.Stroke.Join := StrokeJoin;
        {$ELSE}
        ACanvas.StrokeThickness := StrokeThickness;
        ACanvas.StrokeCap := StrokeCap;
        ACanvas.StrokeDash := StrokeDash;
        ACanvas.StrokeJoin := StrokeJoin;
        {$ENDIF}
        if pt.Color <> claNull then
          ACanvas.Fill.Color := pt.Color;

        InflateRect(rm, -rm.Height / 7.5, -rm.Height / 7.5);
        rm := ModifyRect(rm);
        li := True;
        if c.FVirtualMode then
          c.DoBeforeDrawSerieLegendIconVirtual(ACanvas, Self, pt, rm, li)
        else
          c.DoBeforeDrawSerieLegendIcon(ACanvas, Self, pt.Point, rm, li);

        if li then
        begin
          DrawChartIcon(ACanvas, ChartType, rm);
          if c.FVirtualMode then
            c.DoAfterDrawSerieLegendIconVirtual(ACanvas, Self, pt, rm)
          else
            c.DoAfterDrawSerieLegendIcon(ACanvas, Self, pt.Point, rm)
        end;
      end;
    end;
    c.DoAfterDrawSerieLegend(ACanvas, Self, r);
  end;
  ACanvas.RestoreState(st);
end;

function TTMSFMXChartSerie.DrawLines: TTMSFMXChartDrawLines;
begin
  Result := FDrawLines;
end;

procedure TTMSFMXChartSerie.DrawMarkers(ACanvas: TCanvas);
var
  msw, msh: Double;
  pth: TPathData;
  I: Integer;
  pt: TPointF;
  c: TTMSFMXChart;
  dp: TTMSFMXChartDrawPoint;
  dr: Boolean;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) or FAnimateTimer.Enabled or Revert then
    Exit;

  if not Markers.Visible and (ChartType <> ctXYMarker) and (ChartType <> ctMarker) then
    Exit;

  ACanvas.Fill.Assign(Markers.Fill);
  ACanvas.Stroke.Assign(Markers.Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := Markers.StrokeThickness;
  ACanvas.Stroke.Cap := Markers.StrokeCap;
  ACanvas.Stroke.Dash := Markers.StrokeDash;
  ACanvas.Stroke.Join := Markers.StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := Markers.StrokeThickness;
  ACanvas.StrokeCap := Markers.StrokeCap;
  ACanvas.StrokeDash := Markers.StrokeDash;
  ACanvas.StrokeJoin := Markers.StrokeJoin;
  {$ENDIF}

  msw := Markers.Width / 2;
  msh := Markers.Height / 2;
  pth := TPathData.Create;
  for I := 0 to FDrawPoints.Count - 1 do
  begin
    dp := FDrawPoints[I];
    dr := True;
    st := ACanvas.SaveState;
    c.DoBeforeDrawSerieMarker(ACanvas, Self, dp, dr);
    if dr then
    begin
      pt := dp.Point;
      case Markers.Shape of
        msEllipse:
        begin
          ACanvas.FillEllipse(RectF(pt.X - msw, pt.Y - msh, pt.X + msw, pt.Y + msh), Markers.FillOpacity);
          ACanvas.DrawEllipse(RectF(pt.X - msw, pt.Y - msh, pt.X + msw, pt.Y + msh), Markers.StrokeOpacity);
        end;
        msSquare:
        begin
          ACanvas.FillRect(RectF(pt.X - msw, pt.Y - msh, pt.X + msw, pt.Y + msh), 0, 0, AllCorners, Markers.FillOpacity);
          ACanvas.DrawRect(RectF(pt.X - msw, pt.Y - msh, pt.X + msw, pt.Y + msh), 0, 0, AllCorners, Markers.StrokeOpacity);
        end;
        msDiamond:
        begin
          pth.Clear;
          pth.MoveTo(PointF(pt.X, pt.Y - msh));
          pth.LineTo(PointF(pt.X + msw, pt.Y));
          pth.LineTo(PointF(pt.X, pt.Y + msh));
          pth.LineTo(PointF(pt.X - msw, pt.Y));
          pth.ClosePath;
          ACanvas.FillPath(pth, Markers.FillOpacity);
          ACanvas.DrawPath(pth, Markers.StrokeOpacity);
        end;
        msBitmap:
        begin
          if not Markers.Bitmap.IsEmpty then
            ACanvas.DrawBitmap(Markers.Bitmap, RectF(0, 0, Markers.Bitmap.Width, Markers.Bitmap.Height),
              RectF(pt.X - msw, pt.Y - msh, pt.X + msw, pt.Y + msh), FillOpacity);
        end;
        msTriangle:
        begin
          pth.Clear;
          pth.MoveTo(PointF(pt.X, pt.Y - msh));
          pth.LineTo(PointF(pt.X + msw, pt.Y));
          pth.LineTo(PointF(pt.X - msw, pt.Y));
          pth.ClosePath;
          ACanvas.FillPath(pth, Markers.FillOpacity);
          ACanvas.DrawPath(pth, Markers.StrokeOpacity);
        end;
      end;
      c.DoAfterDrawSerieMarker(ACanvas, Self, dp);
    end;
    ACanvas.RestoreState(st);
  end;
  pth.Free;
end;

function TTMSFMXChartSerie.DrawMultiPoints: TTMSFMXChartDrawMultiPoints;
begin
  Result := FDrawMultiPoints;
end;

function TTMSFMXChartSerie.DrawPoints: TTMSFMXChartDrawPoints;
begin
  Result := FDrawPoints;
end;

function TTMSFMXChartSerie.DrawRects: TTMSFMXChartDrawRects;
begin
  Result := FDrawRects;
end;

function TTMSFMXChartSerie.DrawSlices: TTMSFMXChartDrawSlices;
begin
  Result := FDrawSlices;
end;

procedure TTMSFMXChartSerie.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChartSerie.FindXAxisTextForValue(AValue: Double): String;
var
  l, e, m, cnt: Integer;
  fnd: Boolean;
  dif, dif2: Double;
  pt, ptn: TTMSFMXChartPointVirtual;
begin
  Result := '';
  cnt := GetPointsCount;
  if cnt = 1 then
  begin
    pt := GetPoint(0);
    if pt.XValue = AValue then
      Result := pt.XValueText;

    Exit;
  end;

  l  := 0;
  e := cnt - 1;

  fnd := False;
  while (l <= e) and not fnd do
  begin
    m := (l + e) div 2;
    pt := GetPoint(m);
    ptn := GetPoint(m + 1);

    if (AValue >= pt.XValue) and (m >= cnt) then
    begin
      fnd := True;
      Result := pt.XValueText;
    end
    else
    if (AValue >= pt.XValue) and ((m + 1 < cnt) and (AValue <= ptn.XValue)) then
    begin
      fnd := True;
      dif := Abs(AValue - pt.XValue);
      dif2 := Abs(AValue - ptn.XValue);
      if dif > dif2 then
        Result := ptn.XValueText
      else
        Result := pt.XValueText;
    end
    else if pt.XValue > AValue then
      e := m - 1
    else
      l := m + 1;
  end;
end;

function TTMSFMXChartSerie.Get3DOffsetX: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  gp, l: Integer;
  lgp: TArray<Integer>;
  J: Integer;
  maxgp: Double;
  c: TTMSFMXChart;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;     

  Result := 0;
  for I := c.Series.Count - 1 downto 0 do
  begin
    s := c.Series[I];
    if s = Self then
      Break;
    if s.Visible and s.Is3DEnabled and not s.IsBar and not s.IsStacked then
      Result := Result + s.Offset3DX;
  end;

  gp := -1;
  for I := 0 to c.Series.Count - 1 do
  begin
    s := c.Series[I];
    if s.Visible and s.Is3DEnabled and (s.GroupIndex > gp) and (s.IsBar or s.IsArea) then
    begin
      gp := s.GroupIndex;
      l := Length(lgp);
      SetLength(lgp, l + 1);
      lgp[l] := gp;
    end;
  end;

  maxgp := 0;
  for I := 0 to Length(lgp) - 1 do
  begin
    for J := c.Series.Count - 1 downto 0 do
    begin
      s := c.Series[J];
      if Self.GroupIndex = s.GroupIndex then
        Break;

      if s.Visible and s.Is3DEnabled and (s.GroupIndex = lgp[I]) and ((s.IsBar and not s.IsStacked) or (s.IsArea and s.IsStacked)) then
        maxgp := Max(maxgp, s.Offset3DX);
    end;
    Result := Result + maxgp;
  end;
end;

function TTMSFMXChartSerie.Get3DOffsetY: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  gp, l: Integer;
  lgp: TArray<Integer>;
  J: Integer;
  maxgp: Double;
  c: TTMSFMXChart;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;

  Result := 0;
  for I := c.Series.Count - 1 downto 0 do
  begin
    s := c.Series[I];
    if s = Self then
      Break;
    if s.Visible and s.Is3DEnabled and not s.IsBar and not s.IsStacked then
      Result := Result + s.Offset3DY;
  end;

  gp := -1;
  for I := 0 to c.Series.Count - 1 do
  begin
    s := c.Series[I];
    if s.Visible and s.Is3DEnabled and (s.GroupIndex > gp) and (s.IsBar or s.IsArea) then
    begin
      gp := s.GroupIndex;
      l := Length(lgp);
      SetLength(lgp, l + 1);
      lgp[l] := gp;
    end;
  end;

  maxgp := 0;
  for I := 0 to Length(lgp) - 1 do
  begin
    for J := c.Series.Count - 1 downto 0 do
    begin
      s := c.Series[J];
      if Self.GroupIndex = s.GroupIndex then
        Break;

      if s.Visible and s.Is3DEnabled and (s.GroupIndex = lgp[I]) and ((s.IsBar and not s.IsStacked) or (s.IsArea and s.IsStacked)) then
        maxgp := Max(maxgp, s.Offset3DY);
    end;
    Result := Result + maxgp;
  end;
end;

function TTMSFMXChartSerie.GetBarCount: Integer;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  c: TTMSFMXChart;
  cnt: Integer;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  if IsStacked then
  begin
    cnt := -1;
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.IsBar and (s.Visible) and (s.GroupIndex > cnt) and (s.IsStacked) then
      begin
        Inc(Result);
        Inc(cnt);
      end;
    end;
  end
  else
  begin
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.IsBar and (s.Visible) and (s.GroupIndex = GroupIndex) then
        Inc(Result);
    end;
  end;
end;

function TTMSFMXChartSerie.GetBarWidth: Double;
var
  cnt: Integer;
  xs: Double;
begin
  Result := 0;
  case Bar.WidthType of
    bwtPercentage:
    begin
      cnt := GetBarCount;
      if cnt > 0 then
      begin
        xs := ValueToX(GetXOffset, False);
        Result := xs / cnt;
        Result := Result * Bar.Width / 100;
      end;
    end;
    bwtPixels: Result := Bar.Width;
  end;
end;

function TTMSFMXChartSerie.GetEndCountPoint: Integer;
begin
  Result := Round(Min(GetPointsCount - 1, FEndX));
  case ChartType of
    ctXYLine, ctXYMarker, ctSpider: Result := GetPointsCount - 1;
  end;
end;

function TTMSFMXChartSerie.GetEndPoint: Integer;
begin
  Result := GetPointsCount - 1;
  case ChartType of
    ctBar, ctStackedBar, ctStackedPercentageBar, ctMarker, ctOHLC, ctCandleStick, ctBoxPlot: Result := Round(Min(GetPointsCount - 1, FEndX));
    ctLine, ctDigitalLine, ctArea, ctStackedArea, ctBand, ctStackedPercentageArea: Result := Round(Min(GetPointsCount - 2, FEndX - 1));
    ctXYLine: Result := GetPointsCount - 2;
  end;
end;

function TTMSFMXChartSerie.GetLegendText(APoint: TTMSFMXChartPointVirtual): String;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  Result := FloatToStr(APoint.YValue);
  if APoint.LegendText <> '' then
    Result := APoint.LegendText;

  if c.FVirtualMode then
    c.DoGetSerieLegendTextVirtual(Self, APoint, Result)
  else
    c.DoGetSerieLegendText(Self, APoint.Point, Result);
end;

function TTMSFMXChartSerie.GetMultiPointWidth: Double;
var
  cnt: Integer;
  xs: Double;
begin
  Result := 0;
  case MultiPoint.WidthType of
    bwtPercentage:
    begin
      cnt := 1;
      if cnt > 0 then
      begin
        xs := ValueToX(GetXOffset, False);
        Result := xs / cnt;
        Result := Result * MultiPoint.Width / 100;
      end;
    end;
    bwtPixels: Result := MultiPoint.Width;
  end;
end;

function TTMSFMXChartSerie.GetPieRect(ASerie: Boolean = False): TRectF;
var
  c: TTMSFMXChart;
  w, wp, hp: Double;
  r: TRectF;
  sr: TRectF;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := c.GetSeriesRectangle;
  if c.Series.Count > 0 then
  begin
    w := r.Width;

    if not c.HasStackedPieChart then
    begin
      w := w / c.Series.Count;
      r.Left := r.Left + w * Index;
    end;
    r.Width := w;
    r.Left := r.Left + Pie.Margins.Left;
    r.Top := r.Top + Pie.Margins.Top;
    r.Right := r.Right - Pie.Margins.Right;
    r.Bottom := r.Bottom - Pie.Margins.Bottom;

    if ASerie then
    begin
      Result := r;
      Exit;
    end;

    sr := r;

    wp := r.Width;
    hp := r.Height;

    if Pie.AutoSize then
    begin
      if hp > wp then
        hp := wp
      else
        wp := hp;
    end
    else
    begin
      wp := Pie.Size;
      hp := Pie.Size;
    end;

    case Pie.Position of
      ppTopLeft, ppCenterLeft, ppBottomLeft: r.Left := sr.Left;
      ppTopCenter, ppCenterCenter, ppBottomCenter: r.Left := sr.Left + (sr.Width - wp) / 2;
      ppTopRight, ppCenterRight, ppBottomright: r.Left := sr.Right - wp;
    end;

    case Pie.Position of
      ppTopLeft, ppTopCenter, ppTopRight: r.Top := sr.Top;
      ppCenterLeft, ppCenterCenter, ppCenterRight: r.Top := sr.Top + (sr.Height - hp) / 2;
      ppBottomLeft, ppBottomCenter, ppBottomRight: r.Top := sr.Bottom - hp;
    end;

    r.Width := wp;
    r.Height := hp;

    Result := r;
  end;
end;

function TTMSFMXChartSerie.GetPoint(AIndex: Integer): TTMSFMXChartPointVirtual;
var
  c: TTMSFMXChart;
begin
  Result := GetDefaultVirtualPoint;
  c := Chart;
  if not Assigned(c) then
    Exit;

  Result.XValue := AIndex;

  if (AIndex >= 0) and (AIndex <= Points.Count - 1) then
    Result := Points[AIndex].GetValue;

  Result.Index := AIndex;
  c.DoGetPoint(Self, AIndex, Result);
end;

function TTMSFMXChartSerie.GetPointsCount: Integer;
var
  c: TTMSFMXChart;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;

  if Assigned(Points) then
    Result := Points.Count;

  c.DoGetNumberOfPoints(Self, Result);
end;

function TTMSFMXChartSerie.GetPointsMax: Double;
var
  I: Integer;
  sta, stp: Integer;
  pt: TTMSFMXChartPointVirtual;
begin
  sta := GetStartPoint;
  stp := GetEndPoint;
  Result := 0;
  for I := sta to stp do
  begin
    pt := GetPoint(I);
    Result := Max(Result, pt.YValue);
  end;
end;

function TTMSFMXChartSerie.GetPointsTotal: Double;
var
  I: Integer;
  sta, stp: Integer;
  pt: TTMSFMXChartPointVirtual;
begin
  sta := GetStartPoint;
  stp := GetEndPoint;
  Result := 0;
  for I := sta to stp do
  begin
    pt := GetPoint(I);
    Result := Result + pt.YValue;
  end;
end;

function TTMSFMXChartSerie.GetSpiderLegendText(
  APoint: TTMSFMXChartPointVirtual): String;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  Result := FloatToStr(APoint.YValue);
  if APoint.LegendText <> '' then
    Result := APoint.LegendText;

  if c.FVirtualMode then
    c.DoGetSerieSpiderLegendTextVirtual(Self, APoint, Result)
  else
    c.DoGetSerieSpiderLegendText(Self, APoint.Point, Result);
end;

function TTMSFMXChartSerie.GetStackedDrawLine(ADrawIndex: Integer;
  var ADrawLine: TTMSFMXChartDrawLine): Boolean;
var
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
  idx: Integer;
  I: Integer;
begin
  Result := False;
  c := Chart;
  if not Assigned(c) then
    Exit;

  idx := Index;
  for I := idx - 1 downto 0 do
  begin
    if (I >= 0) and (I <= c.Series.Count - 1) then
    begin
      s := c.Series[I];
      if s.Visible and s.IsStacked and (s.GroupIndex = GroupIndex) and s.IsArea then
      begin
        if (ADrawIndex >= 0) and (ADrawIndex <= s.FDrawLines.Count - 1) then
        begin
          ADrawLine := s.FDrawLines[ADrawIndex];
          Result := True;
          Break;
        end;
      end;
    end;
  end;
end;

function TTMSFMXChartSerie.GetStackedDrawRect(ADrawIndex: Integer; var ADrawRect: TTMSFMXChartDrawRect): Boolean;
var
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
  idx: Integer;
  I: Integer;
begin
  Result := False;
  c := Chart;
  if not Assigned(c) then
    Exit;

  idx := Index;
  for I := idx - 1 downto 0 do
  begin
    if (I >= 0) and (I <= c.Series.Count - 1) then
    begin
      s := c.Series[I];
      if s.Visible and s.IsStacked and (s.GroupIndex = GroupIndex) and s.IsBar then
      begin
        if (ADrawIndex >= 0) and (ADrawIndex <= s.FDrawRects.Count - 1) then
        begin
          ADrawRect := s.FDrawRects[ADrawIndex];
          Result := True;
          Break;
        end;
      end;
    end;
  end;
end;

function TTMSFMXChartSerie.GetStackedMax(APointIndex: Integer): Double;
var
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
  I: Integer;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;

  for I := 0 to c.Series.Count - 1 do
  begin
    s := c.Series[I];
    if s.Visible and s.IsStacked and (s.GroupIndex = GroupIndex) then
    begin
      if (APointIndex >= 0) and (APointIndex <= s.GetPointsCount - 1) then
        Result := Result + s.GetPoint(APointIndex).YValue;
    end;
  end;
end;

function TTMSFMXChartSerie.GetStartCountPoint: Integer;
begin
  Result := 0;
  case ChartType of
    ctLine, ctDigitalLine, ctArea, ctBand, ctStackedArea, ctStackedPercentageArea, ctBar, ctStackedBar, ctStackedPercentageBar,
    ctMarker, ctOHLC, ctCandleStick, ctBoxPlot: Result := Round(Max(0, FStartX));
  end;
end;

function TTMSFMXChartSerie.GetStartPoint: Integer;
begin
  Result := 0;
  case ChartType of
    ctLine, ctDigitalLine, ctArea, ctBand, ctStackedArea, ctStackedPercentageArea, ctBar, ctStackedBar, ctStackedPercentageBar,
    ctMarker, ctOHLC, ctCandleStick, ctBoxPlot: Result := Round(Max(0, FStartX));
  end;
end;

function TTMSFMXChartSerie.GetTotalBarWidth: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  c: TTMSFMXChart;
  cnt: Integer;
  bw: Double;
  K: Integer;
  gp: Integer;
  l, j: Integer;
  lgp: TArray<Integer>;
  xs: Double;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  cnt := GetBarCount;

  if IsStacked then
  begin
    gp := -1;
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.Visible and s.IsStacked and (s.GroupIndex > gp) and s.IsBar then
      begin
        gp := s.GroupIndex;
        l := Length(lgp);
        SetLength(lgp, l + 1);
        lgp[l] := gp;
      end;
    end;

    K := 0;
    for I := 0 to Length(lgp) - 1 do
    begin
      for J := 0 to c.Series.Count - 1 do
      begin
        s := c.Series[J];
        if s.IsBar and (s.Visible) and (s.GroupIndex = lgp[I]) and s.IsStacked then
        begin
          bw := 0;
          case s.Bar.WidthType of
            bwtPercentage:
            begin
              if cnt > 0 then
              begin
                xs := s.ValueToX(s.GetXOffset, False);
                bw := xs / cnt;
                if Length(lgp) - 1 = K then
                  bw := bw * s.Bar.Width / 100
                else
                  bw := bw * (s.Bar.Width + s.Bar.Spacing) / 100
              end;
            end;
            bwtPixels:
            begin
              if Length(lgp) - 1 = K then
                bw := s.Bar.Width
              else
                bw := s.Bar.Width + s.Bar.Spacing
            end;
          end;
          Result := Result + bw;
          Inc(K);
          Break;
        end;
      end;
    end;
  end
  else
  begin
    K := 0;
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.IsBar and (s.Visible) and (s.GroupIndex = GroupIndex) then
      begin
        bw := 0;
        case s.Bar.WidthType of
          bwtPercentage:
          begin
            if cnt > 0 then
            begin
              xs := s.ValueToX(s.GetXOffset, False);
              bw := xs / cnt;
              if cnt - 1 = K then
                bw := bw * s.Bar.Width / 100
              else
                bw := bw * (s.Bar.Width + s.Bar.Spacing) / 100
            end;
          end;
          bwtPixels:
          begin
            if cnt - 1 = K then
              bw := s.Bar.Width
            else
              bw := s.Bar.Width + s.Bar.Spacing
          end;
        end;
        Result := Result + bw;
        Inc(K);
      end;
    end;
  end;
end;

function TTMSFMXChartSerie.GetXOffset: Double;
var
  c: Double;
begin
  Result := 0;
  c := Abs(FXMax - FXMin);
  if c > 0 then
  begin
    if c > 1 then
      Result := 1
    else
    begin
      if GetPointsCount > 1 then
        Result := (GetPoint(1).XValue - GetPoint(0).XValue)
      else
        Result := 1;
    end;
  end;
end;

function TTMSFMXChartSerie.Is3DEnabled: Boolean;
begin
  Result := ((Offset3DX > 0) or (Offset3DY > 0)) and Enable3D;
end;

function TTMSFMXChartSerie.IsArea: Boolean;
begin
  Result := (ChartType = ctArea) or (ChartType = ctBand) or (ChartType = ctStackedArea) or (ChartType = ctStackedPercentageArea);
end;

function TTMSFMXChartSerie.IsBar: Boolean;
begin
  Result := (ChartType = ctBar) or (ChartType = ctStackedBar) or (ChartType = ctStackedPercentageBar);
end;

function TTMSFMXChartSerie.IsPie: Boolean;
begin
  Result := (ChartType = ctPie) or (ChartType = ctVariableRadiusPie) or (ChartType = ctSizedPie) or (ChartType = ctSpider);
end;

function TTMSFMXChartSerie.IsSpider: Boolean;
begin
  Result := (ChartType = ctSpider);
end;

function TTMSFMXChartSerie.IsStacked: Boolean;
begin
  Result := (ChartType = ctStackedBar) or (ChartType = ctStackedPercentageBar) or (ChartType = ctStackedArea) or
    (ChartType = ctStackedPercentageArea);
end;

function TTMSFMXChartSerie.GetBarOffset: Double;
var
  I: Integer;
  s: TTMSFMXChartSerie;
  c: TTMSFMXChart;
  cnt: Integer;
  bw: Double;
  gp: Integer;
  lgp: TArray<Integer>;
  l, J, K: Integer;
  xs: Double;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  cnt := GetBarCount;
  if IsStacked then
  begin
    gp := -1;
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.Visible and s.IsStacked and (s.GroupIndex > gp) and s.IsBar then
      begin
        gp := s.GroupIndex;
        l := Length(lgp);
        SetLength(lgp, l + 1);
        lgp[l] := gp;
      end;
    end;

    K := 0;
    for I := 0 to Length(lgp) - 1 do
    begin
      for J := 0 to c.Series.Count - 1 do
      begin
        s := c.Series[J];
        if s.GroupIndex = GroupIndex then
          Exit;
        if s.IsBar and (s.Visible) and (s.GroupIndex = lgp[I]) and s.IsStacked then
        begin
          bw := 0;
          case s.Bar.WidthType of
            bwtPercentage:
            begin
              if cnt > 0 then
              begin
                xs := s.ValueToX(s.GetXOffset, False);
                bw := xs / cnt;
                if Length(lgp) - 1 = K then
                  bw := bw * s.Bar.Width / 100
                else
                  bw := bw * (s.Bar.Width + s.Bar.Spacing) / 100
              end;
            end;
            bwtPixels:
            begin
              if Length(lgp) - 1 = K then
                bw := s.Bar.Width
              else
                bw := s.Bar.Width + s.Bar.Spacing
            end;
          end;
          Result := Result + bw;
          Inc(K);
          Break;
        end;
      end;
    end;
  end
  else
  begin
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if (s = Self) then
        Exit;
      if s.IsBar and (s.Visible) and (s.GroupIndex = GroupIndex) then
      begin
        bw := 0;
        case s.Bar.WidthType of
          bwtPercentage:
          begin
            if cnt > 0 then
            begin
              xs := s.ValueToX(s.GetXOffset, False);
              bw := xs / cnt;
              bw := bw * (s.Bar.Width + s.Bar.Spacing) / 100;
            end;
          end;
          bwtPixels: bw := s.Bar.Width + s.Bar.Spacing;
        end;
        Result := Result + bw
      end;
    end;
  end;
end;

function TTMSFMXChartSerie.Offset: Boolean;
begin
  Result := Mode = smStatistical;
end;

function TTMSFMXChartSerie.CalculateOffset: Boolean;
var
  I: Integer;
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
begin
  Result := False;
  c := Chart;
  if not Assigned(c) then
    Exit;

  case Mode of
    smMathematical:
    begin
      if IsBar then
        Result := True
      else
      begin
        for I := 0 to c.Series.Count - 1 do
        begin
          s := c.Series[I];
          if s.Visible and ((s.AutoXRange = arCommon) or (s.AutoXRange = arCommonZeroBased)) and s.IsBar then
          begin
            Result := True;
            Break;
          end;
        end;
      end;
    end;
    smStatistical: Result := True;
  end;
end;

procedure TTMSFMXChartSerie.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

procedure TTMSFMXChartSerie.SetAutoXRange(const Value: TTMSFMXChartSerieAutoRange);
begin
  if FAutoXRange <> Value then
  begin
    FAutoXRange := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetAutoYRange(const Value: TTMSFMXChartSerieAutoRange);
begin
  if FAutoYRange <> Value then
  begin
    FAutoYRange := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetBar(const Value: TTMSFMXChartSerieBar);
begin
  if FBar <> Value then
  begin
    FBar.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetChartType(const Value: TTMSFMXChartSerieType);
begin
  if FChartType <> Value then
  begin
    FChartType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetEnable3D(const Value: Boolean);
begin
  if FEnable3D <> Value then
  begin
    FEnable3D := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetGroupIndex(const Value: Integer);
begin
  if FGroupIndex <> Value then
  begin
    FGroupIndex := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetLabels(const Value: TTMSFMXChartSerieLabels);
begin
  if FLabels <> Value then
  begin
    FLabels.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetLegend(const Value: TTMSFMXChartSerieLegend);
begin
  if FLegend <> Value then
  begin
    FLegend.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetLegendText(const Value: String);
begin
  if FLegendText <> Value then
  begin
    FLegendText := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMarkers(const Value: TTMSFMXChartSerieMarkers);
begin
  if FMarkers <> Value then
  begin
    FMarkers.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMaxXOffsetPercentage(const Value: Double);
begin
  if FMaxXOffsetPercentage <> Value then
  begin
    FMaxXOffsetPercentage := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMaxYOffsetPercentage(const Value: Double);
begin
  if FMaxYOffsetPercentage <> Value then
  begin
    FMaxYOffsetPercentage := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMaxX(const Value: Double);
begin
  if FMaxX <> Value then
  begin
    FMaxX := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMaxY(const Value: Double);
begin
  if FMaxY <> Value then
  begin
    FMaxY := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMinXOffsetPercentage(const Value: Double);
begin
  if FMinXOffsetPercentage <> Value then
  begin
    FMinXOffsetPercentage := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMinYOffsetPercentage(const Value: Double);
begin
  if FMinYOffsetPercentage <> Value then
  begin
    FMinYOffsetPercentage := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMinX(const Value: Double);
begin
  if FMinX <> Value then
  begin
    FMinX := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMinY(const Value: Double);
begin
  if FMinY <> Value then
  begin
    FMinY := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMode(const Value: TTMSFMXChartSerieMode);
begin
  if FMode <> Value then
  begin
    FMode := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetOffset3DX(const Value: Double);
begin
  if FOffset3DX <> Value then
  begin
    FOffset3DX := Max(0, Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetOffset3DY(const Value: Double);
begin
  if FOffset3DY <> Value then
  begin
    FOffset3DY := Max(0, Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetMultiPoint(const Value: TTMSFMXChartSerieMultiPoint);
begin
  if FMultiPoint <> Value then
  begin
    FMultiPoint.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetPie(const Value: TTMSFMXChartSeriePie);
begin
  if FPie <> Value then
  begin
    FPie.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetPoints(const Value: TTMSFMXChartPoints);
begin
  if FPoints <> Value then
  begin
    FPoints.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetShowInLegend(const Value: Boolean);
begin
  if FShowInLegend <> Value then
  begin
    FShowInLegend := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerie.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetXGrid(const Value: TTMSFMXChartSerieXYGrid);
begin
  if FXGrid <> Value then
  begin
    FXGrid.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetXValues(const Value: TTMSFMXChartSerieXValues);
begin
  if FXValues <> Value then
  begin
    FXValues.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetYGrid(const Value: TTMSFMXChartSerieXYGrid);
begin
  if FYGrid <> Value then
  begin
    FYGrid.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetYValues(const Value: TTMSFMXChartSerieYValues);
begin
  if FYValues <> Value then
  begin
    FYValues.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.SetZeroReferenceValue(const Value: Double);
begin
  if FZeroReferenceValue <> Value then
  begin
    FZeroReferenceValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerie.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChartSerie.ValueToX(AValue: Double; APosition: Boolean = True): Double;
var
  r: TRectF;
  c: TTMSFMXChart;
  ct: Double;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  if APosition then
  begin
    r := c.GetSeriesRectangle;

    ct := 0;
    if Offset then
      ct := GetXOffset / 2;

    if FXScale > 0 then
      Result := r.Left + (AValue - FXMin + ct) * FXScale
    else
      Result := r.Left + (AValue - FXMin + ct);

    Result := Result + F3DXOffset;
  end
  else
  begin
    if FXScale > 0 then
      Result := AValue * FXScale
    else
      Result := AValue;
  end;
end;

function TTMSFMXChartSerie.ValueToY(AValue: Double; APosition: Boolean = True): Double;
var
  r: TRectF;
  c: TTMSFMXChart;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  r := c.GetSeriesRectangle;
  if APosition then
  begin
    if FYScale > 0 then
      Result := r.Top + r.Height - (AValue - FYMin) * FYScale
    else
      Result := r.Top + r.Height - (AValue - FYMin);

    Result := Result - F3DYOffset;
  end
  else
  begin
    if FYScale > 0 then
      Result := AValue * FYScale
    else
      Result := AValue;
  end;
end;

function TTMSFMXChartSerie.XMax: Double;
begin
  Result := FXMax;
end;

function TTMSFMXChartSerie.XMin: Double;
begin
  Result := FXMin;
end;

function TTMSFMXChartSerie.XScale: Double;
begin
  Result := FXScale;
end;

function TTMSFMXChartSerie.XToValue(AX: Double; APosition: Boolean = True): Double;
var
  r: TRectF;
  c: TTMSFMXChart;
  ct: Double;
  off: Double;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  if APosition then
  begin
    r := c.GetSeriesRectangle;
    ct := 0;
    if Offset then
      ct := GetXOffset / 2;

    off := F3DXOffset;
    if FXScale > 0 then
      Result := ((AX - r.Left - off) / FXScale) + FXMin - ct
    else
      Result := (AX - r.Left - off) + FXMin - ct;
  end
  else
  begin
    if FXScale > 0 then
      Result := AX / FXScale
    else
      Result := AX;
  end;
end;

function TTMSFMXChartSerie.XYToBar(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to FDrawRects.Count - 1 do
  begin
    if PtInRect(FDrawRects[I].Rect, PointF(X, Y)) then
    begin
      Result := FDrawRects[I].Reference;
      Break;
    end;
  end;
end;

function TTMSFMXChartSerie.XYToPoint(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
  r: TRectF;
  dp: TTMSFMXChartDrawPoint;
  c: TTMSFMXChart;
  im: Double;
begin
  Result := nil;
  c := Chart;
  if not Assigned(c) then
    Exit;

  im := c.ClickMargin;
  for I := 0 to FDrawPoints.Count - 1 do
  begin
    dp := FDrawPoints[I];
    r := RectF(dp.Point.X - im / 2, dp.Point.Y - im / 2, dp.point.X + im / 2, dp.Point.Y + im / 2);
    if PtInRect(r, PointF(X, Y)) then
    begin
      Result := FDrawPoints[I].Reference;
      Break;
    end;
  end;
end;

function TTMSFMXChartSerie.XYToSlice(X, Y: Double): TTMSFMXChartPoint;
var
  I: Integer;
  ds: TTMSFMXChartDrawSlice;
  c: TTMSFMXChart;
  pth: TPathData;
  bmp: TBitmap;
begin
  Result := nil;
  c := Chart;
  if not Assigned(c) then
    Exit;

  pth := TPathData.Create;
  bmp := TBitmap.Create(1, 1);
  for I := 0 to FDrawSlices.Count - 1 do
  begin
    ds := FDrawSlices[I];
    pth.Clear;
    pth.MoveTo(ds.CenterPoint);
    pth.AddArc(ds.CenterPoint, ds.Radius, ds.StartAngle, ds.SweepAngle);
    pth.LineTo(ds.CenterPoint);
    pth.ClosePath;
    if bmp.Canvas.PtInPath(PointF(X, Y), pth) then
    begin
      Result := FDrawSlices[I].Reference;
      Break;
    end;
  end;
  bmp.Free;
  pth.Free;
end;

function TTMSFMXChartSerie.XYToBarVirtual(X, Y: Double;
  var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FDrawRects.Count - 1 do
  begin
    if PtInRect(FDrawRects[I].Rect, PointF(X, Y)) then
    begin
      Result := True;
      APoint := FDrawRects[I].VirtualReference;
      Break;
    end;
  end;
end;

function TTMSFMXChartSerie.XYToLegendItem(X, Y: Double): Integer;
var
  c: TTMSFMXChart;
  r, rt: TRectF;
  I, cnt: Integer;
  pt: TTMSFMXChartPointVirtual;
  th, xx, xy: Double;
  dr: Boolean;
  str: String;
  g: TBitmap;
begin
  Result := -1;
  c := Chart;
  if not Assigned(c) then
    Exit;

  cnt := GetPointsCount;
  if cnt = 0 then
    Exit;

  if not Legend.Visible then
    Exit;

  r := Legend.GetRectangle;
  if r.IsEmpty then
    Exit;

  g := TBitmap.Create(1, 1);
  try
    dr := True;
    c.DoBeforeDrawSerieLegend(g.Canvas, Self, r, dr);
    if dr then
    begin
      g.Canvas.Font.Assign(Legend.Font);

      xx := r.Left;
      xy := r.Top + 5;

      if LegendText <> '' then
      begin
        th := g.Canvas.TextHeight(LegendText);
        xy := xy + th + 5;
      end;

      for I := 0 to cnt - 1 do
      begin
        pt := GetPoint(I);
        str := GetLegendText(pt);
        if str <> '' then
        begin
          th := g.Canvas.TextHeight(str);
          rt := RectF(xx, xy, xx + r.Width, xy + th);

          if PtInRect(rt, PointF(X, Y)) then
          begin
            Result := I;
            Break;
          end;

          xy := xy + th + 5;
        end;
      end;
    end;
  finally
    g.Free;
  end;
end;

function TTMSFMXChartSerie.XYToPointVirtual(X,
  Y: Double; var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
  r: TRectF;
  dp: TTMSFMXChartDrawPoint;
  c: TTMSFMXChart;
  im: Double;
begin
  Result := False;
  c := Chart;
  if not Assigned(c) then
    Exit;

  im := c.ClickMargin;
  for I := 0 to FDrawPoints.Count - 1 do
  begin
    dp := FDrawPoints[I];
    r := RectF(dp.Point.X - im / 2, dp.Point.Y - im / 2, dp.point.X + im / 2, dp.Point.Y + im / 2);
    if PtInRect(r, PointF(X, Y)) then
    begin
      Result := True;
      APoint := FDrawPoints[I].VirtualReference;
      Break;
    end;
  end;
end;

function TTMSFMXChartSerie.XYToSliceVirtual(X, Y: Double;
  var APoint: TTMSFMXChartPointVirtual): Boolean;
var
  I: Integer;
  ds: TTMSFMXChartDrawSlice;
  c: TTMSFMXChart;
  pth: TPathData;
  bmp: TBitmap;
begin
  Result := False;
  c := Chart;
  if not Assigned(c) then
    Exit;

  pth := TPathData.Create;
  bmp := TBitmap.Create(1, 1);
  for I := 0 to FDrawSlices.Count - 1 do
  begin
    ds := FDrawSlices[I];
    pth.Clear;
    pth.MoveTo(ds.CenterPoint);
    pth.AddArc(ds.CenterPoint, ds.Radius, ds.StartAngle, ds.SweepAngle);
    pth.LineTo(ds.CenterPoint);
    pth.ClosePath;
    if bmp.Canvas.PtInPath(PointF(X, Y), pth) then
    begin
      Result := True;
      APoint := FDrawSlices[I].VirtualReference;
      Break;
    end;
  end;
  bmp.Free;
  pth.Free;
end;

function TTMSFMXChartSerie.YMax: Double;
begin
  Result := FYMax;
end;

function TTMSFMXChartSerie.YMin: Double;
begin
  Result := FYMin;
end;

function TTMSFMXChartSerie.YScale: Double;
begin
  Result := FYScale;
end;

function TTMSFMXChartSerie.YToValue(AY: Double; APosition: Boolean = True): Double;
var
  r: TRectF;
  c: TTMSFMXChart;
  off: Double;
begin
  c := Chart;
  Result := 0;
  if not Assigned(c) then
    Exit;

  if APosition then
  begin
    r := c.GetSeriesRectangle;

    off := F3DYOffset;

    if FYScale > 0 then
      Result := ((r.Height - AY + r.Top - off) / FYScale) + FYMin
    else
      Result := (r.Height - AY + r.Top - off) + FYMin;
  end
  else
  begin
    if FYScale > 0 then
      Result := AY / FYScale
    else
      Result := AY;
  end;
end;

{ TTMSFMXChartSeries }

function TTMSFMXChartSeries.Add: TTMSFMXChartSerie;
begin
  Result := TTMSFMXChartSerie(inherited Add);
end;

function TTMSFMXChartSeries.Chart: TTMSFMXChart;
begin
  Result := FChart;
end;

constructor TTMSFMXChartSeries.Create(AChart: TTMSFMXChart);
begin
  inherited Create(AChart, TTMSFMXChartSerie);
  FChart := AChart;
end;

procedure TTMSFMXChartSeries.Draw(ACanvas: TCanvas);
var
  I, J: Integer;
  st: TCanvasSaveState;
  c: TTMSFMXChart;
  r, rs: TRectF;
  dr: Boolean;
  offx, offy: Double;
  stopb: Integer;
begin
  inherited;
  c := Chart;
  if not Assigned(c) then
    Exit;

  offx := c.FTotalOffset3DX;
  offy := c.FTotalOffset3DY;
  r := c.GetSeriesRectangle;
  st := ACanvas.SaveState;
  rs.Left := r.Left - c.SeriesMargins.Left;
  rs.Top := r.Top - c.SeriesMargins.Top;
  rs.Bottom := r.Bottom + c.SeriesMargins.Bottom;
  rs.Right := r.Right + c.SeriesMargins.Right;
  rs.Top := rs.Top - offy;
  rs.Right := rs.Right + offx;
  ACanvas.IntersectClipRect(rs);

  dr := True;
  c.DoBeforeDrawSeries(ACanvas, r, dr);
  if dr then
  begin
    stopb := GetBarStop;

    if c.DrawOrder = csdoBarsFirst then
    begin
      for I := 0 to stopb - 1 do
      begin
        for J := 0 to Count - 1 do
        begin
          if I <= Items[J].FDrawRects.Count - 1 then
          begin
            if Items[J].Visible and Items[J].IsBar then
              Items[J].Draw(ACanvas, I);
          end;
        end;
      end;

      for I := 0 to Count - 1 do
      begin
        if Items[I].Visible and Items[I].IsBar then
        begin
          Items[I].DrawMarkers(ACanvas);
          Items[I].DrawLabels(ACanvas);
          Items[I].DrawAnnotations(ACanvas);
        end;
      end;
    end;

    for I := 0 to Count - 1 do
    begin
      if Items[I].Visible and Items[I].IsSpider then
        Items[I].DrawSpiderGrid(ACanvas);
    end;
        
    for I := 0 to Count - 1 do
    begin
      if Items[I].Visible and (not Items[I].IsBar or (c.DrawOrder = csdoNormal)) then
      begin
        if Items[I].IsBar then
        begin
          for J := 0 to Items[I].FDrawRects.Count - 1 do
            Items[I].Draw(ACanvas, J);
        end
        else
          Items[I].Draw(ACanvas);

        Items[I].DrawMarkers(ACanvas);
      end;
    end;

    for I := 0 to Count - 1 do
    begin
      if Items[I].Visible and Items[I].IsSpider then
        Items[I].DrawSpiderValues(ACanvas);        
    end;    

    for I := 0 to Count - 1 do
    begin
      if Items[I].Visible and (not Items[I].IsBar or (c.DrawOrder = csdoNormal)) then
      begin
        Items[I].DrawLabels(ACanvas);
        Items[I].DrawAnnotations(ACanvas);
      end;
    end;

    if c.DrawOrder = csdoBarsLast then
    begin
      for I := 0 to stopb - 1 do
      begin
        for J := 0 to Count - 1 do
        begin
          if I <= Items[J].FDrawRects.Count - 1 then
          begin
            if Items[J].Visible and Items[J].IsBar then
              Items[J].Draw(ACanvas, I);
          end;
        end;
      end;

      for I := 0 to Count - 1 do
      begin
        if Items[I].Visible and Items[I].IsBar then
        begin
          Items[I].DrawMarkers(ACanvas);
          Items[I].DrawLabels(ACanvas);
          Items[I].DrawAnnotations(ACanvas);
        end;
      end;
    end;

    for I := 0 to Count - 1 do
    begin
      if Items[I].Visible then
        Items[I].DrawLegend(ACanvas);
    end;

    c.DoAfterDrawSeries(ACanvas, r);
  end;

  ACanvas.RestoreState(st);
end;

function TTMSFMXChartSeries.GetBarStop: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Count - 1 do
    if Items[I].Visible then
      Result := Max(Result, Items[I].FDrawRects.Count);
end;

function TTMSFMXChartSeries.GetItem(Index: Integer): TTMSFMXChartSerie;
begin
  Result := TTMSFMXChartSerie(inherited Items[Index]);
end;

function TTMSFMXChartSeries.Insert(Index: Integer): TTMSFMXChartSerie;
begin
  Result := TTMSFMXChartSerie(inherited Insert(Index));
end;

procedure TTMSFMXChartSeries.SetItem(Index: Integer;
  const Value: TTMSFMXChartSerie);
begin
  inherited Items[Index] := Value;
end;

{ TTMSFMXChartPoint }

procedure TTMSFMXChartPoint.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartPoint then
  begin
    FTag := (Source as TTMSFMXChartPoint).Tag;
    FXValue := (Source as TTMSFMXChartPoint).XValue;
    FYValue := (Source as TTMSFMXChartPoint).YValue;
    FYValueSecond := (Source as TTMSFMXChartPoint).YValueSecond;
    FYValueOpen := (Source as TTMSFMXChartPoint).YValueOpen;
    FYValueMedian := (Source as TTMSFMXChartPoint).YValueMedian;
    FYValueLow := (Source as TTMSFMXChartPoint).YValueLow;
    FYValueClose := (Source as TTMSFMXChartPoint).YValueClose;
    FYValueVariable := (Source as TTMSFMXChartPoint).YValueVariable;
    FXValueText := (Source as TTMSFMXChartPoint).XValueText;
    FLegendText := (Source as TTMSFMXChartPoint).LegendText;
    FExplode := (Source as TTMSFMXChartPoint).Explode;
    FColor := (Source as TTMSFMXChartPoint).Color;
    FAnnotations.Assign((Source as TTMSFMXChartPoint).Annotations);
  end;
end;

procedure TTMSFMXChartPoint.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

procedure TTMSFMXChartPoint.UpdateSequential;
var
  s: TTMSFMXChartSerie;
begin
  s := Serie;
  if Assigned(s) and s.FSequential then
  begin
    if XValue < s.FPrevXValue then
      s.FSequential := False
    else
      s.FPrevXValue := XValue;
  end;
end;

procedure TTMSFMXChartPoint.BitmapChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChartPoint.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartPoint.Create(Collection: TCollection);
begin
  inherited;
  if Assigned(Collection) then
    FSerie := (Collection as TTMSFMXChartPoints).Serie;

  FTag := 0;
  FExplode := 0;
  if Assigned(FSerie) then
    FXValue := FSerie.Points.Count - 1;

  FYValue := Random(100);
  FYValueSecond := FYValue - 20;
  FYValueVariable := FYValue;
  FYValueLow := FYValue - 20;
  FYValueOpen := FYValue + RandomRange(-10, 10);
  FYValueMedian := FYValueOpen + RandomRange(-5, 5);
  FYValueClose := FYValue - RandomRange(-10, 10);
  FColor := claNull;
  FAnnotations := TTMSFMXChartAnnotations.Create(Self);
  UpdateChart;
end;

destructor TTMSFMXChartPoint.Destroy;
begin
  FAnnotations.Free;
  inherited;
  UpdateChart;
end;

function TTMSFMXChartPoint.GetValue: TTMSFMXChartPointVirtual;
begin
  Result.Color := Color;
  Result.Tag := Tag;
  Result.XValue := XValue;
  Result.XValueText := XValueText;
  Result.YValue := YValue;
  Result.YValueSecond := YValueSecond;
  Result.YValueVariable := YValueVariable;
  Result.LegendText := LegendText;
  Result.YValueHigh := YValueHigh;
  Result.YValueClose := YValueClose;
  Result.YValueOpen := YValueOpen;
  REsult.YValueMedian := YValueMedian;
  Result.YValueLow := YValueLow;
  Result.Explode := Explode;
  Result.Point := Self;
end;

procedure TTMSFMXChartPoint.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartPoint.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartPoint.SetAnnotations(
  const Value: TTMSFMXChartAnnotations);
begin
  if FAnnotations <> Value then
  begin
    FAnnotations.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartPoint.SetColor(const Value: TAlphaColor);
begin
  if FColor <> Value then
  begin
    FColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartPoint.SetExplode(const Value: Double);
begin
  if FExplode <> Value then
  begin
    FExplode := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetLegendText(const Value: String);
begin
  if FLegendText <> value then
  begin
    FLegendText := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartPoint.SetXValue(const Value: Double);
begin
  FXValue := Value;
  UpdateSequential;
  UpdateChart;
end;

procedure TTMSFMXChartPoint.SetXValueText(const Value: String);
begin
  if FXValueText <> Value then
  begin
    FXValueText := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValue(const Value: Double);
begin
  if FYValue <> Value then
  begin
    FYValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueClose(const Value: Double);
begin
  if FYValueClose <> Value then
  begin
    FYValueClose := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueLow(const Value: Double);
begin
  if FYValueLow <> Value then
  begin
    FYValueLow := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueMedian(const Value: Double);
begin
  if FYValueMedian <> Value then
  begin
    FYValueMedian := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueOpen(const Value: Double);
begin
  if FYValueOpen <> Value then
  begin
    FYValueOpen := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueSecond(const Value: Double);
begin
  if FYValueSecond <> Value then
  begin
    FYValueSecond := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartPoint.SetYValueVariable(const Value: Double);
begin
  if FYValueVariable <> Value then
  begin
    FYValueVariable := Value;
    UpdateChart;
  end;
end;

{ TTMSFMXChartPoints }

function TTMSFMXChartPoints.Add: TTMSFMXChartPoint;
begin
  Result := TTMSFMXChartPoint(inherited Add);
end;

constructor TTMSFMXChartPoints.Create(ASerie: TTMSFMXChartSerie);
begin
  inherited Create(ASerie, TTMSFMXChartPoint);
  FSerie := ASerie;
end;

function TTMSFMXChartPoints.GetItem(Index: Integer): TTMSFMXChartPoint;
begin
  Result := TTMSFMXChartPoint(inherited Items[Index]);
end;

function TTMSFMXChartPoints.Insert(Index: Integer): TTMSFMXChartPoint;
begin
  Result := TTMSFMXChartPoint(inherited Insert(Index));
end;

function TTMSFMXChartPoints.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartPoints.SetItem(Index: Integer;
  const Value: TTMSFMXChartPoint);
begin
  inherited Items[Index] := Value;
end;

{ TTMSFMXChartElement }

procedure TTMSFMXChartElement.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

procedure TTMSFMXChartElement.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartElement then
  begin
    FVisible := (Source as TTMSFMXChartElement).Visible;
    FStrokeJoin := (Source as TTMSFMXChartElement).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChartElement).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartElement).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartElement).StrokeDash;
    FStroke.Assign((Source as TTMSFMXChartElement).Stroke);
    FFill.Assign((Source as TTMSFMXChartElement).Fill);
    FFillOpacity := (Source as TTMSFMXChartElement).FillOpacity;
    FStrokeOpacity := (Source as TTMSFMXChartElement).StrokeOpacity;
  end;
end;

function TTMSFMXChartElement.Chart: TTMSFMXChart;
begin
  Result := FChart;
end;

constructor TTMSFMXChartElement.Create(AChart: TTMSFMXChart);
begin
  FChart := AChart;
  FVisible := True;
  FFillOpacity := 1;
  FStrokeOpacity := 1;
  FFill := TBrush.Create(TBrushKind.None, claGhostwhite);
  FFill.OnChanged := FillChanged;
  FStroke := TBrush.Create(TBrushKind.None, claGray);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
end;

destructor TTMSFMXChartElement.Destroy;
begin
  FStroke.Free;
  FFill.Free;
  inherited;
end;

procedure TTMSFMXChartElement.Draw(ACanvas: TCanvas);
begin
  ACanvas.Fill.Assign(Fill);
  ACanvas.Stroke.Assign(Stroke);
  {$IFDEF DELPHIXE11_LVL}
  ACanvas.Stroke.Thickness := StrokeThickness;
  ACanvas.Stroke.Cap := StrokeCap;
  ACanvas.Stroke.Dash := StrokeDash;
  ACanvas.Stroke.Join := StrokeJoin;
  {$ELSE}
  ACanvas.StrokeThickness := StrokeThickness;
  ACanvas.StrokeCap := StrokeCap;
  ACanvas.StrokeDash := StrokeDash;
  ACanvas.StrokeJoin := StrokeJoin;
  {$ENDIF}
end;

procedure TTMSFMXChartElement.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartElement.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

procedure TTMSFMXChartElement.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartElement.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartElement.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

{ TTMSFMXChartTitle }

procedure TTMSFMXChartTitle.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartTitle then
  begin
    FLine := (Source as TTMSFMXChartTitle).Line;
    FBorder := (Source as TTMSFMXChartTitle).Border;
    FFont.Assign((Source as TTMSFMXChartTitle).Font);
    FFontColor := (Source as TTMSFMXChartTitle).FontColor;
    FTextMargins.Assign((Source as TTMSFMXChartTitle).TextMargins);
    FTextVerticalAlignment := (Source as TTMSFMXChartTitle).TextVerticalAlignment;
    FTextHorizontalAlignment := (Source as TTMSFMXChartTitle).TextHorizontalAlignment;
    FText := (Source as TTMSFMXChartTitle).Text;
    FPositions := (Source as TTMSFMXChartTitle).Positions;
    FHeight := (Source as TTMSFMXChartTitle).Height;
  end;
end;

constructor TTMSFMXChartTitle.Create(AChart: TTMSFMXChart);
begin
  inherited;
  FChart := AChart;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFontColor := claBlack;
  FTextMargins := TBounds.Create(RectF(3, 3, 3, 3));
  FTextMargins.OnChange := TextMarginsChanged;
  FHeight := 35;
  FLine := True;
  FBorder := False;
  FPositions := [tipTop];
  FTextVerticalAlignment := TTextAlign.Center;
  FTextHorizontalAlignment := TTextAlign.Center;
  FStroke.Kind := TBrushKind.Solid;
  if Assigned(FChart) then
  begin
    if (csDesigning in FChart.ComponentState) and not ((csReading in FChart.ComponentState) or (csLoading in FChart.ComponentState)) then
      FText := 'Title';
  end;
end;

destructor TTMSFMXChartTitle.Destroy;
begin
  FTextMargins.Free;
  FFont.Free;
  inherited;
end;

procedure TTMSFMXChartTitle.Draw(ACanvas: TCanvas);
var
  pos: TTMSFMXChartTitlePosition;
begin
  inherited;
  if not Visible then
    Exit;

  for pos in Positions do
    DrawPart(ACanvas, pos);
end;

procedure TTMSFMXChartTitle.DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartTitlePosition);
var
  dr: Boolean;
  c: TTMSFMXChart;
  txt: String;
  rText: TRectF;
  r: TRectF;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := GetRectangle(APosition);
  if c.HasPieChart then
    r := ModifyRect(r)
  else
  begin
    case APosition of
      tipTop: r := RectF(int(r.Left) + 0.5, int(r.Top) - 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
      tipBottom: r := RectF(int(r.Left) + 0.5, int(r.Top) - 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
    end;
  end;

  dr := True;
  st := ACanvas.SaveState;
  c.DoBeforeDrawTitle(ACanvas, r, APosition, dr);

  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, FillOpacity);
    if Border then
      ACanvas.DrawRect(r, 0, 0, AllCorners, StrokeOpacity);

    if Line then
    begin
      case APosition of
        tipTop: ACanvas.DrawLine(PointF(R.Left, R.Bottom), PointF(R.Right, R.Bottom), StrokeOpacity);
        tipBottom: ACanvas.DrawLine(PointF(R.Left, R.Top), PointF(R.Right, R.Top), StrokeOpacity);
      end;
    end;

    rText.Left := r.Left + TextMargins.Left;
    rText.Top := r.Top + TextMargins.Top;
    rText.Right := r.Right - TextMargins.Right;
    rText.Bottom := r.Bottom - TextMargins.Bottom;

    ACanvas.Font.Assign(Font);
    ACanvas.Fill.Color := FontColor;
    dr := True;
    txt := Text;
    c.DoDrawTitleText(ACanvas, rText, APosition, txt, dr);
    if dr then
      ACanvas.FillText(rText, txt, False, 1, [], TextHorizontalAlignment, TextVerticalAlignment);

    c.DoAfterDrawTitle(ACanvas, r, APosition);
  end;
  ACanvas.RestoreState(st);
end;

procedure TTMSFMXChartTitle.FontChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChartTitle.GetRectangle(
  APosition: TTMSFMXChartTitlePosition): TRectF;
var
  c: TTMSFMXChart;
  r: TRectF;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := c.GetChartRectangle;
  case APosition of
    tipTop: Result := RectF(R.Left, R.Top, R.Right, R.Top + Height);
    tipBottom: Result := RectF(R.Left, R.Bottom - Height, R.Right, R.Bottom);
  end;

  if (ypLeft in c.YAxis.Positions) and c.YAxis.Visible then
    Result.Left := Result.Left + c.YAxis.GetWidth(ypLeft);

  if (ypRight in c.YAxis.Positions) and c.YAxis.Visible then
    Result.Right := Result.Right - c.YAxis.GetWidth(ypRight);

  Result.Left := Result.Left + c.FTotalOffset3DX;
end;

procedure TTMSFMXChartTitle.SetBorder(const Value: Boolean);
begin
  if FBorder <> Value then
  begin
    FBorder := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetHeight(const Value: Double);
begin
  if FHeight <> Value then
  begin
    FHeight := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartTitle.SetLine(const Value: Boolean);
begin
  if FLine <> Value then
  begin
    FLine := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetPositions(const Value: TTMSFMXChartTitlePositions);
begin
  if FPositions <> Value then
  begin
    FPositions := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartTitle.SetText(const Value: String);
begin
  if FText <> Value then
  begin
    FText := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetTextHorizontalAlignment(const Value: TTextAlign);
begin
  if FTextHorizontalAlignment <> Value then
  begin
    FTextHorizontalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetTextMargins(const Value: TBounds);
begin
  if FTextMargins <> Value then
  begin
    FTextMargins.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.SetTextVerticalAlignment(const Value: TTextAlign);
begin
  if FTextVerticalAlignment <> Value then
  begin
    FTextVerticalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartTitle.TextMarginsChanged(Sender: TObject);
begin
  RepaintChart;
end;

{ TTMSFMXChartXAxis }

procedure TTMSFMXChartXAxis.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartXAxis then
  begin
    FPositions := (Source as TTMSFMXChartXAxis).Positions;
    FHeight := (Source as TTMSFMXChartXAxis).Height;
    FLine := (Source as TTMSFMXChartXAxis).Line;
    FBorder := (Source as TTMSFMXChartXAxis).Border;
    FAutoSize := (Source as TTMSFMXChartXAxis).AutoSize;
    FDisplayAtReferenceValue := (Source as TTMSFMXChartXAxis).DisplayAtReferenceValue;
    FReferenceValue := (Source as TTMSFMXChartXAxis).ReferenceValue;
    FReferenceValueSeriesIndex := (Source as TTMSFMXChartXAxis).ReferenceValueSeriesIndex;
  end;
end;

function TTMSFMXChartXAxis.Chart: TTMSFMXChart;
begin
  Result := FChart;
end;

constructor TTMSFMXChartXAxis.Create(AChart: TTMSFMXChart);
begin
  inherited;
  FChart := AChart;
  FAutoSize := True;
  FLine := True;
  FHeight := 35;
  FBorder := False;
  FPositions := [xpBottom];
  FStroke.Kind := TBrushKind.Solid;
  FDisplayAtReferenceValue := False;
  FReferenceValueSeriesIndex := 0;
  FReferenceValue := 0;
end;

destructor TTMSFMXChartXAxis.Destroy;
begin
  inherited;
end;

procedure TTMSFMXChartXAxis.Draw(ACanvas: TCanvas);
var
  pos: TTMSFMXChartXAxisPosition;
  I: Integer;
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
begin
  inherited;

  if not Visible then
    Exit;

  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.HasPieChart then
    Exit;

  for pos in Positions do
    DrawPart(ACanvas, pos);

  for pos in Positions do
  begin
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.Visible and (pos in s.XValues.Positions) then
        DrawValues(ACanvas, pos, s);
    end;                                           
  end; 
end;

procedure TTMSFMXChartXAxis.DrawKind(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition; AKind: TTMSFMXChartDrawXYValueKind;
  ASerie: TTMSFMXChartSerie);
var
  dv: TTMSFMXChartDrawXYValue;
  I: Integer;
  dr: Boolean;
  c: TTMSFMXChart;
  tc, spc: Double;
  rt: TRectF;
  cx: TPointF;
  m: TMatrix;
  rm: TMatrix;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if AKind = vkMajor then
  begin
    tc := ASerie.XValues.MajorUnitTickMarkSize;
    spc := ASerie.XValues.MajorUnitSpacing;
    ACanvas.Font.Assign(ASerie.XValues.MajorUnitFont);
    ACanvas.Fill.Color := ASerie.XValues.MajorUnitFontColor;
    ACanvas.Stroke.Color := ASerie.XValues.MajorUnitTickMarkColor;
  end
  else
  begin
    tc := ASerie.XValues.MinorUnitTickMarkSize;
    spc := ASerie.XValues.MinorUnitSpacing;
    ACanvas.Font.Assign(ASerie.XValues.MinorUnitFont);
    ACanvas.Fill.Color := ASerie.XValues.MinorUnitFontColor;
    ACanvas.Stroke.Color := ASerie.XValues.MinorUnitTickMarkColor;
  end;

  for I := 0 to ASerie.FDrawXValues.Count - 1 do
  begin
    dv := ASerie.FDrawXValues[I];
    if dv.Kind = AKind then
    begin
      dr := True;
      st := ACanvas.SaveState;
      c.DoBeforeDrawSerieXValue(ACanvas, ASerie, APosition, dv, dr);
      if dr then
      begin
        case APosition of
          xpTop:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionXTop.X) + 0.5, int(dv.ValuePositionXTop.Y - tc) + 0.5), PointF(int(dv.ValuePositionXTop.X + 0.5), int(dv.ValuePositionXTop.Y + 0.5)), 1);
            rt := RectF(int(dv.TextPositionXTop.X - dv.TextWidth / 2) + 0.5, int(dv.TextPositionXTop.Y - dv.TextHeight - tc - spc) + 0.5,
              int(dv.TextPositionXTop.X + dv.TextWidth / 2) + 0.5, int(dv.TextPositionXTop.Y - tc - spc) + 0.5);
          end;
          xpCenter:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionXCenter.X) + 0.5, int(dv.ValuePositionXCenter.Y - tc / 2) + 0.5), PointF(int(dv.ValuePositionXCenter.X) + 0.5,
              int(dv.ValuePositionXCenter.Y + tc / 2) + 0.5), 1);
            rt := RectF(int(dv.TextPositionXCenter.X - dv.TextWidth / 2) + 0.5, int(dv.TextPositionXCenter.Y + tc / 2 + spc) + 0.5,
              int(dv.TextPositionXCenter.X + dv.TextWidth / 2) + 0.5, int(dv.TextPositionXCenter.Y  + tc / 2 + spc + dv.TextHeight) + 0.5);
          end;
          xpBottom:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionXBottom.X) + 0.5, int(dv.ValuePositionXBottom.Y) + 0.5), PointF(int(dv.ValuePositionXBottom.X) + 0.5, int(dv.ValuePositionXBottom.Y + tc) + 0.5),  1);
            rt := RectF(int(dv.TextPositionXBottom.X - dv.TextWidth / 2) + 0.5, int(dv.TextPositionXBottom.Y + tc + spc) + 0.5, int(dv.TextPositionXBottom.X + dv.TextWidth / 2) + 0.5,
              int(dv.TextPositionXBottom.Y + tc + spc + dv.TextHeight) + 0.5);
          end;
        end;

        if ASerie.XValues.Angle <> 0 then
        begin
          cx.X := rt.Left + rt.Width / 2;
          cx.Y := rt.Top + rt.Height / 2;
          m := ACanvas.Matrix;
          rm := TMatrix.CreateRotation(DegToRad(ASerie.XValues.Angle)) * TMatrix.CreateTranslation(cx.X, cx.Y);
          ACanvas.MultiplyMatrix(rm);
          ACanvas.FillText(RectF(-rt.Width / 2, -rt.Height / 2, rt.Width / 2, rt.Height / 2), dv.ValueString, False, 1, [], TTextAlign.Leading);
          ACanvas.SetMatrix(m);
        end
        else
          ACanvas.FillText(rt, dv.ValueString, False, 1, [], TTextAlign.Leading);

        c.DoAfterDrawSerieXValue(ACanvas, ASerie, APosition, dv);
      end;
      ACanvas.RestoreState(st);
    end;
  end;
end;

procedure TTMSFMXChartXAxis.DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartXAxisPosition);
var
  dr: Boolean;
  c: TTMSFMXChart;
  r: TRectF;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := GetRectangle(APosition);

  case APosition of
    xpTop: r := RectF(int(r.Left) + 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
    xpCenter: r := RectF(int(r.Left) + 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) - 0.5);
    xpBottom: r := RectF(int(r.Left) + 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) - 0.5);
  end;

  dr := True;
  st := ACanvas.SaveState;
  c.DoBeforeDrawXAxis(ACanvas, r, APosition, dr);

  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, FillOpacity);
    if Border then
      ACanvas.DrawRect(r, 0, 0, AllCorners, StrokeOpacity);

    if Line then
    begin
      case APosition of
        xpTop: ACanvas.DrawLine(PointF(R.Left, R.Bottom), PointF(R.Right, R.Bottom), StrokeOpacity);
        xpCenter: ACanvas.DrawLine(PointF(R.Left, R.CenterPoint.Y), PointF(R.Right, R.CenterPoint.Y), StrokeOpacity);
        xpBottom:
        begin
          ACanvas.DrawLine(PointF(R.Left, R.Top), PointF(R.Right, R.Top), StrokeOpacity);
          if (c.FTotalOffset3DX > 0) or (c.FTotalOffset3DY > 0) then
          begin
            ACanvas.DrawLine(PointF(R.Left, R.Top), PointF(R.Left + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), StrokeOpacity);
            ACanvas.DrawLine(PointF(R.Right, R.Top), PointF(R.Right + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), StrokeOpacity);
            ACanvas.DrawLine(PointF(R.Left + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), PointF(R.Right + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), StrokeOpacity);
          end;
        end;
      end;
    end;

    c.DoAfterDrawXAxis(ACanvas, r, APosition);
  end;
  ACanvas.RestoreState(st);
end;

procedure TTMSFMXChartXAxis.DrawTitle(ACanvas: TCanvas;
  APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie);
var
  r, rt: TRectF;
  ts: Double;
  txt: String;
  txth: Double;
  tm: TBounds;
  c: TTMSFMXChart;
  dr: Boolean;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  txt := ASerie.XValues.Title.Text;
  if txt <> '' then
  begin
    ACanvas.Font.Assign(ASerie.XValues.Title.Font);
    ACanvas.Fill.Color := ASerie.XValues.Title.FontColor;
    tm := ASerie.XValues.Title.TextMargins;
    r := GetRectangle(APosition);
    ts := ACanvas.TextHeight(txt);
    txth := ts + 4;
    case APosition of
      xpTop: rt := RectF(r.Left + tm.Left, r.Bottom - ASerie.FXAxisTopHeight + tm.Top, r.Right - tm.Right, r.Bottom - ASerie.FXAxisTopHeight + txth + tm.Top + tm.Bottom);
      xpCenter: rt := RectF(r.Left + tm.Left, r.Top + r.Height / 2 + ASerie.FXAxisCenterHeight - tm.Top - tm.Bottom - txth, r.Right - tm.Right, R.Top + r.Height / 2 + ASerie.FXAxisCenterHeight - tm.Top);
      xpBottom: rt := RectF(r.Left + tm.Left, r.Top + ASerie.FXAxisBottomHeight - tm.Top - tm.Bottom - txth, r.Right - tm.Right, r.Top + ASerie.FXAxisBottomHeight - tm.Top);
    end;
    dr := True;
    st := ACanvas.SaveState;
    c.DoBeforeDrawXValuesTitle(ACanvas, r, rt, APosition, ASerie, dr);
    if dr then
    begin
      ACanvas.FillText(rt, txt, False, 1, [], ASerie.XValues.Title.TextHorizontalAlignment, ASerie.XValues.Title.TextVerticalAlignment);
      c.DoAfterDrawXValuesTitle(ACanvas, r, rt, APosition, ASerie);
    end;
    ACanvas.RestoreState(st);
  end;
end;

procedure TTMSFMXChartXAxis.DrawValues(ACanvas: TCanvas;
  APosition: TTMSFMXChartXAxisPosition; ASerie: TTMSFMXChartSerie);
begin
  if not Assigned(ASerie) then
    Exit;

  DrawKind(ACanvas, APosition, vkMinor, ASerie);
  DrawKind(ACanvas, APosition, vkMajor, ASerie);
  DrawTitle(ACanvas, APosition, ASerie);
end;

function TTMSFMXChartXAxis.GetHeight(APosition: TTMSFMXChartXAxisPosition): Double;
var
  c: TTMSFMXChart;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.HasPieChart then
    Exit;

  if not Visible or not (APosition in Positions) then
    Exit;

  Result := FHeight;
  if AutoSize then
  begin
    case APosition of
      xpTop: Result := FTopHeight;
      xpCenter: Result := FCenterHeight;
      xpBottom: Result := FBottomHeight;
    end;
  end;
end;

function TTMSFMXChartXAxis.GetRectangle(
  APosition: TTMSFMXChartXAxisPosition): TRectF;
var
  c: TTMSFMXChart;
  r, rc: TRectF;
  pos: Double;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := c.GetChartRectangle;
  case APosition of
    xpTop:
    begin
      Result := RectF(R.Left, R.Top, R.Right, R.Top + GetHeight(APosition));
      if (tipTop in c.Title.Positions) and c.Title.Visible then
      begin
        Result.Top := Result.Top + c.Title.Height;
        Result.Bottom := Result.Bottom + c.Title.Height;
      end;
    end;
    xpCenter:
    begin
      rc := r;
      if Visible then
      begin
        if xpTop in Positions then
          rc.Top := rc.Top + GetHeight(xpTop);

        if xpBottom in Positions then
          rc.Bottom := rc.Bottom - GetHeight(xpBottom);
      end;

      if c.Title.Visible then
      begin
        if tipTop in c.Title.Positions then
          rc.Top := rc.Top + c.Title.Height;
        if tipBottom in c.Title.Positions then
          rc.Bottom := rc.Bottom - c.Title.Height;
      end;

      if DisplayAtReferenceValue and (ReferenceValueSeriesIndex >= 0) and (ReferenceValueSeriesIndex <= c.Series.Count - 1) then
        pos := c.Series[ReferenceValueSeriesIndex].ValueToY(ReferenceValue) - GetHeight(APosition) / 2
      else
        pos := rc.Top + ((rc.Bottom - rc.Top) - GetHeight(APosition)) / 2;

      Result := RectF(rc.Left, pos, rc.Right, pos + GetHeight(APosition));
    end;
    xpBottom:
    begin
      Result := RectF(R.Left, R.Bottom - GetHeight(APosition), R.Right, R.Bottom);
      if (tipBottom in c.Title.Positions) and (c.Title.Visible) then
      begin
        Result.Top := Result.Top - c.Title.Height;
        Result.Bottom := Result.Bottom - c.Title.Height;
      end;
    end;
  end;

  if (ypLeft in c.YAxis.Positions) and c.YAxis.Visible then
    Result.Left := Result.Left + c.YAxis.GetWidth(ypLeft);

  if (ypRight in c.YAxis.Positions) and c.YAxis.Visible then
    Result.Right := Result.Right - c.YAxis.GetWidth(ypRight);

  if APosition = xpBottom then
    Result.Right := Result.Right - c.FTotalOffset3DX;

  if APosition = xpTop then
    Result.Left := Result.Left + c.FTotalOffset3DX;
end;

function TTMSFMXChartXAxis.IsReferenceValueStored: Boolean;
begin
  Result := ReferenceValue <> 0;
end;

procedure TTMSFMXChartXAxis.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetBorder(const Value: Boolean);
begin
  if FBorder <> Value then
  begin
    FBorder := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetDisplayAtReferenceValue(const Value: Boolean);
begin
  if FDisplayAtReferenceValue <> Value then
  begin
    FDisplayAtReferenceValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetHeight(const Value: Double);
begin
  if FHeight <> Value then
  begin
    FHeight := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetLine(const Value: Boolean);
begin
  if FLine <> Value then
  begin
    FLine := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetPositions(const Value: TTMSFMXChartXAxisPositions);
begin
  if FPositions <> Value then
  begin
    FPositions := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetReferenceValue(const Value: Double);
begin
  if FReferenceValue <> Value then
  begin
    FReferenceValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartXAxis.SetReferenceValueSeriesIndex(const Value: Integer);
begin
  if FReferenceValueSeriesIndex <> Value then
  begin
    FReferenceValueSeriesIndex := Value;
    UpdateChart;
  end;
end;

{ TTMSFMXChartYAxis }

procedure TTMSFMXChartYAxis.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartYAxis then
  begin
    FPositions := (Source as TTMSFMXChartYAxis).Positions;
    FWidth := (Source as TTMSFMXChartYAxis).Width;
    FLine := (Source as TTMSFMXChartYAxis).Line;
    FBorder := (Source as TTMSFMXChartYAxis).Border;
    FAutoSize := (Source as TTMSFMXChartYAxis).AutoSize;
    FDisplayAtReferenceValue := (Source as TTMSFMXChartYAxis).DisplayAtReferenceValue;
    FReferenceValue := (Source as TTMSFMXChartYAxis).ReferenceValue;
    FReferenceValueSeriesIndex := (Source as TTMSFMXChartYAxis).ReferenceValueSeriesIndex;
  end;
end;

constructor TTMSFMXChartYAxis.Create(AChart: TTMSFMXChart);
begin
  inherited;
  FChart := AChart;
  FAutoSize := True;
  FLine := True;
  FBorder := False;
  FWidth := 35;
  FPositions := [ypLeft, ypRight];
  FStroke.Kind := TBrushKind.Solid;
  FDisplayAtReferenceValue := False;
  FReferenceValueSeriesIndex := 0;
  FReferenceValue := 0;
end;

destructor TTMSFMXChartYAxis.Destroy;
begin
  inherited;
end;

procedure TTMSFMXChartYAxis.Draw(ACanvas: TCanvas);
var
  pos: TTMSFMXChartYAxisPosition;
  I: Integer;
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
begin
  inherited;
  if not Visible then
    Exit;

  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.HasPieChart then
    Exit;

  for pos in Positions do
    DrawPart(ACanvas, pos);

  for pos in Positions do
  begin
    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.Visible and (pos in s.YValues.Positions) then
        DrawValues(ACanvas, pos, s);
    end;                                           
  end;    
end;

procedure TTMSFMXChartYAxis.DrawKind(ACanvas: TCanvas;
  APosition: TTMSFMXChartYAxisPosition; AKind: TTMSFMXChartDrawXYValueKind; ASerie: TTMSFMXChartSerie);
var
  tc, spc: Double;
  I: Integer;
  dv: TTMSFMXChartDrawXYValue;
  dr: Boolean;
  c: TTMSFMXChart;
  rt: TRectF;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  if AKind = vkMajor then
  begin
    tc := ASerie.YValues.MajorUnitTickMarkSize;
    spc := ASerie.YValues.MajorUnitSpacing;
    ACanvas.Font.Assign(ASerie.YValues.MajorUnitFont);
    ACanvas.Fill.Color := ASerie.YValues.MajorUnitFontColor;
    ACanvas.Stroke.Color := ASerie.YValues.MajorUnitTickMarkColor;
  end
  else
  begin
    tc := ASerie.YValues.MinorUnitTickMarkSize;
    spc := ASerie.YValues.MinorUnitSpacing;
    ACanvas.Font.Assign(ASerie.YValues.MinorUnitFont);
    ACanvas.Fill.Color := ASerie.YValues.MinorUnitFontColor;
    ACanvas.Stroke.Color := ASerie.YValues.MinorUnitTickMarkColor;
  end;

  for I := 0 to ASerie.FDrawYValues.Count - 1 do
  begin
    dv := ASerie.FDrawYValues[I];
    if dv.Kind = AKind then
    begin
      dr := True;
      st := ACanvas.SaveState;
      c.DoBeforeDrawSerieYValue(ACanvas, ASerie, APosition, dv, dr);
      if dr then
      begin
        case APosition of
          ypLeft:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionYLeft.X) + 0.5, int(dv.ValuePositionYLeft.Y) + 0.5), PointF(int(dv.ValuePositionYLeft.X - tc) + 0.5, int(dv.ValuePositionYLeft.Y) + 0.5), 1);
            rt := RectF(int(dv.TextPositionYLeft.X - dv.TextWidth - tc - spc) + 0.5, int(dv.TextPositionYLeft.Y - dv.TextHeight / 2) + 0.5,
              int(dv.TextPositionYLeft.X - tc - spc) + 0.5, int(dv.TextPositionYLeft.Y + dv.TextHeight / 2) + 0.5);
          end;
          ypCenter:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionYCenter.X - tc / 2) + 0.5, int(dv.ValuePositionYCenter.Y) + 0.5), PointF(int(dv.ValuePositionYCenter.X + tc / 2) + 0.5,
              int(dv.ValuePositionYCenter.Y) + 0.5), 1);
            rt := RectF(int(dv.TextPositionYCenter.X + tc / 2 + spc) + 0.5, int(dv.TextPositionYCenter.Y - dv.TextHeight / 2) + 0.5,
              int(dv.TextPositionYCenter.X + tc / 2 + spc + dv.TextWidth) + 0.5, int(dv.TextPositionYCenter.Y + dv.TextHeight / 2) + 0.5);
          end;
          ypRight:
          begin
            ACanvas.DrawLine(PointF(int(dv.ValuePositionYRight.X) + 0.5, int(dv.ValuePositionYRight.Y) + 0.5), PointF(int(dv.ValuePositionYRight.X + tc) + 0.5, int(dv.ValuePositionYRight.Y) + 0.5), 1);
            rt := RectF(int(tc + spc + dv.TextPositionYRight.X) + 0.5, int(dv.TextPositionYRight.Y - dv.TextHeight / 2) + 0.5, int(tc + spc + dv.TextWidth + dv.TextPositionYRight.X) + 0.5,
              int(dv.TextPositionYRight.Y + dv.TextHeight / 2) + 0.5);
          end;
        end;

        ACanvas.FillText(rt, dv.ValueString, False, 1, [], TTextAlign.Leading);
        c.DoAfterDrawSerieYValue(ACanvas, ASerie, APosition, dv);
      end;
      ACanvas.RestoreState(st);
    end;
  end;
end;

procedure TTMSFMXChartYAxis.DrawPart(ACanvas: TCanvas; APosition: TTMSFMXChartYAxisPosition);
var
  dr: Boolean;
  c: TTMSFMXChart;
  r: TRectF;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := GetRectangle(APosition);
  case APosition of
    ypLeft: r := RectF(int(r.Left) - 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
    ypCenter: r := RectF(int(r.Left) - 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
    ypRight: r := RectF(int(r.Left) + 0.5, int(r.Top) + 0.5, int(r.Right) + 0.5, int(r.Bottom) + 0.5);
  end;

  dr := True;
  st := ACanvas.SaveState;
  c.DoBeforeDrawYAxis(ACanvas, r, APosition, dr);

  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, FillOpacity);
    if Border then
      ACanvas.DrawRect(r, 0, 0, AllCorners, StrokeOpacity);

    if Line then
    begin
      case APosition of
        ypLeft:
        begin
          ACanvas.DrawLine(PointF(R.Right, R.Top), PointF(R.Right, R.Bottom), StrokeOpacity);
          if (c.FTotalOffset3DX > 0) or (c.FTotalOffset3DY > 0) then
          begin
            ACanvas.DrawLine(PointF(R.Right, R.Top), PointF(R.Right + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), StrokeOpacity);
            ACanvas.DrawLine(PointF(R.Right + c.FTotalOffset3DX, R.Top - c.FTotalOffset3DY), PointF(R.Right + c.FTotalOffset3DX, R.Bottom - c.FTotalOffset3DY), StrokeOpacity);
          end;
        end;
        ypCenter: ACanvas.DrawLine(PointF(R.CenterPoint.X, R.Top), PointF(R.CenterPoint.X, R.Bottom), StrokeOpacity);
        ypRight: ACanvas.DrawLine(PointF(R.Left, R.Top), PointF(R.Left, R.Bottom), StrokeOpacity);
      end;
    end;

    c.DoAfterDrawYAxis(ACanvas, r, APosition);
  end;
  ACanvas.RestoreState(st);
end;

procedure TTMSFMXChartYAxis.DrawTitle(ACanvas: TCanvas;
  APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie);
var
  r, rt, txtr: TRectF;
  th, tw: Double;
  txt: String;
  txth: Double;
  tm: TBounds;
  cx: TPointF;
  m, rm: TMatrix;
  dr: Boolean;
  c: TTMSFMXChart;
  st: TCanvasSaveState;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  txt := ASerie.YValues.Title.Text;
  if txt <> '' then
  begin
    ACanvas.Font.Assign(ASerie.YValues.Title.Font);
    ACanvas.Fill.Color := ASerie.YValues.Title.FontColor;
    tm := ASerie.YValues.Title.TextMargins;
    r := GetRectangle(APosition);
    th := ACanvas.TextHeight(txt);
    tw := ACanvas.TextWidth(txt);
    txth := th + 4;
    case APosition of
      ypLeft: rt := RectF(r.Right - ASerie.FYAxisLeftWidth + tm.Left, r.Top + tm.Top, r.Right - ASerie.FYAxisLeftWidth + txth + tm.Left + tm.Right, r.Bottom - tm.Bottom);
      ypCenter: rt := RectF(r.Left + r.Width / 2 + ASerie.FYAxisCenterWidth - tm.Left - txth - tm.Right, r.Top + tm.Top, R.Left + r.Width / 2 + ASerie.FYAxisCenterWidth - tm.Left, r.Bottom - tm.Bottom);
      ypRight: rt := RectF(r.Left + ASerie.FYAxisRightWidth - tm.Left - txth - tm.Right, r.Top + tm.Top, R.Left + ASerie.FYAxisRightWidth - tm.Left, r.Bottom - tm.Bottom);
    end;

    dr := True;
    st := ACanvas.SaveState;
    c.DoBeforeDrawYValuesTitle(ACanvas, r, rt, APosition, ASerie, dr);
    if dr then
    begin
      txtr := rt;

      case ASerie.YValues.Title.TextHorizontalAlignment of
        TTextAlign.Center: txtr.Left := txtr.Left + (txtr.Width - th) / 2;
        TTextAlign.Trailing: txtr.Left := txtr.Right - th;
      end;

      case ASerie.YValues.Title.TextVerticalAlignment of
        TTextAlign.Center: txtr.Top := txtr.Top + (txtr.Height - tw) / 2;
        TTextAlign.Trailing: txtr.Top := txtr.Bottom - tw;
      end;

      txtr.Width := th;
      txtr.Height := tw;
      cx.X := txtr.Left + txtr.Width / 2;
      cx.Y := txtr.Top + txtr.Height / 2;
      m := ACanvas.Matrix;
      case APosition of
        ypLeft: rm := TMatrix.CreateRotation(DegToRad(-90)) * TMatrix.CreateTranslation(cx.X, cx.Y);
        ypCenter, ypRight: rm := TMatrix.CreateRotation(DegToRad(90)) * TMatrix.CreateTranslation(cx.X, cx.Y);
      end;
      ACanvas.MultiplyMatrix(rm);

      ACanvas.FillText(RectF(-txtr.Height / 2, -txtr.Width / 2, txtr.Height / 2, txtr.Width / 2), txt, False, 1, [], TTextAlign.Leading);
      ACanvas.SetMatrix(m);
      c.DoAfterDrawYValuesTitle(ACanvas, r, rt, APosition, ASerie);
    end;
    ACanvas.RestoreState(st);
  end;
end;

procedure TTMSFMXChartYAxis.DrawValues(ACanvas: TCanvas;
  APosition: TTMSFMXChartYAxisPosition; ASerie: TTMSFMXChartSerie);
begin
  if not Assigned(ASerie) then
    Exit;

  DrawKind(ACanvas, APosition, vkMinor, ASerie);
  DrawKind(ACanvas, APosition, vkMajor, ASerie);
  DrawTitle(ACanvas, APosition, ASerie);
end;

function TTMSFMXChartYAxis.GetRectangle(
  APosition: TTMSFMXChartYAxisPosition): TRectF;
var
  c: TTMSFMXChart;
  r, rc: TRectF;
  pos: Double;
begin
  c := Chart;
  if not Assigned(c) then
    Exit;

  r := c.GetChartRectangle;
  case APosition of
    ypLeft: Result := RectF(R.Left, R.Top, R.Left + GetWidth(APosition), R.Bottom);
    ypCenter:
    begin
      rc := r;
      if (ypLeft in Positions) and Visible then
        rc.Left := rc.Left + GetWidth(ypLeft);

      if (ypRight in Positions) and Visible then
        rc.Right := rc.Right - GetWidth(ypRight);

      if DisplayAtReferenceValue and (ReferenceValueSeriesIndex >= 0) and (ReferenceValueSeriesIndex <= c.Series.Count - 1) then
        pos := c.Series[ReferenceValueSeriesIndex].ValueToX(ReferenceValue) - GetWidth(APosition) / 2
      else
        pos := rc.Left + ((rc.Right - rc.Left) - GetWidth(APosition)) / 2;

      Result := RectF(pos, rc.Top, pos + GetWidth(APosition), rc.Bottom);
    end;
    ypright: Result := RectF(R.Right - GetWidth(APosition), R.Top, R.Right, R.Bottom);
  end;

  if (tipTop in c.Title.Positions) and c.Title.Visible then
    Result.Top := Result.Top + c.Title.Height;

  if (xpTop in c.XAxis.Positions) and c.XAxis.Visible then
    Result.Top := Result.Top + c.XAxis.GetHeight(xpTop);

  if (tipBottom in c.Title.Positions) and c.Title.Visible then
    Result.Bottom := Result.Bottom - c.Title.Height;

  if (xpBottom in c.XAxis.Positions) and c.XAxis.Visible then
    Result.Bottom := Result.Bottom - c.XAxis.GetHeight(xpBottom);

  if APosition = ypLeft then
    Result.Top := Result.Top + c.FTotalOffset3DY;

  if APosition = ypRight then
    Result.Bottom := Result.Bottom - c.FTotalOffset3DY;
end;

function TTMSFMXChartYAxis.GetWidth(APosition: TTMSFMXChartYAxisPosition): Double;
var
  c: TTMSFMXChart;
begin
  Result := 0;
  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.HasPieChart then
    Exit;

  if not Visible or not (APosition in Positions) then
    Exit;

  Result := FWidth;
  if AutoSize then
  begin
    case APosition of
      ypLeft: Result := FLeftWidth;
      ypCenter: Result := FCenterWidth;
      ypRight: Result := FRightWidth;
    end;
  end;
end;

function TTMSFMXChartYAxis.IsReferenceValueStored: Boolean;
begin
  Result := ReferenceValue <> 0;
end;

procedure TTMSFMXChartYAxis.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetBorder(const Value: Boolean);
begin
  if FBorder <> Value then
  begin
    FBorder := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetDisplayAtReferenceValue(const Value: Boolean);
begin
  if FDisplayAtReferenceValue <> Value then
  begin
    FDisplayAtReferenceValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetLine(const Value: Boolean);
begin
  if FLine <> Value then
  begin
    FLine := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetWidth(const Value: Double);
begin
  if FWidth <> Value then
  begin
    FWidth := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetReferenceValue(const Value: Double);
begin
  if FReferenceValue <> Value then
  begin
    FReferenceValue := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetReferenceValueSeriesIndex(const Value: Integer);
begin
  if FReferenceValueSeriesIndex <> Value then
  begin
    FReferenceValueSeriesIndex := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartYAxis.SetPositions(const Value: TTMSFMXChartYAxisPositions);
begin
  if FPositions <> Value then
  begin
    FPositions := Value;
    UpdateChart;
  end;
end;

{ TTMSFMXChartSerieMultiPoint }

procedure TTMSFMXChartSerieMultiPoint.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieMultiPoint then
  begin
    FWidth := (Source as TTMSFMXChartSerieMultiPoint).Width;
    FWidthType := (Source as TTMSFMXChartSerieMultiPoint).WidthType;
    FIncreaseStrokeColor := (Source as TTMSFMXChartSerieMultiPoint).IncreaseStrokeColor;
    FDecreaseStrokeColor := (Source as TTMSFMXChartSerieMultiPoint).DecreaseStrokeColor;
    FIncreaseFillColor := (Source as TTMSFMXChartSerieMultiPoint).IncreaseFillColor;
    FDecreaseFillColor := (Source as TTMSFMXChartSerieMultiPoint).DecreaseFillColor;
  end;
end;

function TTMSFMXChartSerieMultiPoint.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieMultiPoint.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FIncreaseStrokeColor := claDarkgreen;
  FDecreaseStrokeColor := claDarkred;
  FIncreaseFillColor := claLightseagreen;
  FDecreaseFillColor := claOrange;
  FWidth := 65;
  FWidthType := bwtPercentage;
end;

destructor TTMSFMXChartSerieMultiPoint.Destroy;
begin

  inherited;
end;

procedure TTMSFMXChartSerieMultiPoint.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieMultiPoint.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieMultiPoint.SetDecreaseFillColor(
  const Value: TAlphaColor);
begin
  if FDecreaseFillColor <> Value then
  begin
    FDecreaseFillColor := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.SetDecreaseStrokeColor(
  const Value: TAlphaColor);
begin
  if FDecreaseStrokeColor <> Value then
  begin
    FDecreaseStrokeColor := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.SetIncreaseFillColor(
  const Value: TAlphaColor);
begin
  if FIncreaseFillColor <> Value then
  begin
    FIncreaseFillColor := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.SetIncreaseStrokeColor(
  const Value: TAlphaColor);
begin
  if FIncreaseStrokeColor <> Value then
  begin
    FIncreaseStrokeColor := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.SetWidth(const Value: Double);
begin
  if FWidth <> Value then
  begin
    FWidth := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.SetWidthType(
  const Value: TTMSFMXChartSerieBarWidthType);
begin
  if FWidthType <> Value then
  begin
    FWidthType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieMultiPoint.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieBar }

procedure TTMSFMXChartSerieBar.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieBar then
  begin
    FWidth := (Source as TTMSFMXChartSerieBar).Width;
    FWidthType := (Source as TTMSFMXChartSerieBar).WidthType;
    FSpacing := (Source as TTMSFMXChartSerieBar).Spacing;
  end;
end;

function TTMSFMXChartSerieBar.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieBar.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FWidth := 65;
  FWidthType := bwtPercentage;
  FSpacing := 20;
end;

destructor TTMSFMXChartSerieBar.Destroy;
begin

  inherited;
end;

procedure TTMSFMXChartSerieBar.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieBar.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieBar.SetSpacing(const Value: Double);
begin
  if FSpacing <> Value then
  begin
    FSpacing := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieBar.SetWidth(const Value: Double);
begin
  if FWidth <> Value then
  begin
    FWidth := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieBar.SetWidthType(
  const Value: TTMSFMXChartSerieBarWidthType);
begin
  if FWidthType <> Value then
  begin
    FWidthType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieBar.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;


{ TTMSFMXChartSeriePie }

procedure TTMSFMXChartSeriePie.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSeriePie then
  begin
    FPosition := (Source as TTMSFMXChartSeriePie).Position;
    FSize := (Source as TTMSFMXChartSeriePie).Size;
    FAutoSize := (Source as TTMSFMXChartSeriePie).AutoSize;
    FMargins.Assign((Source as TTMSFMXChartSeriePie).Margins);
    FStartAngle := (Source as TTMSFMXChartSeriePie).StartAngle;
    FSweepAngle := (Source as TTMSFMXChartSeriePie).SweepAngle;
    FInnerSize := (Source as TTMSFMXChartSeriePie).InnerSize;
    FStacked := (Source as TTMSFMXChartSeriePie).Stacked;
  end;
end;

function TTMSFMXChartSeriePie.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSeriePie.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FPosition := ppCenterCenter;
  FSize := 200;
  FAutoSize := True;
  FStacked := False;
  FStartAngle := 0;
  FSweepAngle := 360;
  FMargins := TBounds.Create(RectF(10, 10, 10, 10));
  FMargins.OnChange := MarginsChanged;
end;

destructor TTMSFMXChartSeriePie.Destroy;
begin
  FMargins.Free;
  inherited;
end;

procedure TTMSFMXChartSeriePie.MarginsChanged(Sender: TObject);
begin
  UpdateChart;
end;

procedure TTMSFMXChartSeriePie.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSeriePie.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSeriePie.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetInnerSize(const Value: Double);
begin
  if FInnerSize <> Value then
  begin
    FInnerSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetMargins(const Value: TBounds);
begin
  if FMargins <> Value then
  begin
    FMargins.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetPosition(
  const Value: TTMSFMXChartSeriePiePosition);
begin
  if FPosition <> Value then
  begin
    FPosition := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetSize(const Value: Double);
begin
  if FSize <> Value then
  begin
    FSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetStacked(const Value: Boolean);
begin
  if FStacked <> value then
  begin
    FStacked := Value;
    Updatechart;
  end;
end;

procedure TTMSFMXChartSeriePie.SetStartAngle(const Value: Double);
begin
  if FStartAngle <> Value then
  begin
    FStartAngle := Value;
    UpdateChart; 
  end;
end;

procedure TTMSFMXChartSeriePie.SetSweepAngle(const Value: Double);
begin
  if FSweepAngle <> Value then
  begin
    FSweepAngle := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSeriePie.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartLegend }

procedure TTMSFMXChartLegend.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartLegend then
  begin
    FTop := (Source as TTMSFMXChartLegend).Top;
    FLeft := (Source as TTMSFMXChartLegend).Left;
    FPosition := (Source as TTMSFMXChartLegend).Position;
    FFont.Assign((Source as TTMSFMXChartLegend).Font);
    FFontColor := (Source as TTMSFMXChartLegend).FontColor;
  end;
end;

constructor TTMSFMXChartLegend.Create(AChart: TTMSFMXChart);
begin
  inherited;
  FTop := 10;
  FLeft := 10;
  FPosition := lpTopLeft;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFontColor := claBlack;
  FStroke.Kind := TBrushKind.Solid;
  FFill.Kind := TBrushKind.Solid;
end;

destructor TTMSFMXChartLegend.Destroy;
begin
  FFont.Free;
  inherited;
end;

procedure TTMSFMXChartLegend.Draw(ACanvas: TCanvas);
var
  c: TTMSFMXChart;
  r, rm, rt: TRectF;
  I: Integer;
  s: TTMSFMXChartSerie;
  th, x, y: Double;
  dr: Boolean;
  st: TCanvasSaveState;
  li: Boolean;
begin
  inherited;
  if not Visible then
    Exit;

  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.Series.Count = 0 then
    Exit;

  r := GetRectangle;
  r := ModifyRect(r);
  if r = TRectF.Empty then
    Exit;

  dr := True;
  st := ACanvas.SaveState;
  c.DoBeforeDrawLegend(ACanvas, r, dr);
  if dr then
  begin
    ACanvas.FillRect(r, 0, 0, AllCorners, FillOpacity);
    ACanvas.DrawRect(r, 0, 0, AllCorners, StrokeOpacity);

    x := r.Left + 5;
    y := r.Top + 5;

    for I := 0 to c.Series.Count - 1 do
    begin
      s := c.Series[I];
      if s.Visible and s.ShowInLegend then
      begin
        ACanvas.Font.Assign(Font);
        th := ACanvas.TextHeight(s.LegendText);
        rt := RectF(x + th + 5, y, x + r.Width - 10, y + th);
        rm := RectF(x, y, x + th, y + th);
        y := y + th + 5;

        ACanvas.Fill.Color := FontColor;

        ACanvas.FillText(rt, s.LegendText, False, 1, [], TTextAlign.Leading, TTextAlign.Leading);

        ACanvas.Fill.Assign(s.Fill);
        ACanvas.Stroke.Assign(s.Stroke);
        {$IFDEF DELPHIXE11_LVL}
        ACanvas.Stroke.Thickness := s.StrokeThickness;
        ACanvas.Stroke.Cap := s.StrokeCap;
        ACanvas.Stroke.Dash := s.StrokeDash;
        ACanvas.Stroke.Join := s.StrokeJoin;
        {$ELSE}
        ACanvas.StrokeThickness := s.StrokeThickness;
        ACanvas.StrokeCap := s.StrokeCap;
        ACanvas.StrokeDash := s.StrokeDash;
        ACanvas.StrokeJoin := s.StrokeJoin;
        {$ENDIF}

        InflateRect(rm, -rm.Height / 7.5, -rm.Height / 7.5);
        rm := ModifyRect(rm);
        li := True;
        c.DoBeforeDrawLegendIcon(ACanvas, s, rm, li);
        if li then        
        begin
          DrawChartIcon(ACanvas, s.ChartType, rm);
          c.DoAfterDrawLegendIcon(ACanvas, s, rm);
        end;        
      end;
    end;
    c.DoAfterDrawLegend(ACanvas, r);
  end;
  ACanvas.RestoreState(st);
end;

procedure TTMSFMXChartLegend.FontChanged(Sender: TObject);
begin
  RepaintChart;
end;

function TTMSFMXChartLegend.GetRectangle: TRectF;
var
  c: TTMSFMXChart;
  I: Integer;
  s: TTMSFMXChartSerie;
  bmp: TBitmap;
  tw, th: Double;
  mxw, mxh, mxhr: Double;
  x, y: Double;
  r: TRectF;
begin
  Result := TRectF.Empty;
  c := Chart;
  if not Assigned(c) then
    Exit;
  if c.Series.Count = 0 then
    Exit;

  mxw := 0;
  mxh := 0;
  mxhr := 0;
  bmp := TBitmap.Create(1, 1);
  bmp.Canvas.Font.Assign(Font);
  for I := 0 to c.Series.Count - 1 do
  begin
    s := c.Series[I];
    if s.Visible and s.ShowInLegend then
    begin
      if s.LegendText <> '' then
      begin
        tw := bmp.Canvas.TextWidth(s.LegendText) + 5;
        if tw > mxw then
          mxw := tw;
        th := bmp.Canvas.TextHeight(s.LegendText) + 5;
        if th > mxh then
          mxhr := th;

        mxh := mxh + th;
      end;
    end;
  end;
  bmp.Free;

  if (mxw = 0) and (mxh = 0) then
  begin
    Result := TRectF.Empty;
    Exit;
  end;


  mxw := mxw + mxhr + 5;
  mxh := mxh + 5;

  r := c.GetSeriesRectangle;
  r.Left := r.Left - c.SeriesMargins.Left + c.FTotalOffset3DX;
  r.Top := r.Top - c.SeriesMargins.Top - c.FTotalOffset3DY;
  r.Bottom := r.Bottom + c.SeriesMargins.Bottom - c.FTotalOffset3DY / 2;
  r.Right := r.Right + c.SeriesMargins.Right + c.FTotalOffset3DX / 2;

  x := r.Left;
  y := r.Top;

  case Position of
    lpTopCenter: x := x + (r.Width - mxw) / 2;
    lpTopRight: x := x + r.Width - mxw;
    lpCenterLeft: y := y + (r.Height - mxh) / 2;
    lpCenterCenter:
    begin
      y := y + (r.Height - mxh) / 2;
      x := x + (r.Width - mxw) / 2;
    end;
    lpCenterRight:
    begin
      x := x + r.Width - mxw;
      y := y + (r.Height - mxh) / 2;
    end;
    lpBottomLeft:
    begin
      y := y + r.Height - mxh;
    end;
    lpBottomCenter:
    begin
      x := x + (r.Width - mxw) / 2;
      y := y + r.Height - mxh;
    end;
    lpBottomRight:
    begin
      x := x + r.Width - mxw;
      y := y + r.Height - mxh;
    end;
  end;

  x := x + Left;
  y := y + Top;

  Result := RectF(x, y, x + mxw, y + mxh);
end;

procedure TTMSFMXChartLegend.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartLegend.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartLegend.SetLeft(const Value: Double);
begin
  if FLeft <> Value then
  begin
    FLeft := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartLegend.SetPosition(
  const Value: TTMSFMXChartLegendPosition);
begin
  if FPosition <> Value then
  begin
    FPosition := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartLegend.SetTop(const Value: Double);
begin
  if FTop <> Value then
  begin
    FTop := Value;
    RepaintChart;
  end;
end;

function TTMSFMXChartLegend.XYToItem(X, Y: Double): Integer;
var
  c: TTMSFMXChart;
  r, rt: TRectF;
  I: Integer;
  s: TTMSFMXChartSerie;
  th, xx, xy: Double;
  dr: Boolean;
  g: TBitmap;
begin
  inherited;
  Result := -1;
  if not Visible then
    Exit;

  c := Chart;
  if not Assigned(c) then
    Exit;

  if c.Series.Count = 0 then
    Exit;

  r := GetRectangle;
  if r.IsEmpty then
    Exit;

  g := TBitmap.Create(1, 1);
  try
    dr := True;
    c.DoBeforeDrawLegend(g.Canvas, r, dr);
    if dr then
    begin
      xx := r.Left;
      xy := r.Top + 5;

      for I := 0 to c.Series.Count - 1 do
      begin
        s := c.Series[I];
        if s.Visible and s.ShowInLegend then
        begin
          g.Canvas.Font.Assign(Font);
          th := g.Canvas.TextHeight(s.LegendText);
          rt := RectF(xx, xy, xx + r.Width, xy + th);
          if PtInRect(rt, PointF(X, Y)) then
          begin
            Result := I;
            Break;
          end;
          xy := xy + th + 5;
        end;
      end;
    end;
  finally
    g.Free;
  end;
end;

{ TTMSFMXChartSerieMarker }

procedure TTMSFMXChartSerieMarkers.Assign(Source: TPersistent);
begin
  if (Source is TTMSFMXChartSerieMarkers) then
  begin
    FWidth := (Source as TTMSFMXChartSerieMarkers).Width;
    FHeight := (Source as TTMSFMXChartSerieMarkers).Height;
    FShape := (Source as TTMSFMXChartSerieMarkers).Shape;
    FFillOpacity := (Source as TTMSFMXChartSerieMarkers).FillOpacity;
    FFill.Assign((Source as TTMSFMXChartSerieMarkers).Fill);
    FStroke.Assign((Source as TTMSFMXChartSerieMarkers).Stroke);
    FStrokeOpacity := (Source as TTMSFMXChartSerieMarkers).StrokeOpacity;
    FStrokeJoin := (Source as TTMSFMXChartSerieMarkers).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChartSerieMarkers).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartSerieMarkers).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartSerieMarkers).StrokeDash;
    FBitmap.Assign((Source as TTMSFMXChartSerieMarkers).Bitmap);
    FVisible := (Source as TTMSFMXChartSerieMarkers).Visible;
  end;
end;

function TTMSFMXChartSerieMarkers.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieMarkers.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FBitmap := TBitmap.Create(0, 0);
  FWidth := 10;
  FHeight := 10;
  FShape := msEllipse;
  FFillOpacity := 1;
  FFill := TBrush.Create(TBrushKind.Solid, claYellowgreen);
  FFill.OnChanged := FillChanged;
  FStrokeOpacity := 1;
  FStroke := TBrush.Create(TbrushKind.Solid, claDarkgreen);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
  FVisible := False;
end;

destructor TTMSFMXChartSerieMarkers.Destroy;
begin
  FBitmap.Free;
  FStroke.Free;
  FFill.Free;
  inherited;
end;

procedure TTMSFMXChartSerieMarkers.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieMarkers.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieMarkers.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieMarkers.SetBitmap(const Value: TBitmap);
begin
  if FBitmap <> Value then
  begin
    FBitmap.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetHeight(const Value: Double);
begin
  if FHeight <> Value then
  begin
    FHeight := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetShape(
  const Value: TTMSFMXChartSerieMarkerShape);
begin
  if FShape <> Value then
  begin
    FShape := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.SetWidth(const Value: Double);
begin
  if FWidth <> Value then
  begin
    FWidth := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieMarkers.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieMarkers.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieXValues }

procedure TTMSFMXChartSerieXValues.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartSerieXValues then
  begin
    FPositions := (Source as TTMSFMXChartSerieXValues).Positions;
    FAngle := (Source as TTMSFMXChartSerieXValues).Angle;
  end;
end;

constructor TTMSFMXChartSerieXValues.Create(ASerie: TTMSFMXChartSerie);
var
  c: TTMSFMXChart;
begin
  inherited;
  FPositions := [xpBottom];
  FAutoUnits := False;
  FMajorUnitFormat := '%.0f';
  FMinorUnitFormat := '%.0f';
  FAngle := 0;
  c := Chart;
  if Assigned(c) then
  begin
    if (csDesigning in c.ComponentState) and not ((csReading in c.ComponentState) or (csLoading in c.ComponentState)) then
      FTitle.FText := 'X-Axis';
  end;
end;

destructor TTMSFMXChartSerieXValues.Destroy;
begin

  inherited;
end;

procedure TTMSFMXChartSerieXValues.SetAngle(const Value: Double);
begin
  if FAngle <> Value then
  begin
    FAngle := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXValues.SetPositions(
  const Value: TTMSFMXChartXAxisPositions);
begin
  if FPositions <> Value then
  begin
    FPositions := Value;
    UpdateChart;
  end;
end;

{ TTMSFMXChartSerieYValues }

procedure TTMSFMXChartSerieYValues.Assign(Source: TPersistent);
begin
  inherited;
  if Source is TTMSFMXChartSerieYValues then
  begin
    FPositions := (Source as TTMSFMXChartSerieYValues).Positions;
  end;
end;

constructor TTMSFMXChartSerieYValues.Create(ASerie: TTMSFMXChartSerie);
var
  c: TTMSFMXChart;
begin
  inherited;
  FPositions := [ypLeft];
  c := Chart;
  if Assigned(c) then
  begin
    if (csDesigning in c.ComponentState) and not ((csReading in c.ComponentState) or (csLoading in c.ComponentState)) then
      FTitle.FText := 'Y-Axis';
  end;
end;

destructor TTMSFMXChartSerieYValues.Destroy;
begin

  inherited;
end;

procedure TTMSFMXChartSerieYValues.SetPositions(
  const Value: TTMSFMXChartYAxisPositions);
begin
  if FPositions <> Value then
  begin
    FPositions := Value;
    UpdateChart;
  end;
end;

{ TTMSFMXChartSerieXYValues }

procedure TTMSFMXChartSerieXYValues.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieXYValues then
  begin
    FMinorUnitTickMarkSize := (Source as TTMSFMXChartSerieXYValues).MinorUnitTickMarkSize;
    FMinorUnitTickMarkColor := (Source as TTMSFMXChartSerieXYValues).MinorUnitTickMarkColor;
    FMinorUnitFontColor := (Source as TTMSFMXChartSerieXYValues).MinorUnitFontColor;
    FMinorUnitFont.Assign((Source as TTMSFMXChartSerieXYValues).MinorUnitFont);
    FMinorUnit := (Source as TTMSFMXChartSerieXYValues).MinorUnit;
    FMinorUnitFormat := (Source as TTMSFMXChartSerieXYValues).MinorUnitFormat;
    FMinorUnitSpacing := (Source as TTMSFMXChartSerieXYValues).MinorUnitSpacing;
    FMinorUnitFormatType := (Source as TTMSFMXChartSerieXYValues).MinorUnitFormatType;
    FMajorUnitTickMarkSize := (Source as TTMSFMXChartSerieXYValues).MajorUnitTickMarkSize;
    FMajorUnitTickMarkColor := (Source as TTMSFMXChartSerieXYValues).MajorUnitTickMarkColor;
    FMajorUnitFontColor := (Source as TTMSFMXChartSerieXYValues).MajorUnitFontColor;
    FMajorUnitFont.Assign((Source as TTMSFMXChartSerieXYValues).MajorUnitFont);
    FMajorUnit := (Source as TTMSFMXChartSerieXYValues).MajorUnit;
    FMajorUnitFormat := (Source as TTMSFMXChartSerieXYValues).MajorUnitFormat;
    FMajorUnitSpacing := (Source as TTMSFMXChartSerieXYValues).MajorUnitSpacing;
    FMajorUnitFormatType := (Source as TTMSFMXChartSerieXYValues).MajorUnitFormatType;
    FAutoUnits := (Source as TTMSFMXChartSerieXYValues).AutoUnits;
    FTitle.Assign((Source as TTMSFMXChartSerieXYValues).Title);
    FSpiderAngle := (Source as TTMSFMXChartSerieXYValues).SpiderAngle;
    FSpiderValues := (Source as TTMSFMXChartSerieXYValues).SpiderValues;
  end;
end;

function TTMSFMXChartSerieXYValues.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieXYValues.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FAutoUnits := True;
  FSpiderValues := True;
  FSpiderAngle := 270;
  FMinorUnitTickMarkSize := 7;
  FMinorUnitFontColor := claBlack;
  FMinorUnitFont := TFont.Create;
  FMinorUnitFont.OnChanged := FontChanged;
  FMinorUnit := 0;
  FMinorUnitFormat := '%.2f';
  FMinorUnitSpacing := 5;
  FMinorUnitFormatType := vftNormal;
  FMajorUnitTickMarkSize := 10;
  FMajorUnitFontColor := claBlack;
  FMajorUnitFont := TFont.Create;
  FMajorUnitFont.OnChanged := FontChanged;
  FMajorUnit := 1;
  FMajorUnitFormat := '%.2f';
  FMajorUnitSpacing := 5;
  FMajorUnitFormatType := vftNormal;
  FMajorUnitTickMarkColor := claSilver;
  FMinorUnitTickMarkColor := claSilver;
  FTitle := TTMSFMXChartSerieXYTitle.Create(FSerie);
end;

destructor TTMSFMXChartSerieXYValues.Destroy;
begin
  FTitle.Free;
  FMajorUnitFont.Free;
  FMinorUnitFont.Free;
  inherited;
end;

procedure TTMSFMXChartSerieXYValues.FontChanged(Sender: TObject);
begin
  UpdateChart;
end;

procedure TTMSFMXChartSerieXYValues.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieXYValues.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitFontColor(
  const Value: TAlphaColor);
begin
  if FMajorUnitFontColor <> Value then
  begin
    FMajorUnitFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetAutoUnits(const Value: Boolean);
begin
  if FAutoUnits <> Value then
  begin
    FAutoUnits := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnit(const Value: Double);
begin
  if FMajorUnit <> Value then
  begin
    FMajorUnit := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitFont(const Value: TFont);
begin
  if FMajorUnitFont <> Value then
  begin
    FMajorUnitFont.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitFormat(const Value: String);
begin
  if FMajorUnitFormat <> Value then
  begin
    FMajorUnitFormat := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitFormatType(
  const Value: TTMSFMXChartSerieValueFormatType);
begin
  if FMajorUnitFormatType <> Value then
  begin
    FMajorUnitFormatType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitSpacing(const Value: Double);
begin
  if FMajorUnitSpacing <> Value then
  begin
    FMajorUnitSpacing := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitTickMarkColor(
  const Value: TAlphaColor);
begin
  if FMajorUnitTickMarkColor <> Value then
  begin
    FMajorUnitTickMarkColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMajorUnitTickMarkSize(
  const Value: Double);
begin
  if FMajorUnitTickMarkSize <> Value then
  begin
    FMajorUnitTickMarkSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnit(const Value: Double);
begin
  if FMinorUnit <> Value then
  begin
    FMinorUnit := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitFont(const Value: TFont);
begin
  if FMinorUnitFont <> Value then
  begin
    FMinorUnitFont.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitFontColor(
  const Value: TAlphaColor);
begin
  if FMinorUnitFontColor <> Value then
  begin
    FMinorUnitFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitFormat(const Value: String);
begin
  if FMinorUnitFormat <> Value then
  begin
    FMinorUnitFormat := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitFormatType(
  const Value: TTMSFMXChartSerieValueFormatType);
begin
  if FMinorUnitFormatType <> Value then
  begin
    FMinorUnitFormatType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitSpacing(const Value: Double);
begin
  if FMinorUnitSpacing <> Value then
  begin
    FMinorUnitSpacing := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitTickMarkColor(
  const Value: TAlphaColor);
begin
  if FMinorUnitTickMarkColor <> Value then
  begin
    FMinorUnitTickMarkColor := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetMinorUnitTickMarkSize(
  const Value: Double);
begin
  if FMinorUnitTickMarkSize <> Value then
  begin
    FMinorUnitTickMarkSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetSpiderAngle(const Value: Single);
begin
  if FSpiderAngle <> Value then
  begin
    FSpiderAngle := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetSpiderValues(const Value: Boolean);
begin
  if FSpiderValues <> Value then
  begin
    FSpiderValues := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.SetTitle(
  const Value: TTMSFMXChartSerieXYTitle);
begin
  if FTitle <> Value then
  begin
    FTitle := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYValues.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieXYGrid }

procedure TTMSFMXChartSerieXYGrid.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieXYGrid then
  begin
    FMajorUnitStroke.Assign((Source as TTMSFMXChartSerieXYGrid).MajorUnitStroke);
    FMajorUnitStrokeOpacity := (Source as TTMSFMXChartSerieXYGrid).MajorUnitStrokeOpacity;
    FMajorUnitStrokeJoin := (Source as TTMSFMXChartSerieXYGrid).MajorUnitStrokeJoin;
    FMajorUnitStrokeThickness := (Source as TTMSFMXChartSerieXYGrid).MajorUnitStrokeThickness;
    FMajorUnitStrokeCap := (Source as TTMSFMXChartSerieXYGrid).MajorUnitStrokeCap;
    FMajorUnitStrokeDash := (Source as TTMSFMXChartSerieXYGrid).MajorUnitStrokeDash;
    FMinorUnitStroke.Assign((Source as TTMSFMXChartSerieXYGrid).MinorUnitStroke);
    FMinorUnitStrokeOpacity := (Source as TTMSFMXChartSerieXYGrid).MinorUnitStrokeOpacity;
    FMinorUnitStrokeJoin := (Source as TTMSFMXChartSerieXYGrid).MinorUnitStrokeJoin;
    FMinorUnitStrokeThickness := (Source as TTMSFMXChartSerieXYGrid).MinorUnitStrokeThickness;
    FMinorUnitStrokeCap := (Source as TTMSFMXChartSerieXYGrid).MinorUnitStrokeCap;
    FMinorUnitStrokeDash := (Source as TTMSFMXChartSerieXYGrid).MinorUnitStrokeDash;
    FVisible := (Source as TTMSFMXChartSerieXYGrid).Visible;
    FExtended := (Source as TTMSFMXChartSerieXYGrid).Extended;
    FSpiderKind := (Source as TTMSFMXChartSerieXYGrid).SpiderKind;
    FSpiderVisible := (Source as TTMSFMXChartSerieXYGrid).SpiderVisible;
    FSpiderLegend := (Source as TTMSFMXChartSerieXYGrid).SpiderLegend;
  end;
end;

function TTMSFMXChartSerieXYGrid.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieXYGrid.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FSpiderVisible := True;
  FSpiderKind := sgkCircles;
  FSpiderLegend := False;
  FMajorUnitStroke := TBrush.Create(TbrushKind.Solid, claSilver);
  FMajorUnitStroke.OnChanged := StrokeChanged;
  FMajorUnitStrokeThickness := 1;
  FMajorUnitStrokeCap := TStrokeCap.Flat;
  FMajorUnitStrokeDash := TStrokeDash.Solid;
  FMajorUnitStrokeJoin := TStrokeJoin.Round;
  FMajorUnitStrokeOpacity := 1;

  FMinorUnitStroke := TBrush.Create(TbrushKind.Solid, claSilver);
  FMinorUnitStroke.OnChanged := StrokeChanged;
  FMinorUnitStrokeThickness := 1;
  FMinorUnitStrokeCap := TStrokeCap.Flat;
  FMinorUnitStrokeDash := TStrokeDash.Solid;
  FMinorUnitStrokeJoin := TStrokeJoin.Round;
  FMinorUnitStrokeOpacity := 1;
  FVisible := False;
  FExtended := True;
end;

destructor TTMSFMXChartSerieXYGrid.Destroy;
begin
  FMinorUnitStroke.Free;
  FMajorUnitStroke.Free;
  inherited;
end;

procedure TTMSFMXChartSerieXYGrid.Draw(ACanvas: TCanvas; AXGrid: Boolean = True);
begin
  if not Visible then
    Exit;

  DrawKind(ACanvas, vkMinor, AXGrid);
  DrawKind(ACanvas, vkMajor, AXGrid);
end;

procedure TTMSFMXChartSerieXYGrid.DrawKind(ACanvas: TCanvas;
  AKind: TTMSFMXChartDrawXYValueKind; AXGrid: Boolean = True);
var
  i: integer;
  opc: Double;
  dgl: TTMSFMXChartDrawXYGridLine;
  r: TRectF;
  dgls: TTMSFMXChartDrawXYGridLines;
  s: TTMSFMXChartSerie;
  c: TTMSFMXChart;
  dr: Boolean;
  st: TCanvasSaveState;
begin
  s := Serie;
  if not Assigned(s) then
    Exit;

  c := Chart;
  if not Assigned(c) then
    Exit;

  r := c.GetSeriesRectangle;
  if AXGrid then
    dgls := s.FDrawXGridLines
  else
    dgls := s.FDrawYGridLines;

  if AKind = vkMajor then
  begin
    ACanvas.Stroke.Assign(MajorUnitStroke);
    {$IFDEF DELPHIXE11_LVL}
    ACanvas.Stroke.Thickness := MajorUnitStrokeThickness;
    ACanvas.Stroke.Cap := MajorUnitStrokeCap;
    ACanvas.Stroke.Dash := MajorUnitStrokeDash;
    ACanvas.Stroke.Join := MajorUnitStrokeJoin;
    {$ELSE}
    ACanvas.StrokeThickness := MajorUnitStrokeThickness;
    ACanvas.StrokeCap := MajorUnitStrokeCap;
    ACanvas.StrokeDash := MajorUnitStrokeDash;
    ACanvas.StrokeJoin := MajorUnitStrokeJoin;
    {$ENDIF}
    opc := MajorUnitStrokeOpacity;
  end
  else
  begin
    ACanvas.Stroke.Assign(MinorUnitStroke);
    {$IFDEF DELPHIXE11_LVL}
    ACanvas.Stroke.Thickness := MinorUnitStrokeThickness;
    ACanvas.Stroke.Cap := MinorUnitStrokeCap;
    ACanvas.Stroke.Dash := MinorUnitStrokeDash;
    ACanvas.Stroke.Join := MinorUnitStrokeJoin;
    {$ELSE}
    ACanvas.StrokeThickness := MinorUnitStrokeThickness;
    ACanvas.StrokeCap := MinorUnitStrokeCap;
    ACanvas.StrokeDash := MinorUnitStrokeDash;
    ACanvas.StrokeJoin := MinorUnitStrokeJoin;
    {$ENDIF}
    opc := MinorUnitStrokeOpacity;
  end;

  for I := 0 to dgls.Count - 1 do
  begin
    dgl := dgls[I];
    if dgl.Kind = AKind then
    begin
      if AXGrid then
      begin
        dr := True;
        st := ACanvas.SaveState;
        c.DoBeforeDrawSerieXGridLine(ACanvas, s, dgl, dr);
        if dr then
        begin
          ACanvas.DrawLine(PointF(int(dgl.XStartPoint.X) + 0.5, int(dgl.XStartPoint.Y) + 0.5), PointF(Int(dgl.XEndPoint.X) + 0.5, Int(dgl.XEndPoint.Y) + 0.5), opc);
          if (c.FTotalOffset3DY > 0) then
            ACanvas.DrawLine(PointF(int(dgl.XEndPoint.X) + 0.5, int(dgl.XEndPoint.Y) + 0.5), PointF(int(dgl.XEndPoint.X - c.FTotalOffset3DX) + 0.5, int(dgl.XEndPoint.Y + c.FTotalOffset3DY) + 0.5), opc);

          c.DoAfterDrawSerieXGridLine(ACanvas, s, dgl);
        end;
        ACanvas.RestoreState(st);
      end
      else
      begin
        dr := True;
        st := ACanvas.SaveState;
        c.DoBeforeDrawSerieYGridLine(ACanvas, s, dgl, dr);
        if dr then
        begin
          ACanvas.DrawLine(PointF(int(dgl.YStartPoint.X) + 0.5, int(dgl.YStartPoint.Y) + 0.5), PointF(int(dgl.YEndPoint.X) + 0.5, int(dgl.YEndPoint.Y) + 0.5), opc);
          if (c.FTotalOffset3DX > 0) then
            ACanvas.DrawLine(PointF(int(dgl.YStartPoint.X) + 0.5, int(dgl.YStartPoint.Y) + 0.5), PointF(int(dgl.YStartPoint.X - c.FTotalOffset3DX) + 0.5, int(dgl.YStartPoint.Y + c.FTotalOffset3DY) + 0.5), opc);

          c.DoAfterDrawSerieYGridLine(ACanvas, s, dgl);
        end;
        ACanvas.RestoreState(st);
      end;
    end;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieXYGrid.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieXYGrid.SetExtended(const Value: Boolean);
begin
  if FExtended <> Value then
  begin
    FExtended := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStroke(const Value: TBrush);
begin
  if FMajorUnitStroke <> Value then
  begin
    FMajorUnitStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStrokeCap(
  const Value: TStrokeCap);
begin
  if FMajorUnitStrokeCap <> Value then
  begin
    FMajorUnitStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStrokeDash(
  const Value: TStrokeDash);
begin
  if FMajorUnitStrokeDash <> Value then
  begin
    FMajorUnitStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStrokeJoin(
  const Value: TStrokeJoin);
begin
  if FMajorUnitStrokeJoin <> Value then
  begin
    FMajorUnitStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStrokeOpacity(
  const Value: Double);
begin
  if FMajorUnitStrokeOpacity <> Value then
  begin
    FMajorUnitStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMajorUnitStrokeThickness(
  const Value: Double);
begin
  if FMajorUnitStrokeThickness <> Value then
  begin
    FMajorUnitStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStroke(const Value: TBrush);
begin
  if FMinorUnitStroke <> Value then
  begin
    FMinorUnitStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStrokeCap(
  const Value: TStrokeCap);
begin
  if FMinorUnitStrokeCap <> Value then
  begin
    FMinorUnitStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStrokeDash(
  const Value: TStrokeDash);
begin
  if FMinorUnitStrokeDash <> Value then
  begin
    FMinorUnitStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStrokeJoin(
  const Value: TStrokeJoin);
begin
  if FMinorUnitStrokeJoin <> Value then
  begin
    FMinorUnitStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStrokeOpacity(
  const Value: Double);
begin
  if FMinorUnitStrokeOpacity <> Value then
  begin
    FMinorUnitStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetMinorUnitStrokeThickness(
  const Value: Double);
begin
  if FMinorUnitStrokeThickness <> Value then
  begin
    FMinorUnitStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetSpiderKind(
  const Value: TTMSFMXChartSerieSpiderGridKind);
begin
  if FSpiderKind <> Value then
  begin
    FSpiderKind := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetSpiderLegend(const Value: Boolean);
begin
  if FSpiderLegend <> Value then
  begin
    FSpiderLegend := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetSpiderVisible(const Value: Boolean);
begin
  if FSpiderVisible <> Value then
  begin
    FSpiderVisible := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYGrid.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieXYGrid.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieXYTitle }

procedure TTMSFMXChartSerieXYTitle.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieXYTitle then
  begin
    FFont.Assign((Source as TTMSFMXChartSerieXYTitle).Font);
    FFontColor := (Source as TTMSFMXChartSerieXYTitle).FontColor;
    FTextMargins.Assign((Source as TTMSFMXChartSerieXYTitle).TextMargins);
    FText := (Source as TTMSFMXChartSerieXYTitle).Text;
    FTextHorizontalAlignment := (Source as TTMSFMXChartSerieXYTitle).TextHorizontalAlignment;
    FTextVerticalAlignment := (Source as TTMSFMXChartSerieXYTitle).TextVerticalAlignment;
  end;
end;

function TTMSFMXChartSerieXYTitle.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieXYTitle.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFontColor := TAlphaColorRec.Black;
  FTextMargins := TBounds.Create(TRectF.Empty);
  FTextMargins.OnChange := TextMarginsChanged;
  FTextHorizontalAlignment := TTextAlign.Center;
  FTextVerticalAlignment := TTextAlign.Center;
end;

destructor TTMSFMXChartSerieXYTitle.Destroy;
begin
  FTextMargins.Free;
  FFont.Free;
  inherited;
end;

procedure TTMSFMXChartSerieXYTitle.FontChanged(Sender: TObject);
begin
  UpdateChart;
end;

procedure TTMSFMXChartSerieXYTitle.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieXYTitle.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieXYTitle.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.SetText(const Value: String);
begin
  if FText <> Value then
  begin
    FText := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.SetTextHorizontalAlignment(const Value: TTextAlign);
begin
  if FTextHorizontalAlignment <> Value then
  begin
    FTextHorizontalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.SetTextVerticalAlignment(const Value: TTextAlign);
begin
  if FTextVerticalAlignment <> Value then
  begin
    FTextVerticalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.SetTextMargins(const Value: TBounds);
begin
  if FTextMargins <> Value then
  begin
    FTextMargins.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieXYTitle.TextMarginsChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieXYTitle.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieLabels }

procedure TTMSFMXChartSerieLabels.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieLabels then
  begin
    FFillOpacity := (Source as TTMSFMXChartSerieLabels).FillOpacity;
    FFill.Assign((Source as TTMSFMXChartSerieLabels).Fill);
    FStroke.Assign((Source as TTMSFMXChartSerieLabels).Stroke);
    FStrokeOpacity := (Source as TTMSFMXChartSerieLabels).StrokeOpacity;
    FStrokeJoin := (Source as TTMSFMXChartSerieLabels).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChartSerieLabels).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartSerieLabels).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartSerieLabels).StrokeDash;
    FVisible := (Source as TTMSFMXChartSerieLabels).Visible;
    FFormat := (Source as TTMSFMXChartSerieLabels).Format;
    FFormatType := (Source as TTMSFMXChartSerieLabels).FormatType;
    FOffsetX := (Source as TTMSFMXChartSerieLabels).OffsetX;
    FOffsetY := (Source as TTMSFMXChartSerieLabels).OffsetY;
    FFont.Assign((Source as TTMSFMXChartSerieLabels).Font);
    FFontColor := (Source as TTMSFMXChartSerieLabels).FontColor;
    FMode := (Source as TTMSFMXChartSerieLabels).Mode;
  end;
end;

function TTMSFMXChartSerieLabels.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieLabels.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FVisible := False;
  FFillOpacity := 1;
  FStrokeOpacity := 1;
  FOffsetX := 0;
  FOffsetY := -10;
  FFontColor := TAlphaColorRec.Black;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFill := TBrush.Create(TBrushKind.Solid, claGhostwhite);
  FFill.OnChanged := FillChanged;
  FStroke := TBrush.Create(TBrushKind.Solid, claGray);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
  FFormat := '%.2f';
  FFormatType := vftNormal;
  FMode := lmNormal;
end;

destructor TTMSFMXChartSerieLabels.Destroy;
begin
  FStroke.Free;
  FFill.Free;
  Ffont.Free;
  inherited;
end;

procedure TTMSFMXChartSerieLabels.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieLabels.FontChanged(Sender: TObject);
begin
  UpdateChart;
end;

procedure TTMSFMXChartSerieLabels.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieLabels.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieLabels.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetFormat(const Value: String);
begin
  if FFormat <> Value then
  begin
    FFormat := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetFormatType(
  const Value: TTMSFMXChartSerieValueFormatType);
begin
  if FFormatType <> Value then
  begin
    FFormatType := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetMode(
  const Value: TTMSFMXChartSerieLabelsMode);
begin
  if FMode <> Value then
  begin
    FMode := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetOffsetX(const Value: Double);
begin
  if FOffsetX <> Value then
  begin
    FOffsetX := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetOffsetY(const Value: Double);
begin
  if FOffsetY <> Value then
  begin
    FOffsetY := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;
procedure TTMSFMXChartSerieLabels.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLabels.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieLabels.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

{ TTMSFMXChartSerieLegend }

procedure TTMSFMXChartSerieLegend.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartSerieLegend then
  begin
    FFillOpacity := (Source as TTMSFMXChartSerieLegend).FillOpacity;
    FFill.Assign((Source as TTMSFMXChartSerieLegend).Fill);
    FStroke.Assign((Source as TTMSFMXChartSerieLegend).Stroke);
    FStrokeOpacity := (Source as TTMSFMXChartSerieLegend).StrokeOpacity;
    FStrokeJoin := (Source as TTMSFMXChartSerieLegend).StrokeJoin;
    FStrokeThickness := (Source as TTMSFMXChartSerieLegend).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartSerieLegend).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartSerieLegend).StrokeDash;
    FVisible := (Source as TTMSFMXChartSerieLegend).Visible;
    FFont.Assign((Source as TTMSFMXChartSerieLegend).Font);
    FFontColor := (Source as TTMSFMXChartSerieLegend).FontColor;
    FPosition := (Source as TTMSFMXChartSerieLegend).Position;
    FLeft := (Source as TTMSFMXChartSerieLegend).Left;
    FTop := (Source as TTMSFMXChartSerieLegend).Top;
  end;
end;

function TTMSFMXChartSerieLegend.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartSerieLegend.Create(ASerie: TTMSFMXChartSerie);
begin
  FSerie := ASerie;
  FVisible := False;
  FFillOpacity := 1;
  FStrokeOpacity := 1;
  FPosition := lpCenterRight;
  FLeft := -10;
  FTop := 0;
  FFontColor := TAlphaColorRec.Black;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFill := TBrush.Create(TBrushKind.Solid, claGhostwhite);
  FFill.OnChanged := FillChanged;
  FStroke := TBrush.Create(TBrushKind.Solid, claGray);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
end;

destructor TTMSFMXChartSerieLegend.Destroy;
begin
  FStroke.Free;
  FFill.Free;
  Ffont.Free;
  inherited;
end;

procedure TTMSFMXChartSerieLegend.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieLegend.FontChanged(Sender: TObject);
begin
  UpdateChart;
end;

function TTMSFMXChartSerieLegend.GetRectangle: TRectF;
var
  I: Integer;
  c: TTMSFMXChart;
  s: TTMSFMXChartSerie;
  pt: TTMSFMXChartPointVirtual;
  bmp: TBitmap;
  tw, th: Double;
  mxw, mxh, mxhr: Double;
  x, y: Double;
  r: TRectF;
  str: String;
  cnt: Integer;
begin
  Result := TRectF.Empty;

  c := Chart;
  if not Assigned(c) then
    Exit;

  s := Serie;
  if not Assigned(s) then
    Exit;

  cnt := s.GetPointsCount;
  if cnt = 0 then
    Exit;

  mxw := 0;
  mxh := 0;
  mxhr := 0;
  bmp := TBitmap.Create(1, 1);
  bmp.Canvas.Font.Assign(Font);
  for I := 0 to cnt - 1 do
  begin
    pt := s.GetPoint(I);
    str := s.GetLegendText(pt);
    if str <> '' then
    begin
      tw := bmp.Canvas.TextWidth(str) + 5;
      if tw > mxw then
        mxw := tw;
      th := bmp.Canvas.TextHeight(str) + 5;
      if th > mxh then
        mxhr := th;

      mxh := mxh + th;
    end;
  end;

  if s.LegendText <> '' then
  begin
    mxw := Max(mxw, bmp.Canvas.TextWidth(s.LegendText) + 5);
    mxh := mxh + bmp.Canvas.TextHeight(s.LegendText) + 5;
  end;

  bmp.Free;

  if (mxw = 0) and (mxh = 0) then
  begin
    Result := TRectF.Empty;
    Exit;
  end;

  mxw := mxw + mxhr + 5;
  mxh := mxh + 5;

  if s.IsPie then
  begin
    r := s.GetPieRect(True);
    r.Left := r.Left - c.SeriesMargins.Left;
    r.Top := r.Top - c.SeriesMargins.Top;
    r.Bottom := r.Bottom + c.SeriesMargins.Bottom;
    r.Right := r.Right + c.SeriesMargins.Right;
  end
  else
  begin
    r := c.GetSeriesRectangle;
    r.Left := r.Left - c.SeriesMargins.Left + c.FTotalOffset3DX;
    r.Top := r.Top - c.SeriesMargins.Top - c.FTotalOffset3DY;
    r.Bottom := r.Bottom + c.SeriesMargins.Bottom - c.FTotalOffset3DY / 2;
    r.Right := r.Right + c.SeriesMargins.Right + c.FTotalOffset3DX / 2;
  end;

  x := r.Left;
  y := r.Top;

  case Position of
    lpTopCenter: x := x + (r.Width - mxw) / 2;
    lpTopRight: x := x + r.Width - mxw;
    lpCenterLeft: y := y + (r.Height - mxh) / 2;
    lpCenterCenter:
    begin
      y := y + (r.Height - mxh) / 2;
      x := x + (r.Width - mxw) / 2;
    end;
    lpCenterRight:
    begin
      x := x + r.Width - mxw;
      y := y + (r.Height - mxh) / 2;
    end;
    lpBottomLeft:
    begin
      y := y + r.Height - mxh;
    end;
    lpBottomCenter:
    begin
      x := x + (r.Width - mxw) / 2;
      y := y + r.Height - mxh;
    end;
    lpBottomRight:
    begin
      x := x + r.Width - mxw;
      y := y + r.Height - mxh;
    end;
  end;

  x := x + Left;
  y := y + Top;

  Result := RectF(x, y, x + mxw, y + mxh);
end;

procedure TTMSFMXChartSerieLegend.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartSerieLegend.Serie: TTMSFMXChartSerie;
begin
  Result := FSerie;
end;

procedure TTMSFMXChartSerieLegend.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetLeft(const Value: Double);
begin
  if FLeft <> Value then
  begin
    FLeft := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetPosition(
  const Value: TTMSFMXChartLegendPosition);
begin
  if FPosition <> Value then
  begin
    FPosition := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;
procedure TTMSFMXChartSerieLegend.SetTop(const Value: Double);
begin
  if FTop <> Value then
  begin
    FTop := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.SetVisible(const Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartSerieLegend.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartSerieLegend.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;


{ TTMSFMXChartAnnotation }

procedure TTMSFMXChartAnnotation.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartAnnotation then
  begin
    FText := (Source as TTMSFMXChartAnnotation).Text;
    FAutoSize := (Source as TTMSFMXChartAnnotation).AutoSize;
    FCornerRadius := (Source as TTMSFMXChartAnnotation).CornerRadius;
    FWidth := (Source as TTMSFMXChartAnnotation).Width;
    FHeight := (Source as TTMSFMXChartAnnotation).Height;
    FLineThickness := (Source as TTMSFMXChartAnnotation).LineThickness;
    FLineColor := (Source as TTMSFMXChartAnnotation).LineColor;
    FFill.Assign((Source as TTMSFMXChartAnnotation).Fill);
    FStroke.Assign((Source as TTMSFMXChartAnnotation).Stroke);
    FStrokeThickness := (Source as TTMSFMXChartAnnotation).StrokeThickness;
    FStrokeCap := (Source as TTMSFMXChartAnnotation).StrokeCap;
    FStrokeDash := (Source as TTMSFMXChartAnnotation).StrokeDash;
    FStrokeJoin := (Source as TTMSFMXChartAnnotation).StrokeJoin;
    FStrokeOpacity := (Source as TTMSFMXChartAnnotation).StrokeOpacity;
    FFillOpacity := (Source as TTMSFMXChartAnnotation).FillOpacity;
    FArrowColor := (Source as TTMSFMXChartAnnotation).ArrowColor;
    FArrowSize := (Source as TTMSFMXChartAnnotation).ArrowSize;
    FBalloonDirection := (Source as TTMSFMXChartAnnotation).BalloonDirection;
    FBalloonArrowSize := (Source as TTMSFMXChartAnnotation).BalloonArrowSize;
    FVisible := (Source as TTMSFMXChartAnnotation).Visible;
    FOffsetX := (Source as TTMSFMXChartAnnotation).OffsetX;
    FOffsetY := (Source as TTMSFMXChartAnnotation).OffsetY;
    FTextVerticalAlignment := (Source as TTMSFMXChartAnnotation).TextVerticalAlignment;
    FTextHorizontalAlignment := (Source as TTMSFMXChartAnnotation).TextHorizontalAlignment;
    FArrow := (Source as TTMSFMXChartAnnotation).Arrow;
    FFont.Assign((Source as TTMSFMXChartAnnotation).Font);
    FFontColor := (Source as TTMSFMXChartAnnotation).FontColor;
    FShape := (Source as TTMSFMXChartAnnotation).Shape;
    FWordWrap := (Source as TTMSFMXChartAnnotation).WordWrap;
    FLineOpacity := (Source as TTMSFMXChartAnnotation).LineOpacity;
    FArrowOpacity := (Source as TTMSFMXChartAnnotation).ArrowOpacity;
  end;
end;

function TTMSFMXChartAnnotation.Chart: TTMSFMXChart;
var
  s: TTMSFMXChartSerie;
begin
  Result := nil;
  s := Serie;
  if Assigned(s) then
    Result := s.Chart;
end;

constructor TTMSFMXChartAnnotation.Create(Collection: TCollection);
begin
  inherited;
  if Assigned(Collection) then
    FPoint := (Collection as TTMSFMXChartAnnotations).Point;

  FAutoSize := True;
  FCornerRadius := 5;
  FLineOpacity := 1;
  FArrowOpacity := 1;
  FWidth := 75;
  FHeight := 35;
  FLineThickness := 1;
  FLineColor := claBlack;
  FFill := TBrush.Create(TBrushKind.Solid, claGhostwhite);
  FFill.OnChanged := FillChanged;
  FStroke := TBrush.Create(TBrushKind.Solid, claGray);
  FStroke.OnChanged := StrokeChanged;
  FStrokeThickness := 1;
  FStrokeCap := TStrokeCap.Flat;
  FStrokeDash := TStrokeDash.Solid;
  FStrokeJoin := TStrokeJoin.Round;
  FStrokeOpacity := 1;
  FFillOpacity := 1;
  FArrowColor := claBlack;
  FArrowSize := 10;
  FBalloonDirection := bdDownLeft;
  FBalloonArrowSize := 10;
  FVisible := True;
  FOffsetX := 0;
  FOffsetY := -10;
  FTextVerticalAlignment := TTextAlign.Center;
  FTextHorizontalAlignment := TTextAlign.Center;
  FArrow := arLine;
  FFont := TFont.Create;
  FFont.OnChanged := FontChanged;
  FFontColor := TAlphaColorRec.Black;
  FShape := asRectangle;
  FWordWrap := True;
  UpdateChart;
end;

destructor TTMSFMXChartAnnotation.Destroy;
begin
  FFont.Free;
  FFill.Free;
  FStroke.Free;
  inherited;
  UpdateChart;
end;

procedure TTMSFMXChartAnnotation.Draw(ACanvas: TCanvas;
  ADrawPoint: TTMSFMXChartDrawPoint);
begin
  if not Visible then
    Exit;
end;

procedure TTMSFMXChartAnnotation.FillChanged(Sender: TObject);
begin
  RepaintChart;
end;

procedure TTMSFMXChartAnnotation.FontChanged(Sender: TObject);
begin
  UpdateChart;
end;

function TTMSFMXChartAnnotation.Point: TTMSFMXChartPoint;
begin
  Result := FPoint;
end;

procedure TTMSFMXChartAnnotation.UpdateChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.UpdateChart;
end;

function TTMSFMXChartAnnotation.GetValue: TTMSFMXChartAnnotationVirtual;
begin
  Result.Text := Text;
  Result.AutoSize := AutoSize;
  Result.CornerRadius := CornerRadius;
  Result.LineOpacity := LineOpacity;
  Result.ArrowOpacity := ArrowOpacity;
  Result.Width := Width;
  Result.Height := Height;
  Result.LineThickness := LineThickness;
  Result.LineColor := LineColor;
  Result.StrokeOpacity := StrokeOpacity;
  Result.FillOpacity := FillOpacity;
  Result.ArrowColor := ArrowColor;
  Result.ArrowSize := ArrowSize;
  Result.BalloonDirection := BalloonDirection;
  Result.BalloonArrowSize := BalloonArrowSize;
  Result.Visible := Visible;
  Result.OffsetX := OffsetX;
  Result.OffsetY := OffsetY;
  Result.TextVerticalAlignment := TextVerticalAlignment;
  Result.TextHorizontalAlignment := TextHorizontalAlignment;
  Result.Arrow := Arrow;
  Result.FontColor := FontColor;
  Result.Shape := Shape;
  Result.WordWrap := WordWrap;
  Result.Annotation := Self;
end;

procedure TTMSFMXChartAnnotation.RepaintChart;
var
  c: TTMSFMXChart;
begin
  c := Chart;
  if Assigned(c) then
    c.RepaintChart;
end;

function TTMSFMXChartAnnotation.Serie: TTMSFMXChartSerie;
var
  p: TTMSFMXChartPoint;
begin
  Result := nil;
  p := Point;
  if Assigned(p) then
    Result := p.Serie;
end;

procedure TTMSFMXChartAnnotation.SetArrow(
  const Value: TTMSFMXChartAnnotationArrow);
begin
  if FArrow <> Value then
  begin
    FArrow := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetArrowColor(const Value: TAlphaColor);
begin
  if FArrowColor <> Value then
  begin
    FArrowColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetArrowOpacity(const Value: Double);
begin
  if FArrowOpacity <> Value then
  begin
    FArrowOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetArrowSize(const Value: Double);
begin
  if FArrowSize <> Value then
  begin
    FArrowSize := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetBalloonArrowSize(const Value: Double);
begin
  if FBalloonArrowSize <> Value then
  begin
    FBalloonArrowSize := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetBalloonDirection(
  const Value: TTMSFMXChartAnnotationBalloonDirection);
begin
  if FBalloonDirection <> Value then
  begin
    FBalloonDirection := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetCornerRadius(const Value: Double);
begin
  if FCornerRadius <> Value then
  begin
    FCornerRadius := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetFill(const Value: TBrush);
begin
  if FFill <> Value then
  begin
    FFill.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetFillOpacity(const Value: Double);
begin
  if FFillOpacity <> Value then
  begin
    FFillOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetFont(const Value: TFont);
begin
  if FFont <> Value then
  begin
    FFont.Assign(Value);
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetFontColor(const Value: TAlphaColor);
begin
  if FFontColor <> Value then
  begin
    FFontColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetHeight(const Value: Double);
begin
  if FHeight <> Value then
  begin
    FHeight := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetLineColor(const Value: TAlphaColor);
begin
  if FLineColor <> Value then
  begin
    FLineColor := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetLineOpacity(const Value: Double);
begin
  if FLineOpacity <> Value then
  begin
    FLineOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetLineThickness(const Value: Double);
begin
  if FLineThickness <> Value then
  begin
    FLineThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetOffSetX(const Value: Double);
begin
  if FOffsetX <> Value then
  begin
    FOffsetX := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetOffSetY(const Value: Double);
begin
  if FOffsetY <> Value then
  begin
    FOffsetY := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetShape(
  const Value: TTMSFMXChartAnnotationShape);
begin
  if FShape <> Value then
  begin
    FShape := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStroke(const Value: TBrush);
begin
  if FStroke <> Value then
  begin
    FStroke.Assign(Value);
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStrokeCap(const Value: TStrokeCap);
begin
  if FStrokeCap <> Value then
  begin
    FStrokeCap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStrokeDash(const Value: TStrokeDash);
begin
  if FStrokeDash <> Value then
  begin
    FStrokeDash := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStrokeJoin(const Value: TStrokeJoin);
begin
  if FStrokeJoin <> Value then
  begin
    FStrokeJoin := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStrokeOpacity(const Value: Double);
begin
  if FStrokeOpacity <> Value then
  begin
    FStrokeOpacity := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetStrokeThickness(const Value: Double);
begin
  if FStrokeThickness <> Value then
  begin
    FStrokeThickness := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetText(const Value: string);
begin
  if FText <> Value then
  begin
    FText := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetTextHorizontalAlignment(
  const Value: TTextAlign);
begin
  if FTextHorizontalAlignment <> Value then
  begin
    FTextHorizontalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetTextVerticalAlignment(
  const Value: TTextAlign);
begin
  if FTextVerticalAlignment <> Value then
  begin
    FTextVerticalAlignment := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetVisible(const Value: Boolean);
begin
  if FVisible <> value then
  begin
    FVisible := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetWidth(const Value: Double);
begin
  if FWidth <> Value then
  begin
    FWidth := Value;
    UpdateChart;
  end;
end;

procedure TTMSFMXChartAnnotation.SetWordWrap(const Value: Boolean);
begin
  if FWordWrap <> Value then
  begin
    FWordWrap := Value;
    RepaintChart;
  end;
end;

procedure TTMSFMXChartAnnotation.StrokeChanged(Sender: TObject);
begin
  RepaintChart;
end;

{ TTMSFMXChartAnnotations }

function TTMSFMXChartAnnotations.Add: TTMSFMXChartAnnotation;
begin
  Result := TTMSFMXChartAnnotation(inherited Add);
end;

constructor TTMSFMXChartAnnotations.Create(APoint: TTMSFMXChartPoint);
begin
  inherited Create(APoint, TTMSFMXChartAnnotation);
  FPoint := APoint;
end;

function TTMSFMXChartAnnotations.GetItem(Index: Integer): TTMSFMXChartAnnotation;
begin
  Result := TTMSFMXChartAnnotation(inherited Items[Index]);
end;

function TTMSFMXChartAnnotations.Insert(Index: Integer): TTMSFMXChartAnnotation;
begin
  Result := TTMSFMXChartAnnotation(inherited Insert(Index));
end;

function TTMSFMXChartAnnotations.Point: TTMSFMXChartPoint;
begin
  Result := FPoint;
end;

procedure TTMSFMXChartAnnotations.SetItem(Index: Integer;
  const Value: TTMSFMXChartAnnotation);
begin
  inherited Items[Index] := Value;
end;

{ TTMSFMXChartSelector }

constructor TTMSFMXChartSelector.Create(AOwner: TComponent);
begin
  inherited;
  ClipChildren := False;
  FSelected := False;
  FChartType := ctLine;
end;

function TTMSFMXChartSelector.GetText: String;
begin
  case ChartType of
    ctSpider: Result := 'Spider';
    ctPie: Result := 'Pie';
    ctSizedPie: Result := 'Sized Pie';
    ctVariableRadiusPie: Result := 'Var Radius Pie';
    ctLine: Result := 'Line';
    ctDigitalLine: Result := 'Digital Line';
    ctArea: Result := 'Area';
    ctBand: Result := 'Band';
    ctStackedArea: Result := 'Stacked Area';
    ctStackedPercentageArea: Result := 'Stacked Area (%)';
    ctBar: Result := 'Bar';
    ctStackedBar: Result := 'Stacked Bar';
    ctStackedPercentageBar: Result := 'Stacked Bar (%)';
    ctMarker: Result := 'Marker';
    ctXYLine: Result := 'X-Y Line';
    ctXYMarker: Result := 'X-Y Marker';
  end;
end;

procedure TTMSFMXChartSelector.Paint;
var
  rm, rt: TRectF;
  str: String;
  st: TCanvasSaveState;
begin
  rm := LocalRect;
  InflateRect(rm, -0.5, -0.5);

  Canvas.Fill.Kind := TBrushKind.Solid;
  Canvas.Stroke.Kind := TBrushKind.Solid;
  {$IFDEF DELPHIXE11_LVL}
  Canvas.Stroke.Dash := TStrokeDash.Solid;
  {$ELSE}
  Canvas.StrokeDash := TStrokeDash.Solid;
  {$ENDIF}

  if Selected then
  begin
    Canvas.Fill.Color := TAlphaColorRec.White;
    Canvas.Stroke.Color := TAlphaColorRec.Darkorange;
    Canvas.FillRect(rm, 5, 5, AllCorners, 1);
    Canvas.DrawRect(rm, 5, 5, AllCorners, 1);
    Canvas.Fill.Color := TAlphaColorRec.Orange;
    Canvas.Stroke.Color := TAlphaColorRec.White;
  end
  else
  begin
    Canvas.Fill.Color := TAlphaColorRec.White;
    Canvas.Stroke.Color := TAlphaColorRec.Slategray;
    Canvas.FillRect(rm, 5, 5, AllCorners, 1);
    Canvas.DrawRect(rm, 5, 5, AllCorners, 1);
    Canvas.Fill.Color := TAlphaColorRec.Slategray;
    Canvas.Stroke.Color := TAlphaColorRec.Slategray;
  end;

  rt := RectF(rm.Left, rm.Top, rm.Right, rm.Top + 20);

  InflateRect(rm, -5, -5);
  Canvas.FillRect(rt, 5, 5, [TCorner.TopLeft, TCorner.TopRight], 1);
  str := GetText;
  st := Canvas.SaveState;
  Canvas.Fill.Color := TAlphaColorRec.White;
  Canvas.Font.Size := 10;
  Canvas.FillText(rt, str, False, 1, [], TTextAlign.Center, TTextAlign.Center);
  Canvas.RestoreState(st);

  rm.Top := rm.Top + 20;

  if Selected then
  begin
    Canvas.Fill.Color := TAlphaColorRec.Orange;
    Canvas.Stroke.Color := TAlphaColorRec.Orange;
  end
  else
  begin
    Canvas.Fill.Color := TAlphaColorRec.Slategray;
    Canvas.Stroke.Color := TAlphaColorRec.Slategray;
  end;

  DrawChartIcon(Canvas, ChartType, rm);
end;

procedure TTMSFMXChartSelector.SetChartType(const Value: TTMSFMXChartSerieType);
begin
  if FChartType <> Value then
  begin
    FChartType := Value;
    Repaint;
  end;
end;

procedure TTMSFMXChartSelector.SetSelected(const Value: Boolean);
begin
  if FSelected <> Value then
  begin
    FSelected := Value;
    Repaint;
  end;
end;

{ TTMSFMXChartInteractionOptions }

procedure TTMSFMXChartInteractionOptions.Assign(Source: TPersistent);
begin
  if Source is TTMSFMXChartInteractionOptions then
  begin
    FScaleMode := (Source as TTMSFMXChartInteractionOptions).ScaleMode;
    FPanning := (Source as TTMSFMXChartInteractionOptions).Panning;
  end;
end;

function TTMSFMXChartInteractionOptions.Chart: TTMSFMXChart;
begin
  Result := FChart;
end;

constructor TTMSFMXChartInteractionOptions.Create(AChart: TTMSFMXChart);
begin
  FChart := AChart;
  FScaleMode := smHorizontal;
  FPanning := True;
end;

destructor TTMSFMXChartInteractionOptions.Destroy;
begin
  inherited;
end;

procedure TTMSFMXChartSerie.DrawSpiderGrid(ACanvas: TCanvas);
var
  r: TRectF;
  st: TCanvasSaveState;
  I, K: Integer;
  ds: TTMSFMXChartDrawSlice;
  yg: TTMSFMXChartDrawXYGridLine;
  opc: Single;
  lastRadius: TPointF;
  rd: TPointF;
  pth: TPathData;
  c: TTMSFMXChart;
  df: Boolean;
  lgt: TTMSFMXChartDrawLabel;
  pt: TTMSFMXChartPointVirtual;
begin
  if not YGrid.SpiderVisible or (FDrawSlices.Count = 0) or (FDrawYGridLines.Count = 0) then
    Exit;

  c := Chart;
  if not Assigned(c) or not IsSpider then
    Exit;

  r := GetPieRect(False);
  pth := TPathData.Create;

  for I := 0 to FDrawYGridLines.Count - 1 do
  begin
    yg := FDrawYGridLines[I];

    if yg.Kind = vkMajor then
    begin
      ACanvas.Stroke.Assign(YGrid.MajorUnitStroke);
      {$IFDEF DELPHIXE11_LVL}
      ACanvas.Stroke.Thickness := YGrid.MajorUnitStrokeThickness;
      ACanvas.Stroke.Cap := YGrid.MajorUnitStrokeCap;
      ACanvas.Stroke.Dash := YGrid.MajorUnitStrokeDash;
      ACanvas.Stroke.Join := YGrid.MajorUnitStrokeJoin;
      {$ELSE}
      ACanvas.StrokeThickness := YGrid.MajorUnitStrokeThickness;
      ACanvas.StrokeCap := YGrid.MajorUnitStrokeCap;
      ACanvas.StrokeDash := YGrid.MajorUnitStrokeDash;
      ACanvas.StrokeJoin := YGrid.MajorUnitStrokeJoin;
     {$ENDIF}
      opc := YGrid.MajorUnitStrokeOpacity;
    end
    else
    begin
      ACanvas.Stroke.Assign(YGrid.MinorUnitStroke);
      {$IFDEF DELPHIXE11_LVL}
      ACanvas.Stroke.Thickness := YGrid.MinorUnitStrokeThickness;
      ACanvas.Stroke.Cap := YGrid.MinorUnitStrokeCap;
      ACanvas.Stroke.Dash := YGrid.MinorUnitStrokeDash;
      ACanvas.Stroke.Join := YGrid.MinorUnitStrokeJoin;
      {$ELSE}
      ACanvas.StrokeThickness := YGrid.MinorUnitStrokeThickness;
      ACanvas.StrokeCap := YGrid.MinorUnitStrokeCap;
      ACanvas.StrokeDash := YGrid.MinorUnitStrokeDash;
      ACanvas.StrokeJoin := YGrid.MinorUnitStrokeJoin;
     {$ENDIF}
      opc := YGrid.MinorUnitStrokeOpacity;
    end;

    st := ACanvas.SaveState;
    df := True;
    c.DoBeforeDrawSerieYGridLine(ACanvas, Self, yg, df);
    if yg.Kind = vkMajor then
      lastRadius := yg.ValueRadius;    
    
    if df then
    begin  
      case YGrid.SpiderKind of
        sgkCircles: ACanvas.DrawArc(r.CenterPoint, yg.ValueRadius, Pie.StartAngle, Pie.SweepAngle, opc);
        sgkPolygon:
        begin
          pth.Clear;
          ds := FDrawSlices[0];
          rd.X := ds.CenterPoint.X + yg.ValueRadius.X * Cos(DegToRad(ds.StartAngle + ds.SweepAngle / 2));
          rd.Y := ds.CenterPoint.Y + yg.ValueRadius.Y * Sin(DegToRad(ds.StartAngle + ds.SweepAngle / 2));
          pth.MoveTo(rd);
          for K := 1 to FDrawSlices.Count - 1 do
          begin
            ds := FDrawSlices[K];
            rd.X := ds.CenterPoint.X + yg.ValueRadius.X * Cos(DegToRad(ds.StartAngle + ds.SweepAngle / 2));
            rd.Y := ds.CenterPoint.Y + yg.ValueRadius.Y * Sin(DegToRad(ds.StartAngle + ds.SweepAngle / 2));
            pth.LineTo(rd);
          end;

          pth.ClosePath;

          ACanvas.DrawPath(pth, opc);
        end;
      end;
      c.DoAfterDrawSerieYGridLine(ACanvas, Self, yg);
    end;

    ACanvas.RestoreState(st);
  end;

  for I := 0 to FDrawSlices.Count - 1 do
  begin
    ds := FDrawSlices[I];
    yg.Kind := vkMajor;
    yg.ValueGridLine := True;

    pt := ds.VirtualReference;

    yg.Value := pt.YValue;

    ACanvas.Stroke.Assign(YGrid.MajorUnitStroke);
    {$IFDEF DELPHIXE11_LVL}
    ACanvas.Stroke.Thickness := YGrid.MajorUnitStrokeThickness;
    ACanvas.Stroke.Cap := YGrid.MajorUnitStrokeCap;
    ACanvas.Stroke.Dash := YGrid.MajorUnitStrokeDash;
    ACanvas.Stroke.Join := YGrid.MajorUnitStrokeJoin;
    {$ELSE}
    ACanvas.StrokeThickness := YGrid.MajorUnitStrokeThickness;
    ACanvas.StrokeCap := YGrid.MajorUnitStrokeCap;
    ACanvas.StrokeDash := YGrid.MajorUnitStrokeDash;
    ACanvas.StrokeJoin := YGrid.MajorUnitStrokeJoin;
   {$ENDIF}
    opc := YGrid.MajorUnitStrokeOpacity;

    rd.X := ds.CenterPoint.X + lastRadius.X * Cos(DegToRad(ds.StartAngle + ds.SweepAngle / 2));
    rd.Y := ds.CenterPoint.Y + lastRadius.Y * Sin(DegToRad(ds.StartAngle + ds.SweepAngle / 2));

    st := ACanvas.SaveState;
    df := True;
    c.DoBeforeDrawSerieYGridLine(ACanvas, Self, yg, df);
    if df then
    begin
      ACanvas.DrawLine(ds.CenterPoint, rd, opc);
      c.DoAfterDrawSerieYGridLine(ACanvas, Self, yg);
    end;
    ACanvas.RestoreState(st);

    if YGrid.SpiderLegend then
    begin
      st := ACanvas.SaveState;
      lgt.ValueString := GetSpiderLegendText(pt);
      lgt.Value := pt.YValue;
      lgt.Reference := pt.Point;
      lgt.VirtualReference := pt;
      lgt.Point := rd;

      ACanvas.Font.Assign(Legend.Font);
      ACanvas.Fill.Color := Legend.FontColor;

      lgt.TextWidth := ACanvas.TextWidth(lgt.ValueString);
      lgt.TextHeight := ACanvas.TextHeight(lgt.ValueString);
      case GetPieQuadrant(ds.CenterPoint, rd) of
        cpq1: lgt.Rect := RectF(lgt.Point.X + 2, lgt.Point.Y - lgt.TextHeight - 2, lgt.Point.X + 2 + lgt.TextWidth, lgt.Point.Y - 2);
        cpq2: lgt.Rect := RectF(lgt.Point.X - lgt.TextWidth - 2, lgt.Point.Y - 2 - lgt.TextHeight, lgt.Point.X - 2, lgt.Point.Y - 2);
        cpq3: lgt.Rect := RectF(lgt.Point.X - lgt.TextWidth - 2, lgt.Point.Y + 2, lgt.Point.X - 2, lgt.Point.Y + lgt.TextHeight + 2);
        cpq4: lgt.Rect := RectF(lgt.Point.X + 2, lgt.Point.Y + 2, lgt.Point.X + lgt.TextWidth + 2, lgt.Point.Y + lgt.TextHeight + 2);
      end;

      df := True;
      c.DoBeforeDrawSerieGridLegendText(ACanvas, Self, lgt, df);
      if df then
      begin
        ACanvas.FillText(lgt.Rect, lgt.ValueString, False, 1, [], TTextAlign.Leading, TTextAlign.Leading);
        c.DoAfterDrawSerieGridLegendText(ACanvas, Self, lgt);
      end;

      ACanvas.RestoreState(st);
    end;
  end;

  pth.Free;
end;

procedure TTMSFMXChartSerie.DrawSpiderValues(ACanvas: TCanvas);
var
  r: TRectF;
  c: TTMSFMXChart;
  I: Integer;
  yv: TTMSFMXChartDrawXYValue;
  tw, th: Single;
  an: Single;
  rd: TPointF;
  rt: TRectF;
  st: TCanvasSaveState;
  df: Boolean;
  cx: TPointF;
  m, rm: TMatrix;
begin
  if not YGrid.SpiderVisible or not YValues.SpiderValues or (FDrawSlices.Count = 0) or (FDrawYGridLines.Count = 0) then
    Exit;

  c := Chart;
  if not Assigned(c) or not IsSpider then
    Exit;

  r := GetPieRect(False);

  for I := 0 to FDrawYValues.Count - 1 do
  begin
    yv := FDrawYValues[I];
    if yv.Kind = vkMajor then
    begin
      ACanvas.Font.Assign(YValues.MajorUnitFont);
      ACanvas.Fill.Color := YValues.MajorUnitFontColor;
      ACanvas.Stroke.Color := YValues.MajorUnitTickMarkColor;
    end
    else
    begin
      ACanvas.Font.Assign(YValues.MinorUnitFont);
      ACanvas.Fill.Color := YValues.MinorUnitFontColor;
      ACanvas.Stroke.Color := YValues.MinorUnitTickMarkColor;
    end;

    tw := ACanvas.TextWidth(yv.ValueString);
    th := ACanvas.TextHeight(yv.ValueString);
    
    an := DegToRad(YValues.SpiderAngle);

    rd.X := r.CenterPoint.X + yv.ValueRadius.X * Cos(an);
    rd.Y := r.CenterPoint.Y + yv.ValueRadius.Y * Sin(an);

    rt := RectF(rd.X - tw / 2, rd.Y - th / 2, rd.X + tw / 2, rd.Y + th / 2);
    st := ACanvas.SaveState;

    if r.CenterPoint = rd then
    begin
      rd.X := r.CenterPoint.X + yv.ValueRadius.X + 1 * Cos(an);
      rd.Y := r.CenterPoint.Y + yv.ValueRadius.Y + 1 * Sin(an);
    end;

    case GetPieQuadrant(r.CenterPoint, rd) of
      cpq1, cpq2: an := an + DegToRad(90);
      cpq3, cpq4: an := an + DegToRad(-90);
    end;

    cx.X := rt.Left + rt.Width / 2;
    cx.Y := rt.Top + rt.Height / 2;
    m := ACanvas.Matrix;
    rm := TMatrix.CreateRotation(an) * TMatrix.CreateTranslation(cx.X, cx.Y);
    ACanvas.MultiplyMatrix(rm);
    rt := RectF(-rt.Width / 2, -rt.Height / 2, rt.Width / 2, rt.Height / 2);

    df := True;
    c.DoBeforeDrawSerieYValue(ACanvas, Self, ypCenter, yv, df);
    if df then
    begin
      ACanvas.FillText(rt, yv.ValueString, False, 1, [], TTextAlign.Leading, TTextAlign.Leading);
      c.DoAfterDrawSerieYValue(ACanvas, Self, ypCenter, yv);
    end;

    ACanvas.SetMatrix(m);

    ACanvas.RestoreState(st);
  end;
end;

end.
